/*****************************************************************
* The software in this file was written by Jessica Gaines,       *
* while working at the University of Edinburgh, under the        *
* direction of Terry Lyons, and using funding                    *
* provided by EC grant SC1*CT92-0784.                            *
* You may use this software for educational and academic         *
* research purposes only.                                        *
* Please do not give copies of the software in this file to      *
* anyone else. Tell anyone interested to email                   *
*  jessica@maths.ed.ac.uk                                        *
* or to write to                                                 *
*  Jessica Gaines                                                *
*  University of Edinburgh                                       *
*  Dept. of Mathematics and Statistics                           *
*  JCMB                                                          *
*  Kings Buildings                                               *
*  Mayfield Road                                                 *
*  Edinburgh EH9 3JZ                                             *
*  UK                                                            *
* and then get copies directly from Edinburgh themselves.        *
******************************************************************
*/

/* This file contains routines for generation and
 handling of Brownian paths.
 Contents:
 brown1,brown2,brown3
 getdw,getdwback,new_getdw
and also subsidiary routines called by the above:
 brownchunk : called by brown3,brown5,getdw,getdwback,new_getdw
 tellprerec : called by getdw,getdw2,new_getdw
 initstack,freestack,push,pop : called by getdw,getdwback,getdw2
 tellnexrec : called by getdwback
*/

#include <math.h>
#include <malloc.h>
#include <stdio.h>
#include <strings.h>

brown1(N,h,func,w)
 float w[],h,(*func)();
 int N;
/* Generates a one-dimensional Brownian path as values
w(t) for t=0,h,2h,....,(N-1)h
*/
{
  float     sum, fastnorm(), mean, sd;
  int       n;

  mean=0.0;
  sd=sqrt(h);
  sum = 0.0;
  for (n = 0; n < N; n++)
      w[n] = sum += fastnorm(mean, sd, func);
}


brown2(NSTEPS,T,func,w)
float w[],(*func)();
int NSTEPS,T;
/* Generates a Brownian path
as a discrete set of points.
Each level is created using the previous
level and a set of normal random numbers. The parameter
NSTEPS specifies the total number of levels, and
T specifies the interval (0,T) on which
the path is to be generated.
*/
{
  float     sum,vari, fastnorm(),mean, *y, sd;
  int       np, k, j;
    mean = 0.0;
    /* step 1 */
    sd = vari = 1.0;
    y=(float *)malloc(sizeof(float));
    if(!y) {printf(" allocation failure in Brown2\n"); exit(1);}
    y[0] = w[0] = sum = 0.0;
    np = T;
    for (k = 1; k <= T; k++)
      {sum += fastnorm(mean, sd, func); w[k]=sum;}
    vari = vari / 2;
    /* subsequent steps */
    for (k = 1; k < NSTEPS; k++)
    {
      vari = vari / 2;
      sd = sqrt(vari);
      y=(float *)realloc((char *)y,(unsigned)(np+1)*sizeof(float));
      if(!y) {printf(" allocation failure in Brown2\n"); exit(1);}
      for (j = 0; j <= np; j++)
        y[j] = w[j];
      for (j = 1; j <= np; j++)
      {
        w[2 * j - 1] = 
          0.5*(y[j - 1] + y[j]) + fastnorm(mean, sd, func);
        w[2 * j] = y[j];
      }
      np = np * 2;
    }
    free((char *)y);
    np++;
    return np;
}

brown2_(NSTEPS,T,func,w)
float w[],(*func)();
int NSTEPS,T;
/* Generates a Brownian path
as a discrete set of points.
Each level is created using the previous
level and a set of normal random numbers. The parameter
NSTEPS specifies the total number of levels, and
T specifies the interval (0,T) on which
the path is to be generated.
 Version to be called from a Fortran program unit.
*/
{
  float     sum,vari, fastnorm(),mean, *y, sd;
  int       np, k, j;
    mean = 0.0;
    /* step 1 */
    sd = vari = 1.0;
    y=(float *)malloc(sizeof(float));
    if(!y) {printf(" allocation failure in Brown2\n"); exit(1);}
    y[0] = w[0] = sum = 0.0;
    np = T;
    for (k = 1; k <= T; k++)
      {sum += fastnorm(mean, sd, func); w[k]=sum;}
    vari = vari / 2;
    /* subsequent steps */
    for (k = 1; k < NSTEPS; k++)
    {
      vari = vari / 2;
      sd = sqrt(vari);
      y=(float *)realloc((char *)y,(unsigned)(np+1)*sizeof(float));
      if(!y) {printf(" allocation failure in Brown2\n"); exit(1);}
      for (j = 0; j <= np; j++)
        y[j] = w[j];
      for (j = 1; j <= np; j++)
      {
        w[2 * j - 1] = 
          0.5*(y[j - 1] + y[j]) + fastnorm(mean, sd, func);
        w[2 * j] = y[j];
      }
      np = np * 2;
    }
    free((char *)y);
    np++;
    return np;
}

brownchunk(wp,np,nw,w1,w2,sd,func)
int np,nw; float **wp,**w1,**w2,sd,(*func)();
/* Creates 2*np intervals on next level of a brownian
tree of dimension nw, given np intervals on previous level. 
wp[][] contains the differences across intervals on
previous level. w1[][] and w2[][] will contain differences
across intervals on this level, each one containing
np differences.
*/
{float fastnorm(),x,ww,**wtemp,**matrix();
 int k,i;
 wtemp=matrix(0,2*np-1,0,nw-1);
 for(k=0;k<np;k++) for(i=0;i<nw;i++)
 {x=fastnorm(0.0, sd, func);
  ww=wp[k][i];
  wtemp[2*k][i]=0.5*ww+x;
  wtemp[2*k+1][i]=0.5*ww-x;
 }
 for(k=0;k<np;k++) for(i=0;i<nw;i++)
 {w1[k][i]=wtemp[k][i];
  w2[k][i]=wtemp[k+np][i];
 }
 free_matrix(wtemp,0,2*np-1,0,nw-1);
} 

 

#define K 5

typedef struct{ int recno,level,prerec,nexrec,parent;
                int lson,rson,spare[7];
                float t0,tp;} record;
typedef struct{ int np,nw,nlevels,nl1,seed,nrecs1,nrecs2;
                int spare[7];
                float sp[2];} header;

getdw(nw,dw,level,t,h,func,idum,cangoup)
int nw,level,idum,*cangoup;
float *dw,t,h,(*func)();
/* Routine to read and return elements from a Brownian
tree stored in a file, also to add elements to the tree 
when necessary.
The external variables wfile and in, that may be used
by other program units, are pointers to the file 
containing the Brownian tree and to the io-unit
connected to the file.
The routine keeps several (K) records in memory at all times,
so that file accesses are not performed during
most calls to the routine.
getdw presumes that it is being used to help solve an sde
forwards through time, ie it presumes that the next call
to getdw will be at the same period or the next one.
*/
{extern char *wfile;
 extern  FILE *in;
 static float ***w;
 static record r[K];
 static header head;
 static int l,np,recsize,init=0;
 static long size;
 float ***array3(),hh,tt,sd,**w2,**matrix();
 int i,j,k,tag;
 record r2;
 /*printf(" Entering getdw: level=%d t=%f h=%f\n",level,t,h);*/
 if(init==0||idum<0) /* Initialize by reading in K records, such that
                the last is on given level and the others are
                all parents. */
 {/*printf(" Initializing getdw: h=%f\n",h);*/
  in=fopen(wfile,"r+");
  if(in==NULL) {printf(" Can't open file in getdw\n"); exit(1);}
  /*else printf(" File %s opened.\n",wfile);*/
  recsize=sizeof(record);
  fread((char *)&head,recsize,1,in);
  if(nw!=head.nw) {printf(" Noise vector wrong size in %s.\n",wfile);
                   printf(" nw=%d head.nw=%d\n",nw,head.nw);
                   exit(1);}
  np=head.np;
  /*printf(" nlevels=%d\n",head.nlevels);
  printf(" nrecs=%d\n",head.nrecs1);*/
  if(head.nlevels<level)
  {printf(" Not enough levels in initial brownian path\n");
   exit(1);}
  if(init==0) w=array3(0,K-1,0,np-1,0,nw-1); 
  size=recsize+np*nw*sizeof(float); 
  fseek(in,size,0);
  readw(w[0],np,nw,r,in);
  while(r[0].level<level-K+1||t<r[0].t0) readw(w[0],np,nw,r,in);
  for(i=1;i<K;i++)
  { fseek(in,size*r[i-1].lson,0);
    readw(w[i],np,nw,r+i,in);}
  /*printf(" At level %d\n",r[K-1].level);*/
  l=level;
  init=1;
  /*printf(" Finished initialising\n");*/
  /* return ; */
 }
 while(level>l) /* Go down one or more levels */
 { for(i=0;i<K-1;i++)
   { r[i]=r[i+1];
     for(j=0;j<np;j++) for(k=0;k<nw;k++) w[i][j][k]=w[i+1][j][k];
   }
  if(r[K-1].lson>0) 
  {fseek(in,size*r[K-1].lson,0);
   readw(w[K-1],np,nw,r+K-1,in);
   l++;
  }
  else /* Create new noise elements */
  {/*printf(" Creating new arrays on lower levels.\n");*/
   i=1<<l;
   /* printf(" previous level=%d 1<<level=%d\n",l,i); */
   hh=1.0/i;
   sd=sqrt(hh*0.5);
   w2=matrix(0,np-1,0,nw-1);
   brownchunk(w[K-2],np,nw,w[K-1],w2,sd,func);
   r[K-1].recno=head.nrecs1++;head.nrecs2++;
   l++; r[K-1].level=l;
   r[K-1].nexrec=r[K-1].recno+1;
   r[K-1].parent=r[K-2].recno;r[K-1].lson=r[K-1].rson=(-1);
   /*r[K-1].prerec=findprerec(r+K-1,size,recsize,in);*/
   r[K-1].prerec=(-1);
   r[K-2].lson=r[K-1].recno;
   r[K-1].t0=r[K-2].t0;
   r[K-1].tp=r[K-1].t0+np*hh;
   fclose(in);
   in=fopen(wfile,"a+");
   i=savew(w[K-1],np,nw,r+K-1,in);
   fclose(in);
   in=fopen(wfile,"r+");
   r2.recno=head.nrecs1++;head.nrecs2++;r2.level=l;
   r2.prerec=r2.recno-1;
   r2.parent=r[K-1].parent;r2.lson=r2.rson=(-1);
   /*r2.nexrec=findnexrec(&r2,size,recsize,in);*/
   r2.nexrec=(-1);
   r[K-2].rson=r2.recno;
   r2.t0=r[K-1].tp;r2.tp=r2.t0+np*hh;
   /*printf(" tp=%f\n",r2.tp);*/
   fclose(in);
   in=fopen(wfile,"a+");
   i=savew(w2,np,nw,&r2,in);
   fclose(in);
   in=fopen(wfile,"r+");
   free_matrix(w2,0,np-1,0,nw-1);
   /* Rewrite header */
   fwrite((char *)&head,recsize,1,in);
   /* Rewrite record(K-2) to update lson,rson */
   fseek(in,r[K-2].recno*size,0);
   savew(w[K-2],np,nw,r+K-2,in);
  }
 }
 
 while(l-K+1>level) /* Go up one or more levels */
 { for(i=K-1;i>0;i--)
   { r[i]=r[i-1];
     for(j=0;j<np;j++) for(k=0;k<nw;k++) w[i][j][k]=w[i-1][j][k];
   }
   fseek(in,size*r[1].parent,0);
   readw(w[0],np,nw,r,in);
   l--;
 }

 if(t<r[0].t0) {printf(" Error: t<t0 in getw.\n"); exit(1);}
 /* Get correct level */ 
   for(i=0;i<K;i++) if(r[i].level==level) {tag=i; break;}

   while(t>=r[tag].tp) /* Get correct time interval */
   { /* Move tag down to K-1 */
    /*printf(" Next time interval needed.\n");
    printf(" t=%f tp=%f tag=%d\n",t,r[tag].tp,tag);*/
    r[K-1]=r[tag];
    for(j=0;j<np;j++) for(k=0;k<nw;k++) w[K-1][j][k]=w[tag][j][k];
    tag=K-1; l=level;
    if(r[tag].nexrec>0)
    {fseek(in,r[tag].nexrec*size,0);
     readw(w[tag],np,nw,r+tag,in);
    }
    else
    {r[0]=r[tag];
     initstack(); 
     while(r[0].nexrec<0)
     {push(r[0].recno); /*printf(" pushing %d\n",r[0].recno);*/
      fseek(in,r[0].parent*size,0);
      readw(w[0],np,nw,r,in);
     }
     fseek(in,r[0].nexrec*size,0);
     readw(w[0],np,nw,r,in);
     while(r[0].level<level) /* Create more tree */
     {
      /* Rewrite parent */
      /*printf(" Creating new arrays to broaden tree.\n");*/
      fseek(in,r[0].recno*size,0);
      r[0].lson=head.nrecs1;r[0].rson=head.nrecs1+1;
            fwrite((char *)r,recsize,1,in);
      hh=1.0/(1<<r[0].level);
      sd = sqrt(0.5*hh);
      brownchunk(w[0],np,nw,w[1],w[2],sd,func);
      r[1]=r[0];
      r[1].recno=head.nrecs1++;head.nrecs2++;
      r[1].parent=r[0].recno;
      r[1].prerec=pop(); printf(" popping %d\n",r[1].prerec);
      r[1].level++;r[1].nexrec=r[1].recno+1;
      r[1].lson=r[1].rson=(-1);
      r[1].tp=r[0].t0+np*hh;
      tellprerec(r[1].prerec,r[1].recno,recsize,size,in);
      fclose(in);
      in=fopen(wfile,"a+");
      savew(w[1],np,nw,r+1,in);
      fclose(in);
      in=fopen(wfile,"r+");
      r[2]=r[1];head.nrecs1++;head.nrecs2++;
      r[2].recno++;r[2].prerec=r[2].recno-1;
      r[2].nexrec=(-1);
      r[2].t0=r[2].tp;r[2].tp+=np*hh;
      fclose(in);
      in=fopen(wfile,"a+");
      savew(w[2],np,nw,r+2,in);
      fclose(in);
      in=fopen(wfile,"r+");
      for(k=0;k<np;k++) for(i=0;i<nw;i++)
       w[0][k][i]=w[1][k][i];
      r[0]=r[1];
      /* Rewrite header */
      fwrite((char *)&head,recsize,1,in);
     }
     r[K-1]=r[0];
     for(j=0;j<np;j++) for(k=0;k<nw;k++) w[K-1][j][k]=w[0][j][k];
     freestack();
    }
/*Fill in w and r */
     for(i=K-2;i>=0;i--)
     {fseek(in,r[i+1].parent*size,0);
      readw(w[i],np,nw,r+i,in);
     }
   } 
/* Return correct dw */
 tt=r[tag].t0; i=0;
 /*printf(" h=%f\n",h);*/
 while(tt<t) { i++; tt=r[tag].t0+i*h;}
 /*printf(" h=%f i=%d\n",h,i);*/
 if(i>np-1) {printf(" Error in time.\n"); exit(1);}
 for(j=0;j<nw;j++) dw[j]=w[tag][i][j];
/* Check if possible to go up a level
   next time. It is only possible when the
   current time index = 3 mod 4. */
  *cangoup=((i%4==3&&level>5)?1:0);
/* printf("Returning w from record %d h=%e
w=%e\n",r[tag].recno,h,dw[0]); */
}

tellprerec(prerec,nexrec,recsize,size,in)
int prerec,nexrec,recsize;
long size;
FILE *in;
/* Updates the record prerec so that the pointer
to the next record points to nexrec.
*/
{ record r;
  fseek(in,prerec*size,0);
  fread((char *)&r,recsize,1,in);
  r.nexrec=nexrec;
  /* printf(" rec=%d nexrec=%d\n",r.recno,r.nexrec); */
  fseek(in,(long)(-recsize),1);
  fwrite((char *)&r,recsize,1,in);
}

int *sp;

initstack()
/* Initialises a stack, pointed to by sp */
{extern int *sp;
 unsigned stacksize=20; /* Max. no of levels */
 sp=(int *)malloc(stacksize*sizeof(int));
 if(!sp) {printf(" allocation failure in initstack\n"); exit(1);}
}

freestack()
/* Releases the memory used by the stack
pointed to by sp */
{extern int *sp;
 free((char *)sp);
}

push(n) int n;
/* Adds the integer n to the stack */
{extern int *sp;
 *(sp++)=n;
}

int pop()
/* Gets an integer from the top of the stack */
{extern int *sp;
 return (*(--sp));
}


getdwback(nw,dw,level,t,h,func,idum,cangoup)
int nw,level,idum,*cangoup;
float *dw,t,h,(*func)();
/* This routine is similar to getdw, but whereas
getdw presumes that an sde is being solved forwards
through time, getdwback presumes that it is being
solved backwards through time.
*/
{extern char *wfile;
 extern  FILE *in;
 static  float ***w;
 static record r[K];
 static header head;
 static int l,np,recsize,init=0;
 static long size;
 float ***array3(),hh,tt,sd,**w2,**matrix();
 int i,j,k,tag;
 record r2;
 /*printf(" Entering getdw: level=%d t=%f\n",level,t);*/
 if(init==0||idum<0) /* Initialize by reading in K records, such that
                the last is on given level and the others are
                all parents. */
 {/*printf(" Initializing getdw\n");*/
  in=fopen(wfile,"r+");
  if(in==NULL) {printf(" Can't open file in getdw\n"); exit(1);}
  /*else printf(" File %s opened.\n",wfile);*/
  recsize=sizeof(record);
  fread((char *)&head,recsize,1,in);
  if(nw!=head.nw) {printf(" Noise vector wrong size in %s.\n",wfile);
                exit(1);}
  np=head.np;
  /*printf(" nlevels=%d\n",head.nlevels);
  printf(" nrecs=%d\n",head.nrecs1);*/
  if(init==0) w=array3(0,K-1,0,np-1,0,nw-1);
  size=recsize+np*nw*sizeof(float);
  /*printf(" size=%d\n",size); 
  printf(" level=%d t=%f\n",level,t);*/ 
  fseek(in,size,0);
  readw(w[K-1],np,nw,r+K-1,in);
  while(r[K-1].level!=level||t<=r[K-1].t0||t>r[K-1].tp)
   readw(w[K-1],np,nw,r+K-1,in);
  /*printf(" At level %d\n",r[K-1].level);*/
  for(i=K-2;i>=0;i--)
  { if(r[i+1].parent<0) {printf(" Need to start deeper.\n");
                         exit(1);}
    fseek(in,size*r[i+1].parent,0);
    readw(w[i],np,nw,r+i,in);
    /*printf(" At level %d\n",r[i].level);*/ }
  l=level;
  init=1;
  /*printf(" Finished initialising\n");*/
  return ;
 }
 while(level>l) /* Go down one or more levels */
 { 
   for(i=0;i<K-1;i++)
   { r[i]=r[i+1];
     for(j=0;j<np;j++) for(k=0;k<nw;k++) w[i][j][k]=w[i+1][j][k];
   }
  if(r[K-1].rson>0) 
  {fseek(in,size*r[K-1].rson,0);
   readw(w[K-1],np,nw,r+K-1,in);
   l++;
  }
  else /* Create new noise elements */
  {/*printf(" Creating new arrays on lower levels.\n");*/
   i=1<<l;
   /*printf(" previous level=%d 1<<level=%d\n",l,i);*/
   hh=1.0/i;
   sd=sqrt(hh*0.5);
   w2=matrix(0,np-1,0,nw-1);
   brownchunk(w[K-2],np,nw,w2,w[K-1],sd,func);
   r2.recno=head.nrecs1++;
   head.nrecs2++;
   l++; r2.level=l;
   r2.nexrec=r2.recno+1;
   r2.parent=r[K-2].recno;r2.lson=r2.rson=(-1);
   r2.prerec=(-1);
   r[K-2].lson=r2.recno;
   r2.t0=r[K-2].t0;
   r2.tp=r2.t0+np*hh;
   fclose(in);
   in=fopen(wfile,"a+");
   i=savew(w2,np,nw,&r2,in);
   fclose(in);
   in=fopen(wfile,"r+");
   r[K-1].recno=head.nrecs1++;head.nrecs2++;r[K-1].level=l;
   r[K-1].prerec=r[K-1].recno-1;
   r[K-1].parent=r2.parent;r[K-1].lson=r[K-1].rson=(-1);
   r[K-1].nexrec=(-1);
   r[K-2].rson=r[K-1].recno;
   r[K-1].t0=r2.tp;r[K-1].tp=r[K-1].t0+np*hh;
   /*printf(" tp=%f\n",r2.tp);*/
   fclose(in);
   in=fopen(wfile,"a+");
   i=savew(w[K-1],np,nw,r+K-1,in);
   fclose(in);
   in=fopen(wfile,"r+");
   free_matrix(w2,0,np-1,0,nw-1);
   /* Rewrite header */
   fwrite((char *)&head,recsize,1,in);
   /* Rewrite record(K-2) to update lson,rson */
   fseek(in,r[K-2].recno*size,0);
   savew(w[K-2],np,nw,r+K-2,in);
  }
 }
 
 while(l-K+1>level) /* Go up one or more levels */
 { for(i=K-1;i>0;i--)
   { r[i]=r[i-1];
     for(j=0;j<np;j++) for(k=0;k<nw;k++) w[i][j][k]=w[i-1][j][k];
   }
   fseek(in,size*r[1].parent,0);
   readw(w[0],np,nw,r,in);
   l--;
 }

 if(t>r[0].tp) 
   {printf(" t=%f r[0].tp=%f\n",t,r[0].tp);
   for(i=0;i<K;i++) printf(" level=%d t0=%f tp=%f\n",r[i].level,r[i].t0,
         r[i].tp);
   printf(" Error: t>tp in getdwback.\n"); exit(1);}
 /* Get correct level */ 
   for(i=0;i<K;i++)  if(r[i].level==level) {tag=i; break;} 
                 
   while(t<=r[tag].t0) /* Get correct time interval */
   { /* Move tag down to K-1 */
    /*printf(" Previous time interval needed.\n");
    printf(" t=%f t0=%f\n",t,r[tag].t0);*/
    r[K-1]=r[tag];
    for(j=0;j<np;j++) for(k=0;k<nw;k++) w[K-1][j][k]=w[tag][j][k];
    tag=K-1; l=level;
    if(r[tag].prerec>0)
    {fseek(in,r[tag].prerec*size,0);
     readw(w[tag],np,nw,r+tag,in);
    }
    else
    {r[0]=r[tag];
     initstack();
     while(r[0].prerec<0)
     {push(r[0].recno);
      fseek(in,r[0].parent*size,0);
      readw(w[0],np,nw,r,in);
     }
     fseek(in,r[0].prerec*size,0);
     r2.t0=r[0].t0;
     readw(w[0],np,nw,r,in);
     if(r[0].tp!=r2.t0) {printf(" oldt0=%f\
 prerec(tp)=%f\n",r2.t0,r[0].tp);
                         printf(" Wrong prerec\n"); exit(1);}
                
     while(r[0].level<level) /* Create more tree */
     {
      /* Rewrite parent */
      /*printf(" Creating new arrays to broaden tree.\n");*/
      fseek(in,r[0].recno*size,0);
      r[0].lson=head.nrecs1;r[0].rson=head.nrecs1+1;
            fwrite((char *)r,recsize,1,in);
      hh=1.0/(1<<r[0].level);
      sd = sqrt(0.5*hh);
      brownchunk(w[0],np,nw,w[1],w[2],sd,func);
      r[1]=r[0];
      r[1].recno=head.nrecs1++;head.nrecs2++;
      r[1].parent=r[0].recno;
      r[1].prerec=(-1);
      r[1].level++;r[1].nexrec=r[1].recno+1;
      r[1].lson=r[1].rson=(-1);
      r[1].tp=r[0].t0+np*hh;
      fclose(in);
      in=fopen(wfile,"a+");
      savew(w[1],np,nw,r+1,in);
      fclose(in);
      in=fopen(wfile,"r+");
      r[2]=r[1];head.nrecs1++;head.nrecs2++;
      r[2].recno++;r[2].prerec=r[2].recno-1;
      r[2].nexrec=pop();
      r[2].t0=r[2].tp;r[2].tp+=np*hh;
      tellnexrec(r[2].nexrec,r[2].recno,recsize,size,in);
      fclose(in);
      in=fopen(wfile,"a+");
      savew(w[2],np,nw,r+2,in);
      fclose(in);
      in=fopen(wfile,"r+");
      for(k=0;k<np;k++) for(i=0;i<nw;i++)
       w[0][k][i]=w[2][k][i];
      r[0]=r[2];
      /* Rewrite header */
      fwrite((char *)&head,recsize,1,in);
     }
     r[K-1]=r[0];
     for(j=0;j<np;j++) for(k=0;k<nw;k++) w[K-1][j][k]=w[0][j][k];
     freestack();
    }
/*Fill in w and r */
     for(i=K-2;i>=0;i--)
     {fseek(in,r[i+1].parent*size,0);
      readw(w[i],np,nw,r+i,in);
     }
   } 
/* Return correct dw */
 tt=r[tag].tp; i=np-1;
 while(tt>t) { tt=r[tag].t0+i*h; i--; }
 /* printf(" i=%d\n",i); */
 if(i<0) {printf(" Error in time.\n"); exit(1);}
 for(j=0;j<nw;j++) dw[j]=w[tag][i][j];
/* Check if possible to go up a level
   next time. It is only possible when the
   current time index = 0 mod 4. */
  *cangoup=((i%4==0&&level>5)?1:0);
}

tellnexrec(nexrec,prerec,recsize,size,in)
int prerec,nexrec,recsize;
long size;
FILE *in;
{ record r;
  fseek(in,nexrec*size,0);
  fread((char *)&r,recsize,1,in);
  /*printf(" prerec=%d\n",r.prerec);*/
  r.prerec=prerec;
  /*printf(" rec=%d prerec=%d\n",r.recno,r.prerec);*/
  fseek(in,(long)(-recsize),1);
  fwrite((char *)&r,recsize,1,in);
}


brown3(nlevels,n1,np,nw,func,seed,file)
float (*func)();
int nlevels,n1,np,nw,seed;
char file[];
/* A new version of brown2.
Creates a disjoint array for each np intervals.
n1=no. of preliminary levels (2 arrays only).
*/
{
  float vari, fastnorm(),mean=0.0,**wp,**w1,**w2,sd;
  float **matrix(),t,h;
  int k,j,i,inrec,prec;
  int spare[7];
  long size;
  FILE *out,*fopenw();
  header head;
  record r,r1;
  size=(long)sizeof(r);
  i=sizeof(head);
  if(size!=i) {printf(" Bad record sizes\n"); exit(1);}
  size+=np*nw*sizeof(float);
  out=fopenw(file); 	/* Open and close file first 	*/
  fclose(out);		/* to ensure it is empty. 	*/
  out=fopen(file,"w+");
  if(out==NULL) {printf(" Can't open %s for direct access.\n",file);
                 exit(1);}
  w1=matrix(0,np-1,0,nw-1);
  w2=matrix(0,np-1,0,nw-1);
  wp=matrix(0,np-1,0,nw-1);
   /* Header */
  for(j=0;j<7;j++) head.spare[j]=r.spare[j]=spare[j]=0;
  for(j=0;j<np;j++) for(i=0;i<nw;i++) wp[j][i]=0.0;
  head.nlevels=nlevels; head.seed=seed;
  head.nl1=n1; head.np=np; head.nw=nw;
  head.nrecs1=head.nrecs2=0;
  head.sp[0]=head.sp[1]=0;
  savew(wp,np,nw,&head,out);
  head.nrecs1++;
    /* Level 1 */
    printf(" Level 1\n");
    t=0.0; h=1.0;
    sd = vari = 1.0;
    for(k=0;k<np;k++) for(i=0;i<nw;i++)
    wp[k][i]=fastnorm(mean, sd, func);
    printf(" w(1)=%f\n",wp[0][0]);
    r.recno=1;r.level=1;
    r.prerec=r.nexrec=(-1);
    r.parent=(-1);
    if(nlevels>1) {r.lson=2;r.rson=3;}
    else r.lson=r.rson=(-1);
    r.t0=t;r.tp=t+np*h;
    savew(wp,np,nw,&r,out);
    head.nrecs1++;
    vari = vari / 2;
    /* subsequent n1-1 levels 
    Only create first 2 arrays of subintervals
    at each level*/
    for (j = 2; j <= n1; j++)
    {
     printf(" Level %d\n",j);
     vari*=0.5; h*=0.5; /*printf(" h=%f\n",h);*/
     sd = sqrt(vari);
     brownchunk(wp,np,nw,w1,w2,sd,func);
     if(j==2) prec=1;
     else     prec=2*(j-2);
     r.recno++;r.level=j;r.prerec=(-1);r.nexrec=r.recno+1;
     r.parent=prec;
     if(j<nlevels) {r.lson=r.recno+2;r.rson=r.recno+3;}
     else r.lson=r.rson=(-1);
     r.t0=t;r.tp=t+np*h;
     savew(w1,np,nw,&r,out);
     head.nrecs1++;
     r.recno++;r.prerec=r.recno-1;
     r.nexrec=(-1);r.t0=r.tp;r.tp+=np*h;
     r.lson=r.rson=(-1);
     savew(w2,np,nw,&r,out);
     head.nrecs1++;
     for(k=0;k<np;k++) for(i=0;i<nw;i++)
      wp[k][i]=w1[k][i];
    }
    /* remaining nlevels-n1 levels:
    Create and store all arrays of
    subintervals. */
    k=2; 
    inrec=r.prerec; /*printf(" inrec=%d\n",inrec);*/
    for(j=n1+1;j<=nlevels;j++)
    {
     printf(" Level %d\n",j);
     vari*=0.5; sd=sqrt(vari);
     h*=0.5; /*printf(" h=%f\n",h);*/
     t=0.0;
     for(i=0;i<k;i++)
     {fseek(out,(long)inrec*size,0);
      readw(wp,np,nw,&r1,out);
      if(r1.level!=j-1) {printf(" Wrong level\n");
                         exit(1);}
      r1.lson=r.recno+1; 
      r1.rson=r1.lson+1;
      fseek(out,-size,1);
      savew(wp,np,nw,&r1,out);
      inrec++;
      brownchunk(wp,np,nw,w1,w2,sd,func);
      fseek(out,(long)0,2);
      r.recno++;r.level=j;
      if(i==0) r.prerec=(-1);
      else r.prerec=r.recno-1;
      r.nexrec=r.recno+1;
      r.parent=r1.recno;r.lson=r.rson=(-1);
      r.t0=t;r.tp=t+np*h;
      savew(w1,np,nw,&r,out);
      head.nrecs1++;
      r.recno++;r.prerec=r.recno-1;
      if(i==(k-1)) r.nexrec=(-1);
      else r.nexrec=r.recno+1;
      r.t0=r.tp;r.tp+=np*h;
      savew(w2,np,nw,&r,out);
      head.nrecs1++;
      t+=2.0*np*h;
     }
     k*=2;
    }
/* Rewrite header */
 fseek(out,(long)0,0);
 fwrite((char *)&head,sizeof(header),1,out);
 /*printf(" nrecs=%d\n",head.nrecs1);*/
    fclose(out);
    free_matrix(wp,0,np-1,0,nw-1);
    free_matrix(w1,0,np-1,0,nw-1);
    free_matrix(w2,0,np-1,0,nw-1);
}

new_getdw(nw,dw,level,t,h,func,idum,cangoup)
int nw,level,idum,*cangoup;
float *dw,t,h,(*func)();
/* Routine to read and return elements from a Brownian
tree stored in a file, also to add elements to the tree 
when necessary.
The external variables wfile and in, that may be used
by other program units, are pointers to the file 
containing the Brownian tree and to the io-unit
connected to the file.
The routine keeps several (K) records in memory at all times,
so that file accesses are not performed during
most calls to the routine.
getdw presumes that it is being used to help solve an sde
forwards through time, ie it presumes that the next call
to getdw will be at the same period or the next one.

This new version does not presume that the Brownian tree
already exists, it can create it.
*/
{extern char *wfile;
 extern  FILE *in;
 static float ***w;
 static record r[K];
 static header head;
 static int l,np,recsize,init=0;
 static long size;
 float ***array3(),hh,tt,sd,vari,**w1,**w2,**wp,**matrix();
 float fastnorm();
 int i,j,k,n,tag;
 record r2;
 /*printf(" Entering getdw: level=%d t=%f\n",level,t);*/
 if(init==0||idum<0) /* Initialize by reading in K records, such that
                the last is on given level and the others are
                all parents. */
 {printf(" Initializing getdw: h=%f\n",h);
  recsize=sizeof(record);
  in=fopen(wfile,"r+");
  if(in==NULL) /* File does not exist, so create it */
  {printf("Starting new Brownian tree in %s\n",wfile);
   in=fopen(wfile,"w+");
   if(in==NULL) {printf(" Can't open %s for direct access.\n",wfile);
                 exit(1);}
   printf("WARNING: the no. of elements in the dw array must be greater\n");
   printf("than the largest final time you will ever set using this Brownian path.\n");
   getint(&np,"no. of elements in each dw array (a power of 2 is recommended)");
   size=recsize+np*nw*sizeof(float); 
    /* Header */
   wp=matrix(0,np-1,0,nw-1);
   w1=matrix(0,np-1,0,nw-1);
   w2=matrix(0,np-1,0,nw-1);
   for(j=0;j<7;j++) head.spare[j]=r[0].spare[j]=0;
   for(j=0;j<np;j++) for(i=0;i<nw;i++)  wp[j][i]=0.0;
   head.nlevels=1; head.seed=0;
   head.nl1=1; head.np=np; head.nw=nw;
   head.nrecs1=head.nrecs2=0;
   head.sp[0]=head.sp[1]=0.0;
   savew(wp,np,nw,&head,in);
   head.nrecs1++;
   /* First record */
   sd = vari = 1.0; hh=1.0; tt=0.0;
   for(k=0;k<np;k++) for(i=0;i<nw;i++)
    wp[k][i]=fastnorm(0.0, sd, func);
   r[0].recno=1;r[0].level=1;
   r[0].prerec=r[0].nexrec=(-1);
   r[0].parent=(-1);
   if(K>1) {r[0].lson=2;r[0].rson=3;}
   else r[0].lson=r[0].rson=(-1);
   r[0].t0=tt; r[0].tp=tt+np*hh;
   savew(wp,np,nw,r,in);
   head.nrecs1++;
   vari = vari / 2;
   /* subsequent K-1 levels :
      Only create first 2 arrays of subintervals at each level
   */
   for (j = 2; j <= K; j++)
   {
    /* printf(" Level %d\n",j); */
    vari*=0.5; hh*=0.5; /*printf(" h=%f\n",hh);*/
    sd = sqrt(vari);
    brownchunk(wp,np,nw,w1,w2,sd,func);
    r[0].recno++;r[0].level=j;r[0].prerec=(-1);r[0].nexrec=r[0].recno+1;
    r[0].parent=(j==2?1:2*(j-2));
    if(j<K) {r[0].lson=r[0].recno+2;r[0].rson=r[0].recno+3;}
    else r[0].lson=r[0].rson=(-1);
    r[0].t0=tt; r[0].tp=tt+np*hh;
    savew(w1,np,nw,r,in);
    head.nrecs1++;
    r[0].prerec=r[0].recno; r[0].recno++;
    r[0].nexrec=(-1);r[0].t0=r[0].tp;r[0].tp+=np*hh;
    r[0].lson=r[0].rson=(-1);
    savew(w2,np,nw,r,in);
    head.nrecs1++;
    for(k=0;k<np;k++) for(i=0;i<nw;i++)
    wp[k][i]=w1[k][i];
   }
  /* Rewrite header */
   fseek(in,(long)0,0);
   fwrite((char *)&head,sizeof(header),1,in);
  /* Tidy up */
   free_matrix(wp,0,np-1,0,nw-1);
   free_matrix(w1,0,np-1,0,nw-1);
   free_matrix(w2,0,np-1,0,nw-1);
   fclose(in);
   in=fopen(wfile,"r+");
  }
/* Now read in first K records */
  {fread((char *)&head,recsize,1,in);
   if(nw!=head.nw) {printf(" Noise vector wrong size in %s.\n",wfile);
                    exit(1);}
   np=head.np;
   /*printf(" nlevels=%d\n",head.nlevels);
   printf(" nrecs=%d\n",head.nrecs1);*/
   if(init==0) w=array3(0,K-1,0,np-1,0,nw-1); 
   size=recsize+np*nw*sizeof(float); 
   fseek(in,size,0);
   readw(w[K-1],np,nw,r+K-1,in);
   for(n=1;n<K;n++)
    if(r[K-1].lson>0)
     {for(i=K-1-n;i<K-1;i++)
      {r[i]=r[i+1];
       for(j=0;j<np;j++) for(k=0;k<nw;k++) w[i][j][k]=w[i+1][j][k];
      }
      fseek(in,size*r[K-1].lson,0);
      readw(w[K-1],np,nw,r+K-1,in);
     }
   l=r[K-1].level; /*printf(" level=%d\n",l);*/
  }
  init=1;
  /*printf(" Finished initialising.\n");*/
  return;
 }
 /*printf("1: h=%f\n",h);*/
 while(level>l) /* Go down one or more levels */
 { for(i=0;i<K-1;i++)
   {r[i]=r[i+1];
    for(j=0;j<np;j++) for(k=0;k<nw;k++) w[i][j][k]=w[i+1][j][k];
   }
  if(r[K-1].lson>0) 
  {/*printf("son=%d\n",r[K-1].lson);*/
   fseek(in,size*r[K-1].lson,0);
   readw(w[K-1],np,nw,r+K-1,in);
   l++;
  }
  else /* Create new noise elements */
  {/*printf(" Creating new arrays on lower levels.\n");*/
   i=1<<l;
   /*printf(" previous level=%d 1<<level=%d\n",l,i);*/
   h=1.0/i;
   sd=sqrt(h*0.5);
   w2=matrix(0,np-1,0,nw-1);
   brownchunk(w[K-2],np,nw,w[K-1],w2,sd,func);
   r[K-1].recno=head.nrecs1++;head.nrecs2++;
   l++; r[K-1].level=l;
   r[K-1].nexrec=r[K-1].recno+1;
   r[K-1].parent=r[K-2].recno;r[K-1].lson=r[K-1].rson=(-1);
   r[K-1].prerec=(-1);
   r[K-2].lson=r[K-1].recno;
   r[K-1].t0=r[K-2].t0;
   r[K-1].tp=r[K-1].t0+np*h;
   fclose(in);
   in=fopen(wfile,"a+");
   i=savew(w[K-1],np,nw,r+K-1,in);
   fclose(in);
   in=fopen(wfile,"r+");
   r2.recno=head.nrecs1++;head.nrecs2++;r2.level=l;
   r2.prerec=r2.recno-1;
   r2.parent=r[K-1].parent;r2.lson=r2.rson=(-1);
   r2.nexrec=(-1);
   r[K-2].rson=r2.recno;
   r2.t0=r[K-1].tp;r2.tp=r2.t0+np*h;
   /*printf(" tp=%f\n",r2.tp);*/
   fclose(in);
   in=fopen(wfile,"a+");
   i=savew(w2,np,nw,&r2,in);
   fclose(in);
   in=fopen(wfile,"r+");
   free_matrix(w2,0,np-1,0,nw-1);
   /* Rewrite header */
   fwrite((char *)&head,recsize,1,in);
   /* Rewrite record(K-2) to update lson,rson */
   fseek(in,r[K-2].recno*size,0);
   savew(w[K-2],np,nw,r+K-2,in);
  }
 }
 
 while(l-K+1>level) /* Go up one or more levels */
 { for(i=K-1;i>0;i--)
   { r[i]=r[i-1];
     for(j=0;j<np;j++) for(k=0;k<nw;k++) w[i][j][k]=w[i-1][j][k];
   }
   fseek(in,size*r[1].parent,0);
   readw(w[0],np,nw,r,in);
   l--;
 }

 if(t<r[0].t0) {printf(" Error: t<t0 in getw.\n"); exit(1);}
 /* Get correct level */ 
   for(i=0;i<K;i++) if(r[i].level==level) {tag=i; break;}

   while(t>=r[tag].tp) /* Get correct time interval */
   { /* Move tag down to K-1 */
    /*printf(" Next time interval needed.\n");
    printf(" t=%f tp=%f tag=%d\n",t,r[tag].tp,tag);*/
    r[K-1]=r[tag];
    for(j=0;j<np;j++) for(k=0;k<nw;k++) w[K-1][j][k]=w[tag][j][k];
    tag=K-1; l=level;
    if(r[tag].nexrec>0)
    {fseek(in,r[tag].nexrec*size,0);
     readw(w[tag],np,nw,r+tag,in);
    }
    else
    {r[0]=r[tag];
     initstack(); 
     while(r[0].nexrec<0)
     {push(r[0].recno); /*printf(" pushing %d\n",r[0].recno);*/
      fseek(in,r[0].parent*size,0);
      readw(w[0],np,nw,r,in);
     }
     fseek(in,r[0].nexrec*size,0);
     readw(w[0],np,nw,r,in);
     while(r[0].level<level) /* Create more tree */
     {
      /* Rewrite parent */
      /*printf(" Creating new arrays to broaden tree.\n");*/
      fseek(in,r[0].recno*size,0);
      r[0].lson=head.nrecs1;r[0].rson=head.nrecs1+1;
            fwrite((char *)r,recsize,1,in);
      h=1.0/(1<<r[0].level);
      sd = sqrt(0.5*h);
      brownchunk(w[0],np,nw,w[1],w[2],sd,func);
      r[1]=r[0];
      r[1].recno=head.nrecs1++;head.nrecs2++;
      r[1].parent=r[0].recno;
      r[1].prerec=pop(); /*printf(" popping %d\n",r[1].prerec);*/
      r[1].level++;r[1].nexrec=r[1].recno+1;
      r[1].lson=r[1].rson=(-1);
      r[1].tp=r[0].t0+np*h;
      tellprerec(r[1].prerec,r[1].recno,recsize,size,in);
      fclose(in);
      in=fopen(wfile,"a+");
      savew(w[1],np,nw,r+1,in);
      fclose(in);
      in=fopen(wfile,"r+");
      r[2]=r[1];head.nrecs1++;head.nrecs2++;
      r[2].recno++;r[2].prerec=r[2].recno-1;
      r[2].nexrec=(-1);
      r[2].t0=r[2].tp;r[2].tp+=np*h;
      fclose(in);
      in=fopen(wfile,"a+");
      savew(w[2],np,nw,r+2,in);
      fclose(in);
      in=fopen(wfile,"r+");
      for(k=0;k<np;k++) for(i=0;i<nw;i++)
       w[0][k][i]=w[1][k][i];
      r[0]=r[1];
      /* Rewrite header */
      fwrite((char *)&head,recsize,1,in);
     }
     r[K-1]=r[0];
     for(j=0;j<np;j++) for(k=0;k<nw;k++) w[K-1][j][k]=w[0][j][k];
     freestack();
    }
/*Fill in w and r */
     for(i=K-2;i>=0;i--)
     {fseek(in,r[i+1].parent*size,0);
      readw(w[i],np,nw,r+i,in);
     }
   } 
/* Return correct dw */
 tt=r[tag].t0; i=0;
 while(tt<t) { i++; tt=r[tag].t0+i*h;}
 /*printf(" i=%d\n",i);*/
 if(i>np-1) {printf(" Error in time: tag=%d t=%f h=%f np=%d tt=%f\n"); exit(1);}
 for(j=0;j<nw;j++) dw[j]=w[tag][i][j];
/* Check if possible to go up a level
   next time. It is only possible when the
   current time index = 3 mod 4. */
  *cangoup=((i%4==3&&level>5)?1:0);
  return ;
}



my_new_getdw(nw,dw,level,t,h,func,idum,cangoup,np)
int nw,level,idum,*cangoup,np;
float *dw,t,h,(*func)();
/* Routine to read and return elements from a Brownian
tree stored in a file, also to add elements to the tree 
when necessary.
The external variables wfile and in, that may be used
by other program units, are pointers to the file 
containing the Brownian tree and to the io-unit
connected to the file.
The routine keeps several (K) records in memory at all times,
so that file accesses are not performed during
most calls to the routine.
getdw presumes that it is being used to help solve an sde
forwards through time, ie it presumes that the next call
to getdw will be at the same period or the next one.

This new version does not presume that the Brownian tree
already exists, it can create it.
*/
{extern char *wfile;
 extern  FILE *in;
 static float ***w;
 static record r[K];
 static header head;
 static int l,recsize,init=0;
 static long size;
 float ***array3(),hh,tt,sd,vari,**w1,**w2,**wp,**matrix();
 float fastnorm();
 int i,j,k,n,tag;
 record r2;
 /*printf(" Entering getdw: level=%d t=%f\n",level,t);*/
 if(init==0||idum<0) /* Initialize by reading in K records, such that
                the last is on given level and the others are
                all parents. */
 {printf(" Initializing getdw: h=%f\n",h);
  recsize=sizeof(record);
  in=fopen(wfile,"r+");
  if(in==NULL) /* File does not exist, so create it */
  {printf("Starting new Brownian tree in %s\n",wfile);
   in=fopen(wfile,"w+");
   if(in==NULL) {printf(" Can't open %s for direct access.\n",wfile);
                 exit(1);}
   /*   printf("WARNING: the no. of elements in the dw array must be greater\n");
   printf("than the largest final time you will ever set using this Brownian path.\n");
   getint(&np,"no. of elements in each dw array (a power of 2 is recommended)");
   */
   size=recsize+np*nw*sizeof(float); 
    /* Header */
   wp=matrix(0,np-1,0,nw-1);
   w1=matrix(0,np-1,0,nw-1);
   w2=matrix(0,np-1,0,nw-1);
   for(j=0;j<7;j++) head.spare[j]=r[0].spare[j]=0;
   for(j=0;j<np;j++) for(i=0;i<nw;i++)  wp[j][i]=0.0;
   head.nlevels=1; head.seed=0;
   head.nl1=1; head.np=np; head.nw=nw;
   head.nrecs1=head.nrecs2=0;
   head.sp[0]=head.sp[1]=0.0;
   savew(wp,np,nw,&head,in);
   head.nrecs1++;
   /* First record */
   sd = vari = 1.0; hh=1.0; tt=0.0;
   for(k=0;k<np;k++) for(i=0;i<nw;i++)
    wp[k][i]=fastnorm(0.0, sd, func);
   r[0].recno=1;r[0].level=1;
   r[0].prerec=r[0].nexrec=(-1);
   r[0].parent=(-1);
   if(K>1) {r[0].lson=2;r[0].rson=3;}
   else r[0].lson=r[0].rson=(-1);
   r[0].t0=tt; r[0].tp=tt+np*hh;
   savew(wp,np,nw,r,in);
   head.nrecs1++;
   vari = vari / 2;
   /* subsequent K-1 levels :
      Only create first 2 arrays of subintervals at each level
   */
   for (j = 2; j <= K; j++)
   {
    /* printf(" Level %d\n",j); */
    vari*=0.5; hh*=0.5; /*printf(" h=%f\n",hh);*/
    sd = sqrt(vari);
    brownchunk(wp,np,nw,w1,w2,sd,func);
    r[0].recno++;r[0].level=j;r[0].prerec=(-1);r[0].nexrec=r[0].recno+1;
    r[0].parent=(j==2?1:2*(j-2));
    if(j<K) {r[0].lson=r[0].recno+2;r[0].rson=r[0].recno+3;}
    else r[0].lson=r[0].rson=(-1);
    r[0].t0=tt; r[0].tp=tt+np*hh;
    savew(w1,np,nw,r,in);
    head.nrecs1++;
    r[0].prerec=r[0].recno; r[0].recno++;
    r[0].nexrec=(-1);r[0].t0=r[0].tp;r[0].tp+=np*hh;
    r[0].lson=r[0].rson=(-1);
    savew(w2,np,nw,r,in);
    head.nrecs1++;
    for(k=0;k<np;k++) for(i=0;i<nw;i++)
    wp[k][i]=w1[k][i];
   }
  /* Rewrite header */
   fseek(in,(long)0,0);
   fwrite((char *)&head,sizeof(header),1,in);
  /* Tidy up */
   free_matrix(wp,0,np-1,0,nw-1);
   free_matrix(w1,0,np-1,0,nw-1);
   free_matrix(w2,0,np-1,0,nw-1);
   fclose(in);
   in=fopen(wfile,"r+");
  }
/* Now read in first K records */
  {fread((char *)&head,recsize,1,in);
   if(nw!=head.nw) {printf(" Noise vector wrong size in %s.\n",wfile);
                    exit(1);}
   np=head.np;
   /*printf(" nlevels=%d\n",head.nlevels);
   printf(" nrecs=%d\n",head.nrecs1);*/
   if(init==0) w=array3(0,K-1,0,np-1,0,nw-1); 
   size=recsize+np*nw*sizeof(float); 
   fseek(in,size,0);
   readw(w[K-1],np,nw,r+K-1,in);
   for(n=1;n<K;n++)
    if(r[K-1].lson>0)
     {for(i=K-1-n;i<K-1;i++)
      {r[i]=r[i+1];
       for(j=0;j<np;j++) for(k=0;k<nw;k++) w[i][j][k]=w[i+1][j][k];
      }
      fseek(in,size*r[K-1].lson,0);
      readw(w[K-1],np,nw,r+K-1,in);
     }
   l=r[K-1].level; /*printf(" level=%d\n",l);*/
  }
  init=1;
  /*printf(" Finished initialising.\n");*/
  return;
 }
 /*printf("1: h=%f\n",h);*/
 while(level>l) /* Go down one or more levels */
 { for(i=0;i<K-1;i++)
   {r[i]=r[i+1];
    for(j=0;j<np;j++) for(k=0;k<nw;k++) w[i][j][k]=w[i+1][j][k];
   }
  if(r[K-1].lson>0) 
  {/*printf("son=%d\n",r[K-1].lson);*/
   fseek(in,size*r[K-1].lson,0);
   readw(w[K-1],np,nw,r+K-1,in);
   l++;
  }
  else /* Create new noise elements */
  {/*printf(" Creating new arrays on lower levels.\n");*/
   i=1<<l;
   /*printf(" previous level=%d 1<<level=%d\n",l,i);*/
   h=1.0/i;
   sd=sqrt(h*0.5);
   w2=matrix(0,np-1,0,nw-1);
   brownchunk(w[K-2],np,nw,w[K-1],w2,sd,func);
   r[K-1].recno=head.nrecs1++;head.nrecs2++;
   l++; r[K-1].level=l;
   r[K-1].nexrec=r[K-1].recno+1;
   r[K-1].parent=r[K-2].recno;r[K-1].lson=r[K-1].rson=(-1);
   r[K-1].prerec=(-1);
   r[K-2].lson=r[K-1].recno;
   r[K-1].t0=r[K-2].t0;
   r[K-1].tp=r[K-1].t0+np*h;
   fclose(in);
   in=fopen(wfile,"a+");
   i=savew(w[K-1],np,nw,r+K-1,in);
   fclose(in);
   in=fopen(wfile,"r+");
   r2.recno=head.nrecs1++;head.nrecs2++;r2.level=l;
   r2.prerec=r2.recno-1;
   r2.parent=r[K-1].parent;r2.lson=r2.rson=(-1);
   r2.nexrec=(-1);
   r[K-2].rson=r2.recno;
   r2.t0=r[K-1].tp;r2.tp=r2.t0+np*h;
   /*printf(" tp=%f\n",r2.tp);*/
   fclose(in);
   in=fopen(wfile,"a+");
   i=savew(w2,np,nw,&r2,in);
   fclose(in);
   in=fopen(wfile,"r+");
   free_matrix(w2,0,np-1,0,nw-1);
   /* Rewrite header */
   fwrite((char *)&head,recsize,1,in);
   /* Rewrite record(K-2) to update lson,rson */
   fseek(in,r[K-2].recno*size,0);
   savew(w[K-2],np,nw,r+K-2,in);
  }
 }
 
 while(l-K+1>level) /* Go up one or more levels */
 { for(i=K-1;i>0;i--)
   { r[i]=r[i-1];
     for(j=0;j<np;j++) for(k=0;k<nw;k++) w[i][j][k]=w[i-1][j][k];
   }
   fseek(in,size*r[1].parent,0);
   readw(w[0],np,nw,r,in);
   l--;
 }

 if(t<r[0].t0) {printf(" Error: t<t0 in getw.\n"); exit(1);}
 /* Get correct level */ 
   for(i=0;i<K;i++) if(r[i].level==level) {tag=i; break;}

   while(t>=r[tag].tp) /* Get correct time interval */
   { /* Move tag down to K-1 */
    /*printf(" Next time interval needed.\n");
    printf(" t=%f tp=%f tag=%d\n",t,r[tag].tp,tag);*/
    r[K-1]=r[tag];
    for(j=0;j<np;j++) for(k=0;k<nw;k++) w[K-1][j][k]=w[tag][j][k];
    tag=K-1; l=level;
    if(r[tag].nexrec>0)
    {fseek(in,r[tag].nexrec*size,0);
     readw(w[tag],np,nw,r+tag,in);
    }
    else
    {r[0]=r[tag];
     initstack(); 
     while(r[0].nexrec<0)
     {push(r[0].recno); /*printf(" pushing %d\n",r[0].recno);*/
      fseek(in,r[0].parent*size,0);
      readw(w[0],np,nw,r,in);
     }
     fseek(in,r[0].nexrec*size,0);
     readw(w[0],np,nw,r,in);
     while(r[0].level<level) /* Create more tree */
     {
      /* Rewrite parent */
      /*printf(" Creating new arrays to broaden tree.\n");*/
      fseek(in,r[0].recno*size,0);
      r[0].lson=head.nrecs1;r[0].rson=head.nrecs1+1;
            fwrite((char *)r,recsize,1,in);
      h=1.0/(1<<r[0].level);
      sd = sqrt(0.5*h);
      brownchunk(w[0],np,nw,w[1],w[2],sd,func);
      r[1]=r[0];
      r[1].recno=head.nrecs1++;head.nrecs2++;
      r[1].parent=r[0].recno;
      r[1].prerec=pop(); /*printf(" popping %d\n",r[1].prerec);*/
      r[1].level++;r[1].nexrec=r[1].recno+1;
      r[1].lson=r[1].rson=(-1);
      r[1].tp=r[0].t0+np*h;
      tellprerec(r[1].prerec,r[1].recno,recsize,size,in);
      fclose(in);
      in=fopen(wfile,"a+");
      savew(w[1],np,nw,r+1,in);
      fclose(in);
      in=fopen(wfile,"r+");
      r[2]=r[1];head.nrecs1++;head.nrecs2++;
      r[2].recno++;r[2].prerec=r[2].recno-1;
      r[2].nexrec=(-1);
      r[2].t0=r[2].tp;r[2].tp+=np*h;
      fclose(in);
      in=fopen(wfile,"a+");
      savew(w[2],np,nw,r+2,in);
      fclose(in);
      in=fopen(wfile,"r+");
      for(k=0;k<np;k++) for(i=0;i<nw;i++)
       w[0][k][i]=w[1][k][i];
      r[0]=r[1];
      /* Rewrite header */
      fwrite((char *)&head,recsize,1,in);
     }
     r[K-1]=r[0];
     for(j=0;j<np;j++) for(k=0;k<nw;k++) w[K-1][j][k]=w[0][j][k];
     freestack();
    }
/*Fill in w and r */
     for(i=K-2;i>=0;i--)
     {fseek(in,r[i+1].parent*size,0);
      readw(w[i],np,nw,r+i,in);
     }
   } 
/* Return correct dw */
 tt=r[tag].t0; i=0;
 while(tt<t) { i++; tt=r[tag].t0+i*h;}
 /*printf(" i=%d\n",i);*/
 if(i>np-1) {printf(" Error in time: tag=%d t=%f h=%f np=%d tt=%f\n"); exit(1);}
 for(j=0;j<nw;j++) dw[j]=w[tag][i][j];
/* Check if possible to go up a level
   next time. It is only possible when the
   current time index = 3 mod 4. */
  *cangoup=((i%4==3&&level>5)?1:0);
  return ;
}



