/*****************************************************************
* The software in this file was written by Jessica Gaines,       *
* while working at the University of Edinburgh, under the        *
* direction of Terry Lyons, and using funding                    *
* provided by EC grant SC1*CT92-0784.                            *
* You may use this software for educational and academic         *
* research purposes only.                                        *
* Please do not give copies of the software in this file to      *
* anyone else. Tell anyone interested to email                   *
*  jessica@maths.ed.ac.uk                                        *
* or to write to                                                 *
*  Jessica Gaines                                                *
*  University of Edinburgh                                       *
*  Dept. of Mathematics and Statistics                           *
*  JCMB                                                          *
*  Kings Buildings                                               *
*  Mayfield Road                                                 *
*  Edinburgh EH9 3JZ                                             *
*  UK                                                            *
* and then get copies directly from Edinburgh themselves.        *
******************************************************************
*/

/*
 * This file contains  basic subroutines used to solve one-dimensional
 * stoch. diff. equations. The routines are all algorithms to approximate a
 * numerical solution to S.D.E.s : euler, eulerback, heun, heunback,
 * milshtein, runkut4, nn_ito, nn_strat, ERKS, heunqc
 */


#include <math.h>
#include <stdio.h>


float 
euler(x, h, dw, t, b, sig)
  float     x, h, dw, t, (*b) (), (*sig) ();
{
  return x + (*b) (x, t) * h + (*sig) (x, t) * dw;
}

float 
eulerback(x, h, dw, t, b, sig)
  float     x, h, dw, t, (*b) (), (*sig) ();
{
  return x - (*b) (x, t) * h - (*sig) (x, t) * dw;
}

float 
heun(x, h, dw, t, a, sig)
  float     x, h, dw, t, (*a) (), (*sig) ();
{
  float     xx;
  xx = x + (*a) (x, t) * h + (*sig) (x, t) * dw;
  return x + 0.5 * ((*a) (x, t) + (*a) (xx, t + h)) * h
    + 0.5 * ((*sig) (x, t) + (*sig) (xx, t + h)) * dw;
}

float 
heunback(x, h, dw, t, a, sig)
  float     x, h, dw, t, (*a) (), (*sig) ();
{
  float     xx;
  xx = x - (*a) (x, t) * h - (*sig) (x, t) * dw;
  return x - 0.5 * ((*a) (x, t) + (*a) (xx, t - h)) * h
    - 0.5 * ((*sig) (x, t) + (*sig) (xx, t - h)) * dw;
}

float 
milshtein(x, h, dw, t, b, sig, dsigdx)
  float     x, h, dw, t, (*b) (), (*sig) (), (*dsigdx) ();
{
  float     g, dgg;
  g = (*sig) (x, t);
  dgg = 0.5 * (*dsigdx) (x, t) * g;
  return x + ((*b) (x, t) - dgg) * h + g * dw + dgg * dw * dw;
}

float 
runkut4(x, h, dw, t, a, sig)
  float     x, h, dw, t, (*a) (), (*sig) ();
{
  float     f0, f1, f2, f3, g0, g1, g2, g3, x1, x2, x3, t1, t2;
  t1 = t + 0.5 * h;
  t2 = t + h;
  f0 = (*a) (x, t);
  g0 = (*sig) (x, t);
  x1 = x + 0.5 * (f0 * h + g0 * dw);
  f1 = (*a) (x1, t1);
  g1 = (*sig) (x1, t1);
  x2 = x + 0.5 * (f1 * h + g1 * dw);
  f2 = (*a) (x2, t1);
  g2 = (*sig) (x2, t1);
  x3 = x + f2 * h + g2 * dw;
  f3 = (*a) (x3, t2);
  g3 = (*sig) (x3, t2);
  return x + ((f0 + 2 * (f1 + f2) + f3) * h + 
              (g0 + 2 * (g1 + g2) + g3) * dw) / 6;
}

float 
nn_ito(x, h, dw, t, b, sig, dbdx, d2bdx, dsigdx, d2sigdx, d3sigdx)
  float     x, h, dw, t, (*d2sigdx) (), (*d3sigdx) ();
  float     (*b) (), (*dbdx) (), (*d2bdx) (), (*sig) (), (*dsigdx) ();
/*
 * Efficient O(h*3) method, converging to Ito solution. See "An
 * Asymptotically Efficient Difference Formula for Solving S.D.E.s" by Nigel
 * Newton in Stochastics 1986.
 */
{
  float     c1, c2, c3, c4, c5, c6, db, dsdx, sig2, d2s, dw2;
  static float sixth = 0.166667;
  dw2 = dw * dw;
  c1 = (*b) (x, t);
  c2 = (*sig) (x, t);
  sig2 = c2 * c2;
  dsdx = (*dsigdx) (x, t);
  c3 = dsdx * c2;
  db = (*dbdx) (x, t);
  d2s = sig2 * (*d2sigdx) (x, t);
  c4 = db * c2 + dsdx * c1 + 0.5 * d2s;
  c5 = c1 * db + 0.5 * (*d2bdx) (x, t) * sig2;
  c6 = sixth * (d2s + c3 * dsdx);
  return x + c1 * h + c2 * dw
    + 0.5 * (c3 * (dw2 - h) + c4 * h * dw + c5 * h * h)
    + c6 * dw * (dw2 - 3 * h);
}

float 
nn_strat(x, h, dw, t, a, sig, dadx, d2adx, dsigdx, d2sigdx, d3sigdx)
  float     x, h, dw, t, (*d2sigdx) (), (*d3sigdx) ();
  float     (*a) (), (*dadx) (), (*d2adx) (), (*sig) (), (*dsigdx) ();
/*
 * Efficient O(h*3) method, converging to Stratonovich solution. Derived
 * similarly to nn_ito.
 */
{
  float     c1, c2, c3, c4, c5, c6, c7, c8;
  float  da, dsdx, sig2, d2s, dw2;
  static float sixth = 0.166667;
  dw2 = dw * dw;
  c1 = (*a) (x, t);
  c2 = (*sig) (x, t);
  sig2 = c2 * c2;
  dsdx = (*dsigdx) (x, t);
  c3 = dsdx * c2;
  da = (*dadx) (x, t);
  d2s = c2 * (*d2sigdx) (x, t);
  c4 = da * c2 + dsdx * c1;
  c5 = c1 * da + 0.5 * (*d2adx) (x, t) * sig2;
  c6 = sixth * (c2 * d2s + c3 * dsdx);
  c7 = c1 * d2s + c4 * dsdx;
  c8 = sixth * c1 * (sig2 * (*d3sigdx) (x, t) 
             + dsdx * (4 * d2s + dsdx * dsdx));
  return x + c1 * h + c2 * dw
    + 0.5 * (c3 * dw2 + c4 * h * dw + c5 * h * h)
    + (c6 * dw + 0.25 * (c7 * h + c8 * dw2)) * dw2;
}

float 
ERKS(x, h, dw, t, a, sig)
  float     x, h, dw, t, (*a) (), (*sig) ();
/*
 * Efficient O(h*3) Runge-Kutta method, 
 * converging to Stratonovitch solution. See 
 * "Asymptotically Efficient Runge-Kutta Methods for a Class of Ito
 * and Stratonovitch Equations" by Nigel
 * Newton in SIAM J. Appl. Math.
 *
 * Takes one step in the solution of the Stratonovich sde
 *  dx=a(x,t)dt+sig(x,t)odw
 * using the discretisation scheme
 * x(k+1)=x(k)+(1/2)*(a0+a1)*h+(1/6)*(sig0+2*sig1+2*sig2+sig3)*dw
 * with 
 *    a0=a(x(k),t)
 *    sig0=sig(x(k),t)
 *    sig1=sig(x(k)+0.5*sig0*dw,t+0.5*h)
 *    sig2=sig(x(k)+0.25*a0*(3h+dw*dw)+0.5*sig1*dw,t+0.5*h)
 *    sig3=sig(x(k)+0.5*a0*(3h-dw*dw)+sig2*dw,t+h)
 *    a1=a(x(k)+0.5*a0*(3h-dw*dw)+sig2*dw,t+h)
 */
{
  float     f0, f1, g0, g1, g2, g3, x1, x2, x3, t1, t2;
  t1 = t + 0.5 * h;
  t2 = t + h;
  f0 = (*a) (x, t);
  g0 = (*sig) (x, t);
  x1 = x + 0.5 * g0 * dw;
  g1 = (*sig) (x1, t1);
  x2 = x + 0.25 * f0 * (3*h + dw * dw) + 0.5*g1*dw;
  g2 = (*sig) (x2, t1);
  x3 = x + 0.5 * f0 * (3*h - dw * dw) + g2*dw;
  f1 = (*a) (x3, t2);
  g3 = (*sig) (x3, t2);
  return x + 0.5 * (f0 + f1) * h + 
              (g0 + 2.0 * (g1 + g2) + g3) * dw / 6.0;
}

#define FCOR 0.06666666		/* 1.0/15.0 */
#define ERRCON 6.0e-4

heunqc(x, htry, hdid, hnext, eps, w, t, wcurr, inc, a, sig)
  float    *x, w[], *t, *hdid, *hnext, eps, htry;
  float     (*a) (), (*sig) ();
  int      *inc, *wcurr;
{
  float     xtemp, xsav, t1, h, hh, dw, errmax, temp;
  float     heun();
  int       iw, sinc, binc;
  xsav = (*x);
  h = htry;
  binc = (*inc);
  t1 = (*t);
  iw = (*wcurr);
  for (;;)
  {
    /* Two small steps */
    hh = 0.5 * h;
    sinc = binc / 2;
    if (sinc < 1)
    {
      printf(" Too few steps in Brown2\n");
      exit(1);
    }
    dw = w[iw + sinc] - w[iw];
    xtemp = heun(xsav, hh, dw, t1, a, sig);
    *t = t1 + hh;
    dw = w[iw + binc] - w[iw + sinc];
    *x = heun(xtemp, hh, dw, *t, a, sig);
    *t = t1 + h;
    /* One large step */
    if (*t == t1)
    {
      printf("Step size too small in heunqcN\n");
      exit(1);
    }
    dw = w[iw + binc] - w[iw];
    xtemp = heun(xsav, h, dw, t1, a, sig);
    errmax = 0.0;
    xtemp = (*x) - xtemp;
    temp = fabs(xtemp);
    if (errmax < temp)
      errmax = temp;
    if (errmax <= eps)
    {
      *hdid = h;
      *wcurr = iw + binc;
      if (errmax > ERRCON)
      {
	*hnext = 2.0 * h;
	*inc = 2.0 * binc;
      } else
      {
	*hnext = 4.0 * h;
	*inc = 4.0 * binc;
      }
      break;
    }
    h = 0.5 * h;		/* reduce stepsize */
    binc = binc / 2;
  }
  (*x) += xtemp * FCOR;
}
