/*****************************************************************
* The software in this file was written by Jessica Gaines,       *
* while working at the University of Edinburgh, under the        *
* direction of Terry Lyons, and using funding                    *
* provided by EC grant SC1*CT92-0784.                            *
* You may use this software for educational and academic         *
* research purposes only.                                        *
* Please do not give copies of the software in this file to      *
* anyone else. Tell anyone interested to email                   *
*  jessica@maths.ed.ac.uk                                        *
* or to write to                                                 *
*  Jessica Gaines                                                *
*  University of Edinburgh                                       *
*  Dept. of Mathematics and Statistics                           *
*  JCMB                                                          *
*  Kings Buildings                                               *
*  Mayfield Road                                                 *
*  Edinburgh EH9 3JZ                                             *
*  UK                                                            *
* and then get copies directly from Edinburgh themselves.        *
******************************************************************
*/

 /* This file contains  basic subroutines used to solve
N-dimensional stoch. diff. equations.
The first set of routines contains algorithms to approximate a
numerical solution for a single step of a given size:
eulerN, eulerNback, heunN, heunNback, heun_iterN, runkut4N, 
 milshteinN, milshteinN_back, ERKSN, FRKSN
The remaining routines call a routine from the first set
while performing one quality controlled step of variable size:
diffqcN, diffbackqcN, diffqcNnew
*/

/* NB : The following 3 routines all have the same no. and type
of arguments, so that any one of them can be called from within diffqcN,
interchangably, depending only on the value of the argument func1 :
        eulerN(x,h,dw,t,nx,nw,b,sig)
         heunN(x,h,dw,t,nx,nw,a,sig)
runkut4N(x,h,dw,t,nx,nw,a,sig) 
*/

#include <math.h>
#include <stdio.h>


eulerN(x,h,dw,t,nx,nw,b,sig)
float x[],h,dw[],t;
int (*b)(),(*sig)();
int nx,nw;
/* Takes one step in the solution of the Ito sde
 dx=b(x,t)dt+sig(x,t)dw
using the discretisation scheme
 x(k+1) = x(k) + b(x(k),t)*h + sig(x(k),t)*dw
nx=dim. of x, nw=dim. of Brownian path, t=t(k) (unchanged on exit),
h=step size,
dw=w(k+1)-w(k)                       (vector of dimension nw),
x= x(k) on entry and x(k+1) on exit  (vector of dimension nx)
*/
{float *B,*S,*vector(); int i,error=0;
 B=vector(0,nx-1);
 S=vector(0,nx-1);
 error=(*b)(x,t,B,nx);
 if(error) return (error);
 error=(*sig)(x,t,dw,S,nx,nw);
 if(error) return (error);
 for(i=0;i<nx;i++) x[i]=x[i]+h*B[i]+S[i];
 free_vector(B,0,nx-1);
 free_vector(S,0,nx-1);
 return (error);
}

eulerNback(x,h,dw,t,nx,nw,b,sig)
float x[],h,dw[],t;
int (*b)(),(*sig)();
int nx,nw;
/* As eulerN, but takes a step backward in time, using
x(k) = x(k+1) - b(x(k+1),t)*h - sig(x(k+1),t)*dw */
{float *B,*S,*vector(); int i,error=0;
 B=vector(0,nx-1);
 S=vector(0,nx-1);
 error=(*b)(x,t,B,nx);
 if(error) return (error);
 error=(*sig)(x,t,dw,S,nx,nw);
 if(error) return (error);
 for(i=0;i<nx;i++) x[i]=x[i]-h*B[i]-S[i];
 free_vector(B,0,nx-1);
 free_vector(S,0,nx-1);
 return (error);
}

heunN(x,h,dw,t,nx,nw,a,sig)
float x[],h,dw[],t;
int (*a)(),(*sig)();
int nx,nw;
/* Takes one step in the solution of the Stratonovich sde
 dx=a(x,t)dt+sig(x,t)odw
using the discretisation scheme
 x(K+1) = x(k) + 0.5*h*(a(x(k),t)+a(xx,t))
                 + 0.5*(sig(x(k),t)+sig(xx,t))*dw 
with
  xx = x(k) + a(x(k),t)*h + sig(x(k),t)*dw .
nx=dim. of x, nw=dim. of Brownian path, t=t(k) (unchanged on exit),
h=step size,
dw=w(k+1)-w(k)                       (vector of dimension nw),
x= x(k) on entry and x(k+1) on exit  (vector of dimension nx)
*/
{float *xx,*A,*S,*AA,*SS,*vector(); int i,error=0;
 A=vector(0,nx-1);
 S=vector(0,nx-1);
 AA=vector(0,nx-1);
 SS=vector(0,nx-1);
 xx=vector(0,nx-1);
 error=(*a)(x,t,A,nx);
 if(error) return (error);
 error=(*sig)(x,t,dw,S,nx,nw);
 if(error) return (error);
 for(i=0;i<nx;i++) xx[i]=x[i]+h*A[i]+S[i];
 error=(*a)(xx,t+h,AA,nx);
 if(error) return (error);
 error=(*sig)(xx,t+h,dw,SS,nx,nw);
 if(error) return (error);
 for(i=0;i<nx;i++)
 x[i]=x[i] + 0.5 * (A[i]+AA[i])*h 
          + 0.5 * (S[i]+SS[i]);
 free_vector(A,0,nx-1);
 free_vector(S,0,nx-1);
 free_vector(AA,0,nx-1);
 free_vector(SS,0,nx-1);
 free_vector(xx,0,nx-1);
 return (error);
}

heunNback(x,h,dw,t,nx,nw,a,sig)
float x[],h,dw[],t;
int (*a)(),(*sig)();
int nx,nw;
/* As heunN, but takes a step backward in time, using
x(k+1) = x(k) - 0.5*h*(a(x(k),t)+a(xx,t))
                 - 0.5*(sig(x(k),t)+sig(xx,t))*dw 
with
  xx = x(k) - b(x(k),t)*h - sig(x(k),t)*dw
*/
{float *xx,*A,*S,*AA,*SS,*vector(); int i,error=0;
 A=vector(0,nx-1);
 S=vector(0,nx-1);
 AA=vector(0,nx-1);
 SS=vector(0,nx-1);
 xx=vector(0,nx-1);
 error=(*a)(x,t,A,nx);
 if(error) return error;
 (*sig)(x,t,dw,S,nx,nw);
 for(i=0;i<nx;i++) xx[i]=x[i]-h*A[i]-S[i];
 error=(*a)(xx,t-h,AA,nx);
 if(error) return error;
 (*sig)(xx,t-h,dw,SS,nx,nw);
 for(i=0;i<nx;i++)
 x[i]=x[i] - 0.5 * (A[i]+AA[i])*h 
          - 0.5 * (S[i]+SS[i]);
 free_vector(A,0,nx-1);
 free_vector(S,0,nx-1);
 free_vector(AA,0,nx-1);
 free_vector(SS,0,nx-1);
 free_vector(xx,0,nx-1);
 return (error);
}

heun_iterN(x,h,dw,t,nx,nw,a,sig,eps)
float x[],h,dw[],t,eps;
int (*a)(),(*sig)();
int nx,nw;
/* Takes one step using the IMPLICIT heun algorithm,
 using iteration until convergence is reached.
 The real number eps (supplied by the calling program)
 is used as the accuracy criterion for convergence.
*/
{float *xsav,*xx,*A,*S,*AA,*SS,*vector(); 
 int i,error=0,notend;
 A=vector(0,nx-1);
 S=vector(0,nx-1);
 AA=vector(0,nx-1);
 SS=vector(0,nx-1);
 xx=vector(0,nx-1);
 xsav=vector(0,nx-1);
 for(i=0;i<nx;i++) xsav[i]=x[i];  /* x(k) stored in xsav */
 error=(*a)(xsav,t,A,nx);         /* a(x(k),t) in A      */
 if(error) return (error);
 error=(*sig)(xsav,t,dw,S,nx,nw); /* sig(x(k),t) in S    */
 if(error) return (error);
 for(i=0;i<nx;i++) xx[i]=xsav[i]+h*A[i]+S[i]; /* initial x(k+1) */
 do
 {error=(*a)(xx,t+h,AA,nx);         /* a(x(k+1),t+h) in AA   */
  if(error) return (error);
  error=(*sig)(xx,t+h,dw,SS,nx,nw); /* sig(x(k+1),t+h) in SS */
  if(error) return (error);
  for(i=0;i<nx;i++)
  x[i]=xsav[i] + 0.5 * (A[i]+AA[i])*h 
            + 0.5 * (S[i]+SS[i]);      /* updated x(k+1) */
  notend=0;
  for(i=0;i<nx;i++) if(fabs(x[i]-xx[i])>eps) notend=1;
  for(i=0;i<nx;i++) xx[i]=x[i];
 } while(notend);
 free_vector(A,0,nx-1);
 free_vector(S,0,nx-1);
 free_vector(AA,0,nx-1);
 free_vector(SS,0,nx-1);
 free_vector(xx,0,nx-1);
 free_vector(xsav,0,nx-1);
 return (error);
}

runkut4N(x,h,dw,t,nx,nw,a,sig) 
float x[],h,dw[],t;
int (*a)(),(*sig)();
int nx,nw;
/* Takes one step in the solution of the Stratonovich sde
 dx=a(x,t)dt+sig(x,t)odw
using the discretisation scheme
 x(k+1)=x(k)+(1/6)*((a0+2*a1+2*a2+a3)*h+(sig0+2*sig1+2*sig2+sig3)*dw)
with a0=a(x(k),t)
     a1=a(x(k)+0.5*a0*h+0.5*sig0*dw,t+0.5*h)
     a2=a(x(k)+0.5*a1*h+0.5*sig1*dw,t+0.5*h)
     a3=a(x(k)+a2*h+sig2*dw,t+h)
and similarly for sig0,..,sig3
Arguments are the same as for heunN.
*/
{float *f,*g,*x1,*dx,t1,t2,*vector(); int i,error=0;
 f=vector(0,nx-1);
 g=vector(0,nx-1);
 x1=vector(0,nx-1);
 dx=vector(0,nx-1);
 t1=t+0.5*h;
 t2=t+h; 
 error=(*a)(x,t,f,nx); 
 if(error) return (error);
 error=(*sig)(x,t,dw,g,nx,nw);
 if(error) return (error);
 for(i=0;i<nx;i++) 
 {dx[i]= f[i]*h + g[i];
  x1[i]=x[i]+0.5*dx[i];}
 error=(*a)(x1,t1,f,nx); 
 if(error) return (error);
 error=(*sig)(x1,t1,dw,g,nx,nw);
 if(error) return (error);
 for(i=0;i<nx;i++) 
 {x1[i]=x[i]+0.5*(f[i]*h+g[i]);
  dx[i]+=2*(f[i]*h+g[i]);}
 error=(*a)(x1,t1,f,nx); 
 if(error) return (error);
 error=(*sig)(x1,t1,dw,g,nx,nw);
 if(error) return (error);
 for(i=0;i<nx;i++) 
 {x1[i]=x[i]+f[i]*h+g[i];
  dx[i]+=2*(f[i]*h+g[i]);}
 error=(*a)(x1,t2,f,nx); 
 if(error) return (error);
 error=(*sig)(x1,t2,dw,g,nx,nw);
 if(error) return (error);
 for(i=0;i<nx;i++) 
 {dx[i]+=f[i]*h + g[i];
  x[i]+= dx[i]/6;}
 free_vector(f,0,nx-1);
 free_vector(g,0,nx-1);
 free_vector(x1,0,nx-1);
 free_vector(dx,0,nx-1);
 return (error);
}  

milshteinN(x,h,dw,t,nx,nw,b,sig,dssig) 
float x[],h,dw[],t;
int (* b)(),(* sig)(),(* dssig)();
int nx,nw;
{ float **dw2,**matrix(),*B,*S,*SS,*vector();
 int i,j,error=0;
/* the scheme used below is the milshtein scheme ;
 for its description, see e.pardoux-d.talay : 
 "discretization and simulation of s.d.e.",
 acta applicandae mathematicae 3,23,1985.
 This scheme is of order 1 only when the
 commutativity condition holds. */
 B=vector(0,nx-1);
 S=vector(0,nx-1);
 SS=vector(0,nx-1);
 dw2=matrix(0,nw-1,0,nw-1);
 for(i=0;i<nw;i++)
  dw2[i][i]=0.5*(dw[i]*dw[i]-h);
 for(i=1;i<nw;i++) for(j=0;j<i;j++)
  dw2[i][j]=dw2[j][i]=0.5*dw[i]*dw[j];
 error=(*b)(x,t,B,nx);
 if(error) return (error);
 error=(*sig)(x,t,dw,S,nx,nw);
 if(error) return (error);
 error=(*dssig)(x,t,dw2,SS,nx,nw);
 if(error) return (error);
 for(i=0;i<nx;i++) x[i]=x[i]+h*B[i]+S[i]+SS[i];
 free_vector(B,0,nx-1);
 free_vector(S,0,nx-1);
 free_vector(SS,0,nx-1);
 free_matrix(dw2,0,nw-1,0,nw-1);
 return (error);
}

milshteinN_back(x,h,dw,t,nx,nw,b,sig,dssig) 
float x[],h,dw[],t;
int (* b)(),(* sig)(),(* dssig)();
int nx,nw;
{ float **dw2,**matrix(),*B,*S,*SS,*vector();
 int i,j,error=0;
/* As milshteinN, but backwards in time.
 For a description, see e.pardoux-d.talay : 
 "discretization and simulation of s.d.e.",
 acta applicandae mathematicae 3,23,1985.
 This scheme is of order 1 only when the
 commutativity condition holds. */
 B=vector(0,nx-1);
 S=vector(0,nx-1);
 SS=vector(0,nx-1);
 dw2=matrix(0,nw-1,0,nw-1);
 for(i=0;i<nw;i++)
  dw2[i][i]=0.5*(dw[i]*dw[i]-h);
 for(i=1;i<nw;i++) for(j=0;j<i;j++)
  dw2[i][j]=dw2[j][i]=0.5*dw[i]*dw[j];
 error=(*b)(x,t,B,nx);
 if(error) return (error);
 error=(*sig)(x,t,dw,S,nx,nw);
 if(error) return (error);
 error=(*dssig)(x,t,dw2,SS,nx,nw);
 if(error) return (error);
 for(i=0;i<nx;i++) x[i]=x[i]-h*B[i]-S[i]-SS[i];
 free_vector(B,0,nx-1);
 free_vector(S,0,nx-1);
 free_vector(SS,0,nx-1);
 free_matrix(dw2,0,nw-1,0,nw-1);
 return (error);
}

ERKSN(x, h, dw, t, nx, nw, a, sig)
  float     x[], h, dw[], t;
  int       nx, nw;
  int       (*a) (), (*sig) ();
/*
 * Efficient O(h*3) Runge-Kutta method, 
 * converging to Stratonovitch solution. See 
 * "Asymptotically Efficient Runge-Kutta Methods for a Class of Ito
 * and Stratonovitch Equations" by Nigel
 * Newton in SIAM J. Appl. Math.
 *
 * Takes one step in the solution of the Stratonovich sde
 *  dx=a(x,t)dt+sig(x,t)odw
 * using the discretisation scheme
 * x(k+1)=x(k)+(1/2)*(a0+a1)*h+(1/6)*(sig0+2*sig1+2*sig2+sig3)*dw
 * with 
 *    a0=a(x(k),t)
 *    sig0=sig(x(k),t)
 *    sig1=sig(x(k)+0.5*sig0*dw,t+0.5*h)
 *    sig2=sig(x(k)+0.25*a0*(3h+dw*dw)+0.5*sig1*dw,t+0.5*h)
 *    sig3=sig(x(k)+0.5*a0*(3h-dw*dw)+sig2*dw,t+h)
 *    a1=a(x(k)+0.5*a0*(3h-dw*dw)+sig2*dw,t+h)
 *
 * Must have dim(w)=1, but can have any dim(x).
 */
{ int i,error=0;
  float     *f0, *f1, *g0, *g1, *g2, *g3, *x1, t1, t2;
  float     *vector();
  t1 = t + 0.5 * h;
  t2 = t + h;
  f0=vector(0,nx-1);
  f1=vector(0,nx-1);
  g0=vector(0,nx-1);
  g1=vector(0,nx-1);
  g2=vector(0,nx-1);
  g3=vector(0,nx-1);
  x1=vector(0,nx-1);
  error=(*a)(x,t,f0,nx); 
  if(error) return (error);
  error=(*sig)(x,t,dw,g0,nx,nw);
  if(error) return (error);
  for(i=0;i<nx;i++)
   x1[i] = x[i] + 0.5 * g0[i];
  error = (*sig) (x1,t1,dw,g1,nx,nw);
  if(error) return (error);
  for(i=0;i<nx;i++)
   x1[i] = x[i] + 0.25 * f0[i] * (3*h + dw[0] * dw[0]) + 0.5*g1[i];
  error = (*sig) (x1,t1,dw,g2,nx,nw);
  if(error) return (error);
  for(i=0;i<nx;i++)
   x1[i] = x[i] + 0.5 * f0[i] * (3*h - dw[0] * dw[0]) + g2[i];
  error = (*a) (x1,t2,f1,nx);
  if(error) return (error);
  error = (*sig) (x1,t2,dw,g3,nx,nw);
  if(error) return (error);
  for(i=0;i<nx;i++)
   x[i] += 0.5 * (f0[i] + f1[i]) * h + 
              (g0[i] + 2.0 * (g1[i] + g2[i]) + g3[i])/ 6.0;
  free_vector(f0,0,nx-1);
  free_vector(f1,0,nx-1);
  free_vector(g0,0,nx-1);
  free_vector(g1,0,nx-1);
  free_vector(g2,0,nx-1);
  free_vector(g3,0,nx-1);
  free_vector(x1,0,nx-1);
  return (error);
}

FRKSN(x,h,dw,t,nx,nw,a,sig)
float x[],h,dw[],t;
int (*a)(),(*sig)();
int nx,nw;
/* Takes one step in the solution of the Stratonovich sde
 dx=a(x,t)dt+sig(x,t)odw
using the discretisation scheme
 x(K+1) = x(k) + h*a(x(k),t) + sig(xx,t)*dw 
with
  xx = x(k) + 1/2*sig(x(k),t)*dw .
nx=dim. of x, nw=dim. of Brownian path, t=t(k) (unchanged on exit),
h=step size,
dw=w(k+1)-w(k)                       (vector of dimension nw),
x= x(k) on entry and x(k+1) on exit  (vector of dimension nx)
*/
{float *xx,*A,*S,*vector(); int i,error=0;
 A=vector(0,nx-1);
 S=vector(0,nx-1);
 xx=vector(0,nx-1);
 error=(*a)(x,t,A,nx);
 if(error) return (error);
 error=(*sig)(x,t,dw,S,nx,nw);
 if(error) return (error);
 for(i=0;i<nx;i++) xx[i]=x[i]+0.5*S[i];
 error=(*sig)(xx,t+h,dw,S,nx,nw);
 if(error) return (error);
 for(i=0;i<nx;i++)
 x[i]=x[i] + A[i]*h + S[i];
 free_vector(A,0,nx-1);
 free_vector(S,0,nx-1);
 free_vector(xx,0,nx-1);
 return (error);
}

diffqcN(x,htry,hdid,hnext,eps,xscal,t,nx,nw,level,
          a,sig,func1,func2) 
float x[],*t,*hdid,*hnext,eps,xscal[],htry,(*func2)();
int (*func1)(),(*a)(),(*sig)();
int nx,nw,*level;
/* Performs one 'quality controlled' step in the numerical soln.
of an sde. Starting with x=x(k), takes one step of size h and two 
half steps and compares the values x1 and x2 (of x) computed. 
If max(i) abs(x1(i)-x2(i))/xscal(i)<eps the value x2 is returned,
otherwise h is halved and the procedure repeated.
Inputs : x     = x(k) on entry, contains x(k+1) on exit
         htry  = initial value of h, remains unchanged
         eps   = max. error to be tolerated, remains unchanged
         xscal(i),i=1,nw = weight on error in x(i), unchanged
         t = t(k) = current time on entry , becomes x(k+1) on exit
         nx = dim. of x, unchanged on exit
         nw = dim. of w (Brownian path), unchanged
         level : htry=2^(-level), updated on exit to hnext=2^(-level)
         a() and sig() = drift and diffusion terms of sde
         func1() = procedure to be used to take steps ie eulerN,heunN etc.
         func2() = procedure to be used as nec. for generation of new 
                   gaussian deviates inside getdw
Outputs: x,t,level (as above)
         hdid = size of successful step taken
         hnext= step size to be tried on next call to diffqcN
*/
{float *xtemp,*xsav,*vector(),t1,h,hh,*dw,*dw2,errmax,temp;
 int i,l,cangoup,error=0;
 xtemp=vector(0,nx-1); 
 xsav=vector(0,nx-1); 
 dw=vector(0,nw-1); 
 dw2=vector(0,nw-1); 
 for(i=0;i<nx;i++) xsav[i]=x[i]; 
 h=htry;
 t1=(*t); 
 for(;;)
 {
/* Two small steps */
    hh=0.5*h; l=(*level)+1;
     for(i=0;i<nx;i++) x[i]=xsav[i]; 
     getdw(nw,dw,l,t1,hh,func2,1,&cangoup);
     error=(*func1)(x,hh,dw,t1,nx,nw,a,sig);
     if(error) return (error);
    *t=t1+hh;
     getdw(nw,dw2,l,*t,hh,func2,2,&cangoup); 
     error=(*func1)(x,hh,dw2,*t,nx,nw,a,sig);
     if(error) return (error);
    *t=t1+h;  
/* One large step */
    if(*t==t1) {printf("Step size too small in N\n"); exit(1);}
    for(i=0;i<nx;i++) xtemp[i]=xsav[i];
    for(i=0;i<nw;i++) dw[i]+=dw2[i]; 
    error=(*func1)(xtemp,h,dw,t1,nx,nw,a,sig);
    if(error) return (error);
    errmax=0.0; 
    for(i=0;i<nx;i++)
    { xtemp[i]=x[i]-xtemp[i];
      temp=fabs(xtemp[i]/xscal[i]);
      if(errmax<temp) errmax=temp;
    }
    if(errmax<=eps || l>=25) 
    { 
      *hdid=h;
      if(cangoup&&l<25) {*hnext=2.0*h; (*level)--;}
      else         *hnext=h;
      /* if(l>=30) printf(" level=%d\n",l); */
      break;
    }
    h=0.5*h; (*level)++; /* reduce stepsize */
  }
 free_vector(xtemp,0,nx-1);
 free_vector(xsav,0,nx-1);
 free_vector(dw,0,nw-1);
 free_vector(dw2,0,nw-1);
 return (error);
} 

diffbackqcN(x,htry,hdid,hnext,eps,xscal,t,nx,nw,level,
          a,sig,func1,func2) 
float x[],*t,*hdid,*hnext,eps,xscal[],htry,(*func2)();
int (*func1)(),(*a)(),(*sig)();
int nx,nw,*level;
/* This routine is identical to diffqcN except that it takes a
step backwards in time.
func1 = eulerNback, heunNback etc.
*/
{float *xtemp,*xsav,*vector(),t1,h,hh,*dw,*dw2,errmax,temp;
 int i,l,cangoup,error=0;
 xtemp=vector(0,nx-1); 
 xsav=vector(0,nx-1); 
 dw=vector(0,nw-1); 
 dw2=vector(0,nw-1); 
 for(i=0;i<nx;i++) xsav[i]=x[i]; 
 h=htry;
 t1=(*t); 
 for(;;)
 {
/* Two small steps */
    hh=0.5*h; l=(*level)+1;
     for(i=0;i<nx;i++) x[i]=xsav[i];
     getdwback(nw,dw,l,t1,hh,func2,1,&cangoup);
     error=(*func1)(x,hh,dw,t1,nx,nw,a,sig);
     if(error) return (error);
    *t=t1-hh; 
     getdwback(nw,dw2,l,*t,hh,func2,2,&cangoup); 
     error=(*func1)(x,hh,dw2,*t,nx,nw,a,sig);
     if(error) return (error);
    *t=t1-h;  
/* One large step */
    if(*t==t1) {printf("Step size too small in diffbackqcN\n"); exit(1);}
    for(i=0;i<nx;i++) xtemp[i]=xsav[i];
    for(i=0;i<nw;i++) dw[i]+=dw2[i]; 
    error=(*func1)(xtemp,h,dw,t1,nx,nw,a,sig);
    if(error) return (error);
    errmax=0.0; 
    for(i=0;i<nx;i++)
    { xtemp[i]=x[i]-xtemp[i];
      temp=fabs(xtemp[i]/xscal[i]);
      if(errmax<temp) errmax=temp;
    }
    if(errmax<=eps || l>=20) 
    { 
      *hdid=h;
      if(cangoup) {*hnext=2.0*h; (*level)--;}
      else         *hnext=h;
      break;
    }
    h=0.5*h; (*level)++; /* reduce stepsize */
  }
 free_vector(xtemp,0,nx-1);
 free_vector(xsav,0,nx-1);
 free_vector(dw,0,nw-1);
 free_vector(dw2,0,nw-1);
 return(error);
} 

diffqcNnew(x,htry,hdid,hnext,t,nx,nw,level,
          a,sig,func1,func2,acc) 
float x[],*t,*hdid,*hnext,htry,(*func2)();
int (*func1)(),(*a)(),(*sig)(),(*acc)();
int nx,nw,*level;
/* Performs one 'quality controlled' step in the numerical soln.
of an sde. Starting with x=x(k), takes one step of size h and
checks to see whether the desired accuracy has been obtained.
If the function acc returns a value of 1 the routine returns
to the calling program,
otherwise h is halved and the procedure repeated.
Inputs : x     = x(k) on entry, contains x(k+1) on exit
         htry  = initial value of h, remains unchanged
         eps   = max. error to be tolerated, remains unchanged
         t = t(k) = current time on entry , becomes x(k+1) on exit
         nx = dim. of x, unchanged on exit
         nw = dim. of w (Brownian path), unchanged
         level : htry=2^(-level), updated on exit to hnext=2^(-level)
         a(x,t) and sig() = drift and diffusion terms of sde
         func1() = procedure to be used to take steps ie euler,heun etc.
         func2() = procedure to be used as nec. for generation of new 
                   gaussian deviates inside getdw
         acc(x,dw,*t,h,nx,nw) = user-supplied procedure to be used to 
                  check accuracy
                  at the end of each step. acc() must return 1 when
                  sufficient accuracy has been obtained and 0 otherwise.

Outputs: x,t,level (as above)
         hdid = size of successful step taken
         hnext= step size to be tried on next call to diffqcNnew
*/
{float *xsav,*vector(),t1,h,*dw;
 int i,l,cangoup,error=0;
 xsav=vector(0,nx-1); 
 dw=vector(0,nw-1); 
 for(i=0;i<nx;i++) xsav[i]=x[i]; 
 h=htry;
 t1=(*t); 
 for(;;) /* loop until accuracy obtained or min. stepsize reached */
 {
/* One step */
     for(i=0;i<nx;i++) x[i]=xsav[i]; 
     l=(*level);
     /* obtain appropriate dw vector */
     getdw(nw,dw,l,t1,h,func2,1,&cangoup); 
     error=(*func1)(x,h,dw,t1,nx,nw,a,sig);
     if(error) return (error);
    *t=t1+h;
    if((*acc)(x,dw,*t,h,nx,nw) || l>=20) /* Check accuracy */
    { 
      *hdid=h;
      if(cangoup) {*hnext=2.0*h; (*level)--;}
      else         *hnext=h;
      break; /* Stop looping */
    }
    h=0.5*h; (*level)++; /* reduce stepsize */
  }
 free_vector(xsav,0,nx-1);
 free_vector(dw,0,nw-1);
 return (error);
} 

diffqcNnew_back(x,htry,hdid,hnext,t,nx,nw,level,
          a,sig,func1,func2,acc) 
float x[],*t,*hdid,*hnext,htry,(*func2)();
int (*func1)(),(*a)(),(*sig)(),(*acc)();
int nx,nw,*level;
/* Performs one 'quality controlled' step backwards in the numerical soln.
of an sde. Starting with x=x(k), takes one step backwards of size h and
checks to see whether the desired accuracy has been obtained.
If the function acc returns a value of 1 the routine returns
to the calling program,
otherwise h is halved and the procedure repeated.
Inputs : x     = x(k) on entry, contains x(k+1) on exit
         htry  = initial value of h, remains unchanged
         eps   = max. error to be tolerated, remains unchanged
         t = t(k) = current time on entry , becomes x(k+1) on exit
         nx = dim. of x, unchanged on exit
         nw = dim. of w (Brownian path), unchanged
         level : htry=2^(-level), updated on exit to hnext=2^(-level)
         a(x,t) and sig() = drift and diffusion terms of sde
         func1() = procedure to be used to take steps ie eulerNback,heunNback
         func2() = procedure to be used as nec. for generation of new 
                   gaussian deviates inside getdw
         acc(x,dw,*t,h,nx,nw) = user-supplied procedure to be used to 
                  check accuracy
                  at the end of each step. acc() must return 1 when
                  sufficient accuracy has been obtained and 0 otherwise.

Outputs: x,t,level (as above)
         hdid = size of successful step taken
         hnext= step size to be tried on next call to diffqcNnew
*/
{float *xsav,*vector(),t1,h,*dw;
 int i,l,cangoup,error=0;
 xsav=vector(0,nx-1); 
 dw=vector(0,nw-1); 
 for(i=0;i<nx;i++) xsav[i]=x[i]; 
 h=htry;
 t1=(*t); 
 for(;;) /* loop until accuracy obtained or min. stepsize reached */
 {
/* One step */
     for(i=0;i<nx;i++) x[i]=xsav[i]; 
     l=(*level);
     /* obtain appropriate dw vector */
     getdwback(nw,dw,l,t1,h,func2,1,&cangoup); 
     error=(*func1)(x,h,dw,t1,nx,nw,a,sig);
     if(error) return (error);
    *t=t1-h;
    if((*acc)(x,dw,*t,h,nx,nw) || l>=20) /* Check accuracy */
    { 
      *hdid=h;
      if(cangoup) {*hnext=2.0*h; (*level)--;}
      else         *hnext=h;
      break; /* Stop looping */
    }
    h=0.5*h; (*level)++; /* reduce stepsize */
  }
 free_vector(xsav,0,nx-1);
 free_vector(dw,0,nw-1);
 return (error);
} 

