/*****************************************************************
* The software in this file was written by Jessica Gaines,       *
* while working at the University of Edinburgh, under the        *
* direction of Terry Lyons, and using funding                    *
* provided by EC grant SC1*CT92-0784.                            *
* You may use this software for educational and academic         *
* research purposes only.                                        *
* Please do not give copies of the software in this file to      *
* anyone else. Tell anyone interested to email                   *
*  jessica@maths.ed.ac.uk                                        *
* or to write to                                                 *
*  Jessica Gaines                                                *
*  University of Edinburgh                                       *
*  Dept. of Mathematics and Statistics                           *
*  JCMB                                                          *
*  Kings Buildings                                               *
*  Mayfield Road                                                 *
*  Edinburgh EH9 3JZ                                             *
*  UK                                                            *
* and then get copies directly from Edinburgh themselves.        *
******************************************************************
*/

/*
 * This file contains functions used when storing various kinds of data in
 * files and when reading the data in again. So far 2 main types of data are
 * catered for: 1) Paths consisting of sets of points (x,y), where x and y
 * are floating-point numbers. 2) Pictures consisting of a matrix whose
 * elements are integer colour indices. Contents of file: fopenw, fopenr,
 * storepaths, readpaths, storearray, readarray, countpics, read_init_vals,
 * float_to_char, int_to_char, reverse, readp, fcopy, savew, readw, savew2,
 * readw2, savew3, readw3, readpaths3d, storepaths3d, storepic_ascii,
 * readpic_ascii, countpics_ascii, readpaths_ascii, storepaths_ascii,
 * readpolar, heapsort, shellsort, getans, getfloat, getdouble, getint,
 * menu, pathno, writepath, storepaths_math, storepaths3D_math
 *
 */

/* Note on conversion to double precision:
 * 1) Replace 'float' with 'double'.
 * 2) Replace '%f' with '%lf' in calls to scanf (only).
 * 3) Make sure that the function getfloat remains unchanged,
 *    since there is already a function getdouble.
 */

#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <strings.h>

typedef struct
{
  float     x, y;
} point;
typedef struct
{
  float     x, y, z;
}         point3d;

FILE     *
fopenw(fname)
/* opens a file for writing and reports on success */
  char      fname[];
{
  FILE     *out;
  out = fopen(fname, "w");
  if (out == NULL)
  {
    printf(" Cannot open file for writing.\n");
    exit(1);
  } else
  {
    printf(" File %s opened.\n", fname);
    return out;
  }
}

FILE     *
fopenr(fname)
/* opens a file for reading and reports on success */
  char      fname[];
{
  FILE     *in;
  in = fopen(fname, "r");
  if (in == NULL)
  {
    printf(" Cannot open file for reading.\n");
    exit(1);
  } else
  {
    printf(" File %s opened.\n", fname);
    return in;
  }
}


storepaths(N, K, KMAX, path, banner, len, out)
/*
 * Stores N paths of K[n] points each in a binary file for plotting. The file
 * *out must be opened for writing prior to entry.
 */
  int       N, K[], len;
  FILE     *out;
  point    *path;
  char     *banner;
{
  int       n;
  fwrite((char *) &N, sizeof(N), 1, out);
  for (n = 0; n < N; n++)
  {
    fwrite((char *) (K + n), sizeof(N), 1, out);
    fwrite((char *) (path + n * KMAX), sizeof(point), K[n], out);
  }
  fwrite((char *) &len, sizeof(int), 1, out);
  fwrite(banner, sizeof(char), len, out);
}

readpaths(K, KMAX, NMAX, path, banner, in)
/*
 * Reads N paths of K[n] points each from a binary file. The first record
 * must be N. Then follow N blocks consisting of    K[n] and the points in
 * path[n].
 */
  int       K[];
  point    *path;
  FILE     *in;
  char     *banner;
{
  int       N, n, len;
  fread((char *) &N, sizeof(N), 1, in);
  if (N > NMAX)
  {
    printf(" N=%d. NMAX too small.\n", N);
    exit(1);
  }
  for (n = 0; n < N; n++)
  {
    fread((char *) (K + n), sizeof(N), 1, in);
    if (K[n] > KMAX)
    {
      printf(" K=%d KMAX=%d : KMAX too small.\n", K[n], KMAX);
      exit(1);
    }
    fread((char *) (path + n * KMAX), sizeof(point), K[n], in);
  }
  fread((char *) &len, sizeof(int), 1, in);
  fread(banner, sizeof(char), len, in);
  return N;
}


storearray(colarray, nx, ny, NY, out)
/*
 * Stores an array of nx*ny colour indices in a binary file for plotting. The
 * file *out must be opened for writing prior to entry. NY=max. rowsize in
 * calling program.
 */
  int       nx, ny, NY;
  int      *colarray;
  FILE     *out;
{
  int       i;
  for (i = 0; i < nx; i++)
    fwrite((char *) (colarray + i * NY), sizeof(i), ny, out);
}

readarray(colarray, nx, ny, NY, in)
/*
 * Reads an array of nx*ny colour indices from a binary file for plotting.
 * The file *in must be opened for reading prior to entry. NY=max. rowsize in
 * calling program
 */
  int       nx, ny, NY;
  int      *colarray;
  FILE     *in;
{
  int       i;
  for (i = 0; i < nx; i++)
    fread((char *) (colarray + i * NY), sizeof(i), ny, in);
}

countpics(nx, ny, in, fname)
/*
 * Reads the dimensions of colour arrays from the top of a file and then
 * counts how many arrays the file contains.
 */
  FILE     *in;
  char      fname[];
  int      *nx, *ny;
{
  int       k, m, n, npic;
  fread((char *) nx, sizeof(int), 1, in);
  fread((char *) ny, sizeof(int), 1, in);
  npic = m = 0;
  n = (*nx) * (*ny);
  while (fread((char *) &k, sizeof(int), 1, in))
  {
    m++;
    if (m == n)
    {
      npic++;
      m = 0;
    }
  }
  fclose(in);
  in = fopen(fname, "r");
  fread((char *) nx, sizeof(int), 1, in);
  fread((char *) ny, sizeof(int), 1, in);
  return npic;
}

read_init_vals(np, nx, x0, rowsize, in)
  float    *x0;
  int       np, nx, rowsize;
  FILE     *in;
/* Initialises the matrix x0 by
 * reading values from
 * the file pointed to by in. The file must
 * be written in ascii format.
 * Reads np sets of nx floating point numbers.
 * ie reads np different initial values
 * for a vector x of dimension nx.
 */
{
  int       i, j;
  for (i = 0; i < np; i++)
    for (j = 0; j < nx; j++)
      fscanf(in, "%f", x0 + i * rowsize + j);
}

float_to_char(p, x, nd)
  char      p[];
  float     x;
  int       nd;
/* Translates a floating point number x
 * into a character string p[], using the
 * first nd decimal places only.
 * Returns the length of the character string.
 */
{
  int       n, i = 0, len = 0;
  char      s[20];
  if (x < 0)
  {
    p[len++] = '-';
    x = (-x);
  }
  n = (int) x;
  x = x - n;
  do
  {
    s[i++] = n % 10 + '0';
  }
  while ((n /= 10) > 0);
  s[i] = '\0';
  reverse(s);
  strcpy(p + len, s);
  len += i;
  p[len++] = '.';
  for (i = 0; i < nd; i++)
  {
    p[len++] = (int) (x *= 10.0) % 10 + '0';
  }
  p[len] = '\0';
  return len;
}

int_to_char(c, n, len)
  char      c[];
  int       n, len;
/* Translates an integer n
 * into a character string p[].
 * Returns the length of the character string.
 */
{
  char      s[20];
  int       i = 0, sign;
  if ((sign = n) < 0)
    n = (-n);
  do
  {
    s[i++] = n % 10 + '0';
  }
  while ((n /= 10) > 0);
  if (sign < 0)
    s[i++] = '-';
  s[i] = '\0';
  reverse(s);
  strcpy(c + len, s);
  return (i);
}

reverse(s)
  char      s[];
/* Reverses the character string s
 * (overwriting s)
 */
{
  int       i, j, c;
  for (i = 0, j = strlen(s) - 1; i < j; i++, j--)
  {
    c = s[i];
    s[i] = s[j];
    s[j] = c;
  }
}

readp(K, KMAX, path, in)
/*
 * Reads a path of K points from a binary file. The first record must be K.
 * Then follow the points in the path.
 */
  int      *K;
  point    *path;
  FILE     *in;
{
  fread((char *) K, sizeof(int), 1, in);
  if (*K > KMAX)
  {
    printf(" KMAX too small.\n");
    exit(1);
  }
  fread((char *) path, sizeof(point), *K, in);
}

fcopy(file1, file2)
  char      file1[], file2[];
/* Copies file1 into file2. */
{
  FILE     *in, *out;
  int       ch;
  in = fopen(file1, "r");
  out = fopen(file2, "w");
  if (in == NULL)
  {
    printf("Can't open %s for input.\n", file1);
    exit(1);
  } else if (out == NULL)
  {
    printf("Can't open %s for ouput.\n", file2);
    exit(1);
  }
  while ((ch = getc(in)) != EOF)
    putc(ch, out);
  fclose(in);
  fclose(out);
}

typedef struct
{
  int       recno, level, prerec, nexrec, parent;
  int       lson, rson, spare[7];
  float     t0, tp;
}         record;

savew(w, np, nw, r, out)
  float   **w;
  int       np, nw;
  FILE     *out;
  record   *r;
/*
 * Stores np*nw float numbers in a file preceded by the contents of the
 * record r.
 */
{
  int       j, n;
  n = fwrite((char *) (r), sizeof(record), 1, out);
  for (j = 0; j < np; j++)
    n += fwrite((char *) (w[j]), sizeof(float), nw, out);
  return n;
}

readw(w, np, nw, r, in)
  float   **w;
  int       np, nw;
  FILE     *in;
  record   *r;
/*
 * Reads the record r and the np vectors in the matrix w from a file. Returns
 * the no. of values read.
 */
{
  int       n, j;
  n = fread((char *) (r), sizeof(record), 1, in);
  if (n > 0)
    for (j = 0; j < np; j++)
      n += fread((char *) (w[j]), sizeof(float), nw, in);
  return n;
}

savew2(w, I, np, nw, r, out)
  float   **w, **I;
  int       np, nw;
  FILE     *out;
  record   *r;
/*
 * Stores 2 sets of np*nw float numbers in a file preceded by the contents of
 * the record r.
 */
{
  int       j, n;
  n = fwrite((char *) (r), sizeof(record), 1, out);
  for (j = 0; j < np; j++)
    n += fwrite((char *) (w[j]), sizeof(float), nw, out);
  for (j = 0; j < np; j++)
    n += fwrite((char *) (I[j]), sizeof(float), nw, out);
  return n;
}

readw2(w, I, np, nw, r, in)
  float   **w, **I;
  int       np, nw;
  FILE     *in;
  record   *r;
/*
 * Reads the record r, the np vectors in the matrix w and the np vectors in
 * the matrix I from a file. Returns the no. of values read.
 */
{
  int       n, j;
  n = fread((char *) (r), sizeof(record), 1, in);
  if (n > 0)
  {
    for (j = 0; j < np; j++)
      n += fread((char *) (w[j]), sizeof(float), nw, in);
    for (j = 0; j < np; j++)
      n += fread((char *) (I[j]), sizeof(float), nw, in);
  }
  return n;
}

savew3(w, I, np, nw, nij, r, out)
  float   **w, **I;
  int       np, nw, nij;
  FILE     *out;
  record   *r;
/*
 * Stores 2 matrices of float numbers 
 * of dims np*nw and np*nij
 * in a file preceded by the contents of
 * the record r.
 */
{
  int       j, n;
  n = fwrite((char *) (r), sizeof(record), 1, out);
  for (j = 0; j < np; j++)
    n += fwrite((char *) (w[j]), sizeof(float), nw, out);
  for (j = 0; j < np; j++)
    n += fwrite((char *) (I[j]), sizeof(float), nij, out);
  return n;
}

readw3(w, I, np, nw, nij, r, in)
  float   **w, **I;
  int       np, nw, nij;
  FILE     *in;
  record   *r;
/*
 * Reads the record r, the np vectors of dim nw in the matrix w
 * and the np vectors of dim nij in
 * the matrix I from a file. Returns the no. of values read.
 */
{
  int       n, j;
  n = fread((char *) (r), sizeof(record), 1, in);
  if (n > 0)
  {
    for (j = 0; j < np; j++)
      n += fread((char *) (w[j]), sizeof(float), nw, in);
    for (j = 0; j < np; j++)
      n += fread((char *) (I[j]), sizeof(float), nij, in);
  }
  return n;
}

readpaths3d(K, KMAX, NMAX, path, banner, in)
/*
 * Reads N paths of K[n] points each from a binary file. The first record
 * must be N. Then follow N blocks consisting of    K[n] and the points in
 * path[n].
 */
  int       K[];
  point3d  *path;
  FILE     *in;
  char     *banner;
{
  int       N, n, len;
  fread((char *) &N, sizeof(N), 1, in);
  if (N > NMAX)
  {
    printf(" NMAX too small.\n");
    exit(1);
  }
  for (n = 0; n < N; n++)
  {
    fread((char *) (K + n), sizeof(N), 1, in);
    if (K[n] > KMAX)
    {
      printf(" KMAX too small.\n");
      exit(1);
    }
    fread((char *) (path + n * KMAX), sizeof(point3d), K[n], in);
  }
  fread((char *) &len, sizeof(int), 1, in);
  fread(banner, sizeof(char), len, in);
  return N;
}

storepaths3d(N, K, KMAX, path, banner, len, out)
/*
 * Stores N paths of K[n] points each in a binary file for plotting. The file
 * *out must be opened for writing prior to entry.
 */
  int       N, K[], len;
  FILE     *out;
  point3d  *path;
  char     *banner;
{
  int       n;
  fwrite((char *) &N, sizeof(N), 1, out);
  for (n = 0; n < N; n++)
  {
    fwrite((char *) (K + n), sizeof(N), 1, out);
    fwrite((char *) (path + n * KMAX), sizeof(point3d), K[n], out);
  }
  fwrite((char *) &len, sizeof(int), 1, out);
  fwrite(banner, sizeof(char), len, out);
}


storepic_ascii(colarray, nx, ny, NY, out)
/*
 * Stores an array of nx*ny colour indices in an ascii file for plotting. The
 * file *out must be opened for writing prior to entry. NY=max. rowsize in
 * calling program.
 */
  int       nx, ny, NY;
  int      *colarray;
  FILE     *out;
{
  int       i, j;
  for (i = 0; i < nx; i++)
    for (j = 0; j < ny; j++)
      fprintf(out, "%d ", *(colarray + i * NY + j));
}

readpic_ascii(colarray, nx, ny, NY, in)
/*
 * Reads an array of nx*ny colour indices from an ascii file for plotting.
 * The file *in must be opened for reading prior to entry. NY=max. rowsize in
 * calling program
 */
  int       nx, ny, NY;
  int      *colarray;
  FILE     *in;
{
  int       i, j;
  for (i = 0; i < nx; i++)
    for (j = 0; j < ny; j++)
      fscanf(in, "%d ", colarray + i * NY + j);
}


countpics_ascii(nx, ny, in, fname)
/*
 * Reads the dimensions of colour arrays from the top of an ascii file and
 * then counts how many arrays the file contains.
 */
  FILE     *in;
  char      fname[];
  int      *nx, *ny;
{
  int       k, m, n, npic;
  fscanf(in, "%d ", nx);
  fscanf(in, "%d ", ny);
  npic = m = 0;
  n = (*nx) * (*ny);
  while (fscanf(in, "%d ", &k) > 0)
  {
    m++;
    if (m == n)
    {
      npic++;
      m = 0;
    }
  }
  fclose(in);
  in = fopen(fname, "r");
  fscanf(in, "%d ", nx);
  fscanf(in, "%d ", ny);
  return npic;
}

readpaths_ascii(K, KMAX, NMAX, path, in)
/*
 * Reads N paths of K[n] points each from an ascii file. The first record
 * must be N. Then follow N blocks consisting of K[n] and the points in
 * path[n].
 */
  int       K[];
  point    *path;
  FILE     *in;
{
  int       N, n, k;
  fscanf(in, "%d", &N);
  printf(" N=%d\n", N);
  if (N > NMAX)
  {
    printf(" N=%d. NMAX too small.\n", N);
    exit(1);
  }
  for (n = 0; n < N; n++)
  {
    fscanf(in, "%d", K + n);
    if (K[n] > KMAX)
    {
      printf(" KMAX too small.");
      printf(" K=%d\n", K[n]);
      exit(1);
    }
    for (k = 0; k < K[n]; k++)
      fscanf(in, "%f %f", &(path[n * KMAX + k].x), &(path[n * KMAX + k].y));
  }
  return N;
}

readpaths3d_ascii(K, KMAX, NMAX, path, in)
/*
 * Reads N paths of K[n] points each from an ascii file. The first record
 * must be N. Then follow N blocks consisting of K[n] and the points in
 * path[n].
 */
  int       K[];
  point3d  *path;
  FILE     *in;
{
  int       N, n, k;
  fscanf(in, "%d", &N);
  if (N > NMAX)
  {
    printf(" N=%d. NMAX too small.\n", N);
    exit(1);
  }
  for (n = 0; n < N; n++)
  {
    fscanf(in, "%d", K + n);
    if (K[n] > KMAX)
    {
      printf(" KMAX too small.\n");
      exit(1);
    }
    for (k = 0; k < K[n]; k++)
      fscanf(in, "%f %f %f", &(path[n * KMAX + k].x),
	     &(path[n * KMAX + k].y), &(path[n * KMAX + k].z));
  }
  return N;
}

storepaths3d_ascii(N, K, KMAX, path, out)
/*
 * Stores N paths of K[n] points each in an ascii file. The first record
 * is N. Then follow N blocks consisting of K[n] and the points in
 * path[n].
 */
  int       N, K[], KMAX;
  point3d  *path;
  FILE     *out;
{
  int       n, k;
  fprintf(out, "%d\n", N);
  for (n = 0; n < N; n++)
  {
    fprintf(out, "%d\n", K[n]);
    for (k = 0; k < K[n]; k++)
      fprintf(out, "%f %f %f\n", path[n * KMAX + k].x,
	     path[n * KMAX + k].y, path[n * KMAX + k].z);
  }
  return N;
}

readpolar(K, KMAX, NMAX, path, banner, in)
/*
 * Reads N paths of K[n] points each from a binary file. The first record
 * must be N. Then follow N blocks consisting of    K[n] and the points in
 * path[n]. The points read in are in polar coordinates and the routine
 * changes them to cartesian coordinates.
 */
  int       K[];
  point    *path;
  FILE     *in;
  char     *banner;
{
  int       N, n, len, i;
  point     p;
  float     r, th;
  fread((char *) &N, sizeof(N), 1, in);
  if (N > NMAX)
  {
    printf(" N=%d. NMAX too small.\n", N);
    exit(1);
  }
  for (n = 0; n < N; n++)
  {
    fread((char *) (K + n), sizeof(N), 1, in);
    if (K[n] > KMAX)
    {
      printf(" KMAX too small.\n");
      exit(1);
    }
    for (i = 0; i < K[n]; i++)
    {
      fread((char *) &p, sizeof(point), 1, in);
      r = p.x;
      th = p.y;
      p.x = r * cos(th);
      p.y = r * sin(th);
      *(path + n * KMAX + i) = p;
    }
  }
  fread((char *) &len, sizeof(int), 1, in);
  fread(banner, sizeof(char), len, in);
  return N;
}

heapsort(p, n)
  float     p[];
  int       n;
/*
 * Sorts the array of real numbers, p, into ascending order
 */
{
  int       l, j, ir, i;
  float     q;
  l = (n >> 1) + 1;
  ir = n;
  for (;;)
  {
    if (l > 1)
    {
      --l;
      q = p[l - 1];
    } else
    {
      q = p[ir - 1];
      p[ir - 1] = p[0];
      if (--ir == 1)
      {
	p[0] = q;
	return;
      }
    }
    i = l;
    j = l << 1;
    while (j <= ir)
    {
      if (j < ir && p[j - 1] < p[j])
	++j;
      if (q < p[j - 1])
      {
	p[i - 1] = p[j - 1];
	j += (i = j);
      } else
	j = ir + 1;
    }
    p[i - 1] = q;
  }
}

void heapsort1(p, n)
  point     p[];
  int       n;
/*
 * Sorts the array of points, p, in ascending order of x, the first
 * coordinate
 */
{
  int       l, j, ir, i;
  point     q;
  l = (n >> 1) + 1;
  ir = n;
  for (;;)
  {
    if (l > 1)
    {
      --l;
      q = p[l - 1];
    } else
    {
      q = p[ir - 1];
      p[ir - 1] = p[0];
      if (--ir == 1)
      {
	p[0] = q;
	return;
      }
    }
    i = l;
    j = l << 1;
    while (j <= ir)
    {
      if (j < ir && p[j - 1].x < p[j].x)
	++j;
      if (q.x < p[j - 1].x)
      {
	p[i - 1] = p[j - 1];
	j += (i = j);
      } else
	j = ir + 1;
    }
    p[i - 1] = q;
  }
}

void heapsort2(p, n)
  point     p[];
  int       n;
/*
 * Sorts the array of points, p, in ascending order of y, the second
 * coordinate
 */
{
  int       l, j, ir, i;
  point     q;
  l = (n >> 1) + 1;
  ir = n;
  for (;;)
  {
    if (l > 1)
    {
      --l;
      q = p[l - 1];
    } else
    {
      q = p[ir - 1];
      p[ir - 1] = p[0];
      if (--ir == 1)
      {
	p[0] = q;
	return;
      }
    }
    i = l;
    j = l << 1;
    while (j <= ir)
    {
      if (j < ir && p[j - 1].y < p[j].y)
	++j;
      if (q.y < p[j - 1].y)
      {
	p[i - 1] = p[j - 1];
	j += (i = j);
      } else
	j = ir + 1;
    }
    p[i - 1] = q;
  }
}

#define ALN2I 1.442695022
#define TINY 1.0e-5

shellsort(p, n)
  point     p[];
  int       n;
/*
 * Sorts the array of points, p, in ascending order of x, the first
 * coordinate
 */
{
  int       nn, m, j, i, lognb2;
  point     t;
  lognb2 = (log((double) n) * ALN2I + TINY);
  m = n;
  for (nn = 0; nn < lognb2; nn++)
  {
    m >>= 1;
    for (j = m; j < n; j++)
    {
      i = j - m;
      t = p[j];
      while (i >= 0 && p[i].x > t.x)
      {
	p[i + m] = p[i];
	i -= m;
      }
      p[i + m] = t;
    }
  }
}


storepaths_ascii(N, K, KMAX, path, out)
/*
 * Stores N paths of K[n] points each in an ascii file. The first record is
 * N. Then follow N blocks consisting of K[n] and the points in path[n].
 */
  int       K[], N;
  point    *path;
  FILE     *out;
{
  int       n, k;
  fprintf(out, "%d\n", N);
  for (n = 0; n < N; n++)
  {
    fprintf(out, "%d\n", K[n]);
    for (k = 0; k < K[n]; k++)
      fprintf(out, "%f %f\n", path[n * KMAX + k].x, path[n * KMAX + k].y);
  }
}

storepaths_math(m,n,kmax,p,out)
/*
 * Stores the array of 2d points, p, in an ascii file
 * in a format suitable for reading using the Mathematica
 * ReadList function, with the character ':' as a record
 * separator.
 */
   int m,n[],kmax;
   point *p;
   FILE *out;
{int i,j;
 for(i=0;i<m;i++)
 {for(j=0;j<n[i];j++)
    fprintf(out,"\n %f %f",p[i*kmax+j].x,p[i*kmax+j].y);
  fprintf(out," :");
 }
 fprintf(out,"\n");
}

getans(message, c1, c2)
  char     *message, c1, c2;
/*
 * Gets an answer to an either/or type question. The question contained in
 * message is printed. This is repeated until the first letter of the answer
 * matches one of the characters c1,c2 (which must be lower case) or upper
 * case c1,c2. When the answer is c1, a value of 1 (true) is returned. When
 * the answer is c2, a value of 0 (false)is returned.
 */
{
  char      answer[81], a;
  int m;
  if (!islower(c1) || !islower(c2))
  {
    printf(" Must use lower case characters \n for arguments 2 and 3 of getans\n");
    exit(1);
  }
  do
  {
    printf("%s (%1c/%1c) ", message, c1, c2);
    /* gets(answer); */
    scanf("%s", answer);
    a = answer[0];
  } while (a != c1 && a != c2 && a != toupper(c1) && a != toupper(c2));
  do{m=getchar();} while(m!='\n');
  return (a == c1 || a == toupper(c1));
}

int 
getint(n, varname)
  int      *n;
  char     *varname;
/*
 * Gets an integer value for the variable at address x and with the name pointed
 * to by varname. Uses the function atoi to translate the input string to
 * integer, but also checks to see whether the user intends to input the value
 * 0 (since atoi also returns 0 when the string does not define an
 * integer).
 */
{
  char      string[80];
  int       m;
  *n = 1;
  do
  {
    m = 1;
    printf(" Give value for %s : ", varname);
    scanf("%s", string);
    if (string[0] == '0' && string[1] == '\0')
      *n = 0;
    else
      m = atoi(string);
  } while (m == 0);
  if (*n > 0)
    *n = m;
  do{m=getchar();} while(m!='\n');
}

int 
getfloat(x, varname)
  float    *x;
  char     *varname;
/*
 * Gets a float value for the variable at address x and with the name pointed
 * to by varname. Uses the function atof to translate the input string to
 * float, but also checks to see whether the user intends to input the value
 * 0.0 (since atof also returns 0.0 when the string does not define a float
 * number).
 */
{
  char      string[80];
  float     y;
  int m;
  *x = 1.0;
  do
  {
    y = 1.0;
    printf(" Give value for %s : ", varname);
    scanf("%s", string);
    if (string[0] == '0' && string[1] == '\0')
      *x = 0.0;
    else if (string[0] == '0' && string[1] == '.' && string[2] == '0' && string[3] == '\0')
      *x = 0.0;
    else if (string[0] == '0' && string[1] == '.' && string[2] == '\0')
      *x = 0.0;
    else if (string[0] == '.' && string[1] == '0' && string[2] == '\0')
      *x = 0.0;
    else
      y = atof(string);
  } while (y == 0.0);
  if (*x > 0.0)
    *x = y;
  do{m=getchar();} while(m!='\n');
}

int 
getdouble(x, varname)
  double    *x;
  char     *varname;
/*
 * Gets a double value for the variable at address x and with the name pointed
 * to by varname. Uses the function atof to translate the input string to
 * float, but also checks to see whether the user intends to input the value
 * 0.0 (since atof also returns 0.0 when the string does not define a float
 * number).
 */
{
  char      string[80];
  double     y;
  int m;
  *x = 1.0;
  do
  {
    y = 1.0;
    printf(" Give value for %s : ", varname);
    scanf("%s", string);
    if (string[0] == '0' && string[1] == '\0')
      *x = (double) 0.0;
    else if (string[0] == '0' && string[1] == '.' && string[2] == '0' && string[3] == '\0')
      *x = (double) 0.0;
    else if (string[0] == '0' && string[1] == '.' && string[2] == '\0')
      *x = (double) 0.0;
    else if (string[0] == '.' && string[1] == '0' && string[2] == '\0')
      *x = (double) 0.0;
    else
      y = atof(string);
  } while (y == 0.0);
  if (*x > 0.0)
    *x = y;
  do{m=getchar();} while(m!='\n');
}

int 
menu(message, n)
  char    **message;
  int       n;
/*
 * Prints out a menu of n items and returns the number of the item chosen The
 * items to appear on the menu are stored in the rows of the char matrix
 * message.
 */
{
  char      answer[81];
  int       i, k;
  printf("\n");
  do
  {
    printf(" Choose a number from the menu:\n");
    for (i = 0; i < n; i++)
      printf("%d : %s\n", i + 1, message[i]);
    printf(" Number: ");
    scanf("%s", answer);
  } while ((k = atoi(answer)) == 0 || k > n);
  printf("\n");
  do{i=getchar();} while(i!='\n');
  return k;
}

int writepath(K, path, out, ascii)
/*
 * Stores a path of K points in an ascii or binary  file. The first
 * record is K. Then follow the points in the path.
 * Use of this function requires less memory than use of storepaths.
 * Prior to the call(s) to writepath, the function pathno must be
 * called once near the beginning of the main program.
 */
  int       K,ascii;
  point    *path;
  FILE     *out;
{
  int       k,ok;
  if(ascii)
  {ok=fprintf(out, "%d\n", K);
   for (k = 0; k < K; k++)
    ok=fprintf(out, "%f %f\n", path[k].x, path[k].y);
  }
  else
  {ok=fwrite((char *) &K, sizeof(K), 1, out);
   ok=fwrite((char *) path, sizeof(point), K, out);
  }
  return (ok);
}

int writepath3d(K, path, out, ascii)
/*
 * Stores a path of K points in R3 in an ascii or binary  file. The first
 * record is K. Then follow the points in the path.
 * Use of this function requires less memory than use of storepaths.
 * Prior to the call(s) to writepath, the function pathno must be
 * called once near the beginning of the main program.
 */
  int       K,ascii;
  point3d    *path;
  FILE     *out;
{
  int       k,ok;
  if(ascii)
  {ok=fprintf(out, "%d\n", K);
   for (k = 0; k < K; k++)
    ok=fprintf(out, "%f %f %f\n", path[k].x, path[k].y, path[k].z);
  }
  else
  {ok=fwrite((char *) &K, sizeof(K), 1, out);
   ok=fwrite((char *) path, sizeof(point3d), K, out);
  }
  return (ok);
}

int pathno(n,out,ascii)
/*
 * Writes number of paths at top of ascii or binary output file.
 */
 int n,ascii;
 FILE *out;
 {int ok;
  if(ascii) ok=fprintf(out,"%d\n",n);
  else      ok=fwrite((char *) &n,sizeof(int),1,out);
  return (ok);
 }

storepaths3d_math(m,n,kmax,p,out)
/*
 * Stores the array of 3d points, p, in an ascii file
 * in a format suitable for reading using the Mathematica
 * ReadList function, with the character ':' as a record
 * separator.
 */
   int m,n;
   point3d *p;
   FILE *out;
{int i,j;
 for(i=0;i<m;i++)
 {for(j=0;j<n;j++)
    fprintf(out,"\n %f %f %f",p[i*kmax+j].x,p[i*kmax+j].y,p[i*kmax+j].z);
  fprintf(out," :");
 }
 fprintf(out,"\n");
}

