/*****************************************************************
* The software in this file was written by Jessica Gaines,       *
* while working at the University of Edinburgh, under the        *
* direction of Terry Lyons, and using funding                    *
* provided by EC grant SC1*CT92-0784.                            *
* You may use this software for educational and academic         *
* research purposes only.                                        *
* Please do not give copies of the software in this file to      *
* anyone else. Tell anyone interested to email                   *
*  jessica@maths.ed.ac.uk                                        *
* or to write to                                                 *
*  Jessica Gaines                                                *
*  University of Edinburgh                                       *
*  Dept. of Mathematics and Statistics                           *
*  JCMB                                                          *
*  Kings Buildings                                               *
*  Mayfield Road                                                 *
*  Edinburgh EH9 3JZ                                             *
*  UK                                                            *
* and then get copies directly from Edinburgh themselves.        *
******************************************************************
*/

#include <math.h>
#include <malloc.h>

/* Contents of file :
ludcmp,lubksb,matinv,
cholesky,chbksb,matinv2,
rtnewt,traprom,
vector,free_vector,
ivector,free_ivector,
matrix,free_matrix,
array3,free_array3,
pointmatrix,free_ptmat,
tridag,tridag0_prep,tridag0_quick,
tridag_plus,tridag_prep,tridag_quick,
ksone,probks
*/

#define TINY 1.0e-20

ludcmp(a,n,indx,d)
int n,*indx;
float **a,*d;
/* Given an n*n matrix a[0..n-1][0..n-1] this routine
 replaces it by the LU decomp. of a rowwise perm. of
 itself. a and n are input. a is overwritten with the LU
 decomp. indx[0..n-1] is an
output vector which records the row permutation effected by
the partial pivoting. d is output as +-1,depending on whether the no. of row interchanges was even or odd.
*/
{
  int i,imax,j,k;
  float big,dum,sum,temp;
  float *vv,*vector(); /* vv stores implicit scaling of
                          each row */
  int free_vector(); 
  vv=vector(0,n-1);
  *d=1.0; /* no row interchanges yet */
  for(i=0;i<n;i++) /* loop over rows to get scaling info. */
  { big=0.0;
    for(j=0;j<n;j++) if((temp=fabs(a[i][j]))>big) big=temp;
    if(big==0.0) {printf(" Singular matrix in LUDCMP\n");
                  exit(1);}
    vv[i]=1.0/big;
  }
  for(j=0;j<n;j++) /* loop over col.s of Crout's method */
  { for(i=0;i<j;i++)
    { sum=a[i][j];
      for(k=0;k<i;k++) sum-=a[i][k]*a[k][j];
      a[i][j]=sum;
    }
    big=0.0; /* search for largest pivot elt. */
    for(i=j;i<n;i++)
    { sum=a[i][j];
      for(k=0;k<j;k++) sum-=a[i][k]*a[k][j];
      a[i][j]=sum;
      if((dum=vv[i]*fabs(sum))>=big)
      { big=dum;
        imax=i;
      }
    }
    if(j!=imax) /* need to interchange rows? */
    { for(k=0;k<n;k++)
      { dum=a[imax][k];
        a[imax][k]=a[j][k];
        a[j][k]=dum;
      }
      *d= (-(*d));
      vv[imax]=vv[j];
    }
    indx[j]=imax;
    if (a[j][j]==0.0) a[j][j]=TINY;
    if(j!=n) /* divide by pivot elt. */
    { dum=1.0/(a[j][j]);
      for(i=j+1;i<n;i++) a[i][j]*=dum;
    }
  }
  free_vector(vv,0,n-1);
}

lubksb(a,n,indx,b)
float **a,b[];
int n,*indx;
/* Solves the set of n linear eqn.s A*X=B. The mat. a is
input as its LU decomp. as given by ludcmp. indx[0..n-1] is
input as returned by ludcmp. b[0..n-1] is input as the right-
hand side vec. B, and returns with the soln. X. a,n and indx
are not changed by this routine and can be used in subsequent
calls with diff. B's.
*/
{ int i,ii=(-1),ip,j;
  float sum;
  for(i=0;i<n;i++)
  { ip=indx[i];
    sum=b[ip];
    b[ip]=b[i];
    if(ii>=0) for(j=ii;j<=i-1;j++) sum-=a[i][j]*b[j];
    else if (sum) ii=i;
    b[i]=sum;
  }
  for(i=n-1;i>=0;i--)
  { sum=b[i];
    for(j=i+1;j<n;j++) sum-=a[i][j]*b[j];
    b[i]=sum/a[i][i];
  }
}

matinv(a,n,ainv)
float **a,**ainv;
int n;
/* inverts the n*n matrix a. Outputs the inverse in ainv,
leaving the original a unchanged. */
{ float **matrix(),**b,*col,*vector(),d;
  int *ivector(),*indx,i,j;
  int free_ivector(),free_matrix(),free_vector();
  indx=ivector(0,n-1);
  col=vector(0,n-1);
  b=matrix(0,n-1,0,n-1);
/* Copy a into b, to leave a unchanged */
  for(i=0;i<n;i++) for(j=0;j<n;j++) b[i][j]=a[i][j];
  ludcmp(b,n,indx,&d); /* LU decomp. of b */
  for(j=0;j<n;j++)
  { for(i=0;i<n;i++) col[i]=0.0;
    col[j]=1.0;
    lubksb(b,n,indx,col); /* invert col. by col. */
    for(i=0;i<n;i++) ainv[i][j]=col[i];
  }
  free_vector(col,0,n-1);
  free_ivector(indx,0,n-1);
  free_matrix(b,0,n-1,0,n-1);
} 

cholesky(a,n)
int n;
float **a;
/* Given a positive definite n*n matrix a[1..n][1..n] 
the routine computes the (unique) lower triangular matrix
L such that a=L*L'. On exit the lower part of a is over-
written with L.
*/
{
  int i,j,k;
  float sum,x;
  for(k=0;k<n;k++)   /* One row at a time */
  { for(i=0;i<k;i++) /* elts below diagonal */
    { x=1.0/a[i][i];
      sum=a[k][i];
      for(j=0;j<i;j++) sum -= a[i][j]*a[k][j];
      a[k][i]=x*sum;
    }
    sum=a[k][k];    /* diagonal elts */
    for(j=0;j<k;j++)
    { x=a[k][j];
      sum -= x*x;
    }
    a[k][k]=sqrt(sum);
  }
}


chbksb(a,n,b)
float **a,b[];
int n;
/* Solves the set of n linear eqn.s A*X=B.
 The positive definite mat. a is
input as its LU decomp. as given by cholesky. 
 b[0..n-1] is input as the right-
hand side vec. B, and returns with the soln. X. a and n
are not changed by this routine and can be used in subsequent
calls with diff. B's.
*/
{ int i,j;
  float sum;
  for(i=0;i<n;i++)
  { sum=b[i];
    for(j=0;j<i;j++) sum -= a[i][j]*b[j];
    b[i]=sum/a[i][i];
  }
  for(i=n-1;i>=0;i--)
  { sum=b[i];
    for(j=i+1;j<n;j++) sum-=a[j][i]*b[j];
    b[i]=sum/a[i][i];
  }
}

matinv2(a,n,ainv)
float **a,**ainv;
int n;
/* inverts the positive definite n*n matrix a.
Outputs the inverse in ainv,
leaving the original a unchanged. */
{ float **matrix(),**b,*col,*vector();
  int i,j;
  int free_matrix(),free_vector();
  col=vector(0,n-1);
  b=matrix(0,n-1,0,n-1);
/* Copy a into b, to leave a unchanged */
  for(i=0;i<n;i++) for(j=0;j<n;j++) b[i][j]=a[i][j];
  cholesky(b,n); /* Cholesky LU decomp. of b */
  for(j=0;j<n;j++)
  { for(i=0;i<n;i++) col[i]=0.0;
    col[j]=1.0;
    chbksb(b,n,col); /* invert col. by col. */
    for(i=0;i<n;i++) ainv[i][j]=col[i];
  }
  free_vector(col,0,n-1);
  free_matrix(b,0,n-1,0,n-1);
} 

#define JMAX 20

float rtnewt(funcd,x1,x2,acc)
float x1,x2,acc;
int (*funcd)();
/* Using Newton-Raphson method find root of a function
known to lie in interval (x1,x2).
The root will be refined until its accuracy
is known within +-acc. funcd is a user supplied
routine providing both the func. value and first
derivative at point x. */
{int j;
 float df,dx,f,rtn;
 rtn=0.5*(x1+x2); /*initial guess*/
 for(j=1;j<=JMAX;j++)
 {(*funcd)(rtn,&f,&df);
  dx=f/df;
  rtn-=dx;
  /*printf(" x=%f ",rtn);*/
  if((x1-rtn)*(rtn-x2)<0.0)
   {printf(" error: jumped out of brackets in RTNEWT\n");
    printf(" x1=%f x2=%f x=%f\n",x1,x2,rtn);
    exit(1);}
  if(fabs(dx)<acc) /*{printf(" %d iterations delta=%f\n",j,rtn);*/
                    return rtn;
 }
 printf(" Maximum no. of iterations exceeded in RTNEWT\n");
 exit(1);
}

#define DIMT 10

float traprom(f,a,b,eps)
float (*f)(),a,b,eps;
/* Calculates the definite integral of f()
between a and b, with accuracy eps,
using the trapezium-rhombus method,
suitable for very smooth functions.
*/
{float T[DIMT][DIMT],sum,h,hh;
 int i=0,j,k,n=1;
 h=b-a;
 T[0][0]=0.5*h*((*f)(a)+(*f)(b)); /* printf(" T(0,0)=%f\n",T[0][0]); */
 do
    {i++; h*=0.5; n*=2; 
     T[i][0]=0.5*T[i-1][0];
     for(k=1,sum=0.0;k<n;k+=2) sum+=(*f)(a+k*h);
     T[i][0]+=(h*sum);
     for(j=1,hh=4.0;j<=i;j++,hh*=4.0) 
      T[i][j]=T[i][j-1]+(1.0/(hh-1.0))*(T[i][j-1]-T[i-1][j-1]);
     /* printf(" T(%d,%d)=%f\n",i,i,T[i][i]); */
    }
 while(i<DIMT&&fabs(T[i][i]-T[i-1][i-1])>eps);
 if(i>=DIMT)
 {printf(" Maximum no. of iterations exceeded in traprom\n");
  exit(1); 
 }
 else return (T[i][i]);
}

float *vector(n1,nh) int n1,nh;
/* Allocates a float vector with range [n1..nh] */
{ float *v;
  v=(float *)malloc((unsigned)(nh-n1+1)*sizeof(float));
  if(!v) {printf(" Allocation failure in vector()\n");
          exit(1);}
  return v-n1;
}

free_vector(v,n1,nh)  float *v; int n1,nh;
/* Frees a float vector allocated by vector(). */
{ free((char*) (v+n1)); }

int *ivector(n1,nh) int n1,nh;
/* Allocates an int vector with range [n1..nh] */
{ int *v;
  v=(int *)malloc((unsigned) (nh-n1+1)*sizeof(int));
  if(!v) {printf(" Allocation failure in ivector()\n");
          exit(1);}
  return v-n1;
}

free_ivector(v,n1,nh)  int *v,n1,nh;
/* Frees an int vector allocated by ivector(). */
{ free((char*) (v+n1)); }

float **matrix(nr1,nrh,nc1,nch) int nr1,nrh,nc1,nch;
/* Allocates a float matrix with range [nr1..nrh][nc1..nch]
*/
{ int i; float **m;
  m=(float**)malloc((unsigned)(nrh-nr1+1)*sizeof(float*));
  if(!m) {printf(" allocation failure 1 in matrix()\n");
           exit(1);}
  m-=nr1;
  for(i=nr1;i<=nrh;i++)
  {
   m[i]=(float *) malloc((unsigned)
                      (nch-nc1+1)*sizeof(float));
   if(!m[i]) {printf(" allocation failure 2 in matrix()\n");
           exit(1);}
   m[i]-=nc1;
  }
  return m;
}

free_matrix(m,nr1,nrh,nc1,nch)
 int nr1,nrh,nc1,nch; float **m;
/* Frees a matrix allocated with matrix. */
{ int i;
  for(i=nrh;i>=nr1;i--) free((char*)(m[i]+nc1));
  free((char*) (m+nr1));
}

float ***array3(nb1,nbh,nr1,nrh,nc1,nch)
int nb1,nbh,nr1,nrh,nc1,nch;
/* Allocates a 3-dim float array with range
[nb1..nbh] [nr1..nrh][nc1..nch]
*/
{ int i; float ***m,**matrix();
  m=(float***)malloc((unsigned)(nbh-nb1+1)*sizeof(float**));
  if(!m) {printf(" allocation failure 1 in array3()\n");
           exit(1);}
  m-=nb1;
  for(i=nb1;i<=nbh;i++)
  {
   m[i]=matrix(nr1,nrh,nc1,nch);
   if(!m[i]) {printf(" allocation failure 2 in array3()\n");
           exit(1);}
   m[i]-=nr1;
  }
  return m;
}

free_array3(m,nb1,nbh,nr1,nrh,nc1,nch)
 int nb1,nbh,nr1,nrh,nc1,nch; float ***m;
/* Frees a 3-dim array allocated with array3. */
{ int i;
  for(i=nbh;i>=nb1;i--) free_matrix(m[i],nr1,nrh,nc1,nch);
  free((char*) (m+nb1));
}

typedef struct
{
  float     x, y;
} point;

point **pointmatrix(nr1,nrh,nc1,nch) int nr1,nrh,nc1,nch;
/* Allocates a  matrix of points(x,y) with range [nr1..nrh][nc1..nch]
*/
{ int i; point **m;
  m=(point**)malloc((unsigned)(nrh-nr1+1)*sizeof(point*));
  if(!m) {printf(" allocation failure 1 in pointmatrix()\n");
           exit(1);}
  m-=nr1;
  for(i=nr1;i<=nrh;i++)
  {
   m[i]=(point *) malloc((unsigned)
                      (nch-nc1+1)*sizeof(point));
   if(!m[i]) {printf(" allocation failure 2 in matrix()\n");
           exit(1);}
   m[i]-=nc1;
  }
  return m;
}

free_ptmat(m,nr1,nrh,nc1,nch)
 int nr1,nrh,nc1,nch; point **m;
/* Frees a matrix allocated with pointmatrix. */
{ int i;
  for(i=nrh;i>=nr1;i--) free((char*)(m[i]+nc1));
  free((char*) (m+nr1));
}

tridag(a,b,c,r,u,n)
float a[],b[],c[],r[],u[];
int n;
/* Solves for a vector u[0..n-1] a tridiagonal linear set of equations.
   a,b,c, and r are input vectors and are not modified.
*/
{int j;
 float bet,*gam,*vector();
 gam=vector(0,n-1);
 if(b[0]==0) {printf("Error 1 in tridag\n"); exit(1);}
 u[0]=r[0]/(bet=b[0]);
 for(j=1;j<n;j++)
 {gam[j]=c[j-1]/bet;
  bet=b[j]-a[j]*gam[j];
/*  printf("bb[%d]=%f cc[%d]=%f\n",j,bet,j,gam[j]); */
  if(bet==0) {printf("Error 2 in tridag\n"); exit(1);}
  u[j]=(r[j]-a[j]*u[j-1])/bet;
 }
 for(j=n-2;j>=0;j--)
  u[j]-=gam[j+1]*u[j+1];
 free_vector(gam,0,n-1);
}

tridag0_prep(a,b,c,bb,cc,n)
float a[],b[],c[];
float bb[],cc[];
int n;
/* Is called before tridag0_quick, which
Solves a set of linear equations given by a
tridiagonal n*n matrix, A.
The input vectors a,b,c contain the non-zero elements of A
(with a(1) and c(n) unused).
a,b,c are unchanged on exit.
The output vectors bb,cc will be used by tridag0_quick
for fast solution of the equations. After the one call to 
tridag0_prep, tridag0_quick may be called many times with different
right-hand sides each time.
*/
{float x,fac;
 int i;
 /* Work downwards, eliminating a */
 bb[0]=b[0];
 for(i=1;i<n;i++)
 {if((x=bb[i-1])==0) {printf("Error in tridag0_prep. bb[%d]=0\n",i-1);
                       exit(1);
                      }
  cc[i]=fac=c[i-1]/x;
  bb[i]=b[i]-a[i]*fac;
 }
}

tridag0_quick(a,bb,cc,r,x,n)
float a[],bb[],cc[],r[],x[];
int n;
/* Solves a set of linear equations 
       A*x=r
where the n*n matrix, A, is tridiagonal except for non-zero
A(1,n) and A(n,1) elements.
 The input vector r
holds the right hand side of the equation and is
unchanged on exit. The solution is output in the vector x.
This version presumes that the function tridag0_prep has
been called once before any calls to tridag0_quick, in
order to set up the necessary bb and cc vectors.
*/
{int i;
 /* Work downwards */
 x[0]=r[0]/bb[0];
 for(i=1;i<n;i++)
  x[i]=(r[i]-a[i]*x[i-1])/bb[i];
 /* Work back up */
 for(i=n-2;i>=0;i--)
  x[i]-=cc[i+1]*x[i+1];
}

tridag_plus(a,b,c,r,x,n)
float a[],b[],c[],r[],x[];
int n;
/* Solves a set of linear equations given by an
n*n matrix, A, that is tridiagonal except for non-zero
A(1,n) and A(n,1) elements.
The input vectors a,b,c contain the non-zero elements of A,
with a(1)=A(1,n) and c(n)=A(n,1). The input vector r
holds the right hand side of the equation. a,b,c,r are
unchanged on exit. The solution is output in the vector x.
*/
{float x0,xn,fac;
 float *vector(),*bb,*e,*d;
 int i;
 bb=vector(0,n-1);
 e=vector(0,n-1);
 d=vector(0,n-1);
 /* Work downwards, eliminating a */
 if(a[0]==0.0) {printf("Error in input: a[0]=0\n"); exit(1);}
 if(c[n-1]==0.0) {printf("Error in input: c[n-1]=0\n"); exit(1);}
 bb[0]=b[0];
 e[0]=a[0];
 x[0]=r[0];
 for(i=1;i<n-2;i++)
 {if((x0=bb[i-1])==0) {printf("Error in tridag_plus. bb[%d]=0\n",i-1);
                       exit(1);
                      }
  fac=a[i]/x0;
  bb[i]=b[i]-fac*c[i-1];
  e[i]=(-fac*e[i-1]);
  x[i]=r[i]-fac*x[i-1];
 }
 if((x0=bb[n-3])==0) {printf("Error in tridag_plus. bb[%d]=0\n",n-3);
                       exit(1);
                      }
 fac=a[n-2]/x0;
 bb[n-2]=b[n-2]-fac*c[n-3];
 e[n-2]=c[n-2]-fac*e[n-3];
 x[n-2]=r[n-2]-fac*x[n-3];
 if((x0=bb[n-2])==0) {printf("Error in tridag_plus. bb[%d]=0\n",n-2);
                       exit(1);
                      }
 fac=a[n-1]/x0;
 bb[n-1]=e[n-1]=b[n-2]-fac*e[n-2];
 x[n-1]=r[n-1]-fac*x[n-2];
 /* Work back up, eliminating c */
 d[n-1]=c[n-1];
 fac=e[n-2]/bb[n-1];
 e[n-2]=0.0;
 d[n-2]=(-fac*d[n-1]);
 x[n-2]-=fac*x[n-1];
 for(i=n-3;i>0;i--)
 {fac=c[i]/bb[i+1];
  d[i]=(-fac*d[i+1]);
  e[i]-=fac*e[i+1];
  x[i]-=fac*x[i+1];
 }
 fac=c[0]/bb[1];
 d[0]=bb[0]-fac*d[1];
 e[0]-=fac*e[1];
 x[0]-=fac*x[1];
 /* Find x[0] and x[n-1] */
 x0=x[0];
 if((xn=d[0]*e[n-1]-d[n-1]*e[0])==0.0) 
  {printf("Error in tridag_plus: d[0]*e[n-1]-d[n-1]*e[0]=0\n");
   exit(1);
  }
 x[0]=(x0*e[n-1]-x[n-1]*e[0])/xn;
 x[n-1]=xn=(x0-d[0]*x[0])/e[0];
 x0=x[0];
 /* Find other elements of x */
 for(i=1;i<n-1;i++) x[i]=(x[i]-x0*d[i]-xn*e[i])/bb[i];
 free_vector(bb,0,n-1);
 free_vector(d,0,n-1);
 free_vector(e,0,n-1);
}

tridag_prep(a,b,c,bb,d,e,fac1,fac2,n)
float a[],b[],c[];
float bb[],d[],e[],fac1[],fac2[];
int n;
/* Is called before tridag_quick, which
Solves a set of linear equations given by an
n*n matrix, A, that is tridiagonal except for non-zero
A(1,n) and A(n,1) elements.
The input vectors a,b,c contain the non-zero elements of A,
with a(1)=A(1,n) and c(n)=A(n,1).
a,b,c are unchanged on exit.
The ouput vectors bb,e,d,fac1,fac2 will be used by tridag_quick
for fast solution of the equations. After the one call to 
tridag_prep, tridag_quick may be called many times with different
right-hand sides each time.
*/
{float x,fac;
 int i;
 /* Work downwards, eliminating a */
 if(a[0]==0.0) {printf("Error in input: a[0]=0\n"); exit(1);}
 if(c[n-1]==0.0) {printf("Error in input: c[n-1]=0\n"); exit(1);}
 bb[0]=b[0];
 e[0]=a[0];
 for(i=1;i<n-2;i++)
 {if((x=bb[i-1])==0) {printf("Error in tridag_prep. bb[%d]=0\n",i-1);
                       exit(1);
                      }
  fac1[i]=fac=a[i]/x;
  bb[i]=b[i]-fac*c[i-1];
  e[i]=(-fac*e[i-1]);
 }
 if((x=bb[n-3])==0) {printf("Error in tridag_prep. bb[%d]=0\n",n-3);
                       exit(1);
                      }
 fac1[n-2]=fac=a[n-2]/x;
 bb[n-2]=b[n-2]-fac*c[n-3];
 e[n-2]=c[n-2]-fac*e[n-3];
 if((x=bb[n-2])==0) {printf("Error in tridag_prep. bb[%d]=0\n",n-2);
                       exit(1);
                      }
 fac1[n-1]=fac=a[n-1]/x;
 bb[n-1]=e[n-1]=b[n-2]-fac*e[n-2];
 /* Work back up, eliminating c */
 d[n-1]=c[n-1];
 fac2[n-2]=fac=e[n-2]/bb[n-1];
 e[n-2]=0.0;
 d[n-2]=(-fac*d[n-1]);
 for(i=n-3;i>0;i--)
 {fac2[i]=fac=c[i]/bb[i+1];
  d[i]=(-fac*d[i+1]);
  e[i]-=fac*e[i+1];
 }
 fac2[0]=fac=c[0]/bb[1];
 d[0]=bb[0]-fac*d[1];
 e[0]-=fac*e[1];
}

tridag_quick(bb,d,e,fac1,fac2,r,x,n)
float bb[],e[],d[],fac1[],fac2[],r[],x[];
int n;
/* Solves a set of linear equations 
       A*x=r
where the n*n matrix, A, is tridiagonal except for non-zero
A(1,n) and A(n,1) elements.
 The input vector r
holds the right hand side of the equation and is
unchanged on exit. The solution is output in the vector x.
This version presumes that the function tridag_prep has
been called once before any calls to tridag_quick, in
order to set up the necessary vectors.
*/
{float x0,xn;
 int i;
 /* Work downwards */
 x[0]=r[0];
 for(i=1;i<n;i++)
  x[i]=r[i]-fac1[i]*x[i-1];
 /* Work back up */
 for(i=n-2;i>=0;i--)
  x[i]-=fac2[i]*x[i+1];
 /* Find x[0] and x[n-1] */
 x0=x[0];
 x[0]=(x0*e[n-1]-x[n-1]*e[0])/(d[0]*e[n-1]-d[n-1]*e[0]);
 x[n-1]=xn=(x0-d[0]*x[0])/e[0];
 x0=x[0];
 /* Find other elements of x */
 for(i=1;i<n-1;i++) x[i]=(x[i]-x0*d[i]-xn*e[i])/bb[i];
}

static float maxarg1,maxarg2;
#define MAX(a,b) (maxarg1=(a),maxarg2=(b),(maxarg1) > (maxarg2) ? \
  (maxarg1) : (maxarg2))

ksone(data,n,func,d,prob)
float data[],*d,*prob;
float (*func)();
int n;
/* Given an array data[0..n-1] and given a user-supplied function of a
single variable func which is a cumulative distribution function ranging
from 0 (for smallest values of its argument) to 1 (for largest values of
its argument), this routine returns the K-S statistic d, and the
significance level prob. Small values of prob shoe that the cumulative
distribution function of data is significantly different from func. The
array data is modified by being sorted into ascending order.
*/
{int j;
 float fo=0.0,fn,ff,en,dt;
 float probks();

 heapsort(data,n); /* needed if the data is not already sorted */
 en=n;
 *d=0.0;
 for(j=0;j<n;j++)
 {fn=j/en;
  ff=(*func)(data[j]);
  dt = MAX(fabs(fo-ff),fabs(fn-ff));
  if(dt>*d) *d=dt;
  fo=fn;
 }
 *prob=probks(sqrt(en)*(*d));
}

#define EPS1 0.001
#define EPS2 1.0e-8

float probks(alam)
float alam;
{int j;
 float a2,fac=2.0,sum=0.0,term,termbf=0.0;

 a2 = -2.0*alam*alam;
 for(j=1;j<=100;j++)
 {term=fac*exp(a2*j*j);
  sum+=term;
  if(fabs(term)<=EPS1*termbf||fabs(term)<EPS2*sum) return sum;
  fac = -fac;
  termbf=fabs(term);
 } /* Get here only by failing to converge */
 return 0.0;
}

