/*****************************************************************
* The software in this file was written by Jessica Gaines,       *
* while working at the University of Edinburgh, under the        *
* direction of Terry Lyons, and using funding                    *
* provided by EC grant SC1*CT92-0784.                            *
* You may use this software for educational and academic         *
* research purposes only.                                        *
* Please do not give copies of the software in this file to      *
* anyone else. Tell anyone interested to email                   *
*  jessica@maths.ed.ac.uk                                        *
* or to write to                                                 *
*  Jessica Gaines                                                *
*  University of Edinburgh                                       *
*  Dept. of Mathematics and Statistics                           *
*  JCMB                                                          *
*  Kings Buildings                                               *
*  Mayfield Road                                                 *
*  Edinburgh EH9 3JZ                                             *
*  UK                                                            *
* and then get copies directly from Edinburgh themselves.        *
******************************************************************
*/

/* This file contains more routines for generation and
 handling of Brownian paths.
 Contents:
 brown4,brown5
 getdw2,getdw3,getdw4
and also
 brownchunk2 : called by brown4,getdw2
*/

#include <math.h>
#include <stdio.h>
#include <strings.h>


/* From here onwards: routines for generating and using brownian
paths consisting of a double tree of differences and integrals
*/


#define K 5

typedef struct{ int recno,level,prerec,nexrec,parent;
                int lson,rson,spare[7];
                float t0,tp;} record;
typedef struct{ int np,nw,nlevels,nl1,seed,nrecs1,nrecs2;
                int spare[7];
                float sp[2];} header;

brown4(nlevels,n1,np,nw,func1,seed1,file)
float (*func1)();
int nlevels,n1,np,nw,seed1;
char file[];
/* A new version of brown3.
Generates differences and I(0,1) type integrals
along a Brownian path.
Creates a disjoint array for each np intervals.
n1=no. of preliminary levels (2 arrays only).
*/
{
  float fastnorm(),mean=0.0,**wp,**w1,**w2,**Ip,**I1,**I2,sd2;
  float **matrix(),t,h;
  int k,j,i,rec=0,inrec,prec;
  int size,spare[7];
  FILE *out,*fopenw(),*fopenr();
  header head;
  record r,r1;
  size=sizeof(r);
  i=sizeof(head);
  if(size!=i) {printf(" Bad record sizes\n"); exit(1);}
  size+=2*np*nw*sizeof(float);
  out=fopen(file,"w+");
  if(out==NULL) printf(" Can't open %s for direct access.\n",file);
  w1=matrix(0,np-1,0,nw-1);
  w2=matrix(0,np-1,0,nw-1);
  wp=matrix(0,np-1,0,nw-1);
  I1=matrix(0,np-1,0,nw-1);
  I2=matrix(0,np-1,0,nw-1);
  Ip=matrix(0,np-1,0,nw-1);
   /* Header */
  for(j=0;j<7;j++) head.spare[j]=r.spare[j]=spare[j]=0;
  for(j=0;j<np;j++) for(i=0;i<nw;i++) wp[j][i]=Ip[j][i]=0.0;
  head.nlevels=nlevels; head.seed=seed1;
  head.nl1=n1; head.np=np; head.nw=nw;
  head.nrecs1=head.nrecs2=0;
  head.sp[0]=head.sp[1]=0;
  savew2(wp,Ip,np,nw,&head,out);
  head.nrecs1++;
    /* Level 1 */
    /*printf(" Level 1\n");*/
    t=0.0; h=1.0;
    sd2 = sqrt(h/3.);
    for(k=0;k<np;k++) for(i=0;i<nw;i++)
    {wp[k][i]=fastnorm(mean, h, func1);
     Ip[k][i]=0.5*h*(wp[k][i]+fastnorm(mean,sd2,func1));
    }
    /*printf(" w(1)=%f\n",wp[0][0]);*/
    r.recno=1;r.level=1;
    r.prerec=r.nexrec=(-1);
    r.parent=(-1);r.lson=2;r.rson=3;
    r.t0=t;r.tp=t+np*h;
    rec+=savew2(wp,Ip,np,nw,&r,out);
    head.nrecs1++;
    /* subsequent n1-1 levels 
    Only create first 2 arrays of subintervals
    at each level*/
    for (j = 2; j <= n1; j++)
    {
     /*printf(" Level %d\n",j);*/
     h*=0.5; /*printf(" h=%f\n",h);*/
     brownchunk2(wp,Ip,np,nw,w1,w2,I1,I2,h,func1);
     if(j==2) prec=1;
     else     prec=2*(j-2);
     r.recno++;r.level=j;r.prerec=(-1);r.nexrec=r.recno+1;
     r.parent=prec;r.lson=r.recno+2;r.rson=r.recno+3;
     r.t0=t;r.tp=t+np*h;
     rec+=savew2(w1,I1,np,nw,&r,out);
     head.nrecs1++;
     r.recno++;r.prerec=r.recno-1;
     r.nexrec=(-1);r.t0=r.tp;r.tp+=np*h;
     r.lson=r.rson=(-1);
     rec+=savew2(w2,I2,np,nw,&r,out);
     head.nrecs1++;
     for(k=0;k<np;k++) for(i=0;i<nw;i++)
      {wp[k][i]=w1[k][i]; Ip[k][i]=I1[k][i];}
    }
    /* remaining nlevels-n1 levels:
    Create and store all arrays of
    subintervals. */
    k=2; 
    inrec=r.prerec; /*printf(" inrec=%d\n",inrec);*/
    for(j=n1+1;j<=nlevels;j++)
    {
     /*printf(" Level %d\n",j);*/
     h*=0.5; /*printf(" h=%f\n",h);*/
     t=0.0;
     for(i=0;i<k;i++)
     {fseek(out,(long)inrec*size,0);
      readw2(wp,Ip,np,nw,&r1,out);
      if(r1.level!=j-1) {printf(" Wrong level\n");
                         exit(1);}
      r1.lson=r.recno+1; 
      r1.rson=r1.lson+1;
      fseek(out,(long)(-size),1);
      savew2(wp,Ip,np,nw,&r1,out);
      inrec++;
      brownchunk2(wp,Ip,np,nw,w1,w2,I1,I2,h,func1);
      fseek(out,(long)0,2);
      r.recno++;r.level=j;
      if(i==0) r.prerec=(-1);
      else r.prerec=r.recno-1;
      r.nexrec=r.recno+1;
      r.parent=r1.recno;r.lson=r.rson=(-1);
      r.t0=t;r.tp=t+np*h;
      rec+=savew2(w1,I1,np,nw,&r,out);
      head.nrecs1++;
      r.recno++;r.prerec=r.recno-1;
      if(i==(k-1)) r.nexrec=(-1);
      else r.nexrec=r.recno+1;
      r.t0=r.tp;r.tp+=np*h;
      rec+=savew2(w2,I2,np,nw,&r,out);
      head.nrecs1++;
      t+=2.0*np*h;
     }
     k*=2;
    }
/* Rewrite header */
 fseek(out,(long)0,0);
 fwrite((char *)&head,sizeof(header),1,out);
 /*printf(" nrecs=%d\n",head.nrecs1);*/
    fclose(out);
    free_matrix(wp,0,np-1,0,nw-1);
    free_matrix(w1,0,np-1,0,nw-1);
    free_matrix(w2,0,np-1,0,nw-1);
    free_matrix(Ip,0,np-1,0,nw-1);
    free_matrix(I1,0,np-1,0,nw-1);
    free_matrix(I2,0,np-1,0,nw-1);
}


brownchunk2(wp,Ip,np,nw,w1,w2,I1,I2,h,func1)
int np,nw; float **wp,**w1,**w2,**Ip,**I1,**I2,h,(*func1)();
/* Generates 2*np differences and integrals on next level of brownian
tree, given np differences and integrals on previous level. 
wp[][] and Ip[][] contain the differences and integrals across intervals on
previous level. w1[][] and w2[][] will contain differences
across intervals on this level, each one containing
np differences. Similarly I1[][] and I2[][] will contain the integrals
on this new level.
Random numbers are generated using func1.
*/
{float fastnorm(),x1,x2,ww,yy,z,**wtemp,**Itemp,**matrix();
 float sd1,sd2;
 int k,i;
 sd1=0.25*sqrt(h*2.); sd2=sqrt(h*6.);
 wtemp=matrix(0,2*np-1,0,nw-1);
 Itemp=matrix(0,2*np-1,0,nw-1);
 for(k=0;k<np;k++) for(i=0;i<nw;i++)
 {x1=fastnorm(0.0, sd1, func1);
  x2=(h/12.0)*fastnorm(0.0, sd2, func1);
  ww=wp[k][i]; yy=Ip[k][i];
  z=(0.75/h)*yy+x1;
  wtemp[2*k][i]=(-0.25)*ww+z;
  wtemp[2*k+1][i]=1.25*ww-z;
  Itemp[2*k][i]=(-0.25)*h*ww+0.5*yy+x2;
  /* Itemp[2*k+1][i]=0.5*h*ww-0.25*yy-x2-h*x1; */
  Itemp[2*k+1][i]=yy-Itemp[2*k][i]-h*wtemp[2*k][i];
 }
 for(k=0;k<np;k++) for(i=0;i<nw;i++)
 {w1[k][i]=wtemp[k][i];
  w2[k][i]=wtemp[k+np][i];
  I1[k][i]=Itemp[k][i];
  I2[k][i]=Itemp[k+np][i];
 }
 free_matrix(wtemp,0,2*np-1,0,nw-1);
 free_matrix(Itemp,0,2*np-1,0,nw-1);
} 

getdw2(nw,dw,I10,level,t,h,func1,idum,cangoup)
int nw,level,idum,*cangoup;
float *dw,*I10,t,h,(*func1)();
{extern char *wfile;
 extern  FILE *in;
 static float ***w,***I;
 static record r[K];
 static header head;
 static int l,np,recsize,init=0;
 static long size;
 float ***array3(),hh,tt,**w2,**I2,**matrix();
 int i,j,k,tag;
 record r2;
 /*printf(" Entering getdw: level=%d t=%f\n",level,t);*/
 if(init==0||idum<0) /* Initialize by reading in K records, such that
                the last is on given level and the others are
                all parents. */
 {/*printf(" Initializing getdw\n");*/
  in=fopen(wfile,"r+");
  if(in==NULL) {printf(" Can't open file in getdw\n"); exit(1);}
  /*else printf(" File %s opened.\n",wfile);*/
  recsize=sizeof(record);
  fread((char *)&head,recsize,1,in);
  if(nw!=head.nw) {printf(" Noise vector wrong size in %s.\n",wfile);
                exit(1);}
  np=head.np;
  /*printf(" nlevels=%d\n",head.nlevels);
  printf(" nrecs=%d\n",head.nrecs1);*/
  if(head.nlevels<level)
  {printf(" Not enough levels in initial brownian path\n");
   exit(1);}
  if(init==0) {w=array3(0,K-1,0,np-1,0,nw-1);
               I=array3(0,K-1,0,np-1,0,nw-1);}
  size=recsize+2*np*nw*sizeof(float); 
  fseek(in,size,0);
  readw2(w[0],I[0],np,nw,r,in);
  while(r[0].level<level-K+1||t<r[0].t0) readw2(w[0],I[0],np,nw,r,in);
  for(i=1;i<K;i++)
  { fseek(in,size*r[i-1].lson,0);
    readw2(w[i],I[i],np,nw,r+i,in);}
  /*printf(" At level %d\n",r[K-1].level);*/
  l=level;
  init=1;
  /*printf(" Finished initialising\n");*/
  return ;
 }
 while(level>l) /* Go down one or more levels */
 { for(i=0;i<K-1;i++)
   { r[i]=r[i+1];
     for(j=0;j<np;j++) for(k=0;k<nw;k++)
     {w[i][j][k]=w[i+1][j][k]; I[i][j][k]=I[i+1][j][k];}
   }
  if(r[K-1].lson>0) 
  {fseek(in,size*r[K-1].lson,0);
   readw2(w[K-1],I[K-1],np,nw,r+K-1,in);
   l++;
  }
  else /* Create new noise elements */
  {/*printf(" Creating new arrays on lower levels.\n");*/
   i=1<<l;
   /*printf(" previous level=%d 1<<level=%d\n",l,i);*/
   hh=1.0/i;
   w2=matrix(0,np-1,0,nw-1);
   I2=matrix(0,np-1,0,nw-1);
   brownchunk2(w[K-2],I[K-2],np,nw,w[K-1],w2,I[K-1],I2,hh,func1);
   r[K-1].recno=head.nrecs1++;head.nrecs2++;
   l++; r[K-1].level=l;
   r[K-1].nexrec=r[K-1].recno+1;
   r[K-1].parent=r[K-2].recno;r[K-1].lson=r[K-1].rson=(-1);
   /*r[K-1].prerec=findprerec(r+K-1,size,recsize,in);*/
   r[K-1].prerec=(-1);
   r[K-2].lson=r[K-1].recno;
   r[K-1].t0=r[K-2].t0;
   r[K-1].tp=r[K-1].t0+np*hh;
   fclose(in);
   in=fopen(wfile,"a+");
   i=savew2(w[K-1],I[K-1],np,nw,r+K-1,in);
   fclose(in);
   in=fopen(wfile,"r+");
   r2.recno=head.nrecs1++;head.nrecs2++;r2.level=l;
   r2.prerec=r2.recno-1;
   r2.parent=r[K-1].parent;r2.lson=r2.rson=(-1);
   /*r2.nexrec=findnexrec(&r2,size,recsize,in);*/
   r2.nexrec=(-1);
   r[K-2].rson=r2.recno;
   r2.t0=r[K-1].tp;r2.tp=r2.t0+np*hh;
   /*printf(" tp=%f\n",r2.tp);*/
   fclose(in);
   in=fopen(wfile,"a+");
   i=savew2(w2,I2,np,nw,&r2,in);
   fclose(in);
   in=fopen(wfile,"r+");
   free_matrix(w2,0,np-1,0,nw-1);
   free_matrix(I2,0,np-1,0,nw-1);
   /* Rewrite header */
   fwrite((char *)&head,recsize,1,in);
   /* Rewrite record(K-2) to update lson,rson */
   fseek(in,r[K-2].recno*size,0);
   savew2(w[K-2],I[K-2],np,nw,r+K-2,in);
  }
 }
 
 while(l-K+1>level) /* Go up one or more levels */
 { for(i=K-1;i>0;i--)
   { r[i]=r[i-1];
     for(j=0;j<np;j++) for(k=0;k<nw;k++) 
     {w[i][j][k]=w[i-1][j][k]; I[i][j][k]=I[i-1][j][k];}
   }
   fseek(in,size*r[1].parent,0);
   readw2(w[0],I[0],np,nw,r,in);
   l--;
 }

 if(t<r[0].t0) {printf(" Error: t<t0 in getw.\n"); exit(1);}
 /* Get correct level */ 
   for(i=0;i<K;i++) if(r[i].level==level) {tag=i; break;}

   while(t>=r[tag].tp) /* Get correct time interval */
   { /* Move tag down to K-1 */
    /*printf(" Next time interval needed.\n");*/
    /*printf(" t=%f tp=%f\n",t,r[tag].tp);*/
    r[K-1]=r[tag];
    for(j=0;j<np;j++) for(k=0;k<nw;k++) 
     {w[K-1][j][k]=w[tag][j][k]; I[K-1][j][k]=I[tag][j][k];}
    tag=K-1; l=level;
    if(r[tag].nexrec>0)
    {fseek(in,r[tag].nexrec*size,0);
     readw2(w[tag],I[tag],np,nw,r+tag,in);
    }
    else
    {r[0]=r[tag];
     initstack();
     while(r[0].nexrec<0)
     {push(r[0].recno);
      fseek(in,r[0].parent*size,0);
      readw2(w[0],I[0],np,nw,r,in);
     }
     fseek(in,r[0].nexrec*size,0);
     readw2(w[0],I[0],np,nw,r,in);
     while(r[0].level<level) /* Create more tree */
     {
      /* Rewrite parent */
      /*printf(" Creating new arrays to broaden tree.\n");*/
      fseek(in,r[0].recno*size,0);
      r[0].lson=head.nrecs1;r[0].rson=head.nrecs1+1;
            fwrite((char *)r,recsize,1,in);
      hh=1.0/(1<<r[0].level);
      brownchunk2(w[0],I[0],np,nw,w[1],w[2],I[1],I[2],hh,func1);
      r[1]=r[0];
      r[1].recno=head.nrecs1++;head.nrecs2++;
      r[1].parent=r[0].recno;
      r[1].prerec=pop();
      r[1].level++;r[1].nexrec=r[1].recno+1;
      r[1].lson=r[1].rson=(-1);
      r[1].tp=r[0].t0+np*hh;
      tellprerec(r[1].prerec,r[1].recno,recsize,size,in);
      fclose(in);
      in=fopen(wfile,"a+");
      savew2(w[1],I[1],np,nw,r+1,in);
      fclose(in);
      in=fopen(wfile,"r+");
      r[2]=r[1];head.nrecs1++;head.nrecs2++;
      r[2].recno++;r[2].prerec=r[2].recno-1;
      /*r[2].nexrec=findnexrec(r+2,size,recsize,in); */
      r[2].nexrec=(-1);
      r[2].t0=r[2].tp;r[2].tp+=np*hh;
      fclose(in);
      in=fopen(wfile,"a+");
      savew2(w[2],I[2],np,nw,r+2,in);
      fclose(in);
      in=fopen(wfile,"r+");
      for(k=0;k<np;k++) for(i=0;i<nw;i++)
       {w[0][k][i]=w[1][k][i]; I[0][k][i]=I[1][k][i];}
      r[0]=r[1];
      /* Rewrite header */
      fwrite((char *)&head,recsize,1,in);
     }
     r[K-1]=r[0];
     for(j=0;j<np;j++) for(k=0;k<nw;k++) 
      {w[K-1][j][k]=w[0][j][k]; I[K-1][j][k]=I[0][j][k];}
     freestack();
    }
/*Fill in w and r */
     for(i=K-2;i>=0;i--)
     {fseek(in,r[i+1].parent*size,0);
      readw2(w[i],I[i],np,nw,r+i,in);
     }
   } 
/* Return correct dw */
 tt=r[tag].t0; i=0;
 while(tt<t) { i++; tt=r[tag].t0+i*h;}
 /*printf(" i=%d\n",i);*/
 if(i>np-1) {printf(" Error in time.\n"); exit(1);}
 for(j=0;j<nw;j++) {dw[j]=w[tag][i][j]; I10[j]=I[tag][i][j];}
/* Check if possible to go up a level
   next time. It is only possible when the
   current time index = 3 mod 4. */
  *cangoup=(i%4==3?1:0);
  return ;
}

new_getdw2(nw,dw,I10,level,t,h,func1,idum,cangoup)
int nw,level,idum,*cangoup;
float *dw,*I10,t,h,(*func1)();
{extern char *wfile;
 extern  FILE *in;
 static float ***w,***I;
 static record r[K];
 static header head;
 static int l,np,recsize,init=0;
 static long size;
 float ***array3(),hh,tt,**w1,**w2,**wp,**I1,**I2,**Ip,**matrix();
 int i,j,k,n,tag;
 record r2;
 /* printf(" Entering getdw: level=%d t=%f\n",level,t); */
 if(init==0||idum<0) /* Initialize by reading in K records, such that
                the last is on given level and the others are
                all parents. */
 {/* printf(" Initializing getdw2\n"); */
  recsize=sizeof(record);
  in=fopen(wfile,"r+");
  if(in==NULL)
  {printf("Starting new Brownian tree in %s\n",wfile);
   in=fopen(wfile,"w+");
   if(in==NULL) {printf(" Can't open %s for direct access.\n",wfile);
                 exit(1);}
   getint(&np,"no. of elements in each array (a power of 2 is recommended)");
   size=recsize+2*np*nw*sizeof(float); 
    /* Header */
   wp=matrix(0,np-1,0,nw-1);
   w1=matrix(0,np-1,0,nw-1);
   w2=matrix(0,np-1,0,nw-1);
   Ip=matrix(0,np-1,0,nw-1);
   I1=matrix(0,np-1,0,nw-1);
   I2=matrix(0,np-1,0,nw-1);
   for(j=0;j<7;j++) head.spare[j]=r[0].spare[j]=0;
   for(j=0;j<np;j++) for(i=0;i<nw;i++)  wp[j][i]=Ip[j][i]=0.0;
   head.nlevels=1; head.seed=0;
   head.nl1=1; head.np=np; head.nw=nw;
   head.nrecs1=head.nrecs2=0;
   head.sp[0]=head.sp[1]=0.0;
   savew2(wp,Ip,np,nw,&head,in);
   head.nrecs1++;
   /* First record */
   hh=1.0; tt=0.0;
   for(k=0;k<np;k++) for(i=0;i<nw;i++)
   {wp[k][i]=fastnorm(0.0, hh, func1);
    Ip[k][i]=0.5*hh*(wp[k][i]+fastnorm(0.0,sqrt(hh/3.),func1));
   }
   r[0].recno=1;r[0].level=1;
   r[0].prerec=r[0].nexrec=(-1);
   r[0].parent=(-1);
   if(K>1) {r[0].lson=2;r[0].rson=3;}
   else r[0].lson=r[0].rson=(-1);
   r[0].t0=tt; r[0].tp=tt+np*hh;
   savew2(wp,Ip,np,nw,r,in);
   head.nrecs1++;
   /* subsequent K-1 levels :
      Only create first 2 arrays of subintervals at each level
   */
   for (j = 2; j <= K; j++)
   {
    /* printf(" Level %d\n",j); */
    hh*=0.5; /* printf(" h=%f\n",hh); */
    brownchunk2(wp,Ip,np,nw,w1,w2,I1,I2,hh,func1);
    r[0].recno++;r[0].level=j;r[0].prerec=(-1);r[0].nexrec=r[0].recno+1;
    r[0].parent=(j==2?1:2*(j-2));
    if(j<K) {r[0].lson=r[0].recno+2;r[0].rson=r[0].recno+3;}
    else r[0].lson=r[0].rson=(-1);
    r[0].t0=tt; r[0].tp=tt+np*hh;
    savew2(w1,I1,np,nw,r,in);
    head.nrecs1++;
    r[0].prerec=r[0].recno; r[0].recno++;
    r[0].nexrec=(-1);r[0].t0=r[0].tp;r[0].tp+=np*hh;
    r[0].lson=r[0].rson=(-1);
    savew2(w2,I2,np,nw,r,in);
    head.nrecs1++;
    for(k=0;k<np;k++) for(i=0;i<nw;i++)
    {wp[k][i]=w1[k][i]; Ip[k][i]=I1[k][i];}
   }
  /* Rewrite header */
   fseek(in,(long)0,0);
   fwrite((char *)&head,sizeof(header),1,in);
  /* Tidy up */
   free_matrix(wp,0,np-1,0,nw-1);
   free_matrix(w1,0,np-1,0,nw-1);
   free_matrix(w2,0,np-1,0,nw-1);
   free_matrix(Ip,0,np-1,0,nw-1);
   free_matrix(I1,0,np-1,0,nw-1);
   free_matrix(I2,0,np-1,0,nw-1);
   fclose(in);
   in=fopen(wfile,"r+");
  }
/* Now read in first K records */
  {fread((char *)&head,recsize,1,in);
   if(nw!=head.nw) {printf(" Noise vector wrong size in %s.\n",wfile);
                    exit(1);}
   np=head.np;
   /*printf(" nlevels=%d\n",head.nlevels);
   printf(" nrecs=%d\n",head.nrecs1);*/
   if(init==0) {w=array3(0,K-1,0,np-1,0,nw-1); 
                I=array3(0,K-1,0,np-1,0,nw-1); 
               }
   size=recsize+2*np*nw*sizeof(float); 
   fseek(in,size,0);
   readw2(w[K-1],I[K-1],np,nw,r+K-1,in);
   for(n=1;n<K;n++)
    if(r[K-1].lson>0)
     {for(i=K-1-n;i<K-1;i++)
      {r[i]=r[i+1];
       for(j=0;j<np;j++) for(k=0;k<nw;k++) 
       {w[i][j][k]=w[i+1][j][k]; I[i][j][k]=I[i+1][j][k];}
      }
      fseek(in,size*r[K-1].lson,0);
      readw2(w[K-1],I[K-1],np,nw,r+K-1,in);
     }
   l=r[K-1].level; /* printf(" level=%d\n",l); */
  }
  init=1;
  /* printf(" Finished initialising\n"); */
  return ;
 }
 while(level>l) /* Go down one or more levels */
 { for(i=0;i<K-1;i++)
   { r[i]=r[i+1];
     for(j=0;j<np;j++) for(k=0;k<nw;k++) 
     {w[i][j][k]=w[i+1][j][k]; I[i][j][k]=I[i+1][j][k];}
   }
  if(r[K-1].lson>0) 
  {fseek(in,size*r[K-1].lson,0);
   readw2(w[K-1],I[K-1],np,nw,r+K-1,in);
   l++;
  }
  else /* Create new noise elements */
  {/*printf(" Creating new arrays on lower levels.\n");*/
   i=1<<l;
   /*printf(" previous level=%d 1<<level=%d\n",l,i);*/
   hh=1.0/i;
   w2=matrix(0,np-1,0,nw-1);
   I2=matrix(0,np-1,0,nw-1);
   brownchunk2(w[K-2],I[K-2],np,nw,w[K-1],w2,I[K-1],I2,hh,func1);
   r[K-1].recno=head.nrecs1++;head.nrecs2++;
   l++; r[K-1].level=l;
   r[K-1].nexrec=r[K-1].recno+1;
   r[K-1].parent=r[K-2].recno;r[K-1].lson=r[K-1].rson=(-1);
   /*r[K-1].prerec=findprerec(r+K-1,size,recsize,in);*/
   r[K-1].prerec=(-1);
   r[K-2].lson=r[K-1].recno;
   r[K-1].t0=r[K-2].t0;
   r[K-1].tp=r[K-1].t0+np*hh;
   fclose(in);
   in=fopen(wfile,"a+");
   i=savew2(w[K-1],I[K-1],np,nw,r+K-1,in);
   fclose(in);
   in=fopen(wfile,"r+");
   r2.recno=head.nrecs1++;head.nrecs2++;r2.level=l;
   r2.prerec=r2.recno-1;
   r2.parent=r[K-1].parent;r2.lson=r2.rson=(-1);
   /*r2.nexrec=findnexrec(&r2,size,recsize,in);*/
   r2.nexrec=(-1);
   r[K-2].rson=r2.recno;
   r2.t0=r[K-1].tp;r2.tp=r2.t0+np*hh;
   /*printf(" tp=%f\n",r2.tp);*/
   fclose(in);
   in=fopen(wfile,"a+");
   i=savew2(w2,I2,np,nw,&r2,in);
   fclose(in);
   in=fopen(wfile,"r+");
   free_matrix(w2,0,np-1,0,nw-1);
   free_matrix(I2,0,np-1,0,nw-1);
   /* Rewrite header */
   fwrite((char *)&head,recsize,1,in);
   /* Rewrite record(K-2) to update lson,rson */
   fseek(in,r[K-2].recno*size,0);
   savew2(w[K-2],I[K-2],np,nw,r+K-2,in);
  }
 }
 
 while(l-K+1>level) /* Go up one or more levels */
 { for(i=K-1;i>0;i--)
   { r[i]=r[i-1];
     for(j=0;j<np;j++) for(k=0;k<nw;k++) 
     {w[i][j][k]=w[i-1][j][k]; I[i][j][k]=I[i-1][j][k];}
   }
   fseek(in,size*r[1].parent,0);
   readw2(w[0],I[0],np,nw,r,in);
   l--;
 }

 if(t<r[0].t0) {printf(" Error: t<t0 in getw.\n"); exit(1);}
 /* Get correct level */ 
   for(i=0;i<K;i++) if(r[i].level==level) {tag=i; break;}

   while(t>=r[tag].tp) /* Get correct time interval */
   { /* Move tag down to K-1 */
    /*printf(" Next time interval needed.\n");*/
    /*printf(" t=%f tp=%f\n",t,r[tag].tp);*/
    r[K-1]=r[tag];
    for(j=0;j<np;j++) for(k=0;k<nw;k++) 
    {w[K-1][j][k]=w[tag][j][k]; I[K-1][j][k]=I[tag][j][k];}
    tag=K-1; l=level;
    if(r[tag].nexrec>0)
    {fseek(in,r[tag].nexrec*size,0);
     readw2(w[tag],I[tag],np,nw,r+tag,in);
    }
    else
    {r[0]=r[tag];
     initstack();
     while(r[0].nexrec<0)
     {push(r[0].recno);
      fseek(in,r[0].parent*size,0);
      readw2(w[0],I[0],np,nw,r,in);
     }
     fseek(in,r[0].nexrec*size,0);
     readw2(w[0],I[0],np,nw,r,in);
     while(r[0].level<level) /* Create more tree */
     {
      /* Rewrite parent */
      /*printf(" Creating new arrays to broaden tree.\n");*/
      fseek(in,r[0].recno*size,0);
      r[0].lson=head.nrecs1;r[0].rson=head.nrecs1+1;
            fwrite((char *)r,recsize,1,in);
      hh=1.0/(1<<r[0].level);
      brownchunk2(w[0],I[0],np,nw,w[1],w[2],I[1],I[2],hh,func1);
      r[1]=r[0];
      r[1].recno=head.nrecs1++;head.nrecs2++;
      r[1].parent=r[0].recno;
      r[1].prerec=pop();
      r[1].level++;r[1].nexrec=r[1].recno+1;
      r[1].lson=r[1].rson=(-1);
      r[1].tp=r[0].t0+np*hh;
      tellprerec(r[1].prerec,r[1].recno,recsize,size,in);
      fclose(in);
      in=fopen(wfile,"a+");
      savew2(w[1],I[1],np,nw,r+1,in);
      fclose(in);
      in=fopen(wfile,"r+");
      r[2]=r[1];head.nrecs1++;head.nrecs2++;
      r[2].recno++;r[2].prerec=r[2].recno-1;
      /*r[2].nexrec=findnexrec(r+2,size,recsize,in); */
      r[2].nexrec=(-1);
      r[2].t0=r[2].tp;r[2].tp+=np*hh;
      fclose(in);
      in=fopen(wfile,"a+");
      savew2(w[2],I[2],np,nw,r+2,in);
      fclose(in);
      in=fopen(wfile,"r+");
      for(k=0;k<np;k++) for(i=0;i<nw;i++)
       {w[0][k][i]=w[1][k][i]; I[0][k][i]=I[1][k][i];}
      r[0]=r[1];
      /* Rewrite header */
      fwrite((char *)&head,recsize,1,in);
     }
     r[K-1]=r[0];
     for(j=0;j<np;j++) for(k=0;k<nw;k++) 
     {w[K-1][j][k]=w[0][j][k]; I[K-1][j][k]=I[0][j][k];}
     freestack();
    }
/*Fill in w and r */
     for(i=K-2;i>=0;i--)
     {fseek(in,r[i+1].parent*size,0);
      readw2(w[i],I[i],np,nw,r+i,in);
     }
   } 
/* Return correct dw */
 tt=r[tag].t0; i=0;
 while(tt<t) { i++; tt=r[tag].t0+i*h;}
 /*printf(" i=%d\n",i);*/
 if(i>np-1) {printf(" Error in time.\n"); exit(1);}
 for(j=0;j<nw;j++) {dw[j]=w[tag][i][j]; I10[j]=I[tag][i][j];}
/* Check if possible to go up a level
   next time. It is only possible when the
   current time index = 3 mod 4. */
  *cangoup=(i%4==3?1:0);
  return ;
}

/* From here onwards: routines for generating and using brownian
paths consisting of a double tree of differences and integrals of dwi*dwj
*/

brown5(nlevels,n1,np,nw,func,seed1,file1,file2)
float (*func)();
int nlevels,n1,np,nw,seed1;
char file1[],file2[];
/* Another new version of brown3.
Generates differences and I(i,j) type integrals
along a Brownian path.
Creates a disjoint array for each np intervals.
n1=no. of preliminary levels (2 arrays only).
*/
{
  extern char *wfile;
  extern FILE *in;
  float sd,vari,hh,tt;
  float fastnorm(),mean=0.0,**wp,**w1,**w2,**Iij;
  float *vector(),**matrix(),t,h,*dw,*sum1,*sum2;
  int i,j,k,ni,nj,ip,rec=0,inrec,prec,nij,KK,L;
  int dum,lowl,nk,spare[7];
  long size;
  FILE *out1,*out2,*out3,*in2,*fopenw(),*fopenr();
  char string[80];
  header head;
  record r,r1;
  if(nw>2) {printf("This program is not correct for nw>2\n"); exit(1);}
  size=sizeof(r);
  i=sizeof(head);
  if(size!=i) {printf(" Bad record sizes\n"); exit(1);}
  nij=(nw*(nw-1))/2; /* Dimension of vector of integrals */
  printf("nij=%d\n",nij);
  size+=np*nw*sizeof(float);
  out1=fopen(file1,"w+");
  if(out1==NULL) printf(" Can't open %s for direct access.\n",file1);
  /*out3=fopenw("intij.dat","w");*/
  L=2*nlevels-1;
  w1=matrix(0,np-1,0,nw-1);
  w2=matrix(0,np-1,0,nw-1);
  wp=matrix(0,np-1,0,nw-1);
  Iij=matrix(0,np-1,0,nij-1);
  dw=vector(0,nw-1);
  sum1=vector(0,nw-1);
  sum2=vector(0,nw-1);
   /* Header */
  for(j=0;j<7;j++) head.spare[j]=r.spare[j]=spare[j]=0;
  for(j=0;j<np;j++) for(i=0;i<nw;i++) wp[j][i]=0.0;
  head.nlevels=L; head.seed=seed1;
  head.nl1=n1; head.np=np; head.nw=nw;
  head.nrecs1=head.nrecs2=0;
  head.sp[0]=head.sp[1]=0;
  savew(wp,np,nw,&head,out1);
  head.nrecs1++;
  /* Pass 1  : levels of dw  down to level L 
      records stored in tempfile for now.   */
  printf(" Pass 1\n");
    /* Level 1 */
    printf(" Level 1\n");
    t=0.0; h=1.0;
    sd = vari = 1.0;
    for(k=0;k<np;k++) for(i=0;i<nw;i++)
    wp[k][i]=fastnorm(mean, sd, func);
    printf(" w(1)=%f\n",wp[0][0]);
    r.recno=1;r.level=1;
    r.prerec=r.nexrec=(-1);
    r.parent=(-1);r.lson=2;r.rson=3;
    r.t0=t;r.tp=t+np*h;
    rec+=savew(wp,np,nw,&r,out1);
    head.nrecs1++;
    vari = vari / 2;
    /* subsequent n1-1 levels 
    Only create first 2 arrays of subintervals
    at each level*/
    for (j = 2; j <= n1; j++)
    {
     printf(" Level %d\n",j);
     vari*=0.5; h*=0.5; /*printf(" h=%f\n",h);*/
     sd = sqrt(vari);
     brownchunk(wp,np,nw,w1,w2,sd,func);
     if(j==2) prec=1;
     else     prec=2*(j-2);
     r.recno++;r.level=j;r.prerec=(-1);r.nexrec=r.recno+1;
     r.parent=prec;r.lson=r.recno+2;r.rson=r.recno+3;
     r.t0=t;r.tp=t+np*h;
     rec+=savew(w1,np,nw,&r,out1);
     head.nrecs1++;
     r.recno++;r.prerec=r.recno-1;
     r.nexrec=(-1);r.t0=r.tp;r.tp+=np*h;
     r.lson=r.rson=(-1);
     rec+=savew(w2,np,nw,&r,out1);
     head.nrecs1++;
     for(k=0;k<np;k++) for(i=0;i<nw;i++)
      wp[k][i]=w1[k][i];
    }
    /* remaining L-n1 levels:
    Create and store all arrays of
    subintervals. */
    k=2; 
    inrec=r.prerec; /*printf(" inrec=%d\n",inrec);*/
    for(j=n1+1;j<=L;j++)
    {
     printf(" Level %d\n",j);
     vari*=0.5; sd=sqrt(vari);
     h*=0.5; /*printf(" h=%f\n",h);*/
     t=0.0;
     for(i=0;i<k;i++)
     {fseek(out1,inrec*size,0);
      readw(wp,np,nw,&r1,out1);
      if(r1.level!=j-1) {printf(" Wrong level\n");
                         exit(1);}
      r1.lson=r.recno+1; 
      r1.rson=r1.lson+1;
      fseek(out1,-size,1);
      savew(wp,np,nw,&r1,out1);
      inrec++;
      brownchunk(wp,np,nw,w1,w2,sd,func);
      fseek(out1,(long)0,2);
      r.recno++;r.level=j;
      if(i==0) r.prerec=(-1);
      else r.prerec=r.recno-1;
      r.nexrec=r.recno+1;
      r.parent=r1.recno;r.lson=r.rson=(-1);
      r.t0=t;r.tp=t+np*h;
      rec+=savew(w1,np,nw,&r,out1);
      head.nrecs1++;
      r.recno++;r.prerec=r.recno-1;
      if(i==(k-1)) r.nexrec=(-1);
      else r.nexrec=r.recno+1;
      r.t0=r.tp;r.tp+=np*h;
      rec+=savew(w2,np,nw,&r,out1);
      head.nrecs1++;
      t+=2.0*np*h;
     }
     k*=2;
    }
/* Rewrite header */
 fseek(out1,(long)0,0);
 fwrite((char *)&head,sizeof(header),1,out1);
 fclose(out1);

 /* Pass 2  : levels of Iij down to level nlevels */
 printf(" Pass 2\n");
 strcpy(string,"cp ");
 strcat(string,file1);
 strcat(string," temp.dat\0");
 printf("%s\n",string);
 system(string);
 in2=fopenr(file1);
 out2=fopenw(file2);
 wfile="temp.dat";

  /* Header */
  for(j=0;j<np;j++) for(i=0;i<nw;i++) wp[j][i]=0.0;
  for(j=0;j<np;j++) for(i=0;i<nij;i++) Iij[j][i]=0.0;
  savew3(wp,Iij,np,nw,nij,&head,out2);
  /* Level 1 */
  /*fprintf(out3," Level 1\n");*/
  rec=0; h=hh=1.0; k=1; inrec=1;
  readw(wp,np,nw,&r,in2);
  readw(wp,np,nw,&r,in2);
  /*fprintf(out3," w(1)=%f\n",wp[0][0]);*/
  rec+=savew3(wp,Iij,np,nw,nij,&r,out2);
  head.nrecs1++;
  /* Subsequent levels */
  for(j=2;j<=nlevels;j++) /* Loop over levels */
  {/*fprintf(out3," Level %d\n",j);*/
   t=tt=0.0;
   k*=2;
   h*=0.5; hh*=0.25;
   lowl=2*j-1; KK=1<<(j-1); printf(" lowl=%d K=%d\n",lowl,KK);
   getdw(nw,dw,lowl,tt,hh,func,-1,&dum); /* Initialize getdw */
   for(ni=0;ni<nw;ni++) sum1[ni]=sum2[ni]=0.0;
   for(i=0;i<k;i++) /* Loop over records at level j */
   {readw(wp,np,nw,&r,in2);
    for(ip=0;ip<np;ip++) /* Loop over time periods in each record */
    {for(ni=0;ni<nw;ni++) for(nj=0;nj<ni;nj++) 
       Iij[ip][(ni-1)*nw+nj]=(-sum1[ni])*wp[ip][nj];
     for(nk=0;nk<KK;nk++) /* Loop over KK vectors from level 2*j-1 */
     {getdw(nw,dw,lowl,tt,hh,func,1,&dum);
      for(ni=0;ni<nw;ni++) for(nj=0;nj<ni;nj++) 
       Iij[ip][(ni-1)*nw+nj]+=(sum2[ni]*dw[nj]);
      for(ni=0;ni<nw;ni++) sum2[ni]+=dw[ni];
      tt+=hh;
     }
     t+=h;
     for(ni=0;ni<nw;ni++) sum1[ni]+=wp[ip][ni];
    }
    rec+=savew3(wp,Iij,np,nw,nij,&r,out2);
   }
  fclose(in);
  }
    fclose(in2);
    fclose(out2);
    /*fclose(out3);*/
    strcpy(string,"rm");
    strcat(string," temp.dat\0");
    printf("%s\n",string);
    system(string);
    free_matrix(wp,0,np-1,0,nw-1);
    free_matrix(w1,0,np-1,0,nw-1);
    free_matrix(w2,0,np-1,0,nw-1);
    free_matrix(Iij,0,np-1,0,nw-1);
    free_vector(sum1,0,nw-1);
    free_vector(sum2,0,nw-1);
    free_vector(dw,0,nw-1);
}


getdw3(nw,nij,dw,I,level,t,h,func,idum,cangoup)
int nw,nij,level,idum,*cangoup;
float *dw,*I,t,h,(*func)();
{extern char *wfile;
 extern  FILE *in;
 static float ***w,***Iij;
 static record r[K];
 static header head;
 static int l,np,recsize,init=0;
 static long size;
 float ***array3(),tt,**matrix();
 int i,j,k,tag;
 /* float hh,sd,**w2,
 record r2; */
 /*printf(" Entering getdw: level=%d t=%f\n",level,t);*/
 if(init==0||idum<0) /* Initialize by reading in K records, such that
                the last is on given level and the others are
                all parents. */
 {/*printf(" Initializing getdw\n");*/
  in=fopen(wfile,"r+");
  if(in==NULL) {printf(" Can't open file in getdw\n"); exit(1);}
  /*else printf(" File %s opened.\n",wfile);*/
  recsize=sizeof(record);
  fread((char *)&head,recsize,1,in);
  if(nw!=head.nw) {printf(" Noise vector wrong size in %s.\n",wfile);
                exit(1);}
  np=head.np;
  /*printf(" nlevels=%d\n",head.nlevels);
  printf(" nrecs=%d\n",head.nrecs1);*/
  if(head.nlevels<level)
  {printf(" Not enough levels in initial brownian path\n");
   exit(1);}
  if(init==0) {w=array3(0,K-1,0,np-1,0,nw-1);
               Iij=array3(0,K-1,0,np-1,0,nij-1);
              }
  size=recsize+np*(nw+nij)*sizeof(float); 
  fseek(in,size,0);
  readw3(w[0],Iij[0],np,nw,nij,r,in);
  while(r[0].level<level-K+1||t<r[0].t0) readw3(w[0],Iij[0],np,nw,nij,r,in);
  for(i=1;i<K;i++)
  { fseek(in,size*r[i-1].lson,0);
    readw3(w[i],Iij[i],np,nw,nij,r+i,in);}
  /*printf(" At level %d\n",r[K-1].level);*/
  l=level;
  init=1;
  /*printf(" Finished initialising\n");*/
  return ;
 }
 while(level>l) /* Go down one or more levels */
 { for(i=0;i<K-1;i++)
   { r[i]=r[i+1];
     for(j=0;j<np;j++) for(k=0;k<nw;k++) w[i][j][k]=w[i+1][j][k];
     for(j=0;j<np;j++) for(k=0;k<nij;k++) Iij[i][j][k]=Iij[i+1][j][k];
   }
  if(r[K-1].lson>0) 
  {fseek(in,size*r[K-1].lson,0);
   readw3(w[K-1],Iij[K-1],np,nw,nij,r+K-1,in);
   l++;
  }
  else /* Creation of new elements not yet implemented*/
  {printf(" Arrays needed on lower levels.\n");
   exit(1);
  }
 }
 
 while(l-K+1>level) /* Go up one or more levels */
 { for(i=K-1;i>0;i--)
   { r[i]=r[i-1];
     for(j=0;j<np;j++) for(k=0;k<nw;k++) w[i][j][k]=w[i-1][j][k];
     for(j=0;j<np;j++) for(k=0;k<nij;k++) Iij[i][j][k]=Iij[i-1][j][k];
   }
   fseek(in,size*r[1].parent,0);
   readw3(w[0],Iij[0],np,nw,nij,r,in);
   l--;
 }

 if(t<r[0].t0) {printf(" Error: t<t0 in getw.\n"); exit(1);}
 /* Get correct level */ 
   for(i=0;i<K;i++) if(r[i].level==level) {tag=i; break;}

   while(t>=r[tag].tp) /* Get correct time interval */
   { /* Move tag down to K-1 */
    /*printf(" Next time interval needed.\n");
    printf(" t=%f tp=%f tag=%d\n",t,r[tag].tp,tag);*/
    r[K-1]=r[tag];
    for(j=0;j<np;j++) for(k=0;k<nw;k++) w[K-1][j][k]=w[tag][j][k];
    for(j=0;j<np;j++) for(k=0;k<nij;k++) Iij[K-1][j][k]=Iij[tag][j][k];
    tag=K-1; l=level;
    if(r[tag].nexrec>0)
    {fseek(in,r[tag].nexrec*size,0);
     readw3(w[tag],Iij[tag],np,nw,nij,r+tag,in);
    }
    else
    {printf(" Time horizon too short in brownian tree.\n");
     exit(1);
    }
/*Fill in w and r */
     for(i=K-2;i>=0;i--)
     {fseek(in,r[i+1].parent*size,0);
      readw3(w[i],Iij[i],np,nw,nij,r+i,in);
     }
   } 
/* Return correct dw */
 tt=r[tag].t0; i=0;
 while(tt<t) { i++; tt=r[tag].t0+i*h;}
 /*printf(" i=%d\n",i);*/
 if(i>np-1) {printf(" Error in time.\n"); exit(1);}
 for(j=0;j<nw;j++) dw[j]=w[tag][i][j];
 for(j=0;j<nij;j++) I[j]=Iij[tag][i][j];
/* Check if possible to go up a level
   next time. It is only possible when the
   current time index = 3 mod 4. */
  *cangoup=(i%4==3?1:0);
  return ;
}


getdw4(nw,dw,level,t,h,func,idum,cangoup)
int nw,level,idum,*cangoup;
float *dw,t,h,(*func)();
/* Routine to read and return elements from a Brownian
tree stored in a file.
This version cannot generate new nodes to add to the tree.
It is therefore suitable for a tree that includes the Levy area
integrals.
The external variables wfile and in, that may be used
by other program units, are pointers to the file 
containing the Brownian tree and to the io-unit
connected to the file.
The routine keeps a record in memory at all times,
so that file accesses are not performed during
most calls to the routine.
getdw presumes that it is being used to help solve an sde
forwards through time, ie it presumes that the next call
to getdw will be at the same period or the next one.
*/
{extern char *wfile;
 extern  FILE *in;
 static float **w;
 static record r;
 static header head;
 static int l,np,recsize,init=0;
 static long size;
 float hh,tt,**w2,**matrix();
 int i,j;

 /* printf(" Entering getdw: level=%d t=%f h=%f\n",level,t,h); */
 if(init==0||idum<0) /* Initialize by reading in a record */
 {/* printf(" Initializing getdw: h=%f\n",h); */
  in=fopen(wfile,"r+");
  if(in==NULL) {printf(" Can't open file in getdw\n"); exit(1);}
  recsize=sizeof(record);
  fread((char *)&head,recsize,1,in);
  if(nw!=head.nw) {printf(" Noise vector wrong size in %s.\n",wfile);
                   exit(1);}
  np=head.np;
  /* printf(" nlevels=%d\n",head.nlevels);
  printf(" nrecs=%d\n",head.nrecs1); */
  if(head.nlevels<level)
  {printf(" Not enough levels in initial brownian path\n");
   exit(1);}
  if(init==0) w=matrix(0,np-1,0,nw-1); 
  size=recsize+np*nw*sizeof(float); 
  fseek(in,size,0);
  readw(w,np,nw,&r,in);
  l=r.level;
  init=1;
  /* printf(" Finished initialising\n"); */
 }

 while(level>l) /* Go down one or more levels */
 {if(r.lson>0) 
  {fseek(in,size*r.lson,0);
   readw(w,np,nw,&r,in);
   l++;
  }
  else 
  {printf("Error in getdw4: can't go down. Required step size too small.\n");
   exit(1);
  }
 }
 
 while(level<l) /* Go up one or more levels */
 {if(r.parent>0)
  {fseek(in,size*r.parent,0);
   readw(w,np,nw,&r,in);
   l--;
  }
  else 
  {printf("Error in getdw4: can't go up. Required step size too large\n");
   printf("level=%d l=%d\n",level,l);
   exit(1);
  }
 }

 if(t<r.t0) {printf(" Error: t<t0 in getw.\n"); exit(1);}

   while(t>=r.tp) /* Get correct time interval */
   { 
    if(r.nexrec>0)
    {fseek(in,r.nexrec*size,0);
     readw(w,np,nw,&r,in);
    }
    else
    {printf("Final time too short in Brownian path\n");
     exit(1);
    }
   }

/* Return correct dw */
 tt=r.t0; i=0;
 while(tt<t) { i++; tt=r.t0+i*h;}
 /*printf(" h=%f i=%d\n",h,i);*/
 if(i>np-1) {printf(" Error in time.\n"); exit(1);}
 for(j=0;j<nw;j++) dw[j]=w[i][j];
/* Check if possible to go up a level
   next time. It is only possible when the
   current time index = 3 mod 4. */
  *cangoup=((i%4==3&&level>5)?1:0);

/* printf("In getdw4 l=%d level=%d recno=%d r.level=%d\n",
  l,level,r.recno,r.level); */
}



