/*****************************************************************
* The software in this file was written by Jessica Gaines,       *
* while working at the University of Edinburgh, under the        *
* direction of Terry Lyons, and using funding                    *
* provided by EC grant SC1*CT92-0784.                            *
* You may use this software for educational and academic         *
* research purposes only.                                        *
* Please do not give copies of the software in this file to      *
* anyone else. Tell anyone interested to email                   *
*  jessica@maths.ed.ac.uk                                        *
* or to write to                                                 *
*  Jessica Gaines                                                *
*  University of Edinburgh                                       *
*  Dept. of Mathematics and Statistics                           *
*  JCMB                                                          *
*  Kings Buildings                                               *
*  Mayfield Road                                                 *
*  Edinburgh EH9 3JZ                                             *
*  UK                                                            *
* and then get copies directly from Edinburgh themselves.        *
******************************************************************
*/

/* This file contains routines for generation of random
numbers. Contents:
 ran00, ran0, ran1, ran2, ran3,
 ranvec, snormvec, fnormvec,
 slownorm, fastnorm,
 randexp,cauchy,gamm,
 bnldev,

and also subsidiary routines called by the above:
 findj,tail,wedge : called by fastnorm
 
*/


#include <math.h>
#include <stdio.h>
#include <strings.h>

float ran00(idum)
int idum;
/* Returns a uniform random deviate between 0.0 and 0.1 
   using the system-supplied routine rand(). No shuffling.
   Set idum to any negative value to initialize or
   reinitialize the sequence.*/
{
  static float y,maxran;
  static int iff=0;
  int j; float dum;
  unsigned i,k;
  if (idum < 0 || iff == 0) {
      printf(" Initializing\n");
      iff=1;
      i=2;
      do {
          k=i;
          i<<=1;
         } while (i);
         maxran=k;
         if(idum<0) idum=(-idum);
         srand(idum);
         for (j=1;j<=97;j++) dum=rand();
   }
    y=rand();
   return y/maxran;
}

float ran0(idum)
int idum;
/* Returns a uniform random deviate between 0.0 and 0.1 
   using the system-supplied routine rand() and storing and
   shuffling 97 numbers. Set idum to any negative value to
   initialize or reinitialize the sequence.*/
{
  static float y,maxran,v[98];
  static int iff=0;
  int j; float dum;
  unsigned i,k;
  if (idum < 0 || iff == 0) {
      printf(" Initializing\n");
      iff=1;
      i=2;
      do {
          k=i;
          i<<=1;
         } while (i);
         maxran=k;
         srand(-idum);
         for (j=1;j<=97;j++) dum=rand();
         for (j=1;j<=97;j++) v[j]=rand();
         y=rand();
   }
   j=1+97.0*y/maxran;
   if (j > 97 || j < 1) printf("RAN0: This cannot happen.\n j=%d maxran=%f"
        ,j,maxran);
   y=v[j];
   v[j]=rand();
   return y/maxran;
}

#define M1 259200
#define IA1 7141
#define IC1 54773
#define RM1 (1.0/M1)
#define M2 134456
#define IA2 8121
#define IC2 28411
#define RM2 (1.0/M2)
#define M3 243000
#define IA3 4561
#define IC3 51349

float ran1(idum)
int idum;
/* Returns a uniform random deviate between 0.0 and 1.0. Set idum to any
 * negative value to initialize or reinitialize the sequence.
 * WARNING : no longer in Numerical Recipes, so could be a problem with this.
 */
{static long ix1,ix2,ix3;
 static float r[97];
 float temp;
 static int iff=0;
 int j;

 if(idum<0 || iff==0) /* Initialise on 1st call even if idum>=0 */
 {iff=1;  printf(" Initializing\n");
  ix1=(IC1-idum) % M1; /* seed the 1st routine */
  ix1=(IA1*ix1+IC1) % M1;
  ix2=ix1 % M2;       /* and use it to seed the second */
  ix1=(IA1*ix1+IC1) % M1;
  ix3=ix1 % M3;        /* and 3rd routines */
  for(j=0;j<97;j++) /* Fill table with sequential uniform deviates
                       generated by the 1st two routines */
  {ix1=(IA1*ix1+IC1) % M1;
   ix2=(IA2*ix2+IC2) % M2;
   r[j]=(ix1+ix2*RM2)*RM1; /* Low and high order pieces combined here */
  }
 }
 /* Except when initializing, this is where we start.
    Generate the next number for each sequence.
  */
 ix1=(IA1*ix1+IC1) % M1;
 ix2=(IA2*ix2+IC2) % M2;
 ix3=(IA3*ix3+IC3) % M3;
 j=(97*ix3)/M3; /* Use 3rd sequence to get an integer between 1 and 97 */
 if(j>96||j<0) {printf("Error in ran1\n"); exit(1);}
 temp=r[j];               /* Return that table entry */
 r[j]=(ix1+ix2*RM2)*RM1;  /* and refill it */
 return temp;
}

#define M 714025
#define IA 1366
#define IC 150889

float ran2(idum)
long idum;
/* Returns a uniform random deviate between 0.0 and 1.0.
   Set idum to any negative value to initialize or
   reinitialize the sequence. Uses no system-supplied 
   routines. Shuffles. */
{
  static long iy,ir[98],newdum;
  static int iff=0;
  int j;
  if (idum<0 || iff==0)
   {  iff=1;
      printf(" Initializing\n");
      if ((idum=(IC-(idum))%M)<0) idum=(-(idum));
      for (j=1;j<=97;j++)
        { idum=(IA*(idum)+IC)%M;
          ir[j]=(idum);
        }
      idum=(IA*(idum)+IC)%M;
      iy=newdum=(idum);
   }
  j=1+97.0*iy/M;
  if (j>97||j<1) printf("RAN2: This cannot happen.");
  iy=ir[j];
  newdum=(IA*newdum+IC)%M;
  ir[j]=newdum;
  return (float) iy/M;
}
 

#define MBIG 1000000000
#define MSEED 161803398
#define MZ 0
#define FAC (1.0/MBIG)

float ran3(idum)
int idum;
/* Portable routine to return a uniform random deviate 
 * between 0.0 and 1.0. Set idum to any negative value to
 * initialize or reinitialize the sequence.
 */
{
  static int inext,inextp;
  static long ma[56];
  static int iff=0;
  long mj,mk;
  int i,ii,k;

  if (idum<0||iff==0)
    {iff=1; printf(" Initializing\n");
     mj=MSEED-(idum<0?-idum:idum);
     mj%=MBIG;
     ma[55]=mj;
     mk=1;
     for (i=1;i<=54;i++)
        {ii=(21*i)%55;
         ma[ii]=mk;
         mk=mj-mk;
         if (mk<MZ) mk+=MBIG;
         mj=ma[ii];
        }
     for (k=1;k<=4;k++)
        for (i=1;i<=55;i++)
           {ma[i]-=ma[1+(i+30)%55];
            if (ma[i]<MZ) ma[i]+=MBIG;
           }
     inext=0;
     inextp=31;
    }
  if (++inext == 56) inext=1;
  if (++inextp== 56) inextp=1;
  mj=ma[inext]-ma[inextp];
  if(mj<MZ) mj+=MBIG;
  ma[inext]=mj;
  return mj*FAC;
}

int ran3_(idum)
int idum;
/* Portable routine to return a uniform random deviate 
   between 0.0 and 1.0. Set idum to any negative value to
   initialize or reinitialize the sequence.
   Version to be called from a Fortran program unit.
*/

{
  static int inext,inextp;
  static long ma[56];
  static int iff=0;
  long mj,mk;
  int i,ii,k;

  if (idum<0||iff==0)
    {iff=1; printf(" Initializing\n");
     mj=MSEED-(idum<0?-idum:idum);
     mj%=MBIG;
     ma[55]=mj;
     mk=1;
     for (i=1;i<=54;i++)
        {ii=(21*i)%55;
         ma[ii]=mk;
         mk=mj-mk;
         if (mk<MZ) mk+=MBIG;
         mj=ma[ii];
        }
     for (k=1;k<=4;k++)
        for (i=1;i<=55;i++)
           {ma[i]-=ma[1+(i+30)%55];
            if (ma[i]<MZ) ma[i]+=MBIG;
           }
     inext=0;
     inextp=31;
    }
  if (++inext == 56) inext=1;
  if (++inextp== 56) inextp=1;
  mj=ma[inext]-ma[inextp];
  if(mj<MZ) mj+=MBIG;
  ma[inext]=mj;
  return 1;
}



ranvec(n,x,func) /* gives a vector of n uniform random no.s 
                  in interval [0,1] using routine "func" as
                  specified by calling routine */  
int n;
float x[],(*func)();
{
  int i;
  for (i=0;i<n;i++) x[i]=(*func)(1);
}

snormvec(n,x,sd,mean,func)
 /* gives a vector of n normally distributed deviates with 
   mean=mean and sd=standard deviation, using func(=user 
   supplied function) as the source of uniform deviates 
   and the routine slownorm to normalise the distribution.*/  
int n;
float x[],(*func)(),sd,mean;
{
  int i;
  float slownorm();
  for (i=0;i<n;i++) x[i]=slownorm(mean,sd,func);
}

fnormvec(n,x,sd,mean,func)
 /* gives a vector of n normally distributed deviates with 
   mean=mean and sd=standard deviation, using func(=user 
   supplied function) as the source of uniform deviates 
   and the routine fastnorm to normalise distribution. */  
int n;
float x[],(*func)(),sd,mean;
{
  int i;
  float fastnorm();
  for (i=0;i<n;i++) x[i]=fastnorm(mean,sd,func);
}

float slownorm(mean,sd,func)
float mean,sd,(*func)();
/* Returns a normally distributed deviate with mean=mean and
   sd=standard deviation, using func(=user supplied function)
   as the source of uniform deviates */

{
  static int iset=0;
  static float gset;
  float fac,r,v1,v2;

  if(iset==0)
  { do
      { v1=2.0*(*func)(1)-1.0;
        v2=2.0*(*func)(1)-1.0;
        r=v1*v1+v2*v2;
      } while (r>=1.0);
    fac=sqrt(-2.0*log(r)/r);
    gset=sd*v1*fac+mean;
    iset=1;
    return sd*v2*fac+mean;
  } else
  { iset=0;
    return gset;
  }
}

float randexp(lam,func)
 /* Returns an exponentially distributed random variable:
    p(x)dx=lam*exp(-lam*x)dx, x>=0
    using func as source of uniform variates
 */
float lam,(*func)();
{ return (-1.0/lam * log( (*func)(1) ) ); }

float pi=3.141592653589793;

float cauchy(a,b,func)
 /* Returns a random variable from a Cauchy distribution:
    p(x)dx=(1/pi)*b/((x-a)^2+b^2)
    using func as source of uniform variates
 */
float a,b,(*func)();
{ return (a+b*tan(pi*(*func)(1))); }

float gamm(a,b,func)
 /* Returns a random variable from a Gamma distribution:
    p(x)dx= ( exp(-b*x)*b^a*x^(a-1) )/ gamm(a) dx
    using func as source of uniform variates
 */
float a,b,(*func)();
{int j,ia;
 float am,e,s,v1,v2,x,y,eps=1e-6;
 if(a<1) {printf("Error in gamma: a<1\n"); exit(1);}
 ia=(int)(a+eps);
 if(ia<6 && ia<a+eps && ia>a-eps) /* Direct method: adding waiting times */
 { x=1.0;
   for (j=1;j<=ia;j++) x*=(*func)(1);
   x= (-log(x));
 }
 else
 {do
  {do
   {do
    {v1=2.0*(*func)(1)-1.0;
     v2=2.0*(*func)(1)-1.0;
    }while (v1*v1+v2*v2>1.0);
    y=v2/v1;
    am=ia-1;
    s=sqrt(2.0*am+1.0);
    x=s*y+am;
   }while (x<=0.0);
   e=(1.0+y*y)*exp(am*log(x/am)-s*y);
  }while ((*func)(1)>e);
 }
 return x/b;
}


/* ***************************************************** */
/* The following group of routines consists of fastnorm()
   and those routines called by fastnorm.
*/

#define R1 0.3814697266e-5   /* 2^-18 */
#define R2 0.953674316e-6    /* 2^-20 */
#define R3 0.238418579e-6    /* 2^-22 */
#define R4 0.59604645e-7     /* 2^-24 */
#define MAXRAN 4.294967296e9 /* 2^32  */

 static int A[10]={0,0,0,1,1,2,2,3,4,5},
  B[12]={1,1,1,2,3,3,3,4,6,6,7,8},
  C[17]={0,1,2,2,3,3,4,4,4,6,6,6,7,7,9,9,10};
static double P[13]=
{       0.884912651365847,
        0.894643672548026,
        0.910026545246263,
        0.928783242137085,
        0.944950452666289,
        0.960135597462283,
        0.970662936272365,
        0.981843486103641,
        0.987269804689910,
        0.991486919348447,
        0.994040473529890,
        0.997300203936739,
        1.0
};
static double Q[12]=
{
        0.880834058401424,
        0.885164064747997,
        0.896774888625428,
        0.913824407505834,
        0.930264034831596,
        0.947209250499235,
        0.960305506875539,
        0.974220919528959,
        0.982088312021475,
        0.988221454936694,
        0.992127450599242,
        0.996256397735859
};
static double D[12]=
{
        0.507894519985643,
        0.786281122462607,
        0.901965046237357,
        0.978080431064381,
        0.992530768577612,
        0.979815257323371,
        0.968564825442572,
        0.958250115973873,
        0.948554192970468,
        0.939291236052193,
        0.930346394387841,
        0.921645337293069
};
static double E[12]=
{
        16.0,
        8.0,
        5.333333333333333,
        4.0,
        3.078962214587208,
        2.437680485542790,
        1.995299871621917,
        1.672253526481772,
        1.426416923928201,
        1.233401389609144,
        1.078108473581538,
        0.950690738793861
};

float fastnorm(mean,sd,func)
/* Returns a normally distibuted random deviate with mean=
   mean and standard deviation=sd, using the caller-supplied
   routine "func" as source of uniform deviates. Works faster
   than slownorm. */
float mean,sd,(*func)();
{
  unsigned long u,ubits;
  int j;
  unsigned short sign;
  float x;
  float wedge(),tail();
  double xx;

   x=(*func)(1);
   u=x*MAXRAN;
   sign=(u&0X80000000)>>31; 
   ubits=(u & 0X78000000)>>27;
   if(ubits<10) x=A[ubits]*0.25+((u & 0X7FFFF80)>>7)*R3;
   else {ubits=(u & 0X7E000000)>>25; /* Changed D to E here */
         if(ubits<52)
            x=B[ubits-40]*0.25+((u & 0X1FFFF80)>>7)*R2;
         else {ubits=(u & 0X7F800000)>>23;
               if(ubits<225)
                  x=C[ubits-208]*0.25+((u&0X7FFF80)>>7)*R1;
               else { /* not an easy case */
                     xx=((u&0X7FFFFF80)>>7)*R4;
                     j=findj(xx); 
                     if(j<12) /* wedge or skinny rectangle*/
                       {if(xx<Q[j]) x=0.25*(j+(*func)(j));
                        else x=wedge(j,func);
                       }
                     else x=tail(func);
                     }
                }
          }
    if(sign) x=(-x);
    return sd*x+mean;
}

findj(x) double x; /* Used by fastnorm */
{int j=0;
 while(x>P[j]) j++;
 if(j>12) printf(" Mistake in findj : j>12\n");
 return j;
}

float tail(func) /* Used by fastnorm */
float (*func)();
{
  float fac,r,v1,v2;
  /*printf(" entering tail");*/

    do
      { v1=(*func)(1); 
        v2=(*func)(1); 
        r=v1*v1+v2*v2; 
      } while (r>=1.0);
    fac=sqrt((9.0-2.0*log(r))/r);
    v1=v1*fac; 
    if(v1>=3.0) return v1;
    else { v2=v2*fac; if(v2>=3.0) return v2;
                      else return tail(func);
         }
}

float wedge(j,func) /* Used by fastnorm */
int j; float (*func)();
{
 float x,y,temp,xx;
 x=(*func)(j);
 y=(*func)(j);
 if(x>y) {temp=x; x=y; y=temp;} /* need x<y */
 xx=0.25*(j+x);
 if(y<=D[j]) return xx;
 else
  {temp=x+E[j]*(exp((j+1)*(j+1)*0.03125-xx*xx*0.5)-1.0);
   if(y<=temp) return xx;
   else return wedge(j,func);
  }
}

/*******************************************************/
 
/* The following group of routines consists of fastnorm2()
   and those routines called by fastnorm2.
*/

#define S1 0.3814697266e-5
#define S2 0.59604645e-7  
#define NW 13      /* Number of wedges */
#define NWEDGE 12  /* Number of wedges - 1 */
#define DIMW 25    /* Dimension of wtree: 2*NWEDGE+1 */
#define DIMW1 24   /* Dimension of wtree-1: 2*NWEDGE */

 static float AA[225]={
  0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,
  0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,
  0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,
  0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,
  0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,
  0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25,
  0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25,
  0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25,
  0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25,
  0.25,0.25,0.25,0.25,0.25,
  0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,
  0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,
  0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,
  0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,
  0.75,0.75,0.75,0.75,0.75,0.75,0.75,0.75,0.75,0.75,
  0.75,0.75,0.75,0.75,0.75,0.75,0.75,0.75,0.75,0.75,
  0.75,0.75,0.75,0.75,0.75,0.75,0.75,0.75,0.75,0.75,
  1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,
  1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,
  1.0,1.0,1.0,
  1.25,1.25,1.25,1.25,1.25,1.25,1.25,1.25,1.25,1.25,
  1.25,1.25,1.25,1.25,1.25,1.25,
  1.5,1.5,1.5,1.5,1.5,1.5,1.5,1.5,1.5,1.5,
  1.5,
  1.75,1.75,1.75,1.75,1.75,1.75,
  2.0,2.0,2.0,2.0,
  2.25,2.25,
  2.50};

typedef struct {double p; short int rson,lson;} wedgenode;
static wedgenode wtree[DIMW]=
{
  {1.000000000000000e+00,2,1},
  {9.494713901105239e-01,4,3},
  {1.000000000000000e+00,6,5},
  {9.179213068830829e-01,21,7},
  {9.494713901105239e-01,22,20},
  {9.781397813377510e-01,8,19},
  {1.000000000000000e+00,9,17},
  {8.991646099922610e-01,23,18},
  {9.781397813377510e-01,24,10},
  {1.000000000000000e+00,11,16},
  {9.721333799719041e-01,13,15},
  {1.000000000000000e+00,14,12},
  {9.974464458185570e-01,-1,-1},
  {9.721333799719041e-01,-1,-1},
  {1.000000000000000e+00,-1,-1},
  {9.688736495650551e-01,-1,-1},
  {9.947466497552960e-01,-1,-1},
  {9.893203311690271e-01,-1,-1},
  {8.894335888100819e-01,-1,-1},
  {9.646565349065180e-01,-1,-1},
  {9.340885174122869e-01,-1,-1},
  {9.179213068830829e-01,-1,-1},
  {9.494713901105239e-01,-1,-1},
  {8.991646099922610e-01,-1,-1},
  {9.781397813377510e-01,-1,-1}
};

float fastnorm2(mean,sd,func)
/* Returns a normally distibuted random deviate with mean=
   mean and standard deviation=sd, using the caller-supplied
   routine "func" as source of uniform deviates. Works faster
   than slownorm. */
float mean,sd,(*func)();
{
  unsigned long u,ubits;
  int j;
  unsigned short sign;
  float x;
  float wedge2(),tail2();
  double xx;

   u=(*func)(1)*MAXRAN;
   sign=(u&0X80000000)>>31; 
   if((ubits=(u & 0X7F800000)>>23)<225)
        x=AA[ubits]+((u&0X7FFF80)>>7)*S1; 
   else { /* not an easy case */
         xx=((u&0X7FFFFF80)>>7)*S2;
         j=choose_wedge(xx); 
         if(j<12) /* wedge or skinny rectangle */
         {if(xx<Q[j]) x=0.25*(j+(*func)(j));
          else x=wedge(j,func);
         }
         else x=tail(func);
        }
    if(sign) x=(-x);
    return sd*x+mean;
}

choose_wedge(x) double x;
{int i,rson,lson;
 lson=wtree[0].lson;
 rson=wtree[0].rson;
/* wtime++; */
 while(lson>=0)
 {if(x<=wtree[lson].p) 
   {i=lson; lson=wtree[i].lson; rson=wtree[i].rson;}
  else                
   {i=rson; lson=wtree[i].lson; rson=wtree[i].rson;}
/*  wtime++; */
 }
 if(i<NWEDGE) {printf(" Mistake in choose_wedge: i=%d\n",i);
                exit(1);
               }
 i=DIMW1-i; /* in wtree the wedges are in the 2nd half and in reverse
              order */
/* printf(" Wedge %d\n",i); */
 return i;
}

/*******************************************************/


