/***************************************************************
 * stoparams.c                                                 *
 ***************************************************************

 ************************************************************************
 * The programming for this Mathematica module was developed through a  *
 * generous contribution to the UCSD Mathematics Department from the    *
 * GenCorp Foundation of Aerojet, a corporate donor committed to        *
 * supporting educational improvement.  It was written by David         *
 * Glickenstein under the supervision of Ruth Williams at University of *
 * California, San Diego.  The author disclaims any responsibility for  *
 * problems experienced by users as a result of using this software.    *
 * The software may not be distributed without written permission from  *
 * the author.  For more information contact:                           *
 *     David Glickenstein       or   Ruth Williams                      *
 *     glicken @math.ucsd.edu        williams@math.ucsd.edu             *
 *   at:       UCSD Mathematics Department                              *
 *             9500 Gilman Drive 0112                                   *
 *             La Jolla, CA 92093-0112                                  *
 *                                                                      *
 * The module uses C functions developed by by Jessica Gaines, while    *
 * working at the University of Edinburgh, under the direction of Terry *
 * Lyons, and using funding provided by EC grant SC1*CT92-0784.  The    *
 * software is to be used for educational and academic research         *
 * purposes only.  Anyone interested in more information on these       *
 * C functions should email jessica@maths.ed.ac.uk or to write to       *
 *        Jessica Gaines                                                * 
 *        University of Edinburgh                                       *
 *        Dept. of Mathematics and Statistics                           *
 *        JCMB                                                          *
 *        Kings Buildings                                               *
 *        Mayfield Road                                                 *
 *        Edinburgh EH9 3JZ UK                                          *
 *  and then get original copies directly from Edinburgh themselves.    *
 ************************************************************************/

/***************************************************************
 *  This file is used to specify the parameters for the        *
 *  program "stosolve."  Immediately following are             *
 *  instructions.  DO NOT change any of the assignments which  *
 *  are not explained.  This file contains some code which     *
 *  is necessary for interfacing with the rest of the program. *
 *  All variables and functions which can be changed are       *
 *  shown in parentheses next to an explanation of what they   *
 *  are.                                                       *
 ***************************************************************

 ***************************************************************
 *  Instructions:                                              *
 *    1. Each of the paramters which you can change are        *
 *       preceded by boxes of ****** which explain what the    *
 *       variable or function represents.  You should adjust   *
 *       each value appropriately, paying attention to the     *
 *       restrictions on each variable.                        *
 *                                                             *
 *    2. Do not change anything which is not explained.  This  *
 *       file contains declarations, flow control statements,  *
 *       and other parts which need to remain the way they are *
 *       for the program to function properly.                 *
 *                                                             *
 *    3. The coordinates of a vector v are specified using     *
 *       square brackets like v[2] and are indexed starting    *
 *       with 0, so if v is a 3 dimensional vector, its        *
 *       coordinates are v[0], v[1], and v[2].  Two            *
 *       dimensional arrays are specified similarly, for       *
 *       instance w[1][3].  It is important to understand      *
 *       what each index represents when assigning values.     *
 *                                                             *
 *    4. When specifying functions, be sure to use parentheses *
 *       and always type * for multiplication (unlike          *
 *       mathematica, in C it is necessary to specify          *
 *       multiplication explicitly).                           *
 *                                                             *
 *    5. All lines should end in a semicolon (;) to tell C     *
 *       that it is the end of a line.                         *
 *                                                             *
 *    6. Some variable assignments have a * before it (for     *
 *       instance, when we assign NX, the statement is         *
 *       *NX = 2;).  Do not change these *'s, since they are   *
 *       important, but do not add any * if it was not already *
 *       there.                                                *
 *                                                             *
 *    7. The end of the file contains specifications for each  *
 *       path.  You do not need to change information for      *
 *       for any paths other than the first NPATH sets (NPATH  *
 *       is the number of paths which you are generating).     *
 *                                                             *
 *    8. When all values are set, you must save the file       *
 *       before running the next line in Mathematica (which    *
 *       should be "!make").  You can save either by moving    *
 *       the mouse to the "Files" menu and choosing "Save      *
 *       Buffer" or by pressing <Control>-x <Control>-s.       *
 *                                                             *
 *    9. The program uses noise paths which must already be    *
 *       generated by the program brown3 and stored in the     *
 *       file "bpaths."                                        *
 ***************************************************************/

#include "stosolve.h"


void initparams(NX, NPATH, endpt, x0)
  int *NX, *NPATH;
  float *endpt, x0[NMAX][DIMX];
{
/***************************************************************
 * Number of equations in the system ( NX )                    *
 *   This integer must be less than 8.                         *
 ***************************************************************/

  *NX = 2;

/***************************************************************
 * Number of starting points ( NPATH )                         *
 *   This integer is the number of paths which will be         *         
 *   solved.  It must be less than 5.                          *
 ***************************************************************/

  *NPATH = 2;

/***************************************************************
 * Final time ( endpt )                                        *
 *   The program begins at time 0 and ends at this floating    *
 *   point number.  The number of steps between 1 unit of      *
 *   time is specified by the variable nsteps below and can    *
 *   be different for each path.                               *
 *                                                             *
 *   The value must be between 0 and 100.                      *
 ***************************************************************/

  *endpt = 25.0;

/***************************************************************
 * Starting points ( x0[][] )                                  *
 *   The two dimensional array x0[i][j] refers to the initial  *
 *   point of path number i-1 and dimension j-1.               *
 *                                                             *
 *   There first coordinate should go from 0 to NPATHS-1 and   *
 *   the second coordinate should go from 0 to NX-1 (the       *
 *   number of equations minus one).  You must specify all     *
 *   values of x0[i][j] in those ranges.                       *
 ***************************************************************/

  x0[0][0] = -1.9;
  x0[0][1] = 1.2;
  x0[1][0] = -1.9;
  x0[1][1] = 1.2;

}


int a(x, t, A, nx)
  float  x[], t, A[];
  int    nx;
{
/***************************************************************
 * Stratonovich drift  ( A[] )                                 *
 *   The array A[i] where A[i] is the drift of equation i+1 as *
 *   a function of x[j] (the solution vector) and t (time).    *
 *                                                             *
 *   There should be nx of these, i.e. the same as the number  *
 *   of equations.                                             *
 ***************************************************************/

  A[0] = 1.0/x[0];


  return(0);
}


int b(x, t, B, nx)
  float  x[], t, B[];
  int    nx;
{

/***************************************************************
 * Ito drift ( B[] )                                           *
 *   The array B[i] where B[i] is the drift of equation i+1 as *
 *   a function of x[j] (the solution vector) and t (time).    *
 *                                                             *
 *   There should be nx of these, i.e. the same as the number  *
 *   of equations.                                             *
 ***************************************************************/
 
  B[0] = 3.0*(x[0]+x[1] - (1.0/3.0)*(x[0]*x[0]*x[0]) - 0.34);
  B[1] = (-1.0/3.0)*(x[0] + 0.8*x[1] - 0.7);


  return(0);
}


int sig(x, t, w, S, nx, nw)
  float  x[], t, w[], S[];
  int    nx, nw;
{

/***************************************************************
 * Diffusion  ( S[] )                                          *
 *   The array S[i] where S[i] is the diffusion of equation    *
 *   i+1 as a function of x[j] (the solution vector), t        *
 *   (time), and w[k] (a vector of noise terms, w[k] being the *
 *   kth noise vector, and we have a total of 4 to work with). *
 *   Actually, S is the product of the matrix sig(x,t) and     *
 *   the vector dw.                                            *  
 *                                                             *
 *   There should be nx of these, i.e. the same as the number  *
 *   of equations.                                             *
 ***************************************************************/

  S[0] = .1*w[0];
  S[1] = 0;


  return(0);
}



/***************************************************************
 * The following options are done for each path.  You only     *
 * need to worry about changing the first NPATH sets of data   *
 * since those are the only ones seen by the program.          *
 ***************************************************************/

void getpathparams(path, funcno, nsteps)
  int path, *funcno, *nsteps;
{
  switch(path) {

    case 0: {                                                        
                                                               
    /*********************
     *   First Path      *
     *********************/

/***************************************************************
 * Choose integration method ( funcno )                        *
 *   Be sure to use the Ito (b) or Stratonovich (a) drift      *
 *   according to which the following solution methods         *
 *   uses.                                                     *
 *                                                             *
 *   The choices for funcno with solution type are:            *
 *        1 Euler (Ito)                                        *
 *        2 Heun  (Stratonovich)                               *
 *        3 Runge-Kutta (Stratonovich)                         *
 ***************************************************************/

    *funcno = 1;

/***************************************************************
 * Number of steps used in 1 unit of time ( nsteps )           *
 *   The time step is 2^(-nsteps)                              *
 *                                                             *
 *  The value must be between 3 and 10 (or the noise must be   *
 *  generated again with brown3).                              *
 ***************************************************************/

    *nsteps = 6;


    return;
  }
    

    case 1: {

    /*********************
     *   Second Path     *
     *********************/

/***************************************************************
 * Choose integration method ( funcno )                        *
 *   Be sure to use the Ito (b) or Stratonovich (a) drift      *
 *   according to which the following solution methods         *
 *   uses.                                                     *
 *                                                             *
 *   The choices for funcno with solution type are:            *
 *        1 Euler (Ito)                                        *
 *        2 Heun  (Stratonovich)                               *
 *        3 Runge-Kutta (Stratonovich)                         *
 ***************************************************************/

    *funcno = 1;

/***************************************************************
 * Number of steps used in 1 unit of tim ( nsteps )            *
 *   The time step is 2^(-nsteps)                              *
 *                                                             *
 *  The value must be between 3 and 10 (or the noise must be   *
 *  generated again with brown3).                              *
 ***************************************************************/

    *nsteps = 6;


    return;
    }


    case 2: {

    /*********************
     *   Third Path      *
     *********************/

/***************************************************************
 * Choose integration method ( funcno )                        *
 *   Be sure to use the Ito (b) or Stratonovich (a) drift      *
 *   according to which the following solution methods         *
 *   uses.                                                     *
 *                                                             *
 *   The choices for funcno with solution type are:            *
 *        1 Euler (Ito)                                        *
 *        2 Heun  (Stratonovich)                               *
 *        3 Runge-Kutta (Stratonovich)                         *
 ***************************************************************/

    *funcno = 1;

/***************************************************************
 * Number of steps used in 1 unit of time ( nsteps )           *
 *   The time step is 2^(-nsteps)                              *
 *                                                             *
 *  The value must be between 3 and 10 (or the noise must be   *
 *  generated again with brown3).                              *
 ***************************************************************/

    *nsteps = 4;


    return;
    }


    case 3: {

    /*********************
     *   Fourth Path     *
     *********************/

/***************************************************************
 * Choose integration method ( funcno )                        *
 *   Be sure to use the Ito (b) or Stratonovich (a) drift      *
 *   according to which the following solution methods         *
 *   uses.                                                     *
 *                                                             *
 *   The choices for funcno with solution type are:            *
 *        1 Euler (Ito)                                        *
 *        2 Heun  (Stratonovich)                               *
 *        3 Runge-Kutta (Stratonovich)                         *
 ***************************************************************/

    *funcno = 1;

/***************************************************************
 * Number of steps used in 1 unit of time ( nsteps )           *
 *   The time step is 2^(-nsteps)                              *
 *                                                             *
 *  The value must be between 3 and 10 (or the noise must be   *
 *  generated again with brown3).                              *
 ***************************************************************/

    *nsteps = 4;


    return;
    }


    case 4: {

    /*********************
     *   Fifth Path      *
     *********************/

/***************************************************************
 * Choose integration method ( funcno )                        *
 *   Be sure to use the Ito (b) or Stratonovich (a) drift      *
 *   according to which the following solution methods         *
 *   uses.                                                     *
 *                                                             *
 *   The choices for funcno with solution type are:            *
 *        1 Euler (Ito)                                        *
 *        2 Heun  (Stratonovich)                               *
 *        3 Runge-Kutta (Stratonovich)                         *
 ***************************************************************/

    *funcno = 1;

/***************************************************************
 * Number of steps used in 1 unit of time ( nsteps )           *
 *   The time step is 2^(-nsteps)                              *
 *                                                             *
 *  The value must be between 3 and 10 (or the noise must be   *
 *  generated again with brown3).                              *
 ***************************************************************/

    *nsteps = 4;


    return;
    }
  }

}



