/************************************************************************
 *   stosolve2.c                                                        *
 ************************************************************************

 ************************************************************************
 * The programming for this Mathematica module was developed through a  *
 * generous contribution to the UCSD Mathematics Department from the    *
 * GenCorp Foundation of Aerojet, a corporate donor committed to        *
 * supporting educational improvement.  It was written by David         *
 * Glickenstein under the supervision of Ruth Williams at University of *
 * California, San Diego.  The author disclaims any responsibility for  *
 * problems experienced by users as a result of using this software.    *
 * The software may not be distributed without written permission from  *
 * the author.  For more information contact:                           *
 *     David Glickenstein       or   Ruth Williams                      *
 *     glicken @math.ucsd.edu        williams@math.ucsd.edu             *
 *   at:       UCSD Mathematics Department                              *
 *             9500 Gilman Drive 0112                                   *
 *             La Jolla, CA 92093-0112                                  *
 *                                                                      *
 * The module uses C functions developed by by Jessica Gaines, while    *
 * working at the University of Edinburgh, under the direction of Terry *
 * Lyons, and using funding provided by EC grant SC1*CT92-0784.  The    *
 * software is to be used for educational and academic research         *
 * purposes only.  Anyone interested in more information on these       *
 * C functions should email jessica@maths.ed.ac.uk or to write to       *
 *        Jessica Gaines                                                * 
 *        University of Edinburgh                                       *
 *        Dept. of Mathematics and Statistics                           *
 *        JCMB                                                          *
 *        Kings Buildings                                               *
 *        Mayfield Road                                                 *
 *        Edinburgh EH9 3JZ UK                                          *
 *  and then get original copies directly from Edinburgh themselves.    *
 ************************************************************************/

/* This is an edited version of an example program designed by
 * Jessica Gaines for use with her stochastic differential equation
 * package.  Her notes follow.
 *
 * This program uses another file called stoparams2.c which contains
 * the data which would need to be changed to study these systems.
 *
 *                David Glickenstein
 */

/*
 * This program solves an N-dimensional stoch. diff. equation numerically
 * forwards through time, using  any of eulerN, heunN, runkut4N, heun_iterN
 * or milshteinN. The method uses a fixed step size, but reads in a Brownian
 * path previously generated and stored in a file (using brown3),
 * or creates the Brownian path if the file is empty. (This
 * allows for comparison with solutions generated by diff2qc.c using the same
 * Brownian path.) The program has two arguments: 1) The name of the file for
 * storing the solution(s). 2) The name of the file containing the Brownian
 * path.
 *
 * To change the sde, edit the routines a(),b(),sig(),dssig().
 * For N>2 change DIMX, DIMW.
 */

#include <stdio.h>
#include <math.h>
#include <string.h>
#include "stosolve2.h"

typedef struct
{
  float     x, y;
} point;
char     *wfile;
FILE     *in;
float     beta;
int       count;
char     *menuitems[FMAX] = {"euler", "heun",
 "runge-kutta", "heun-iter", "milshtein"};

int main(argc, argv)
  int argc; 
  char *argv[];
{
  return stosolve(argc, argv);
}


int stosolve()

{
  float     x[DIMX], h, t, dw[DIMW], x0[NMAX][DIMX];
  float     xscal[DIMX], ran3(), w = 0.0, endpt, eps;
  int       mm, n, i, j,  K[NMAX], seed = -99, JUMP, NX, NW, NPATH,
            error, ascii, funcno;
  int       nsteps, len, a(), sig(), dssig(), b(), heunN(), readp = 0,
            NK = 1, k;
  FILE     *out, *fopenw(), *in2, *fopenr();
  char      banner[1000], num[10];

  out = fopenw("soln");
  wfile = "bpaths1";
  printf(" Brownian paths in %s.\n", wfile);
  ascii= 1; /* getans(" Store paths in ascii or binary file?",'a','b');*/

  initparams(&NX, &NPATH, &endpt, x0, &seed);
  NW = 4;  /* To change this, we need to alter the bpaths file by
	      generating a new noise file with brown3  */

  ran3(seed);
  for (k = 0; k < NK; k++)
  {
    for (i = 0; i < NPATH; i++)
    {
      sprintf(wfile,"bpaths%d",i+1);
      if (i > 0)
	fprintf(out, "  :  \n");
      getpathparams(&funcno, &nsteps);
      nsteps++;
      if (funcno == 4)
	getfloat(&eps, "epsilon (accuracy in heun-iter)");
      JUMP = 1;
      for (n = 0; n < NX; n++)
      {
	x[n] = x0[0][n];
	xscal[n] = 1.0;
      }
      K[i] = 1;
      count = 0;
      t = w = 0.0;
      h = 1.0 / (1 << (nsteps - 1));
      mm = n = 0;
      /* Initialize getdw */
      my_new_getdw(NW, dw, nsteps, t, h, ran3, -1, &len, 512);
      /* OR getdw(NW, dw, nsteps, t, h, ran3, -1, &len); */
      while (t < endpt && mm < KMAX)
      {
	n++;
	my_new_getdw(NW, dw, nsteps, t, h, ran3, 1, &len, 512);
	/* OR getdw(NW, dw, nsteps, t, h, ran3, 1, &len); */
	w += dw[0];
	if (funcno == 1)
	  error = eulerN(x, h, dw, t, NX, NW, b, sig);
	else if (funcno == 2)
	  error = heunN(x, h, dw, t, NX, NW, a, sig);
	else if (funcno == 3)
	  error = runkut4N(x, h, dw, t, NX, NW, a, sig);
	else if (funcno == 4)
	  error = heun_iterN(x, h, dw, t, NX, NW, a, sig, eps);
	else if (funcno == 5)
	  error = milshteinN(x, h, dw, t, NX, NW, b, sig, dssig);
	if (error)
	{
	  printf(" Numerical error.\n");
	  break;
	}
	t += h;
	count++;
	if (n % JUMP == 0)
	{
	  mm++;
	  K[i]++;
	}
	fprintf(out, "%f ", t);
	for (j=0; j<NX; j++)
	  fprintf(out, "%f ", x[j]);
	fprintf(out, "\n");
      }
      printf(" Approximate x(1)=%f\n", x[0]);
      printf(" Approximate x(2)=%f\n", x[1]);
      printf(" Step size=%f No. of steps=%d\n", h, count);

      /* Close file with Brownian paths */
      fclose(in);
    }
  }
  fclose(out);
}


dssig(x, t, dw2, SS, nx, nw) 
  float     x[], t, **dw2, SS[];
  int       nx, nw;
/* 0.5*sig*d(sig)/dx  */
/* SS[i] = d(sig(i,j))/dx(m) * sig(m,k) * dw2[j][k] */
{
  extern float beta;
  SS[0] = beta * beta * x[0] * dw2[0][0];
  return (0);
}





