(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 3.0,
MathReader 3.0, or any compatible application. The data for the notebook 
starts with the line of stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     71669,       3392]*)
(*NotebookOutlinePosition[     89636,       4068]*)
(*  CellTagsIndexPosition[     89592,       4064]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Stochastic Differential Equations", "Title",
  FontColor->RGBColor[1, 0, 0]],

Cell[TextData[StyleBox["Stochastic Flows",
  FontColor->RGBColor[0, 0, 1]]], "Subtitle"],

Cell[TextData[{
  "This notebook is used to approximate solutions to stochastic differential \
equations.  There are two version of this notebook.  This version fixes the \
noise data so that it is easy to look at the flow by varying the initial \
conditions.   The other version (in a notebook called \"stosolve2.nb\" and \
found in a separate directory called \"/noise\") evaluates the same system \
with different noise data.  The program uses a C library to do the numerical \
approximation and ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " as a front end and for plotting and doing statistics with the results."
}], "Text"],

Cell[CellGroupData[{

Cell[TextData[StyleBox["Acknowledgements",
  FontSlant->"Italic"]], "Section"],

Cell["\<\
The programming for this Mathematica module was done by David \
Glickenstein under the supervision of Ruth Williams.  It was developed \
through a generous contribution to the UCSD Mathematics Department from the \
GenCorp Foundation of Aerojet, a corporate donor committed to supporting \
educational improvement.\
\>", "Text"],

Cell["\<\
The underlying C software written for working with the stochastic \
differential equations was written by Jessica Gaines, while working at the \
University of Edinburgh, under the direction of Terry Lyons, and using \
funding provided by EC grant SC1*CT92-0784.  The software is to be used for \
educational and academic research purposes only.  Anyone interested in more \
information should email jessica@maths.ed.ac.uk or to write to Jessica \
Gaines, University of Edinburgh, Dept. of Mathematics and Statistics, JCMB, \
Kings Buildings, Mayfield Road, Edinburgh EH9 3JZ UK  and then get original \
copies directly from Edinburgh themselves.  \
\>", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["Disclaimer",
  FontSlant->"Italic"]], "Section"],

Cell["\<\
The author disclaims any responsibility for problems experienced by \
users as a result of using this software.  The software may not be \
distributed without written permission from the author.  For more information \
contact:
David Glickenstein\t\tor\t\tRuth Williams
glicken @math.ucsd.edu\t\t\twilliams@math.ucsd.edu

at:\tUCSD Mathematics Department
\t9500 Gilman Drive 0112
\tLa Jolla, CA  USA  92093-0112
\
\>", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Introduction", "Section"],

Cell[TextData[{
  "This notebook is used to approximate the Ito solution of ",
  Cell[BoxData[
      \(TraditionalForm
      \`dX\  = \ \(b(X, t)\)\ dt\  + \ \(\[Sigma](X, t)\) dW\)]],
  " where ",
  Cell[BoxData[
      \(TraditionalForm\`b\)]],
  " is the Ito drift, ",
  Cell[BoxData[
      \(TraditionalForm\`\[Sigma]\)]],
  " is the dispersion coefficient, ",
  Cell[BoxData[
      \(TraditionalForm\`\(t\ \)\)]],
  "is time, and ",
  Cell[BoxData[
      \(TraditionalForm\`W\)]],
  " is the noise, and ",
  Cell[BoxData[
      \(TraditionalForm\`X\)]],
  " is a vector (",
  Cell[BoxData[
      \(TraditionalForm\`b\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`\[Sigma]\)]],
  " are matrix valued).  It also can approximate Stratonovich solutions of ",
  Cell[BoxData[
      \(TraditionalForm
      \`dX\  = \ \(a(X, t)\) dt\  + \ \(\[Sigma](X, t)\)\[SmallCircle]dW\)]],
  ".  It uses a combination of C and ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  "."
}], "Text"],

Cell[TextData[{
  "It is important to be careful when entering input, since C is slightly \
different from ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  ".  Make sure to read the directions in each section, as they are a \
guideline as to how to use the program.  When working with vectors, remember \
that the jth element of a vector v is v[j] in C and v[[j]] in ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  "."
}], "Text"],

Cell["\<\
The C library was entirely developed by Jessica Gaines and the \
programs are largely based upon her example programs.\
\>", "Text"],

Cell["\t\t\tDavid Glickenstein, UCSD", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Initialization", "Section"],

Cell[TextData[{
  "These initialization routines must be run each time a new ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " kernel is run, but the code should not be altered at all."
}], "Text"],

Cell[CellGroupData[{

Cell["Load Packages", "Subsubsection"],

Cell[BoxData[
    \(<< Graphics`Graphics3D`\)], "Input",
  CellLabel->"In[1]:="]
}, Closed]],

Cell[CellGroupData[{

Cell["Turn off unnecessary warnings", "Subsubsection"],

Cell[BoxData[
    \(\(Off[General::spell1]; \)\)], "Input",
  CellLabel->"In[2]:="]
}, Closed]],

Cell[CellGroupData[{

Cell["Set colors", "Subsubsection"],

Cell[BoxData[
    \(Blue\  = \ RGBColor[0, 0, 1]; \nGreen\  = \ RGBColor[0, 1, 0]; \n
    Red\  = \ RGBColor[1, 0, 0]; \nYellow\  = \ RGBColor[1, 1, 0]; \n
    Cyan\  = \ RGBColor[0, 1, 1]; \nMagenta\  = \ RGBColor[1, 0, 1]; \n
    White\  = \ RGBColor[1, 1, 1]; \nBlack\  = \ RGBColor[0, 0, 0]; \)], 
  "Input",
  CellLabel->"In[3]:="]
}, Closed]],

Cell[CellGroupData[{

Cell["Define function PathGraph2D", "Subsubsection"],

Cell[BoxData[
    \(\(PathGraph2D\  = \ 
      Function[\n\t\t{l, pathnumber, \ varx, \ vary, \ color}, \n\t\t
        If[\((pathnumber\  <= \ 0)\)\  || \ \((pathnumber\  > \ Length[l])\), 
          \n\t\t\tPrint["\<Error: There are \>", \ Length[l], "\< paths.\>"], 
          \ \n\t\t\t
          If[\((Min[varx, vary]\  <= \ 0)\)\  || 
              \((Max[varx, \ vary]\  > \ 
                  Length[\(l[\([pathnumber]\)]\)[\([1]\)]])\)\ , \n
            \t\t\t\t\t\t
            Print["\<Error: There are \>", \ 
              Length[\(l[\([pathnumber]\)]\)[\([1]\)]], \ "\< variables.\>"], 
            \n\t\t\t\t\t
            ListPlot[
              Map[Function[q, \ q[\([{varx, \ vary}]\)]], \n\t\t\t\t\t\t\t\ 
                l[\([pathnumber]\)]], \n\t\t\t\t\t\tPlotJoined -> True, \ \n
              \t\t\t\t\t\tPlotStyle\  -> \ color, \n\t\t\t\t\t\t
              DisplayFunction -> Identity]]]]; \)\)], "Input",
  CellLabel->"In[4]:="]
}, Closed]],

Cell[CellGroupData[{

Cell["Define function PathGraph3D", "Subsubsection"],

Cell[BoxData[
    \(\(PathGraph3D\  = \ 
      Function[{l, \ pathnumber, \ varx, \ vary, \ varz, \ color}, \ \t\t
        If[\((pathnumber\  <= \ 0)\)\  || \ \((pathnumber\  > \ Length[l])\), 
          \n\t\t\tPrint["\<Error: There are \>", \ Length[l], "\< paths.\>"], 
          \ \n\t\t\t
          If[\((Min[varx, vary, \ varz]\  <= \ 0)\)\  || 
              \((Max[varx, \ vary, \ varz]\  > \ 
                  Length[\(l[\([pathnumber]\)]\)[\([1]\)]])\)\ , \n
            \t\t\t\t\t\t
            Print["\<Error: There are \>", \ 
              Length[\(l[\([pathnumber]\)]\)[\([1]\)]], \ "\< variables.\>"], 
            \n\t\t\t\t
            ScatterPlot3D[
              Map[Function[q, \ q[\([{varx, \ vary, \ varz}]\)]], \ 
                l[\([pathnumber]\)]], \ \n\t\t\t\t\tPlotJoined -> True, \ \n
              \t\t\t\t\tPlotStyle -> color, \ \n\t\t\t\t\t
              DisplayFunction -> Identity]]]]; \)\)], "Input",
  CellLabel->"In[12]:="]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Open parameters file", "Section"],

Cell["\<\
We first need to open the file which contains the parameters which \
determine the system to solve and how we go about solving it.  The file is \
called \"stoparams.c\" and is a C file.  It must be in your current \
directory.  The command I included here opens the file using emacs, but it is \
not important which text editor is used.  The next step is to edit this file \
appropriately and MAKE SURE TO SAVE  the file before continuing.\
\>", "Text"],

Cell[BoxData[
    \(\(! "\<emacs stoparams.c&\>"\)\)], "Input",
  CellLabel->"In[1]:="]
}, Closed]],

Cell[CellGroupData[{

Cell["Compile and run stosolve", "Section"],

Cell[CellGroupData[{

Cell["\<\
This command compiles the C code with the parameters you have just \
entered.  Any time you change the parameters be sure to save the file and \
then run this command.  It is important that \"makefile\" be in the current \
directory.\
\>", "Text"],

Cell[BoxData[
    \(\(! make\)\)], "Input",
  CellLabel->"In[38]:="]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
This command runs the C program.  It reads noise from the (binary) \
file \"bpaths\" and stores the solutions in a (text) file called \
\"soln.\"\
\>", "Text"],

Cell[BoxData[
    \(\(! stosolve\)\)], "Input",
  CellLabel->"In[39]:="]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Graph paths in 2D and 3D", "Section"],

Cell[CellGroupData[{

Cell["Load solutions", "Subsection"],

Cell[CellGroupData[{

Cell["\<\
We need to create a list of the points in the solution paths.  This \
produces list (called p) of each path, reading from the file \"soln.\"  Each \
path consists of  a list of points.  Each point has first coordinate time and \
then the coordinates of the path in order.\
\>", "Text"],

Cell[BoxData[
    \(\(\nsolnfile\  = \ OpenRead["\<soln\>"]; \nml\  = \ {}; \n
    r\  = \ empty; \n
    While[r\  =!= \ EndOfFile, \n\tr\  = \ empty; \n\tl\  = \ {}; \n\t
      While[\((r\  =!= \ EndOfFile)\)\  && \ \((r\  =!= \ "\<:\>")\), \n\t\t
        r\  = \ Read[solnfile, \ Word]; \n\t\t
        If[\((r\  === \ "\<:\>")\)\  || \ \((r\  === \ EndOfFile)\), \ 
          ml\  = \ Append[ml, l], l\  = \ Append[l, ToExpression[r]]]\n\t]]; 
    \nClose[solnfile]; \np\  = \ {}; \n
    For[j\  = \ 1, \ j\  <= \ Length[ml], \ \(j++\), \n\tq\  = \ {}; \n\t
      For[i\  = \ 1, \ i < \ Length[ml[\([j]\)]], \ i\  += 3, \ 
        q\  = \ Append[q, \ \(ml[\([j]\)]\)[\([{i, \ i + 1, \ i + 2}]\)]\ ]]; 
      \n\t\tp\  = \ Append[p, q]]; \)\)], "Input",
  CellLabel->"In[7]:="]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Introduction to graphing commands", "Subsection"],

Cell[CellGroupData[{

Cell["PathGraph2D, PathGraph3D", "Subsubsection"],

Cell["\<\
PathGraph2D and PathGraph3D graph the paths in 2D or 3D and have \
the following options:
PathGraph2D[l, pathnumber, varx, vary, color], 
Path3D[l, pathnumber, varx, vary, varz, color]
\t1. The parameter l is the list of paths which are loaded above in the \
\"Load solutions\" section.
\t2. The parameter pathnumber is the number of the path which you want to \
graph (e.g. 1 is the first path).
\t3. The parameters varx, vary, varz are the numbers of the coordinates you \
want to graph, with 1 being time, 2 being the first coordinate of the \
solution, 3 being the second coordinate in the solution, etc. (e.g. for the \
phase space of the first two solution coordinates, we use varx = 2 and vary = \
3).
\t4. The parameter color is the color you want the path.  The options are \
Red, Blue, Green, Yellow, Cyan, Magenta, White, and Black.  You can also \
specify RGB values by setting the color to be RGBColor[a,b,c] where a,b,c are \
the RGB values.\
\>", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Show", "Subsubsection"],

Cell["\<\
The Show command displays the graphs.  The first parameters are the \
graphs (generated with PathGraph2D and PathGraph3D) which you want to show on \
the same axes.  The next parameters are options.  Among the options you must \
specify DisplayFunction->$DisplayFunction or the graphs will not be \
displayed.  
When viewing 3D graphs, it is often advisable to include the option \
ViewPoint->[a,b,c] where [a,b,c] is the point from which you are viewing the \
graph.  As this point is not at all obvious, the best way to use this option \
is to go to the Input Menu and select  \"3D ViewPoint selector\" (or \
<Shift>-<Mod1>-V).  This window allows you to select your viewpoint by the \
mouse.  Then press the Paste button to copy the viewpoint to the notebook and \
use it in the Show command.\
\>", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Some Tips and Shortcuts", "Subsubsection"],

Cell[TextData[{
  "Note that graphics objects can be assigned to variables, so the command \n\
\tgraph1 = PathGraph2D[p, 1, 1, 2, Blue];\nmakes perfect sense to ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " and allows you now to run commands like Show[graph1, \
DisplayFunction->$DisplayFunction].\nIf you want to show the same thing as \
you displayed in the previous command (perhaps with a change in ViewPoint or \
some other option), you can use the command Show[%, \
DisplayFunction->$DisplayFunction] (and you can change or add options) and \
the % simply takes same graphs that were shown last time."
}], "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Graph", "Subsection"],

Cell[CellGroupData[{

Cell["An example of graphing 2 paths in 2D.", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[PathGraph2D[p, 1, 1, 3, Blue], \ PathGraph2D[p, 1, 1, 2, Red], 
      DisplayFunction -> $DisplayFunction]; \)\)], "Input",
  CellLabel->"In[54]:="],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0380952 0.281946 0.151494 [
[.21429 .26945 -3 -9 ]
[.21429 .26945 3 0 ]
[.40476 .26945 -6 -9 ]
[.40476 .26945 6 0 ]
[.59524 .26945 -6 -9 ]
[.59524 .26945 6 0 ]
[.78571 .26945 -6 -9 ]
[.78571 .26945 6 0 ]
[.97619 .26945 -6 -9 ]
[.97619 .26945 6 0 ]
[.01131 .13045 -12 -4.5 ]
[.01131 .13045 0 4.5 ]
[.01131 .43344 -6 -4.5 ]
[.01131 .43344 0 4.5 ]
[.01131 .58493 -6 -4.5 ]
[.01131 .58493 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.21429 .28195 m
.21429 .2882 L
s
[(5)] .21429 .26945 0 1 Mshowa
.40476 .28195 m
.40476 .2882 L
s
[(10)] .40476 .26945 0 1 Mshowa
.59524 .28195 m
.59524 .2882 L
s
[(15)] .59524 .26945 0 1 Mshowa
.78571 .28195 m
.78571 .2882 L
s
[(20)] .78571 .26945 0 1 Mshowa
.97619 .28195 m
.97619 .2882 L
s
[(25)] .97619 .26945 0 1 Mshowa
.125 Mabswid
.0619 .28195 m
.0619 .2857 L
s
.1 .28195 m
.1 .2857 L
s
.1381 .28195 m
.1381 .2857 L
s
.17619 .28195 m
.17619 .2857 L
s
.25238 .28195 m
.25238 .2857 L
s
.29048 .28195 m
.29048 .2857 L
s
.32857 .28195 m
.32857 .2857 L
s
.36667 .28195 m
.36667 .2857 L
s
.44286 .28195 m
.44286 .2857 L
s
.48095 .28195 m
.48095 .2857 L
s
.51905 .28195 m
.51905 .2857 L
s
.55714 .28195 m
.55714 .2857 L
s
.63333 .28195 m
.63333 .2857 L
s
.67143 .28195 m
.67143 .2857 L
s
.70952 .28195 m
.70952 .2857 L
s
.74762 .28195 m
.74762 .2857 L
s
.82381 .28195 m
.82381 .2857 L
s
.8619 .28195 m
.8619 .2857 L
s
.9 .28195 m
.9 .2857 L
s
.9381 .28195 m
.9381 .2857 L
s
.25 Mabswid
0 .28195 m
1 .28195 L
s
.02381 .13045 m
.03006 .13045 L
s
[(-1)] .01131 .13045 1 0 Mshowa
.02381 .43344 m
.03006 .43344 L
s
[(1)] .01131 .43344 1 0 Mshowa
.02381 .58493 m
.03006 .58493 L
s
[(2)] .01131 .58493 1 0 Mshowa
.125 Mabswid
.02381 .16075 m
.02756 .16075 L
s
.02381 .19105 m
.02756 .19105 L
s
.02381 .22135 m
.02756 .22135 L
s
.02381 .25165 m
.02756 .25165 L
s
.02381 .31225 m
.02756 .31225 L
s
.02381 .34254 m
.02756 .34254 L
s
.02381 .37284 m
.02756 .37284 L
s
.02381 .40314 m
.02756 .40314 L
s
.02381 .46374 m
.02756 .46374 L
s
.02381 .49404 m
.02756 .49404 L
s
.02381 .52434 m
.02756 .52434 L
s
.02381 .55464 m
.02756 .55464 L
s
.02381 .10015 m
.02756 .10015 L
s
.02381 .06986 m
.02756 .06986 L
s
.02381 .03956 m
.02756 .03956 L
s
.02381 .00926 m
.02756 .00926 L
s
.02381 .61523 m
.02756 .61523 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
0 0 1 r
.5 Mabswid
.02619 .46892 m
.02857 .47321 L
.03095 .47705 L
.03333 .48046 L
.03571 .4836 L
.0381 .48644 L
.04048 .48903 L
.04286 .49147 L
.04524 .49357 L
.04762 .49522 L
.05 .4966 L
.05238 .49784 L
.05476 .49869 L
.05714 .49922 L
.05952 .4996 L
.0619 .49967 L
.06429 .49949 L
.06667 .49879 L
.06905 .4976 L
.07143 .49573 L
.07381 .49321 L
.07619 .48973 L
.07857 .48531 L
.08095 .47993 L
.08333 .47374 L
.08571 .4669 L
.0881 .45984 L
.09048 .45277 L
.09286 .44557 L
.09524 .43836 L
.09762 .43147 L
.1 .4248 L
.10238 .41822 L
.10476 .41176 L
.10714 .40543 L
.10952 .3993 L
.1119 .39328 L
.11429 .38754 L
.11667 .38188 L
.11905 .37639 L
.12143 .37093 L
.12381 .36572 L
.12619 .36071 L
.12857 .35577 L
.13095 .35101 L
.13333 .34633 L
.13571 .34192 L
.1381 .3375 L
.14048 .33317 L
.14286 .32898 L
Mistroke
.14524 .32494 L
.14762 .321 L
.15 .31714 L
.15238 .31323 L
.15476 .30929 L
.15714 .3055 L
.15952 .30177 L
.1619 .29827 L
.16429 .29496 L
.16667 .29171 L
.16905 .2886 L
.17143 .28562 L
.17381 .28275 L
.17619 .28001 L
.17857 .27741 L
.18095 .27479 L
.18333 .27221 L
.18571 .26977 L
.1881 .26752 L
.19048 .26523 L
.19286 .26295 L
.19524 .26068 L
.19762 .25846 L
.2 .25639 L
.20238 .25438 L
.20476 .25242 L
.20714 .25054 L
.20952 .24878 L
.2119 .24706 L
.21429 .24523 L
.21667 .24355 L
.21905 .24197 L
.22143 .24045 L
.22381 .239 L
.22619 .23755 L
.22857 .23617 L
.23095 .23492 L
.23333 .23387 L
.23571 .23294 L
.2381 .23214 L
.24048 .23122 L
.24286 .23043 L
.24524 .22968 L
.24762 .22898 L
.25 .22833 L
.25238 .22782 L
.25476 .2274 L
.25714 .22699 L
.25952 .22657 L
.2619 .22622 L
Mistroke
.26429 .22586 L
.26667 .22558 L
.26905 .22522 L
.27143 .22478 L
.27381 .22424 L
.27619 .22376 L
.27857 .22327 L
.28095 .22288 L
.28333 .22253 L
.28571 .22231 L
.2881 .22219 L
.29048 .22206 L
.29286 .22186 L
.29524 .22175 L
.29762 .22164 L
.3 .22156 L
.30238 .22146 L
.30476 .22143 L
.30714 .22137 L
.30952 .22124 L
.3119 .22124 L
.31429 .22132 L
.31667 .22138 L
.31905 .22153 L
.32143 .2218 L
.32381 .22197 L
.32619 .22221 L
.32857 .22268 L
.33095 .22317 L
.33333 .22365 L
.33571 .22423 L
.3381 .22498 L
.34048 .22581 L
.34286 .22661 L
.34524 .2275 L
.34762 .22847 L
.35 .22942 L
.35238 .2304 L
.35476 .23141 L
.35714 .23247 L
.35952 .23356 L
.3619 .23466 L
.36429 .23569 L
.36667 .23691 L
.36905 .23824 L
.37143 .23974 L
.37381 .24115 L
.37619 .24264 L
.37857 .24408 L
.38095 .24558 L
Mistroke
.38333 .24723 L
.38571 .249 L
.3881 .25096 L
.39048 .2531 L
.39286 .25538 L
.39524 .25774 L
.39762 .26024 L
.4 .26309 L
.40238 .26626 L
.40476 .26985 L
.40714 .27388 L
.40952 .27841 L
.4119 .28361 L
.41429 .28942 L
.41667 .29564 L
.41905 .30217 L
.42143 .30904 L
.42381 .31599 L
.42619 .32306 L
.42857 .33015 L
.43095 .33711 L
.43333 .34376 L
.43571 .35029 L
.4381 .35668 L
.44048 .36294 L
.44286 .36915 L
.44524 .37521 L
.44762 .38108 L
.45 .38674 L
.45238 .39255 L
.45476 .39811 L
.45714 .40327 L
.45952 .40827 L
.4619 .41307 L
.46429 .41785 L
.46667 .42253 L
.46905 .42703 L
.47143 .43135 L
.47381 .43561 L
.47619 .43965 L
.47857 .44355 L
.48095 .4472 L
.48333 .45078 L
.48571 .45415 L
.4881 .45731 L
.49048 .46029 L
.49286 .46315 L
.49524 .46572 L
.49762 .46806 L
.5 .47039 L
Mistroke
.50238 .47261 L
.50476 .47463 L
.50714 .47648 L
.50952 .47811 L
.5119 .47952 L
.51429 .48074 L
.51667 .48179 L
.51905 .48266 L
.52143 .48331 L
.52381 .48356 L
.52619 .48353 L
.52857 .48304 L
.53095 .48217 L
.53333 .48086 L
.53571 .47889 L
.5381 .4763 L
.54048 .47295 L
.54286 .46877 L
.54524 .46376 L
.54762 .45816 L
.55 .4519 L
.55238 .44538 L
.55476 .4388 L
.55714 .43216 L
.55952 .42567 L
.5619 .41923 L
.56429 .41286 L
.56667 .40662 L
.56905 .40054 L
.57143 .39458 L
.57381 .3888 L
.57619 .38313 L
.57857 .37762 L
.58095 .37227 L
.58333 .36707 L
.58571 .36197 L
.5881 .35709 L
.59048 .35237 L
.59286 .34766 L
.59524 .34294 L
.59762 .33842 L
.6 .33415 L
.60238 .33005 L
.60476 .32611 L
.60714 .32245 L
.60952 .31884 L
.6119 .31505 L
.61429 .31139 L
.61667 .30784 L
.61905 .30448 L
Mistroke
.62143 .30129 L
.62381 .29822 L
.62619 .29511 L
.62857 .29213 L
.63095 .28908 L
.63333 .28616 L
.63571 .28346 L
.6381 .28071 L
.64048 .27801 L
.64286 .27528 L
.64524 .27258 L
.64762 .26994 L
.65 .26741 L
.65238 .26506 L
.65476 .26293 L
.65714 .26089 L
.65952 .25899 L
.6619 .25701 L
.66429 .25506 L
.66667 .25327 L
.66905 .25168 L
.67143 .25013 L
.67381 .2486 L
.67619 .24728 L
.67857 .24606 L
.68095 .24485 L
.68333 .24358 L
.68571 .2422 L
.6881 .2409 L
.69048 .23961 L
.69286 .23844 L
.69524 .23739 L
.69762 .23631 L
.7 .23532 L
.70238 .2344 L
.70476 .23357 L
.70714 .23268 L
.70952 .23188 L
.7119 .23117 L
.71429 .23041 L
.71667 .22972 L
.71905 .229 L
.72143 .22837 L
.72381 .22775 L
.72619 .22715 L
.72857 .22661 L
.73095 .22603 L
.73333 .22544 L
.73571 .22488 L
.7381 .22445 L
Mistroke
.74048 .22408 L
.74286 .22373 L
.74524 .22337 L
.74762 .22309 L
.75 .22284 L
.75238 .2227 L
.75476 .22272 L
.75714 .22282 L
.75952 .22302 L
.7619 .22316 L
.76429 .22329 L
.76667 .22358 L
.76905 .22387 L
.77143 .22416 L
.77381 .22448 L
.77619 .22492 L
.77857 .22547 L
.78095 .22609 L
.78333 .22685 L
.78571 .22768 L
.7881 .22868 L
.79048 .22973 L
.79286 .23074 L
.79524 .23181 L
.79762 .23312 L
.8 .2344 L
.80238 .23569 L
.80476 .23717 L
.80714 .23861 L
.80952 .24005 L
.8119 .24153 L
.81429 .24306 L
.81667 .24466 L
.81905 .24645 L
.82143 .24839 L
.82381 .25043 L
.82619 .25273 L
.82857 .25521 L
.83095 .25789 L
.83333 .26093 L
.83571 .26439 L
.8381 .26826 L
.84048 .27254 L
.84286 .27728 L
.84524 .28275 L
.84762 .28858 L
.85 .29497 L
.85238 .30168 L
.85476 .30866 L
.85714 .31583 L
Mistroke
.85952 .32296 L
.8619 .32993 L
.86429 .33685 L
.86667 .3436 L
.86905 .35026 L
.87143 .35678 L
.87381 .36312 L
.87619 .36929 L
.87857 .37534 L
.88095 .38114 L
.88333 .38679 L
.88571 .39231 L
.8881 .39758 L
.89048 .40269 L
.89286 .40766 L
.89524 .41236 L
.89762 .41694 L
.9 .42141 L
.90238 .4257 L
.90476 .43004 L
.90714 .43412 L
.90952 .43801 L
.9119 .44175 L
.91429 .44542 L
.91667 .44899 L
.91905 .45251 L
.92143 .45609 L
.92381 .45961 L
.92619 .46293 L
.92857 .46601 L
.93095 .46892 L
.93333 .47151 L
.93571 .47397 L
.9381 .47629 L
.94048 .47846 L
.94286 .48052 L
.94524 .48241 L
.94762 .48406 L
.95 .48549 L
.95238 .48669 L
.95476 .48777 L
.95714 .48856 L
.95952 .48903 L
.9619 .48918 L
.96429 .48904 L
.96667 .48835 L
.96905 .48724 L
.97143 .48548 L
.97381 .48293 L
.97619 .47946 L
Mistroke
Mfstroke
1 0 0 r
.02619 .03208 m
.02857 .05063 L
.03095 .06852 L
.03333 .07855 L
.03571 .09008 L
.0381 .10021 L
.04048 .10525 L
.04286 .11956 L
.04524 .13949 L
.04762 .15091 L
.05 .15684 L
.05238 .17442 L
.05476 .18916 L
.05714 .1961 L
.05952 .21028 L
.0619 .22272 L
.06429 .24741 L
.06667 .27172 L
.06905 .30527 L
.07143 .33781 L
.07381 .38587 L
.07619 .43391 L
.07857 .48366 L
.08095 .52687 L
.08333 .56276 L
.08571 .57899 L
.0881 .58498 L
.09048 .59678 L
.09286 .60332 L
.09524 .59353 L
.09762 .58874 L
.1 .58965 L
.10238 .58892 L
.10476 .58789 L
.10714 .58325 L
.10952 .58351 L
.1119 .57423 L
.11429 .57503 L
.11667 .57149 L
.11905 .57487 L
.12143 .56689 L
.12381 .56141 L
.12619 .56203 L
.12857 .55732 L
.13095 .55739 L
.13333 .54826 L
.13571 .55212 L
.1381 .55166 L
.14048 .5481 L
.14286 .54417 L
Mistroke
.14524 .54246 L
.14762 .54232 L
.15 .54761 L
.15238 .55207 L
.15476 .54784 L
.15714 .54831 L
.15952 .54005 L
.1619 .534 L
.16429 .53356 L
.16667 .52932 L
.16905 .52556 L
.17143 .52272 L
.17381 .51927 L
.17619 .51407 L
.17857 .51751 L
.18095 .51768 L
.18333 .51286 L
.18571 .50552 L
.1881 .50966 L
.19048 .5108 L
.19286 .51224 L
.19524 .51112 L
.19762 .50656 L
.2 .50486 L
.20238 .50398 L
.20476 .50205 L
.20714 .49731 L
.20952 .49699 L
.2119 .50375 L
.21429 .49825 L
.21667 .49447 L
.21905 .49288 L
.22143 .49089 L
.22381 .49202 L
.22619 .48977 L
.22857 .4844 L
.23095 .47599 L
.23333 .47141 L
.23571 .46547 L
.2381 .472 L
.24048 .46634 L
.24286 .4655 L
.24524 .46305 L
.24762 .46167 L
.25 .45531 L
.25238 .45135 L
.25476 .45136 L
.25714 .45235 L
.25952 .44896 L
.2619 .44983 L
Mistroke
.26429 .44639 L
.26667 .45037 L
.26905 .45467 L
.27143 .45955 L
.27381 .45716 L
.27619 .45792 L
.27857 .45384 L
.28095 .45171 L
.28333 .44643 L
.28571 .44121 L
.2881 .44216 L
.29048 .44569 L
.29286 .44127 L
.29524 .44139 L
.29762 .4402 L
.3 .44117 L
.30238 .43754 L
.30476 .43937 L
.30714 .44278 L
.30952 .43643 L
.3119 .43256 L
.31429 .43393 L
.31667 .42914 L
.31905 .42346 L
.32143 .42785 L
.32381 .42442 L
.32619 .41338 L
.32857 .41156 L
.33095 .41223 L
.33333 .40673 L
.33571 .39827 L
.3381 .39339 L
.34048 .39493 L
.34286 .38941 L
.34524 .3851 L
.34762 .38506 L
.35 .38306 L
.35238 .38079 L
.35476 .37727 L
.35714 .37523 L
.35952 .37415 L
.3619 .3762 L
.36429 .36634 L
.36667 .36034 L
.36905 .35112 L
.37143 .35377 L
.37381 .34918 L
.37619 .35021 L
.37857 .34642 L
.38095 .33772 L
Mistroke
.38333 .33116 L
.38571 .32015 L
.3881 .31018 L
.39048 .30163 L
.39286 .2959 L
.39524 .28706 L
.39762 .26866 L
.4 .25082 L
.40238 .22834 L
.40476 .20413 L
.40714 .17733 L
.40952 .14083 L
.4119 .10799 L
.41429 .08363 L
.41667 .06353 L
.41905 .04211 L
.42143 .03242 L
.42381 .02136 L
.42619 .01472 L
.42857 .01542 L
.43095 .02474 L
.43333 .02507 L
.43571 .02675 L
.4381 .02747 L
.44048 .02545 L
.44286 .02741 L
.44524 .03163 L
.44762 .03667 L
.45 .02559 L
.45238 .03268 L
.45476 .04735 L
.45714 .05097 L
.45952 .05649 L
.4619 .0535 L
.46429 .05474 L
.46667 .05962 L
.46905 .0643 L
.47143 .06413 L
.47381 .07083 L
.47619 .07491 L
.47857 .08332 L
.48095 .08388 L
.48333 .09125 L
.48571 .09857 L
.4881 .10481 L
.49048 .10802 L
.49286 .11941 L
.49524 .12887 L
.49762 .12704 L
.5 .13091 L
Mistroke
.50238 .13831 L
.50476 .14518 L
.50714 .15418 L
.50952 .16333 L
.5119 .17119 L
.51429 .17859 L
.51667 .18631 L
.51905 .19659 L
.52143 .21453 L
.52381 .22851 L
.52619 .25017 L
.52857 .269 L
.53095 .29033 L
.53333 .32372 L
.53571 .35479 L
.5381 .39329 L
.54048 .43575 L
.54286 .47904 L
.54524 .51114 L
.54762 .5474 L
.55 .56495 L
.55238 .5732 L
.55476 .58126 L
.55714 .57955 L
.55952 .58192 L
.5619 .5841 L
.56429 .58273 L
.56667 .57989 L
.56905 .57937 L
.57143 .57541 L
.57381 .57448 L
.57619 .57178 L
.57857 .5681 L
.58095 .5654 L
.58333 .5645 L
.58571 .55832 L
.5881 .55444 L
.59048 .55778 L
.59286 .56214 L
.59524 .55614 L
.59762 .5475 L
.6 .54309 L
.60238 .53856 L
.60476 .52834 L
.60714 .52902 L
.60952 .54016 L
.6119 .53754 L
.61429 .53473 L
.61667 .52846 L
.61905 .52306 L
Mistroke
.62143 .51994 L
.62381 .52427 L
.62619 .52066 L
.62857 .5263 L
.63095 .52232 L
.63333 .51407 L
.63571 .51885 L
.6381 .51836 L
.64048 .52221 L
.64286 .52286 L
.64524 .52228 L
.64762 .51919 L
.65 .51262 L
.65238 .50365 L
.65476 .50108 L
.65714 .49584 L
.65952 .50133 L
.6619 .50167 L
.66429 .49549 L
.66667 .48721 L
.66905 .48682 L
.67143 .48651 L
.67381 .47839 L
.67619 .4742 L
.67857 .47491 L
.68095 .47855 L
.68333 .48464 L
.68571 .48248 L
.6881 .48283 L
.69048 .47801 L
.69286 .47318 L
.69524 .47539 L
.69762 .47192 L
.7 .46926 L
.70238 .46627 L
.70476 .46928 L
.70714 .46584 L
.70952 .46227 L
.7119 .46469 L
.71429 .46232 L
.71667 .46468 L
.71905 .46022 L
.72143 .46066 L
.72381 .46039 L
.72619 .4578 L
.72857 .45981 L
.73095 .46122 L
.73333 .46013 L
.73571 .45445 L
.7381 .45168 L
Mistroke
.74048 .45101 L
.74286 .45159 L
.74524 .44865 L
.74762 .44712 L
.75 .44201 L
.75238 .43405 L
.75476 .43056 L
.75714 .42577 L
.75952 .42869 L
.7619 .42853 L
.76429 .42111 L
.76667 .42079 L
.76905 .42041 L
.77143 .41874 L
.77381 .41317 L
.77619 .40687 L
.77857 .40369 L
.78095 .39623 L
.78333 .39193 L
.78571 .38349 L
.7881 .38033 L
.79048 .3812 L
.79286 .3776 L
.79524 .36512 L
.79762 .36597 L
.8 .3637 L
.80238 .3543 L
.80476 .35434 L
.80714 .35368 L
.80952 .35044 L
.8119 .34716 L
.81429 .34213 L
.81667 .33206 L
.81905 .32293 L
.82143 .31726 L
.82381 .3025 L
.82619 .29244 L
.82857 .28089 L
.83095 .26112 L
.83333 .23901 L
.83571 .21591 L
.8381 .19362 L
.84048 .16779 L
.84286 .1293 L
.84524 .1075 L
.84762 .07599 L
.85 .05555 L
.85238 .03732 L
.85476 .02254 L
.85714 .01826 L
Mistroke
.85952 .02062 L
.8619 .01782 L
.86429 .01997 L
.86667 .01892 L
.86905 .02035 L
.87143 .02373 L
.87381 .02714 L
.87619 .02763 L
.87857 .03478 L
.88095 .03724 L
.88333 .03956 L
.88571 .04633 L
.8881 .05052 L
.89048 .05285 L
.89286 .06151 L
.89524 .064 L
.89762 .06532 L
.9 .07079 L
.90238 .06456 L
.90476 .07369 L
.90714 .07968 L
.90952 .0835 L
.9119 .08416 L
.91429 .08588 L
.91667 .08494 L
.91905 .08015 L
.92143 .07972 L
.92381 .08633 L
.92619 .09537 L
.92857 .10091 L
.93095 .11416 L
.93333 .11829 L
.93571 .12296 L
.9381 .1282 L
.94048 .13201 L
.94286 .13863 L
.94524 .14849 L
.94762 .15766 L
.95 .16757 L
.95238 .17202 L
.95476 .18538 L
.95714 .20028 L
.95952 .21507 L
.9619 .22886 L
.96429 .25567 L
.96667 .27615 L
.96905 .308 L
.97143 .3479 L
.97381 .39342 L
.97619 .43782 L
Mistroke
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[54]:=",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/@P10`40000100000?ooog1`L<c<c;Ng
]k34gVIVIZo^k_Cdm42P`0000000E@00ZP00o`1E001EE@1EZP1Eo`2Z002ZE@2ZZP2Zo`3o003oE@3o
ZP3ooeD005D0EED0ZUD0oeEE05EEEEEEZUEEoeFZ05FZEEFZZUFZoeGo05GoEEGoZUGoojX00:X0EJX0
ZZX0ojYE0:YEEJYEZZYEojZZ0:ZZEJZZZZZZoj[o0:[oEJ[oZZ[oool00?l0EOl0Z_l0oomE0?mEEOmE
Z_mEoonZ0?nZEOnZZ_nZoooo0?ooEOooZ_ooo`0001lO7clo?emOGgmoOinOWkno_moOgoooon?U]knH
][OUiBgUiL30`9T0I_ooc>OWifm_Kh0003<cIVIVc820P9VIo`00P28R8X200020P4:JY`0g?0200?mV
<omVc1YOI`00kUDJRj>SXiVIVIT00000<`00IP00V@00c00c000c<`0cIP0cV@0cc00co`1V001V<`1V
IP1VV@1Vc01Vo`2I002I<`2IIP2IV@2Ic02Io`3<003<<`3<IP3<V@3<c03<o`3o<`3oIP3oV@3oc3<0
03<0<c<0IS<0VC<0c3<0oc<c03<c<c<cVC<cc3<coc=V03=V<c=VIS=VVC=Vc3=Voc>I03>I<c>IIS>I
VC>Ic3>Ioc?<03?<<c?<IS?<VC?<c3?<oc?o03?o<c?oIS?oVC?oc3?oofH006H0<fH0IVH0VFH0c6H0
ofHc06Hc<fHcIVHcVFHcc6HcofIV06IV<fIVVFIVofJI06JI<fJIIVJIVFJIc6JIofK<06K<<fK<IVK<
VFK<c6K<ofKo06Ko<fKoIVKoVFKoc6KooiT0<iT0VIT0c9T0oiTc09Tc<iTcIYTcVITcc9TcoiUV09UV
<iUVIYUVVIUVc9UVoiVI09VI<iVIIYVIc9W<09W<<iW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wo
ol`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc?m98DT00?m98DT000m900<:BDT0odT?B@00
3dT00`Y9B@3oB@m9000?B@032TU90?m93dT000m90P[oBA19000?B@032TU906e90SZOB@003dT00`Y9
B@1]B@8jMTT3>RI9000?B@032TU906a900<jBCX0MDT2>P03BCXj02E9000?B@032TU906a900HjBCXj
BCX2B@03>TU906e900<jBDT00TT00cY9B@0RB@003dT00`Y9B@1/B@09>TU9BCY9>TTj06m900<jBDT0
0TT2>R=9000?B@032SY906a900<jBDT00dT3>Vi900<jBDT01DT00cY9B@0PB@003dT00`XjB@1/B@03
>TU900A90SY^B@03>TU900E900<jBDT084T000m900<:>TT0JdT00cY9B@07B@03>TU906]900<jBDT0
1TT00cY9B@0OB@003dT00`XjB@1[B@03>TU900M900<jBDT0JdT00cY9B@07B@03>TU901i9000?B@03
2SY906]900<jBDT01dT00cY9B@1[B@03>TU900M900<jBDT07TT000m900<:>TT0JdT00cY9B@08B@03
>TU906Y900<jBDT01dT00cY9B@0NB@003dT00`XjB@1[B@03>TU900Q90SY[B@03>TU900Q900<jBDT0
7DT000m900<:>TT0JTT00cY9B@09B@03>TTj06Y900<jBDT02DT00cY9B@0LB@003dT00`Y9>P1ZB@03
>TU900]900<jBDT0J4T00cY9B@09B@03>TU901a9000?B@032TTj06Y900<jBDT034T00cY9B@1WB@03
>TU900U90cXLB@003dT00`Y9>P1ZB@03>TU900e900<jBDT0IDT00cY9B@0<B@03>TU901Y9000?B@8:
00<jBDT0J4T00cY9B@0=B@03>TU906E900<jBDT034T2>Q]9000?B@042TU9>VU900<jBDT03TT00cY9
B@1TB@03>TU900e900<jBDT06DT000m900@:BDTjJDT00cY9B@0>B@03>TU906A900<jBDT03TT00cY9
B@02>QI9000?B@042TU9>VU900<jBDT03TT2>VE900<jBDT03TT2>P03BCXj01I9000?B@042TU9>VU9
00<jBDT044T00cY9B@1RB@03>TU9011900<jBCX05TT000m900@:BDTjJDT00cY9B@0@B@03>TU90699
00<jBDT04TT00cY9B@0DB@003dT010Y9BCYYB@03>TU9011900<jBDT0HDT00cY9B@0DB@03>TU901=9
000?B@052TU9BCX0IdT00cY9B@0AB@03>TU9065900<jBDT054T00cY9B@0CB@003dT22P9900<jBDT0
IDT00cY9B@0BB@03>TU9061900<jBDT05DT00cY9B@0BB@003dT00`Y9B@02B@03>TU906A900<jBDT0
4TT00cY9B@1PB@03>TU901E900<jBDT04TT000m900<:BDT00TT00cY9B@1TB@03>TU901=900<jBDT0
GdT00cY9B@0EB@03>TU90199000?B@032TU9009900<jBDT0I4T00cY9B@0CB@03>TU905m900<jBDT0
5DT00cY9B@0BB@003dT00`Y9B@02B@03>TU906A900<jBDT04dT00cY9B@1OB@03>TU901E900<jBDT0
4TT000m900<:BDT00TT00cY9B@1TB@03>TU901=900<jBDT0GdT00cY9B@0FB@03>TU90159000?B@03
2TU9009900<jBDT0I4T00cY9B@0CB@03>TU905m900<jBDT05dT00cY9B@0@B@001dT42PA900<:BDT0
0TT00cY9B@1SB@03>TU901E90SYOB@03>TU901M900<jBDT044T000U900<:BDT00dT32P=900<jBDT0
HTT00cY9B@0EB@8jGdT00cY9B@0GB@03>TU901190009B@032TU900=900<:BDT00dT00cY9B@1RB@03
>TU901I900<jBDT0GDT00cY9B@0HB@03>TU900m900000dT:2P022PA900<:BDT00dT00`Y9B@03B@03
>TU9069900<jBDT05dT00cY9B@1LB@03>TU901U900<jBDT03TT000U900<:BDT00dT00`Y9B@03B@03
>TU9069900<jBDT05dT00cY9B@1LB@03>TU901U900<jBDT03TT000M90`X5B@032TU900=900<jBDT0
HTT00cY9B@0HB@03>TU905]900<jBDT06DT00cY9B@0>B@002DT00`Y9B@03B@032TU900A900<jBDT0
HDT00cY9B@0HB@03>TU905]900<jBDT06DT00cY9B@0>B@003dT00`Y9B@04B@03>TU9065900<jBDT0
64T00cY9B@1JB@03>TU901Y900<jBDT03TT000m900<:BDT014T00cY9B@1QB@03>TU901Q900<jBDT0
FTT00cY9B@0KB@03>TU900e9000?B@032TU900A900<jBDT0HDT00cY9B@0HB@03>TU905Y900<jBDT0
6dT00cY9B@0=B@003dT22PE900<jBDT0HDT00cY9B@0IB@03>TU905U900<jBDT06dT00cY9B@0=B@00
3dT00`Y9B@04B@03>TU9065900<jBDT06DT00cY9B@1IB@03>TU901a900<jBDT034T000m900<:BDT0
1DT00cY9B@1PB@03>TU901Y900<jBDT0F4T00cY9B@0LB@03>TU900a9000?B@032TU900E900<jBDT0
H4T00cY9B@0JB@03>TU905Q900<jBDT074T00cY9B@0<B@003dT00`Y9B@05B@03>TU9061900<jBDT0
6TT00cY9B@1HB@03>TU901a900<jBDT034T000m900<:BDT01DT00cY9B@1PB@03>TU901Y900<jBDT0
F4T00cY9B@0MB@03>TU900]9000?B@032TU900I900<jBDT0GdT00cY9B@0KB@03>TU905I900<jBDT0
7TT00cY9B@0;B@003dT00`Y9B@06B@03>TU905m900<jBDT06dT00cY9B@1FB@03>TU901i900<jBDT0
2dT000m90PX7B@03>TU905m900<jBDT06dT00cY9B@1FB@03>TU901i900<jBDT02dT000m900<:BDT0
1TT00cY9B@1NB@03>TU901e900<jBDT0EDT00cY9B@0OB@03>TU900Y9000?B@032TU900I900<jBDT0
GTT00cY9B@0MB@03>TU905E900<jBDT07dT00cY9B@0:B@003dT00`Y9B@06B@03>TU905i900<jBDT0
7DT00cY9B@1EB@03>TU901m900<jBDT02TT000m900<:BDT01dT00cY9B@1MB@03>TU901e900<jBDT0
EDT00cY9B@0OB@03>TU900Y9000?B@032TU900M900<jBDT0GDT00cY9B@0MB@03>TU905E900<jBDT0
7dT00cY9B@0:B@003dT00`Y9B@07B@03>TU905e900<jBDT07DT00cY9B@1EB@03>TU901m900<jBDT0
2TT000m900<:BDT01dT00cY9B@1MB@03>TU901e900<jBDT0EDT00cY9B@0OB@03>TU900Y9000?B@03
2TU900Q900<jBDT0FdT00cY9B@0NB@03>TU905E900<jBDT07dT00cY9B@0:B@003dT22PU900<jBDT0
>4TB3A5900<jBDT07dT00cY9B@0iB@X=4DT00cY9B@0PB@03>TU900U9000?B@032TU900Q900<jBDT0
<TT63A590`d?B@03>TU901m900<jBDT0=DT53@U910d=B@03>TU9021900<jBDT02DT000m900<:BDT0
24T00cY9B@0^B@@=6TT33@a900<jBDT07dT00cY9B@0aB@@=4TT23@Y900<jBDT08DT00cY9B@09B@00
3dT00`Y9B@08B@03>TU902a90`dPB@8=2TT00cY9B@0OB@03>TU902m90`dGB@8=24T00cY9B@0QB@03
>TU900U9000?B@032TU900Q900<jBDT0:TT23BE90Pd8B@03>TU901m900<jBDT0;4T33Aa90Pd6B@03
>TU9025900<jBDT02DT000m900<:BDT024T00cY9B@0XB@<=:4T00`e9B@05B@03>TU9021900<jBDT0
:DT23B5900<=BDT00dT00cY9B@0QB@03>TU900U9000?B@032TU900Q900<jBDT09dT23B]90Pd5B@03
>TU9021900<jBDT0:4T00`e9B@0RB@8=0dT00cY9B@0RB@03>TU900Q9000?B@032TU900Q900<jBDT0
9TT00`e9B@0]B@8=0dT00cY9B@0PB@03>TU902I90PdVB@8=0TT00cY9B@0RB@03>TU900Q9000?B@8:
2DT00cY9B@0TB@8=0PX_B@8=0PX00cY9B@022Qi900<jBDT02dT42P990`X@B@8=6TT42P=90PX6B@8=
00=9>TT08DT22P04>PY9B@<:14T000m900<:BDT024T00cY9B@0SB@8=0dT00`Y9B@0^B@053DTjB@X0
0TT00`Y9B@0KB@03>TU900e900<:BDT014T00`Y9B@0<B@8=6dT00`Y9B@03B@042TU92PM900<jBDT0
8DT00`Y9>P06B@032TU90059000?B@032TU900Q900<jBDT08TT00`e9B@03B@032TU902i90Pd00cY9
2P02B@032TU901]900<jBDT03DT00`Y9B@04B@032TU900]900<=BDT074T00`Y9B@02B@042TU92PM9
00<jBDT08TT00`XjB@05B@032TU90059000?B@032TU900U900<jBDT084T23@990`XaB@052SX=B@X0
0TT00`Y9B@0KB@03>TU900e900@:BDU90`X<B@8=84T01@Y9BDT:009900<:BDT01DT00cX=B@0SB@04
>TU9B@<:14T000m900<:BDT02DT00cY9B@0OB@033DU9009900<:BDT0<DT01@Xj3DT:009900<:BDT0
6dT00cY9B@0=B@052TU9B@X03DT23B9900@:BDT:0TT00`Y9B@05B@03>TT=02=900Dj2TU92P06B@00
3dT00`Y9B@09B@03>TU901i900<=BDT00dT00`Y9B@0_B@<:00@jB@d:0TT00`Y9B@0KB@03>TU900]9
0`X3B@032TU900Y900<=BDT07dT010Y9B@X2B@042TU92PM900<jB@d08DT010Y9>PX2B@032TU900A9
000?B@032TU900U900<jBDT07DT00`e9B@04B@@:<4T01@XjBDT=008:7TT00cY9B@0=B@042TU9B@@:
2DT00`e9B@0PB@8:14T22PQ900@jBDT=8DT00`XjB@02B@@:0dT000m900<:BDT02DT00cY9B@0LB@8=
>dT013Y9B@dPB@03>TU901e900<=BDT0<DT01CY9BDT=025900<jBDT024T000m900<:BDT02DT00cY9
B@0KB@033DU903Y900DjBDU93@0QB@03>TU901]900<=BDT0<TT01CY9BDT=025900<jBDT024T000Q9
4`X00cX:2P0K2P033@X:03X:00<j2PX00PX00`d:2P0N2P03>PX:01X:00<=2PX0<`X01CX:2PX=024:
00<j2PX020X000m900<:BDT024T00`XjB@07B@032TU900Q900<:BDT01DT23@032TU900Q900<:BDT0
24T00`Y9B@07B@032TU900Q900<:BDT01dT00`Y9B@05B@04>TU92P9900<=BDT01DT00`Y9B@07B@03
2TU900Q900@:BDTj1TT00`Y9B@08B@032TU900M90Pd9B@032TU900Q900<:BDT01dT00`Y9B@08B@03
2TU900M900<:>TT014T01@e9BDT:00Y900<:BDT01dT00`Y9B@05B@04>TU92PM9000?B@032TU900Y9
00<jBDT064T00`e9B@0lB@03>TU900=900<=BDT07DT00cY9B@0HB@033DU903A900<jBDT014T00`e9
B@0MB@03>TU900Q9000?B@032TU900Y900<jBDT05dT00`e9B@0mB@03>TU900=900<=BDT07DT00cY9
B@0HB@033DU903A900<jBDT014T00`e9B@0NB@03>TU900M9000?B@032TU900Y900<jBDT05TT00`e9
B@0nB@03>TU900A900<=BDT074T00cY9B@0GB@033DU903A900<jBDT01DT00`e9B@0NB@03>TU900M9
000?B@032TU900Y900<jBDT05TT00`e9B@0mB@03>TU900E900<=BDT074T00cY9B@0FB@033DU903E9
00<jBDT01DT00`e9B@0NB@03>TU900M9000?B@032TU900Y900<jBDT05DT00`e9B@0nB@03>TU900I9
00<=BDT06dT00cY9B@0FB@033DU903E900<jBDT01TT00`e9B@0MB@03>TU900M9000?B@032TU900Y9
00<jBDT054T00`e9B@0nB@03>TU900M900<=BDT074T00cY9B@0DB@033DU903I900<jBDT01TT00`e9
B@0MB@03>TU900M9000?B@8:2dT00cY9B@0DB@033DU903i900<jBDT01dT00`e9B@0LB@03>TU901=9
00<=BDT0=dT00cY9B@07B@033DU901a900<jBDT01dT000m900<:BDT02TT00cY9B@0CB@033DU903m9
00<jBDT01dT00`e9B@0LB@03>TU901=900<=BDT0=dT00cY9B@07B@033DU901a900<jBDT01dT000m9
00<:BDT02TT00cY9B@0BB@033DU9041900<jBDT024T00`e9B@0KB@03>TU9019900<=BDT0>4T00cY9
B@07B@033DU901a900<jBDT01dT000m900<:BDT02TT00cY9B@0BB@033DU9041900<jBDT024T00`e9
B@0KB@03>TU9015900<=BDT0>4T00cY9B@08B@033DU901a900<jBDT01dT000m900<:BDT02TT00cY9
B@0BB@033DU903m900<jBDT02TT00`e9B@0JB@03>TU9015900<=BDT0>4T00cY9B@09B@033DU901]9
00<jBDT01dT000m900<:BDT02TT00cY9B@0AB@033DU9041900<jBDT02TT00`e9B@0JB@03>TU90119
00<=BDT0>DT00cY9B@09B@033DU901a900<jBDT01TT000m900<:BDT02TT00cY9B@0@B@033DU90459
00<jBDT02TT00`e9B@0JB@03>TU900m900<=BDT0>TT00cY9B@0:B@033DU901]900<jBDT01TT000m9
00<:BDT02TT00cY9B@0@B@033DU9041900<jBDT02dT00`e9B@0JB@03>TU900m900<=BDT0>DT00cY9
B@0;B@033DU901]900<jBDT01TT000m90PX;B@03>TU900m900<=BDT0@DT00cY9B@0<B@033DU901U9
00<jBDT03TT00`e9B@0jB@03>TU900]900<=BDT06dT00cY9B@06B@003dT00`Y9B@0:B@03>TU900i9
00<=BDT0@TT00cY9B@0<B@033DU901U900<jBDT03TT00`e9B@0jB@03>TU900]900<=BDT06dT00cY9
B@06B@003dT00`Y9B@0:B@03>TU900i900<=BDT0@4T2>Q1900<=BDT064T00cY9B@0=B@033DU903Y9
00<jBDT03DT00`e9B@0JB@03>TU900I9000?B@032TU900Y900<jBDT03TT00`e9B@0oB@8j4DT00`e9
B@0HB@03>TU900e900<=BDT0>DT2>Pm900<=BDT06TT00cY9B@06B@003dT00`Y9B@0:B@03>TU900e9
00<=BDT0@4T00cY9B@0@B@033DU901Q900<jBDT034T00`e9B@0iB@03>TU9011900<=BDT06DT00cY9
B@06B@003dT00`Y9B@0:B@03>TU900a900<=BDT0@4T00cY9B@0BB@033DU901M900<jBDT034T00`e9
B@0iB@03>TU9011900<=BDT06DT00cY9B@06B@003dT00`Y9B@0;B@03>TU900]900<=BDT0@4T00cY9
B@0BB@033DU901M900<jBDT02dT00`e9B@0iB@03>TU9015900<=BDT06DT00cY9B@06B@003dT00`Y9
B@0;B@03>TU900]900<=BDT0@4T00cY9B@0CB@033DU901I900<jBDT02dT00`e9B@0hB@03>TU901=9
00<=BDT064T00cY9B@06B@003dT00`Y9B@0;B@03>TU900Y900<=BDT0@DT00cY9B@0CB@033DU901I9
00<jBDT02dT00`e9B@0hB@03>TU901=900<=BDT064T00cY9B@06B@003dT22Pa900<jBDT02TT00`e9
B@0oB@8j5TT00`e9B@0GB@03>TU900U900<=BDT0>4T00cY9B@0EB@033DU901M900<jBDT01TT000m9
00<:BDT02dT00cY9B@09B@033DU903m900<jBCX05dT00`e9B@0FB@03>TU900Q900<=BDT0>DT00cY9
B@0EB@033DU901M900<jBDT01TT000m900<:BDT02dT00cY9B@09B@033DU903i900<jBDT064T00`e9
B@0FB@03>TU900Q900<=BDT0>4T00cY9B@0FB@033DU901M900<jBDT01TT000m900<:BDT02dT00cY9
B@09B@033DU903a90cXKB@033DU901E900<jBDT024T00`e9B@0gB@03>TU901Q900<=BDT05TT00cY9
B@06B@003dT00`Y9B@0;B@03>TU900Q900<=BDT0?DT00cY9B@0KB@033DU901E900<jBDT01dT00`e9
B@0hB@03>TU901Q900<=BDT05TT00cY9B@06B@003dT00`Y9B@0;B@03>TU900M900<=BDT0?4T00cY9
>P0MB@033DU901E900<jBDT01dT00`e9B@0gB@03>TU901Y900<=BDT05DT00cY9B@06B@003dT00`Y9
B@0;B@03>TU900M900<=BDT0?4T2>Qm900<=BDT054T00cY9B@06B@033DU903Q900<jBDT06TT00`e9
B@0EB@03>TU900I9000?B@032TU900]900<jBDT01dT00`e9B@0lB@03>TU901i900<=BDT054T00cY9
B@06B@033DU903Q900<jBDT06TT00`e9B@0EB@03>TU900I9000?B@8:34T00cY9B@07B@033DU903a9
00<jBDT07dT00`e9B@0CB@03>TU900I900<=BDT0=dT00cY9B@0LB@033DU901A900<jBDT01TT000m9
00<:BDT02dT00cY9B@06B@033DU903a900<jBDT084T00`e9B@0CB@03>TU900E900<=BDT0=dT00cY9
B@0NB@033DU901=900<jBDT01TT000m900<:BDT034T00cY9B@05B@033DU903a900<jBDT084T00`e9
B@0CB@03>TU900E900<=BDT0=dT00cY9B@0NB@033DU901=900<jBDT01TT000m900<:BDT034T00cY9
B@04B@033DU903]90SXTB@033DU901=900<jBDT00dT00`e9B@0hB@03>TU901i900<=BDT04dT00cY9
B@06B@003dT00`Y9B@0<B@03>TU900A900<=BDT0>dT00cY9B@0SB@033DU901=900<jBDT00dT00`e9
B@0hB@03>TU901m900<=BDT04dT00cY9B@05B@003dT00`Y9B@0<B@03>TU900A900<=BDT0>TT00cY9
B@0UB@033DU9019900<jBDT00dT00`e9B@0eB@<j8dT00`e9B@0BB@03>TU900E9000?B@032TU900a9
00<jBDT00dT00`e9B@0jB@8j9dT00`e9B@0BB@03>TU9009900<=BDT0=TT00cY9B@0SB@033DU90199
00<jBDT01DT000M910X4B@032TU900a900<jBDT00dT00`e9B@0iB@03>TTj02Q900<=BDT04DT00cY9
B@02B@033DU903A90SXWB@033DU9015900<jBDT01DT000U900<:BDT00dT00`Y9B@0<B@03>TU900=9
00<=BDT0>4T2>RY900<=BDT04DT01CY9BDT=03I900<jBDT09dT00`e9B@0AB@03>TU900E90009B@03
2TU900=90`X<B@03>TU900=900<=BDT0=dT00cY9>P0[B@033DU9011900DjBDU93@0fB@03>TU902Q9
00<=BDT044T00cY9B@05B@002DT00`Y9B@03B@032TU900a900<jBDT00TT00`e9B@0fB@<j;TT00`e9
B@0?B@05>TU9B@d0=TT00cY9B@0XB@033DU9011900<jBDT01DT000U900<:BDT00dT00`Y9B@0<B@03
>TU9009900<=BDT0<TT00cY9>P02>P03BCY902e900<=BDT03dT01CY9BDT=03E900<jBDT0:TT00`e9
B@0GB@001dT32PE900<:BDT034T01CY9BDT=02m900<jBDT00TT00cY9>P0dB@033DU900i900@jBDT=
=DT00cY9B@0/B@033DU901I90009B@032TU900=900<:BDT034T01CY9BDT=02i900<jBCX00dT00cY9
B@0eB@033DU900e900@jBDT==DT00cY9B@0/B@033DU901I9000?B@032TU900a900DjBDU93@0/B@8j
0TT00cY9B@02>SM900<=BDT03DT00cY93@0cB@<j<4T00`e9B@0EB@003dT00`Y9B@0<B@05>TU9B@d0
:dT00cY9B@02B@03>TTj03Y900<=BDT034T00cY93@0`B@04>TU9>S=900<=BDT054T000m900<:BDT0
34T013Y9B@d[B@03>TU900A900<jBDT0>TT00`e9B@0;B@03>TT=02i91CXdB@033DU901A9000?B@8:
3DT013Y9B@d[B@03>TU9045900<=BDT02dT00cX=B@0/B@<j>DT00`e9B@0CB@003dT00`Y9B@0<B@03
>TT=02U90cY5B@8=2dT00cY9B@0ZB@8j?DT00`e9B@0BB@003dT00`X=B@0<B@03>TT=02Q900<jBCX0
AdT00`e9B@09B@03>TU902Q90cXnB@8=4dT000m900<:3DT034T00cY93@0XB@03>TU904Q90Pd9B@03
>TU902990SX3B@8j@TT00`e9B@0@B@003dT00`Y93@0<B@03>TT=02Q900<jBDT0BDT33@E90Pd00cY9
B@0RB@05>TTjBCX0A4T23@U900<=BDT01DT000m900@:BDT=2dT00cX=B@0WB@03>TU904a90`d01TT=
3@e9>RA900<jBCX00SY6B@033DU900E900<=BDT01TT000m900<:BDT00Pd:B@03>TU902M900<jBDT0
CTT33@=900<jBDT084T2>P9900<jBDT0AdT23@E900<=BDT01TT000m900<:BDT00TT00`e9B@07B@03
>TU902I900<jBDT0EDT00cY9B@0PB@03>TU904e91@d9B@003dT22P=90Pd7B@033CY902E90cYGB@03
>TU9021900<jBDT0FdT000m900<:BDT014T00`e9B@03B@033DTj02=90SX00dTjB@1IB@03>TU901a9
00<jBCX0GTT000m900<:BDT01DT53@9900<jBDT08DT3>U]900<jBDT06dT013Y9>SYNB@003dT00`Y9
B@0<B@03>TU9021900<jBCX0G4T00cY9B@0KB@03>TTj05m9000?B@032TU900e900<jBDT06TT00cY9
B@02>Um900<jBDT06dT00cY9B@1OB@003dT00`Y9B@0=B@03>TU901Y90SX00dTjB@1OB@03>TU901Y9
00<jBDT0H4T000m900<:BDT03DT00cY9B@0GB@03>TTj009900<jBDT0GdT00cY9B@0FB@05>TU9BCX0
HTT000m900<:BDT03DT00cY9B@0GB@<jI4T00cY9B@0FB@8j00=9>TT0HTT000m900<:BDT03DT00cY9
B@0FB@8jITT00cY9B@0BB@04>TTj>P990SYSB@003dT22Pi900<jBDT05DT00cY9B@1VB@03>TU90199
0cX3B@03>TU90699000?B@032TU900e900<jBDT05DT00cY9B@1WB@03>TU900e900@jBDTjJdT000m9
00<:BDT03DT00cY9B@0DB@03>TU906Q900<jBDT03DT013Y9BCY[B@003dT00`Y9B@0=B@03>TU901=9
00<jBDT0JDT00cY9B@0<B@04>TTj>Va9000?B@032TU900e900<jBDT04dT00cY9B@1YB@03>TU900a9
00@jBCXjK4T000m900<:BDT03DT00cY9B@0>B@8j0dT00cY9B@1YB@03>TU900]900<jBDT0KTT000m9
00<:BDT03TT00cY9B@0=B@03>TTj008jK4T00cY9B@0;B@03>TU906i9000?B@032TU900i900<jBDT0
2TT00cY9>P03B@8jK4T00cY9B@0;B@03>TU906i9000?B@8:3dT00cY9B@0:B@<j0dT00cY9B@1[B@03
>TU900Q900<jBCX0LDT000m900<:BDT03TT00cY9B@09B@8jLdT00cY9B@07B@04>TTj>W59000?B@03
2TU900i900<jBDT024T00cY9B@1cB@03>TU900M900<jBCX0LTT000m900<:BDT03TT00cY9B@08B@03
>TU907=900<jBDT01TT2>WA9000?B@032TU900i900<jBDT01dT00cY9B@1dB@03>TU900E900<jBDT0
M4T000m900<:BDT03TT00cY9B@06B@8jMdT00cY9B@04B@03>TU907A9000?B@032TU900i900<jBDT0
14T3>WQ900<jBDT00TT2>WM90007B@@:14T00`Y9B@0>B@03>TU900A900<jBDT0NDT00cY9B@02>WQ9
0007B@032TU900E900<:BDT03dT00cY9B@02B@8jNdT3>WY90008B@032TU900A90`X?B@03>TU90099
00<jBDT0O4T00cY9B@1hB@002DT00`Y9B@03B@032TU900m900<jBDT00S[jB@002TT00`Y9B@02B@03
2TU901190S[lB@001dT010Y9B@X4B@032TU901190S[lB@0024T22PE900<:BDT044T00cY9B@3kB@00
3dT00`Y9B@0@B@03>TU90?]9000?B@032TU90?m93dT000m900<:BDT0odT?B@003dT22_m944T000m9
00<:BDT0odT?B@00odTQB@00odTQB@00\
\>"],
  ImageRangeCache->{{{0, 287}, {176.938, 0}} -> {-1.46075, -1.92087, 0.094376,
   0.0237321}}]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["An example of a graph in 3D", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[PathGraph3D[p, 1, 1, 2, 3, Magenta], 
      ViewPoint -> {1.300, \ \(-2.400\), \ 2.000}, \n\t
      DisplayFunction -> $DisplayFunction]; \)\)], "Input",
  CellLabel->"In[13]:="],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .44061 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics3D
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.0407423 1.04074 -0.285384 1.04074 [
[.02673 .28582 -2.72019 -9 ]
[.02673 .28582 3.27981 0 ]
[.35445 .17705 -5.84178 -9 ]
[.35445 .17705 6.15822 0 ]
[.71172 .05852 -6.30086 -9 ]
[.71172 .05852 5.69914 0 ]
[.95485 .01653 0 -5.97566 ]
[.95485 .01653 12 3.02434 ]
[.97088 .0401 0 -5.96049 ]
[.97088 .0401 6 3.03951 ]
[.98663 .06327 0 -5.94563 ]
[.98663 .06327 6 3.05437 ]
[1.00213 .08607 0 -5.93107 ]
[1.00213 .08607 6 3.06893 ]
[1.00707 .10541 0 -5.9201 ]
[1.00707 .10541 6 3.0799 ]
[1.00887 .12166 0 -5.91129 ]
[1.00887 .12166 18 3.08871 ]
[1.01067 .13802 0 -5.90243 ]
[1.01067 .13802 6 3.09757 ]
[ 0 0 0 0 ]
[ 1 .44061 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.00588 .30527 m
.92848 0 L
s
.02552 .29878 m
.02491 .30525 L
s
[(0)] .02673 .28582 -0.09327 1 Mshowa
.3541 .19005 m
.35393 .19656 L
s
[(10)] .35445 .17705 -0.02637 1 Mshowa
.71237 .07151 m
.71269 .07801 L
s
[(20)] .71172 .05852 .05014 1 Mshowa
.125 Mabswid
.08904 .27776 m
.08873 .28165 L
s
.15363 .25639 m
.15337 .26028 L
s
.21932 .23465 m
.21911 .23855 L
s
.28613 .21254 m
.28598 .21644 L
s
.42326 .16717 m
.42321 .17107 L
s
.49363 .14389 m
.49364 .14779 L
s
.56525 .12019 m
.56532 .12409 L
s
.63815 .09607 m
.63828 .09997 L
s
.78794 .0465 m
.7882 .0504 L
s
.8649 .02104 m
.86523 .02493 L
s
.25 Mabswid
.92848 0 m
.99308 .09493 L
s
.94249 .02059 m
.93631 .02261 L
s
[(-1)] .95485 .01653 -1 .32792 Mshowa
.9585 .04412 m
.95232 .04612 L
s
[(0)] .97088 .0401 -1 .32455 Mshowa
.97425 .06725 m
.96806 .06924 L
s
[(1)] .98663 .06327 -1 .32125 Mshowa
.98973 .09001 m
.98353 .09198 L
s
[(2)] 1.00213 .08607 -1 .31801 Mshowa
.125 Mabswid
.94572 .02532 m
.94201 .02654 L
s
.94893 .03005 m
.94522 .03126 L
s
.95213 .03475 m
.94842 .03596 L
s
.95532 .03944 m
.95161 .04065 L
s
.96167 .04877 m
.95796 .04998 L
s
.96483 .05342 m
.96112 .05462 L
s
.96798 .05805 m
.96427 .05924 L
s
.97112 .06266 m
.96741 .06385 L
s
.97737 .07183 m
.97365 .07303 L
s
.98047 .0764 m
.97676 .07759 L
s
.98357 .08095 m
.97985 .08214 L
s
.98666 .08549 m
.98294 .08667 L
s
.93926 .01583 m
.93555 .01705 L
s
.93601 .01106 m
.9323 .01228 L
s
.93275 .00627 m
.92905 .0075 L
s
.92949 .00147 m
.92578 .0027 L
s
.9928 .09451 m
.98908 .09569 L
s
.25 Mabswid
.99308 .09493 m
1 .15778 L
s
.99467 .10933 m
.98846 .11129 L
s
[(0)] 1.00707 .10541 -1 .31558 Mshowa
.99645 .12555 m
.99025 .1275 L
s
[(0.5)] 1.00887 .12166 -1 .31362 Mshowa
.99825 .14189 m
.99204 .14382 L
s
[(1)] 1.01067 .13802 -1 .31165 Mshowa
.125 Mabswid
.99502 .11256 m
.9913 .11374 L
s
.99538 .1158 m
.99166 .11698 L
s
.99574 .11905 m
.99201 .12022 L
s
.99609 .1223 m
.99237 .12347 L
s
.99681 .12881 m
.99309 .12998 L
s
.99717 .13207 m
.99345 .13324 L
s
.99753 .13534 m
.9938 .1365 L
s
.99789 .13861 m
.99416 .13977 L
s
.99861 .14517 m
.99489 .14633 L
s
.99897 .14846 m
.99525 .14961 L
s
.99934 .15175 m
.99561 .1529 L
s
.9997 .15504 m
.99597 .1562 L
s
.99431 .1061 m
.99059 .10727 L
s
.99396 .10287 m
.99024 .10405 L
s
.9936 .09965 m
.98988 .10083 L
s
.99325 .09643 m
.98953 .09761 L
s
.25 Mabswid
.00588 .30527 m
0 .36582 L
s
0 .36582 m
.08686 .44061 L
s
.08686 .44061 m
.09153 .38101 L
s
.09153 .38101 m
.00588 .30527 L
s
.92848 0 m
.99308 .09493 L
s
.99308 .09493 m
1 .15778 L
s
1 .15778 m
.93475 .06375 L
s
.93475 .06375 m
.92848 0 L
s
.00588 .30527 m
0 .36582 L
s
0 .36582 m
.93475 .06375 L
s
.93475 .06375 m
.92848 0 L
s
.92848 0 m
.00588 .30527 L
s
.09153 .38101 m
.99308 .09493 L
s
.99308 .09493 m
1 .15778 L
s
1 .15778 m
.08686 .44061 L
s
.08686 .44061 m
.09153 .38101 L
s
0 0 m
1 0 L
1 .44061 L
0 .44061 L
closepath
clip
newpath
1 0 1 r
.5 Mabswid
.16222 .40251 m
.16298 .39923 L
s
.15927 .40383 m
.16222 .40251 L
s
.15562 .40449 m
.15927 .40383 L
s
.16298 .39923 m
.16441 .39661 L
s
.1666 .39473 m
.16857 .39266 L
s
.16441 .39661 m
.1666 .39473 L
s
.16857 .39266 m
.17049 .39059 L
s
.17049 .39059 m
.17194 .3881 L
s
.17403 .38625 m
.17486 .38326 L
s
.17194 .3881 m
.17403 .38625 L
s
.15274 .40585 m
.15562 .40449 L
s
.17702 .38156 m
.1786 .37934 L
s
.17486 .38326 m
.17702 .38156 L
s
.14848 .40593 m
.15274 .40585 L
s
.18111 .378 m
.1821 .37528 L
s
.1786 .37934 m
.18111 .378 L
s
.1821 .37528 m
.18342 .37292 L
s
.18342 .37292 m
.18556 .37134 L
s
.18556 .37134 m
.18698 .36912 L
s
.14158 .4035 m
.14848 .40593 L
s
.18698 .36912 m
.18904 .36752 L
s
.18904 .36752 m
.18987 .36482 L
s
.19243 .36376 m
.19441 .36216 L
s
.18987 .36482 m
.19243 .36376 L
s
.19441 .36216 m
.19598 .36021 L
s
.19598 .36021 m
.19751 .35824 L
s
.20671 .35345 m
.20819 .35149 L
s
.20408 .35434 m
.20671 .35345 L
s
.19751 .35824 m
.19932 .35657 L
s
.20134 .35511 m
.20408 .35434 L
s
.19932 .35657 m
.20134 .35511 L
s
.20819 .35149 m
.2103 .35014 L
s
.2103 .35014 m
.21125 .34773 L
s
.13366 .40002 m
.14158 .4035 L
s
.21125 .34773 m
.21249 .34563 L
s
.21249 .34563 m
.21447 .34426 L
s
.21447 .34426 m
.21595 .34243 L
s
.21595 .34243 m
.21749 .3407 L
s
.21749 .3407 m
.21915 .3391 L
s
.21915 .3391 m
.22074 .33744 L
s
.12484 .39553 m
.13366 .40002 L
s
.22074 .33744 m
.22209 .3356 L
s
.22457 .33486 m
.22663 .3337 L
s
.22209 .3356 m
.22457 .33486 L
s
.22663 .3337 m
.22803 .33193 L
s
.22803 .33193 m
.22911 .32988 L
s
.23168 .3293 m
.23387 .32833 L
s
.22911 .32988 m
.23168 .3293 L
s
.23387 .32833 m
.23609 .3274 L
s
.23609 .3274 m
.23798 .32615 L
s
.23798 .32615 m
.23942 .32449 L
s
.23942 .32449 m
.24123 .32321 L
s
.11622 .39102 m
.12484 .39553 L
s
.24123 .32321 m
.24315 .32204 L
s
.24315 .32204 m
.24494 .32076 L
s
.24494 .32076 m
.24635 .31914 L
s
.24635 .31914 m
.24835 .31809 L
s
.25127 .31794 m
.25259 .31623 L
s
.24835 .31809 m
.25127 .31794 L
s
.25259 .31623 m
.25413 .31477 L
s
.25413 .31477 m
.25596 .3136 L
s
.25596 .3136 m
.25774 .31239 L
s
.10757 .38629 m
.11622 .39102 L
s
.25774 .31239 m
.25993 .31159 L
s
.25993 .31159 m
.26167 .31036 L
s
.26167 .31036 m
.26301 .30875 L
s
.26301 .30875 m
.26396 .30678 L
s
.26396 .30678 m
.2654 .30534 L
s
.2654 .30534 m
.26667 .30375 L
s
.10105 .38332 m
.10757 .38629 L
s
.26956 .30376 m
.27086 .3022 L
s
.26667 .30375 m
.26956 .30376 L
s
.27086 .3022 m
.27279 .30128 L
s
.27279 .30128 m
.27452 .30017 L
s
.27452 .30017 m
.27638 .2992 L
s
.09438 .38007 m
.10105 .38332 L
s
.27638 .2992 m
.2776 .2976 L
s
.2776 .2976 m
.27913 .29634 L
s
.27913 .29634 m
.28117 .2956 L
s
.28117 .2956 m
.28334 .29499 L
s
.08898 .37785 m
.09438 .38007 L
s
.28334 .29499 m
.28495 .29381 L
s
.28495 .29381 m
.28711 .2932 L
s
.28711 .2932 m
.28871 .29203 L
s
.08353 .37547 m
.08898 .37785 L
s
.28871 .29203 m
.29127 .29182 L
s
.29127 .29182 m
.29388 .29164 L
s
.29656 .29152 m
.2983 .29045 L
s
.29388 .29164 m
.29656 .29152 L
s
.2983 .29045 m
.30045 .28979 L
s
.30045 .28979 m
.30198 .28851 L
s
.0798 .37455 m
.08353 .37547 L
s
.30198 .28851 m
.30376 .2875 L
s
.07582 .37335 m
.0798 .37455 L
s
.30376 .2875 m
.30514 .28609 L
s
.07286 .37301 m
.07582 .37335 L
s
.06881 .37164 m
.07286 .37301 L
s
.30514 .28609 m
.30652 .28472 L
s
.06436 .36984 m
.06881 .37164 L
s
.30652 .28472 m
.3087 .28415 L
s
.3087 .28415 m
.31121 .28392 L
s
.06155 .36945 m
.06436 .36984 L
s
.05797 .36832 m
.06155 .36945 L
s
.31121 .28392 m
.3127 .28265 L
s
.05319 .36604 m
.05797 .36832 L
s
.3127 .28265 m
.31477 .28198 L
s
.31477 .28198 m
.31668 .28115 L
s
.04921 .36438 m
.05319 .36604 L
s
.31668 .28115 m
.31887 .28059 L
s
.04655 .36385 m
.04921 .36438 L
s
.04318 .36264 m
.04655 .36385 L
s
.31887 .28059 m
.32046 .27944 L
s
.0396 .36118 m
.04318 .36264 L
s
.03624 .35986 m
.0396 .36118 L
s
.03177 .35746 m
.03624 .35986 L
s
.32046 .27944 m
.32276 .27901 L
s
.32276 .27901 m
.32526 .27878 L
s
.32526 .27878 m
.32651 .27727 L
s
.0272 .35488 m
.03177 .35746 L
s
.32651 .27727 m
.32808 .2761 L
s
.32808 .2761 m
.33033 .27563 L
s
.33033 .27563 m
.33178 .27436 L
s
.33178 .27436 m
.33312 .27299 L
s
.33312 .27299 m
.33576 .27295 L
s
.33576 .27295 m
.33739 .27187 L
s
.33739 .27187 m
.33805 .26982 L
s
.33805 .26982 m
.33989 .26901 L
s
.33989 .26901 m
.34205 .26854 L
s
.34205 .26854 m
.34342 .26725 L
s
.34342 .26725 m
.34441 .2656 L
s
.34441 .2656 m
.34587 .26444 L
s
.34587 .26444 m
.34814 .26415 L
s
.34814 .26415 m
.34951 .26292 L
s
.34951 .26292 m
.35104 .26187 L
s
.35104 .26187 m
.35312 .26139 L
s
.35312 .26139 m
.35495 .26065 L
s
.35495 .26065 m
.35674 .25988 L
s
.35674 .25988 m
.35838 .25896 L
s
.35838 .25896 m
.3602 .25824 L
s
.3602 .25824 m
.36215 .25764 L
s
.36215 .25764 m
.36451 .25747 L
s
.36451 .25747 m
.36534 .25571 L
s
.36534 .25571 m
.36667 .25449 L
s
.36667 .25449 m
.36758 .25287 L
s
.36758 .25287 m
.37001 .25286 L
s
.37001 .25286 m
.37152 .25187 L
s
.37152 .25187 m
.37375 .25163 L
s
.37375 .25163 m
.37537 .25075 L
s
.37537 .25075 m
.37636 .24923 L
s
.37636 .24923 m
.37763 .24802 L
s
.37763 .24802 m
.37832 .24624 L
s
.37832 .24624 m
.37916 .24464 L
s
.37916 .24464 m
.38017 .24327 L
s
.38017 .24327 m
.38155 .24229 L
s
.38155 .24229 m
.38253 .24092 L
s
.38253 .24092 m
.38228 .23829 L
s
.38228 .23829 m
.38211 .23581 L
s
.38211 .23581 m
.38134 .23275 L
s
.38134 .23275 m
.38034 .22955 L
s
.38034 .22955 m
.37901 .22607 L
s
.37901 .22607 m
.37641 .22137 L
s
.37641 .22137 m
.37427 .21729 L
s
.57011 .26651 m
.57211 .2643 L
s
.56768 .26827 m
.57011 .26651 L
s
.57211 .2643 m
.57394 .26191 L
s
.56523 .27001 m
.56768 .26827 L
s
.56327 .27233 m
.56523 .27001 L
s
.56013 .27333 m
.56327 .27233 L
s
.57394 .26191 m
.57604 .25987 L
s
.57604 .25987 m
.57774 .25739 L
s
.57774 .25739 m
.5798 .25536 L
s
.5798 .25536 m
.58165 .25311 L
s
.55697 .2743 m
.56013 .27333 L
s
.58165 .25311 m
.58339 .25076 L
s
.58339 .25076 m
.58524 .24858 L
s
.58524 .24858 m
.58731 .24667 L
s
.55269 .27397 m
.55697 .2743 L
s
.58731 .24667 m
.58875 .24407 L
s
.59574 .23949 m
.59721 .23699 L
s
.59304 .24059 m
.59574 .23949 L
s
.58875 .24407 m
.59047 .24183 L
s
.59047 .24183 m
.59304 .24059 L
s
.59721 .23699 m
.59837 .23417 L
s
.54614 .27099 m
.55269 .27397 L
s
.37427 .21729 m
.37322 .21447 L
s
.59837 .23417 m
.60003 .23198 L
s
.60003 .23198 m
.60168 .2298 L
s
.60168 .2298 m
.60266 .22689 L
s
.60844 .22557 m
.61032 .22372 L
s
.60493 .22551 m
.60844 .22557 L
s
.61032 .22372 m
.61218 .22186 L
s
.60266 .22689 m
.60493 .22551 L
s
.61218 .22186 m
.61363 .21956 L
s
.54007 .26844 m
.54614 .27099 L
s
.37322 .21447 m
.37273 .21232 L
s
.61363 .21956 m
.61519 .21741 L
s
.61519 .21741 m
.61702 .21561 L
s
.61972 .21486 m
.6215 .21301 L
s
.61702 .21561 m
.61972 .21486 L
s
.62436 .21245 m
.62609 .21056 L
s
.6215 .21301 m
.62436 .21245 L
s
.62609 .21056 m
.62733 .20811 L
s
.37273 .21232 m
.37205 .21004 L
s
.53261 .26419 m
.54007 .26844 L
s
.63009 .20749 m
.63224 .20614 L
s
.62733 .20811 m
.63009 .20749 L
s
.63489 .2054 m
.63717 .20421 L
s
.63224 .20614 m
.63489 .2054 L
s
.63717 .20421 m
.63931 .20286 L
s
.63931 .20286 m
.64115 .20117 L
s
.64115 .20117 m
.6426 .19902 L
s
.6426 .19902 m
.64377 .19658 L
s
.37205 .21004 m
.37291 .20945 L
s
.64377 .19658 m
.64568 .19506 L
s
.52521 .25986 m
.53261 .26419 L
s
.64568 .19506 m
.64729 .1932 L
s
.37291 .20945 m
.37359 .20868 L
s
.65014 .19285 m
.6524 .19177 L
s
.64729 .1932 m
.65014 .19285 L
s
.6524 .19177 m
.6539 .18979 L
s
.6539 .18979 m
.65517 .18755 L
s
.37359 .20868 m
.37485 .20855 L
s
.65517 .18755 m
.65735 .18645 L
s
.51827 .25589 m
.52521 .25986 L
s
.65735 .18645 m
.65954 .18536 L
s
.65954 .18536 m
.66082 .1832 L
s
.37485 .20855 m
.37707 .20944 L
s
.66082 .1832 m
.66257 .18162 L
s
.37707 .20944 m
.38042 .21149 L
s
.38042 .21149 m
.3826 .21223 L
s
.66257 .18162 m
.66488 .18074 L
s
.67048 .18015 m
.67246 .17884 L
s
.66754 .18028 m
.67048 .18015 L
s
.66488 .18074 m
.66754 .18028 L
s
.3826 .21223 m
.38496 .21313 L
s
.67246 .17884 m
.67473 .1779 L
s
.51223 .25278 m
.51827 .25589 L
s
.67473 .1779 m
.67642 .17623 L
s
.38496 .21313 m
.3872 .21388 L
s
.67642 .17623 m
.6781 .17459 L
s
.3872 .21388 m
.38908 .21422 L
s
.6781 .17459 m
.68059 .17396 L
s
.68059 .17396 m
.68244 .17253 L
s
.38908 .21422 m
.39148 .21509 L
s
.50588 .24921 m
.51223 .25278 L
s
.68244 .17253 m
.68438 .17123 L
s
.39148 .21509 m
.39419 .21625 L
s
.68438 .17123 m
.68628 .16989 L
s
.39419 .21625 m
.397 .21749 L
s
.68628 .16989 m
.68887 .16942 L
s
.68887 .16942 m
.69072 .16802 L
s
.397 .21749 m
.39771 .21645 L
s
.69072 .16802 m
.69257 .16663 L
s
.50101 .24718 m
.50588 .24921 L
s
.69257 .16663 m
.6951 .1661 L
s
.6951 .1661 m
.69708 .16488 L
s
.40486 .22045 m
.4075 .22134 L
s
.40079 .21795 m
.40486 .22045 L
s
.39771 .21645 m
.40079 .21795 L
s
.69708 .16488 m
.6996 .16435 L
s
.4075 .22134 m
.41038 .22246 L
s
.6996 .16435 m
.70135 .16284 L
s
.49645 .24541 m
.50101 .24718 L
s
.41038 .22246 m
.41216 .22236 L
s
.70135 .16284 m
.70366 .16204 L
s
.70366 .16204 m
.70588 .16115 L
s
.41216 .22236 m
.41449 .22284 L
s
.49153 .24315 m
.49645 .24541 L
s
.70588 .16115 m
.70785 .15993 L
s
.41449 .22284 m
.4173 .2238 L
s
.70785 .15993 m
.71034 .15937 L
s
.4173 .2238 m
.42008 .2247 L
s
.71034 .15937 m
.71276 .15873 L
s
.71276 .15873 m
.7149 .15772 L
s
.42008 .2247 m
.42224 .22489 L
s
.48756 .24188 m
.49153 .24315 L
s
.7149 .15772 m
.71652 .15607 L
s
.42224 .22489 m
.42528 .22602 L
s
.48309 .23998 m
.48756 .24188 L
s
.42528 .22602 m
.42799 .22673 L
s
.43126 .22802 m
.43352 .22817 L
s
.42799 .22673 m
.43126 .22802 L
s
.71652 .15607 m
.71848 .15486 L
s
.47959 .23909 m
.48309 .23998 L
s
.43352 .22817 m
.43666 .22924 L
s
.71848 .15486 m
.72067 .15395 L
s
.43666 .22924 m
.43979 .23027 L
s
.45245 .23409 m
.45442 .23363 L
s
.44905 .23294 m
.45245 .23409 L
s
.43979 .23027 m
.44278 .23109 L
s
.44278 .23109 m
.44539 .23147 L
s
.4764 .2385 m
.47959 .23909 L
s
.44539 .23147 m
.44905 .23294 L
s
.47326 .23793 m
.4764 .2385 L
s
.72067 .15395 m
.72301 .15323 L
s
.47005 .23725 m
.47326 .23793 L
s
.46669 .23635 m
.47005 .23725 L
s
.45442 .23363 m
.45711 .23396 L
s
.72301 .15323 m
.72495 .15201 L
s
.46334 .23542 m
.46669 .23635 L
s
.46026 .23474 m
.46334 .23542 L
s
.45711 .23396 m
.46026 .23474 L
s
.72495 .15201 m
.72705 .151 L
s
.72705 .151 m
.72875 .14949 L
s
.72875 .14949 m
.73013 .14759 L
s
.73013 .14759 m
.73202 .14636 L
s
.73202 .14636 m
.73376 .14496 L
s
.73376 .14496 m
.73638 .14469 L
s
.73638 .14469 m
.73865 .14396 L
s
.73865 .14396 m
.7401 .14219 L
s
.7401 .14219 m
.74236 .14147 L
s
.74236 .14147 m
.74461 .14074 L
s
.74461 .14074 m
.74672 .13982 L
s
.74672 .13982 m
.74839 .13835 L
s
.74839 .13835 m
.74999 .1368 L
s
.74999 .1368 m
.75193 .13572 L
s
.75193 .13572 m
.7534 .13403 L
s
.7534 .13403 m
.75523 .13283 L
s
.75523 .13283 m
.75659 .13105 L
s
.75659 .13105 m
.75856 .13006 L
s
.75856 .13006 m
.76098 .12966 L
s
.76098 .12966 m
.76289 .12861 L
s
.76289 .12861 m
.76382 .12628 L
s
.76382 .12628 m
.76624 .12593 L
s
.76624 .12593 m
.76832 .12513 L
s
.76832 .12513 m
.7696 .12329 L
s
.7696 .12329 m
.77194 .12286 L
s
.77194 .12286 m
.77421 .12232 L
s
.77421 .12232 m
.77618 .1214 L
s
.77618 .1214 m
.77816 .12049 L
s
.77816 .12049 m
.77994 .11933 L
s
.77994 .11933 m
.78116 .11744 L
s
.78116 .11744 m
.7825 .11573 L
s
.7825 .11573 m
.78422 .11456 L
s
.78422 .11456 m
.78493 .11206 L
s
.78493 .11206 m
.78617 .11032 L
s
.78617 .11032 m
.78725 .10838 L
s
.78725 .10838 m
.78741 .10527 L
s
.78741 .10527 m
.78731 .10188 L
s
.78731 .10188 m
.78711 .09842 L
s
.78711 .09842 m
.787 .09515 L
s
.787 .09515 m
.7865 .09143 L
s
.7865 .09143 m
.78456 .0859 L
s
.78456 .0859 m
.78453 .08301 L
s
.78453 .08301 m
.78339 .07871 L
s
.78339 .07871 m
.78352 .07621 L
s
.78352 .07621 m
.78391 .07409 L
s
.78391 .07409 m
.7847 .07256 L
s
.94935 .12765 m
.95638 .13277 L
s
.7847 .07256 m
.7867 .07267 L
s
.7867 .07267 m
.78947 .07378 L
s
.78947 .07378 m
.79164 .07408 L
s
.94214 .12215 m
.94935 .12765 L
s
.79164 .07408 m
.79438 .07511 L
s
.79438 .07511 m
.79676 .07563 L
s
.79676 .07563 m
.79942 .0765 L
s
.79942 .0765 m
.80231 .07765 L
s
.93548 .11729 m
.94214 .12215 L
s
.80231 .07765 m
.8052 .07876 L
s
.8052 .07876 m
.80775 .07938 L
s
.80775 .07938 m
.81107 .081 L
s
.81107 .081 m
.81384 .08186 L
s
.81384 .08186 m
.81661 .08265 L
s
.92964 .11349 m
.93548 .11729 L
s
.81661 .08265 m
.81987 .0841 L
s
.81987 .0841 m
.82285 .0851 L
s
.82285 .0851 m
.82561 .08579 L
s
.82561 .08579 m
.82908 .08741 L
s
.82908 .08741 m
.83186 .08803 L
s
.83186 .08803 m
.8345 .08844 L
s
.92501 .11131 m
.92964 .11349 L
s
.8345 .08844 m
.83762 .08947 L
s
.83762 .08947 m
.83941 .08866 L
s
.91967 .10805 m
.92501 .11131 L
s
.84294 .09022 m
.8461 .09124 L
s
.83941 .08866 m
.84294 .09022 L
s
.8461 .09124 m
.84903 .09189 L
s
.84903 .09189 m
.8516 .09202 L
s
.8516 .09202 m
.85428 .0923 L
s
.91572 .10669 m
.91967 .10805 L
s
.85428 .0923 m
.85667 .09216 L
s
.91165 .10512 m
.91572 .10669 L
s
.85667 .09216 m
.85864 .09141 L
s
.90756 .10346 m
.91165 .10512 L
s
.85864 .09141 m
.86109 .09134 L
s
.90363 .10196 m
.90756 .10346 L
s
.87491 .09628 m
.87786 .09669 L
s
.87095 .0944 m
.87491 .09628 L
s
.90067 .1018 m
.90363 .10196 L
s
.89711 .10076 m
.90067 .1018 L
s
.87786 .09669 m
.88087 .09716 L
s
.86109 .09134 m
.86433 .09233 L
s
.86783 .09366 m
.87095 .0944 L
s
.86433 .09233 m
.86783 .09366 L
s
.89363 .09978 m
.89711 .10076 L
s
.88087 .09716 m
.88393 .09769 L
s
.88393 .09769 m
.88684 .09796 L
s
.89006 .09864 m
.89363 .09978 L
s
.88684 .09796 m
.89006 .09864 L
s
0 g
.25 Mabswid
.92848 0 m
.99308 .09493 L
s
.99308 .09493 m
1 .15778 L
s
1 .15778 m
.93475 .06375 L
s
.93475 .06375 m
.92848 0 L
s
.00588 .30527 m
0 .36582 L
s
0 .36582 m
.93475 .06375 L
s
.93475 .06375 m
.92848 0 L
s
.92848 0 m
.00588 .30527 L
s
.00588 .30527 m
.92848 0 L
s
.02552 .29878 m
.02491 .30525 L
s
[(0)] .02673 .28582 -0.09327 1 Mshowa
.3541 .19005 m
.35393 .19656 L
s
[(10)] .35445 .17705 -0.02637 1 Mshowa
.71237 .07151 m
.71269 .07801 L
s
[(20)] .71172 .05852 .05014 1 Mshowa
.125 Mabswid
.08904 .27776 m
.08873 .28165 L
s
.15363 .25639 m
.15337 .26028 L
s
.21932 .23465 m
.21911 .23855 L
s
.28613 .21254 m
.28598 .21644 L
s
.42326 .16717 m
.42321 .17107 L
s
.49363 .14389 m
.49364 .14779 L
s
.56525 .12019 m
.56532 .12409 L
s
.63815 .09607 m
.63828 .09997 L
s
.78794 .0465 m
.7882 .0504 L
s
.8649 .02104 m
.86523 .02493 L
s
.25 Mabswid
.99308 .09493 m
1 .15778 L
s
.99467 .10933 m
.98846 .11129 L
s
[(0)] 1.00707 .10541 -1 .31558 Mshowa
.99645 .12555 m
.99025 .1275 L
s
[(0.5)] 1.00887 .12166 -1 .31362 Mshowa
.99825 .14189 m
.99204 .14382 L
s
[(1)] 1.01067 .13802 -1 .31165 Mshowa
.125 Mabswid
.99502 .11256 m
.9913 .11374 L
s
.99538 .1158 m
.99166 .11698 L
s
.99574 .11905 m
.99201 .12022 L
s
.99609 .1223 m
.99237 .12347 L
s
.99681 .12881 m
.99309 .12998 L
s
.99717 .13207 m
.99345 .13324 L
s
.99753 .13534 m
.9938 .1365 L
s
.99789 .13861 m
.99416 .13977 L
s
.99861 .14517 m
.99489 .14633 L
s
.99897 .14846 m
.99525 .14961 L
s
.99934 .15175 m
.99561 .1529 L
s
.9997 .15504 m
.99597 .1562 L
s
.99431 .1061 m
.99059 .10727 L
s
.99396 .10287 m
.99024 .10405 L
s
.9936 .09965 m
.98988 .10083 L
s
.99325 .09643 m
.98953 .09761 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[13]:=",
  ImageSize->{288, 126.875},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000OPP10`40000100000?ooog1`L<c<c;Ng
]k34gVIVIZo^k_Cdm42P`0000000E@00ZP00o`1E001EE@1EZP1Eo`2Z002ZE@2ZZP2Zo`3o003oE@3o
ZP3ooeD005D0EED0ZUD0oeEE05EEEEEEZUEEoeFZ05FZEEFZZUFZoeGo05GoEEGoZUGoojX00:X0EJX0
ZZX0ojYE0:YEEJYEZZYEojZZ0:ZZEJZZZZZZoj[o0:[oEJ[oZZ[oool00?l0EOl0Z_l0oomE0?mEEOmE
Z_mEoonZ0?nZEOnZZ_nZoooo0?ooEOooZ_ooo`0001lO7clo?emOGgmoOinOWkno_moOgoooon?U]knH
][OUiBgUiL30`9T0I_ooc>OWifm_Kh0003<cIVIVc820P9VIo`00P28R8X200020P4:JY`0g?0200?mV
<omVc1YOI`00kUDJRj>SXiVIVIT00000<`00IP00V@00c00c000c<`0cIP0cV@0cc00co`1V001V<`1V
IP1VV@1Vc01Vo`2I002I<`2IIP2IV@2Ic02Io`3<003<<`3<IP3<V@3<c03<o`3o<`3oIP3oV@3oc3<0
03<0<c<0IS<0VC<0c3<0oc<c03<c<c<cVC<cc3<coc=V03=V<c=VIS=VVC=Vc3=Voc>I03>I<c>IIS>I
VC>Ic3>Ioc?<03?<<c?<IS?<VC?<c3?<oc?o03?o<c?oIS?oVC?oc3?oofH006H0<fH0IVH0VFH0c6H0
ofHc06Hc<fHcIVHcVFHcc6HcofIV06IV<fIVVFIVofJI06JI<fJIIVJIVFJIc6JIofK<06K<<fK<IVK<
VFK<c6K<ofKo06Ko<fKoIVKoVFKoc6KooiT0<iT0VIT0c9T0oiTc09Tc<iTcIYTcVITcc9TcoiUV09UV
<iUVIYUVVIUVc9UVoiVI09VI<iVIIYVIc9W<09W<<iW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wo
ol`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc?m98DT00?m98DT00?m98DT00?m98DT00?m9
8DT00?I90PXXB@00ldT32P03B@X:00a910XGB@00l4T32PA90PX>B@032TU901I9003]B@<:1dT32Pe9
00<:BDT05TT00>Y90`X:B@042TT:2PA910X4B@032TU901I9003VB@@:3TT010Y92PX;B@032TU901I9
003TB@<:4DT22P03B@X:00Q90`XHB@00hDT32QA900<:BDT00PX:B@032TU901I9003NB@<:5dT010Y9
BDT22PI90PXJB@00fdT32QY900@:BDU90PX5B@042TU92QU9002hB@@:0dT22QM90`XMB@032TU90099
0PX4B@042TU92QU9002hB@032TU900=900@:BDT:4dT32R1900<:BDT00dT22P=900@:BDT:6DT00;U9
00<:BDT00TT010Y9B@X@B@<:8dT00`Y9B@02B@032TT:00=900@:BDT:6DT00;Y900D:BDU92P02B@03
2TU900Y91@XVB@032TU900=90PX2B@042TU92QU9002kB@042TU92P9900<:BDT01dT32R]900<:BDT0
14T22P990PX2B@@:54T00;Q900@:BDT:0TT010Y9B@X6B@<:;TT00`Y9B@05B@8:1dT00`Y9B@0CB@00
^DT22PA90PX4B@<:<DT00`Y9B@06B@032TU900E900<:BDT04dT00<990`XcB@8:1dT00`Y92P06B@03
2TU901=9002oB@<:<dT32P9900<:BDT01DT22PI900<:BDT04dT00;a90`XBB@<m7TT32PE900<:BDT0
1TT22P=90`XEB@00^DT32P9900<:BDT03dT00ce9B@03?AU90`X9B@032TU900I90PX4B@032TT:00<:
44T00;I90`XGB@03?DU900=90cdBB@@:3DT00`Y9B@06B@8:1DT00`Y9B@0AB@00/dT32Q]900<mBDT0
1DT4?@]90`XAB@032TU900I90`X5B@032TU90119002`B@<:7TT00ce9B@09B@<m1DT32QE900<:BDT0
1TT22PI900<:BDT03dT00:e90`XQB@03?DU900a90Sd32P8m5dT00`Y9B@05B@8:1dT00`Y9B@0>B@00
ZTT32RA900<mBDT02dT32PE913dCB@032TU900I90PX3B@042TT:2Q19002WB@@:9TT00ce9B@08B@<:
2dT6?@m900<:BDT00TT52PA90PX00dT:B@0>B@00Y4T32RY900<mBDT01DT32QA90Sd>B@@:14T22P=9
00@:BDT:3dT00:590`X]B@03?DU900990`XIB@@m1dT42PM90PX3B@042TU92Pm9002NB@<:<4T00ce9
2P022R1900@mBDU90`X4B@032TU900=90`X3B@042TT:2PA900<:BDT00TT32P=9002KB@<:<DT42RA9
00<m2PX024T00`Y9B@03B@8:0dT22P03B@X:00]90@X1B@59002HB@<:<DT32P8m8dT32P03BCdm00U9
00<:BDT00TT22PA90PX00dT:B@0:B@4:0DT1B@00UDT32S590`X014Tm?CdQB@<:1TT2?@M900<:BDT0
0TT22PA900@:BDT:24T32P=9002BB@<:00=92TT0;TT32P990SdQB@<:2dT00ce9B@05B@042TU9B@8:
14T010Y9B@X8B@032TU900=9002?B@<:<DT32PA90SdNB@@:3dT00ce9B@05B@032TU9008:14T42PQ9
00<:BDT00dT008a90`XaB@<:1DT2?Ae90`XDB@8m1DT00`Y9B@022PE90PX9B@@:0TT008U90`XaB@<:
1dT2?A]90`XOB@042TT:2PI900<:BDT03DT008I90`XaB@<:24T3?AU90`XSB@042TT:2PE900<:BDT0
3DT008=90`XaB@<:2DT3?AQ90`XVB@8:00=92TT014T00`Y9B@0=B@00P4T42Rm910X;B@8m5dT32RY9
0`X3B@<:3dT007a910X`B@<:3DT2?AI90`X^B@8:1DT00`Y9B@0=B@00NDT32S590`X?B@8m4dT42S59
0PXEB@00FDT42P=90PXDB@<:<DT32Pm90cdBB@<:=DT22QE9001KB@052TU9B@X00TT00`Y9B@0>B@<:
<DT32Q190cdAB@<:=DT32QM9001KB@052TU9B@X00TT00`Y9B@0;B@<:<DT32Q590SdAB@<:=4T42QY9
001KB@052TU9B@X00TT00`Y9B@08B@@:<4T32Q990Sd@B@<:=4T32Qi9001KB@052TU9B@X00TT00`Y9
B@05B@<:<DT32Q=90Sd?B@<:=4T32R59001IB@<:0dT010Y9B@X4B@<:<DT32QE90Sd<B@@:=4T32RA9
001KB@032TU900990PX2B@<:<4T42QE90cd;B@<:=4T42RM9001QB@<:<4T32QQ900<mBDT02DT32SA9
0`X[B@00GTT32S190`XJB@03?DU900M90`XdB@<:;TT005]910X_B@<:6dT2?@M90`XdB@<:<DT005Q9
0`X3B@032TU902Y90`XMB@8m1DT32S=910XdB@00EDT32S190`XOB@8m0TT42S=90`XhB@00DTT32S19
0`XQB@8m0`XdB@<:>dT004m90`XAB@8m7DT32R990Sd22SA90`XnB@00C4T32QA900@mBCdm64T32R=9
0Sd00`Y9B@0bB@<:@DT004U910XFB@04?DU9B@@m44T42R=90PX2?C9910Y4B@00ATT32QY900<mBDT0
0dT3?@a90`XTB@<:00=9?Cd0<4T32TQ90013B@<:7TT00ce9B@05B@@m1DT32RA90`X3B@8m;TT32T]9
0010B@<:8TT00ce9B@08B@8m0`XSB@@:14T3?Ba90`Y>B@00?DT32RE900<mBDT01dT32P990cdNB@<:
1dT2?BY910YAB@00>TT32RQ900<mBDT014T32PQ91CdFB@<:2TT00ce9B@0VB@<:EDT003M910XZB@04
?DU9B@<:44T4?@m90`X<B@8m9DT32UQ9000dB@<:;TT013d:2PXGB@@m24T32Pi90SdSB@<:FdT00359
0`X_B@<:7TT00ce9B@02B@<:44T00ce9B@0OB@@:GTT002i90`X^B@@:00=9?DT07dT2?@8:4TT2?Ai9
0`YRB@00:TT42Ri90`X2B@<m7TT32P990SdBB@03?DU901Y90`YUB@009dT32Rm90`X4B@8m7DT32PM9
0Sd?B@03?DU901Q90`YXB@0094T32P03B@Y902a90`X4B@<m74T32Pa90Sd<B@03?DU901E910Y[B@00
8DT32Rm90`X5B@<m6dT32Q590Sd9B@03?DU901=90`Y_B@001TT22QI90`X_B@<:1dT2?AY90`XFB@8m
1TT2?A990`YbB@001DT010Y9B@XBB@<:;dT32PQ90cdGB@@:6dT013e9BDT2?A590`YeB@001DT010Y9
B@X?B@<:;dT32PY90SdFB@<:84T3?@m910YhB@001DT010Y9B@X<B@<:;dT32P]90cdDB@<:<TT32Wa9
0005B@042TU92PU90`X2B@032TU902U910X;B@<m54T32S990`YoB@001DT010Y9B@X6B@<:;TT32Pe9
0SdDB@<:<TT32X990006B@8:14T32Ri90`X<B@Dm4TT32S5910Z5B@002DT32Ri90`X=B@<m4dT32S59
0`Z9B@001TT32Ri90`X?B@8m4DT42S590`Z<B@000dT42Re90`XAB@8m3dT32S990`Z?B@0000=92PX0
0dT00`Y9B@0XB@<:4TT2?@i90`XbB@<:TTT00003B@X:02]90`XDB@8m34T32S5910ZEB@0000A92TT:
9dT32QI90Sd:B@<:<DT32YU9000014T:BDT22R5910XGB@8m2DT32S590`ZLB@0000=92TT00dT00`Y9
B@0KB@<:6DT2?@M910XaB@<:WdT00003B@Y900A900<:BDT05dT32Q]90Sd5B@<:<DT42Z9900000dT:
B@05B@032TU901=90`XLB@<m0dT32S590`ZVB@0000=92TT01TT00`Y9B@0?B@<:7TT2?@990`XaB@<:
ZDT00003B@Y900M900<:BDT02dT32R190Sd32S590`Z/B@0000<:BDT024T00`Y9B@07B@<:8DT2?@8:
<4T42Zm900000`Y9B@09B@8:14T32R590`X00ce9B@0^B@<:/dT000032TU900]910XQB@<:0TT00ce9
B@0/B@<:]TT000032TU900Q91@XPB@<:14T2?B]90`ZiB@0000<:BDT014T013d:2PX5B@032TU901Y9
0`X5B@<m:4T42[a900000`Y9B@02B@<:00<mBDT01TT00`Y9B@0FB@<:1dT2?BM90`[0B@0000<:B@X0
0PX4B@<m1TT00`Y9B@0BB@<:2DT2?BE90`[3B@000PX:B@<m14T00`Y9B@0>B@<:2dT2?B=90`[6B@00
00=92TT034T3?@990PX:B@@:3DT2?B1910[9B@000TT00`Y9B@0=B@@m00<:BDT014T32Q190SdNB@<:
cDT000=90PXAB@8m0TT32Q=900<mBDT06TT32]190005B@032TU901190SdEB@03?DU901Q90`[CB@00
1TT00`Y9B@0?B@032TTm01=900<mBDT05DT42]I90007B@032TU900i900<:BDT00Sd@B@8m54T32]Y9
0008B@032TU900e900<:BDT00TT3?@a90SdBB@<:gDT000U900<:BDT034T00`Y9B@05B@8m2DT2?A19
0`[PB@002TT00`Y9B@0;B@032TU900M90Sd7B@03?DU900a90`[SB@002dT22P]900<:BDT02DT2?@=9
0cd:B@@:iTT000e900<:BDT024T00`Y9B@0;B@<m2TT32^Y9000>B@032TU900M900<:BDT05DT32^e9
000?B@032TU900E900<:BDT04dT32_19000@B@032TU900A900<:BDT03dT42_=9000AB@032TU900=9
00<:BDT034T32_M9000BB@032TU9009900<:BDT02DT32_Y9000CB@8:0TT00`Y9B@06B@<:oDT001E9
00<:B@X014T42_m90DT001I90PX014T:2P[oB@E9000GB@8:odT8B@00odTQB@00odTQB@00odTQB@00
\
\>"],
  ImageRangeCache->{{{0, 287}, {125.875, 0}} -> {0.0391426, 0.256262, 
  0.00360367, 0.00360367}}]
}, Open  ]]
}, Closed]]
}, Closed]]
}, Closed]]
},
FrontEndVersion->"X 3.0",
ScreenRectangle->{{0, 1280}, {0, 1024}},
CellGrouping->Manual,
WindowSize->{544, 597},
WindowMargins->{{113, Automatic}, {Automatic, 109}},
StyleDefinitions -> Notebook[{
  
  Cell[CellGroupData[{
  Cell["Style Definitions", "Subtitle"],
  
  Cell["\<\
These styles are carefully tuned for all aspects of the online \
help. It's recommended that they remain unchanged.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell["Style Environment Names", "Subsection"],
  
  Cell[StyleData[All, "Working"],
    CellBracketOptions->{"Color"->RGBColor[0, 0.2, 1]},
    ScriptMinSize->9],
  
  Cell[StyleData[All, "Printout"],
    PrivateFontOptions->{"FontType"->"Outline"}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Notebook Options", "Subsection"],
  
  Cell["\<\
  The options defined for the style below will be used at the \
Notebook level.  \
\>", "Text"],
  
  Cell[StyleData["Notebook"],
    PageHeaders->{{Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"], None, Cell[ 
            TextData[ {
              OptionValueBox[ "FileName"]}], "Header"]}, {Cell[ 
            TextData[ {
              OptionValueBox[ "FileName"]}], "Header"], None, Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"]}},
    PrintingOptions->{"FirstPageHeader"->False,
    "FacingPages"->True},
    CellLabelAutoDelete->False,
    CellFrameLabelMargins->6,
    StyleMenuListing->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Headings", "Subsection"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Title"],
    CellMargins->{{5, Inherited}, {2, 0}},
    CellGroupingRules->{"TitleGrouping", 0},
    PageBreakBelow->False,
    CounterIncrements->"Title",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subtitle", 0}, {"Subsubtitle", 0}},
    FontFamily->"Times",
    FontSize->36,
    FontWeight->"Bold",
    FontColor->GrayLevel[1],
    Background->RGBColor[0, 0, 1]],
  
  Cell[StyleData["Title", "Printout"],
    CellMargins->{{0, 10}, {2, 0}},
    FontSize->28,
    Background->GrayLevel[0.6]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subtitle"],
    CellMargins->{{12, Inherited}, {2, 2}},
    CellGroupingRules->{"TitleGrouping", 10},
    PageBreakBelow->False,
    CounterIncrements->"Subtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subsubtitle", 0}},
    FontFamily->"Helvetica",
    FontSize->24],
  
  Cell[StyleData["Subtitle", "Printout"],
    CellMargins->{{7, 10}, {2, 4}},
    FontSize->18]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubtitle"],
    CellMargins->{{12, Inherited}, {2, 12}},
    CellGroupingRules->{"TitleGrouping", 20},
    PageBreakBelow->False,
    CounterIncrements->"Subsubtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}},
    FontFamily->"Helvetica",
    FontSize->14,
    FontSlant->"Italic"],
  
  Cell[StyleData["Subsubtitle", "Printout"],
    CellMargins->{{7, 10}, {2, 8}},
    FontSize->12]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Section"],
    CellFrame->{{6, 0.5}, {0.5, 0.5}},
    CellMargins->{{8, 4}, {2, 50}},
    CellElementSpacings->{"ClosedGroupTopMargin"->18},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CellFrameColor->RGBColor[0, 0, 1],
    CounterIncrements->"Section",
    CounterAssignments->{{"Subsection", 0}, {"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->18,
    FontWeight->"Bold",
    FontColor->RGBColor[0, 0, 1]],
  
  Cell[StyleData["Section", "Printout"],
    CellFrame->{{6, 0.5}, {0.5, 0.5}},
    CellMargins->{{6, 4}, {2, 70}},
    CellFrameColor->GrayLevel[0],
    FontSize->14,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsection"],
    CellDingbat->"\[FilledSquare]",
    CellMargins->{{24, 4}, {2, 18}},
    CellElementSpacings->{"ClosedGroupTopMargin"->12},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CellFrameLabelMargins->6,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->15,
    FontWeight->"Bold",
    FontColor->RGBColor[0, 0, 1]],
  
  Cell[StyleData["Subsection", "Printout"],
    CellMargins->{{19, 4}, {2, 18}},
    FontSize->12,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubsection"],
    CellMargins->{{10, 4}, {2, 18}},
    CellElementSpacings->{"ClosedGroupTopMargin"->12},
    CellGroupingRules->{"SectionGrouping", 50},
    PageBreakBelow->False,
    CounterIncrements->"Subsubsection",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontColor->RGBColor[0, 0, 1]],
  
  Cell[StyleData["Subsubsection", "Printout"],
    CellMargins->{{7, 4}, {2, 18}},
    FontSize->10,
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Text", "Subsection"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Text"],
    CellMargins->{{10, 4}, {0, 8}},
    CellHorizontalScrolling->True,
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 12},
    CounterIncrements->"Text",
    FontSize->14],
  
  Cell[StyleData["Text", "Printout"],
    CellMargins->{{7, 0}, {0, 8}},
    FontSize->11]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MathCaption"],
    CellFrame->{{3, 0}, {0, 0.5}},
    CellMargins->{{47, 14}, {0, 14}},
    CellFrameColor->RGBColor[0, 0, 1],
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 8},
    FontFamily->"Helvetica",
    FontSize->12],
  
  Cell[StyleData["MathCaption", "Printout"],
    CellMargins->{{34, 54}, {0, 14}},
    CellFrameColor->GrayLevel[0],
    FontSize->10]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Input/Output", "Subsection"],
  
  Cell["\<\
The cells in this section define styles used for input and output \
to the kernel.  Be careful when modifying, renaming, or removing these \
styles, because the front end associates special meanings with these style \
names.    \
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Input"],
    CellMargins->{{54, 14}, {3, 8}},
    Evaluatable->True,
    CellGroupingRules->"InputGrouping",
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelAutoDelete->False,
    CellLabelMargins->{{21, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultInputFormatType,
    AutoItalicWords->{},
    FormatType->StandardForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    CounterIncrements->"Input",
    FontSize->12,
    FontWeight->"Bold"],
  
  Cell[StyleData["Input", "Printout"],
    ShowCellBracket->False,
    CellMargins->{{43, 4}, {3, 8}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Output"],
    CellMargins->{{55, 4}, {8, 3}},
    CellEditDuplicate->True,
    CellGroupingRules->"OutputGrouping",
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    CellLabelAutoDelete->False,
    CellLabelMargins->{{21, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultOutputFormatType,
    AutoItalicWords->{},
    FormatType->StandardForm,
    CounterIncrements->"Output",
    FontSize->12],
  
  Cell[StyleData["Output", "Printout"],
    ShowCellBracket->False,
    CellMargins->{{43, 4}, {4, 2}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Message"],
    CellMargins->{{56, 4}, {3, 8}},
    CellGroupingRules->"OutputGrouping",
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    CellLabelMargins->{{21, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultOutputFormatType,
    AutoItalicWords->{},
    FormatType->InputForm,
    CounterIncrements->"Message",
    StyleMenuListing->None,
    FontColor->RGBColor[0, 0.2, 1]],
  
  Cell[StyleData["Message", "Printout"],
    ShowCellBracket->False,
    CellMargins->{{44, 4}, {3, 8}},
    FontSize->10,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Print"],
    CellMargins->{{56, 4}, {3, 8}},
    CellGroupingRules->"OutputGrouping",
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    CellLabelMargins->{{21, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultOutputFormatType,
    AutoItalicWords->{},
    FormatType->InputForm,
    CounterIncrements->"Print",
    StyleMenuListing->None],
  
  Cell[StyleData["Print", "Printout"],
    ShowCellBracket->False,
    CellMargins->{{44, 4}, {3, 8}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Graphics"],
    CellMargins->{{55, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Graphics",
    StyleMenuListing->None],
  
  Cell[StyleData["Graphics", "Printout"],
    CellMargins->{{42, 4}, {3, 8}},
    ImageSize->{250, 250},
    FontSize->9]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["CellLabel"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontColor->RGBColor[0, 0.2, 1]],
  
  Cell[StyleData["CellLabel", "Printout"],
    FontSize->7,
    FontSlant->"Oblique",
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Open  ]],
  
  Cell[CellGroupData[{
  
  Cell["Unique Styles", "Subsection"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemizedText"],
    CellMargins->{{26, 4}, {0, 8}},
    ShowSpecialCharacters->Automatic,
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 8},
    ParagraphIndent->-19,
    CounterIncrements->"Text",
    FontSize->14],
  
  Cell[StyleData["ItemizedText", "Printout"],
    CellMargins->{{20, 4}, {0, 8}},
    ParagraphIndent->-14,
    FontSize->11]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemizedTextNote"],
    CellMargins->{{45, 4}, {0, 5}},
    ShowSpecialCharacters->Automatic,
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 5},
    CounterIncrements->"Text",
    FontSize->14],
  
  Cell[StyleData["ItemizedTextNote", "Printout"],
    CellMargins->{{34, 4}, {0, 4}},
    ParagraphSpacing->{0, 4},
    FontSize->11]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["IndentedText"],
    CellMargins->{{26, 4}, {0, 8}},
    ShowSpecialCharacters->Automatic,
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 8},
    CounterIncrements->"Text",
    FontSize->14],
  
  Cell[StyleData["IndentedText", "Printout"],
    CellMargins->{{20, 4}, {0, 8}},
    FontSize->11]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Note"],
    CellFrame->True,
    CellMargins->{{10, 4}, {0, 8}},
    CellHorizontalScrolling->True,
    CellFrameColor->RGBColor[0, 0, 1],
    ShowSpecialCharacters->Automatic,
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 8},
    CounterIncrements->"Text",
    FontFamily->"Helvetica",
    FontSize->10],
  
  Cell[StyleData["Note", "Printout"],
    CellMargins->{{7, 4}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->8]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Picture"],
    CellMargins->{{10, Inherited}, {0, 8}},
    CellHorizontalScrolling->True],
  
  Cell[StyleData["Picture", "Printout"],
    CellMargins->{{7, Inherited}, {0, 8}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Caption"],
    CellMargins->{{10, 50}, {0, 3}},
    FontFamily->"Helvetica",
    FontSize->9,
    FontColor->RGBColor[0, 0, 1]],
  
  Cell[StyleData["Caption", "Printout"],
    CellMargins->{{7, 50}, {2, 4}},
    FontSize->7,
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Tables", "Subsection"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["2ColumnTable"],
    CellMargins->{{10, 4}, {0, 8}},
    CellHorizontalScrolling->True,
    GridBoxOptions->{ColumnWidths->{0.39, 0.59},
    ColumnAlignments->{Left}}],
  
  Cell[StyleData["2ColumnTable", "Printout"],
    CellMargins->{{7, 4}, {0, 8}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["3ColumnTable"],
    CellMargins->{{10, 4}, {0, 8}},
    CellHorizontalScrolling->True,
    StyleMenuListing->None,
    GridBoxOptions->{ColumnWidths->0.325,
    ColumnAlignments->{Left}}],
  
  Cell[StyleData["3ColumnTable", "Printout"],
    CellMargins->{{7, 4}, {0, 8}},
    FontSize->10]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Formulas and Programming", "Subsection"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ChemicalFormula"],
    CellMargins->{{10, 4}, {0, 8}},
    CellHorizontalScrolling->True,
    AutoSpacing->False,
    ScriptLevel->1,
    ScriptBaselineShifts->{0.6, Automatic},
    SingleLetterItalics->False,
    ZeroWidthTimes->True],
  
  Cell[StyleData["ChemicalFormula", "Printout"],
    CellMargins->{{7, 4}, {0, 8}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["DisplayFormula"],
    CellMargins->{{55, Inherited}, {Inherited, Inherited}},
    CellHorizontalScrolling->True,
    ScriptLevel->0,
    SingleLetterItalics->True,
    UnderoverscriptBoxOptions->{LimitsPositioning->True},
    GridBoxOptions->{ColumnWidths->Automatic}],
  
  Cell[StyleData["DisplayFormula", "Printout"],
    CellMargins->{{43, Inherited}, {Inherited, Inherited}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Program"],
    CellMargins->{{10, Inherited}, {Inherited, Inherited}},
    CellHorizontalScrolling->True,
    FontFamily->"Courier"],
  
  Cell[StyleData["Program", "Printout"],
    CellMargins->{{7, Inherited}, {Inherited, Inherited}},
    FontSize->9.5]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Hyperlink Styles", "Subsection"],
  
  Cell["\<\
The cells below define styles useful for making hypertext \
ButtonBoxes.  The \"Hyperlink\" style is for links within the same Notebook, \
or between Notebooks.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Hyperlink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontColor->GrayLevel[1],
    Background->RGBColor[0, 0.6, 1],
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`NotebookLocate[ #2]}]&),
    Active->True,
    ButtonFrame->"None",
    ButtonNote->ButtonData}],
  
  Cell[StyleData["Hyperlink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell["\<\
The following styles are for linking automatically to the on-line \
help system.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MainBookLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontColor->GrayLevel[1],
    Background->RGBColor[0, 0.6, 1],
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "MainBook", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["MainBookLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["AddOnsLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontColor->GrayLevel[1],
    Background->RGBColor[0, 0.6, 1],
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "AddOns", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["AddOnLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["RefGuideLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontColor->GrayLevel[1],
    Background->RGBColor[0, 0.6, 1],
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "RefGuideLink", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["RefGuideLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["GettingStartedLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontColor->GrayLevel[1],
    Background->RGBColor[0, 0.6, 1],
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "GettingStarted", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["GettingStartedLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["OtherInformationLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontColor->GrayLevel[1],
    Background->RGBColor[0, 0.6, 1],
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "OtherInformation", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["OtherInformationLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1]]
  }, Closed]]
  }, Closed]]
  }, Open  ]]
  }]
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1709, 49, 82, 1, 116, "Title"],
Cell[1794, 52, 88, 1, 33, "Subtitle"],
Cell[1885, 55, 641, 11, 141, "Text"],

Cell[CellGroupData[{
Cell[2551, 70, 78, 1, 92, "Section"],
Cell[2632, 73, 338, 6, 84, "Text"],
Cell[2973, 81, 671, 10, 160, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[3681, 96, 72, 1, 60, "Section"],
Cell[3756, 99, 436, 12, 282, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[4229, 116, 31, 0, 60, "Section"],
Cell[4263, 118, 997, 34, 84, "Text"],
Cell[5263, 154, 447, 11, 84, "Text"],
Cell[5713, 167, 142, 3, 46, "Text"],
Cell[5858, 172, 46, 0, 27, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[5941, 177, 33, 0, 60, "Section"],
Cell[5977, 179, 206, 5, 46, "Text"],

Cell[CellGroupData[{
Cell[6208, 188, 38, 0, 36, "Subsubsection"],
Cell[6249, 190, 80, 2, 26, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[6366, 197, 54, 0, 30, "Subsubsection"],
Cell[6423, 199, 83, 2, 26, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[6543, 206, 35, 0, 30, "Subsubsection"],
Cell[6581, 208, 336, 6, 138, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[6954, 219, 52, 0, 30, "Subsubsection"],
Cell[7009, 221, 952, 18, 218, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[7998, 244, 52, 0, 30, "Subsubsection"],
Cell[8053, 246, 962, 18, 218, "Input"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[9064, 270, 39, 0, 60, "Section"],
Cell[9106, 272, 463, 7, 103, "Text"],
Cell[9572, 281, 87, 2, 26, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[9696, 288, 43, 0, 60, "Section"],

Cell[CellGroupData[{
Cell[9764, 292, 257, 5, 65, "Text"],
Cell[10024, 299, 68, 2, 26, "Input"]
}, Open  ]],

Cell[CellGroupData[{
Cell[10129, 306, 169, 4, 46, "Text"],
Cell[10301, 312, 72, 2, 26, "Input"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[10422, 320, 43, 0, 60, "Section"],

Cell[CellGroupData[{
Cell[10490, 324, 36, 0, 40, "Subsection"],

Cell[CellGroupData[{
Cell[10551, 328, 295, 5, 84, "Text"],
Cell[10849, 335, 778, 13, 314, "Input"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[11676, 354, 55, 0, 34, "Subsection"],

Cell[CellGroupData[{
Cell[11756, 358, 49, 0, 36, "Subsubsection"],
Cell[11808, 360, 979, 18, 365, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[12824, 383, 29, 0, 30, "Subsubsection"],
Cell[12856, 385, 819, 13, 210, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[13712, 403, 48, 0, 30, "Subsubsection"],
Cell[13763, 405, 632, 11, 196, "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[14444, 422, 27, 0, 34, "Subsection"],

Cell[CellGroupData[{
Cell[14496, 426, 62, 0, 36, "Subsubsection"],

Cell[CellGroupData[{
Cell[14583, 430, 175, 3, 58, "Input"],
Cell[14761, 435, 30767, 1221, 186, 15479, 1027, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[45577, 1662, 52, 0, 30, "Subsubsection"],

Cell[CellGroupData[{
Cell[45654, 1666, 205, 4, 58, "Input"],
Cell[45862, 1672, 25755, 1714, 135, 19000, 1625, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]]
}, Closed]]
}, Closed]]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

