(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 3.0,
MathReader 3.0, or any compatible application. The data for the notebook 
starts with the line of stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     82802,       3925]*)
(*NotebookOutlinePosition[    100768,       4601]*)
(*  CellTagsIndexPosition[    100724,       4597]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Stochastic Differential Equations", "Title",
  FontColor->RGBColor[1, 0, 0]],

Cell["Multiple Noise Realizations", "Subtitle",
  FontColor->RGBColor[0, 0, 1]],

Cell[TextData[{
  "This notebook is used to approximate solutions to stochastic differential \
equations.  There are two version of this notebook.    This version \
approximates the solution of a fixed system with different sets of noise \
data.  The other version (called \"stosolve.nb\" and found in a separate \
directory called \"/flow\")  fixes the noise data so that it is easy to look \
at the flow by varying the initial conditions. The program uses a C library \
to do the numerical approximation and ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " as a front end and for plotting and doing statistics with the results."
}], "Text"],

Cell[CellGroupData[{

Cell[TextData[StyleBox["Acknowledgements",
  FontSlant->"Italic"]], "Section"],

Cell["\<\
The programming for this Mathematica module was done by David \
Glickenstein under the supervision of Ruth Williams.  It was developed \
through a generous contribution to the UCSD Mathematics Department from the \
GenCorp Foundation of Aerojet, a corporate donor committed to supporting \
educational improvement.\
\>", "Text"],

Cell["\<\
The underlying C software written for working with the stochastic \
differential equations was written by Jessica Gaines, while working at the \
University of Edinburgh, under the direction of Terry Lyons, and using \
funding provided by EC grant SC1*CT92-0784.  The software is to be used for \
educational and academic research purposes only.  Anyone interested in more \
information should email jessica@maths.ed.ac.uk or to write to Jessica \
Gaines, University of Edinburgh, Dept. of Mathematics and Statistics, JCMB, \
Kings Buildings, Mayfield Road, Edinburgh EH9 3JZ UK  and then get original \
copies directly from Edinburgh themselves.  \
\>", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["Disclaimer",
  FontSlant->"Italic"]], "Section"],

Cell["\<\
The author disclaims any responsibility for problems experienced by \
users as a result of using this software.  The software may not be \
distributed without written permission from the author.  For more information \
contact:
David Glickenstein\t\tor\t\tRuth Williams
glicken @math.ucsd.edu\t\t\twilliams@math.ucsd.edu

at:\tUCSD Mathematics Department
\t9500 Gilman Drive 0112
\tLa Jolla, CA 92093-0112
\
\>", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Introduction", "Section"],

Cell[TextData[{
  "This notebook is used to find the Ito solution of ",
  Cell[BoxData[
      \(TraditionalForm
      \`dX\  = \ \(b(X, t)\)\ dt\  + \ \(\[Sigma](X, t)\) dW\)]],
  " where ",
  Cell[BoxData[
      \(TraditionalForm\`b\)]],
  " is the Ito drift, ",
  Cell[BoxData[
      \(TraditionalForm\`\[Sigma]\)]],
  " is the diffusion, ",
  Cell[BoxData[
      \(TraditionalForm\`\(t\ \)\)]],
  "is time, and ",
  Cell[BoxData[
      \(TraditionalForm\`W\)]],
  " is the noise, and ",
  Cell[BoxData[
      \(TraditionalForm\`X\)]],
  " is a vector (",
  Cell[BoxData[
      \(TraditionalForm\`b\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`\[Sigma]\)]],
  " are matrix valued).  It also can find Stratonovich solutions of ",
  Cell[BoxData[
      \(TraditionalForm
      \`dX\  = \ \(a(X, t)\) dt\  + \ \(\[Sigma](X, t)\)\[SmallCircle]dW\)]],
  ".  It uses a combination of C and ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  "."
}], "Text"],

Cell[TextData[{
  "It is important to be careful when entering input, since C is slightly \
different from ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  ".  Make sure to read the directions in each section, as they are a \
guideline as to how to use the program.  When working with vectors, remember \
that the jth element of a vector v is v[j] in C and v[[j]] in ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  "."
}], "Text"],

Cell["\<\
The C library was entirely developed by Jessica Gaines and the \
programs are largely based upon her example programs.\
\>", "Text"],

Cell["\t\t\tDavid Glickenstein, UCSD", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Initialization", "Section"],

Cell[TextData[{
  "These initialization routines must be run each time a new ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " kernel is run, but the code should not be altered at all."
}], "Text"],

Cell[CellGroupData[{

Cell["Load Packages", "Subsubsection"],

Cell[BoxData[
    \(<< Graphics`Graphics3D`\)], "Input",
  CellLabel->"In[1]:="],

Cell[BoxData[
    \(<< Statistics`DescriptiveStatistics`\)], "Input",
  CellLabel->"In[2]:="]
}, Closed]],

Cell[CellGroupData[{

Cell["Turn off unnecessary warnings", "Subsubsection"],

Cell[BoxData[
    \(\(Off[General::spell1]; \)\)], "Input",
  CellLabel->"In[3]:="]
}, Closed]],

Cell[CellGroupData[{

Cell["Set colors", "Subsubsection"],

Cell[BoxData[
    \(Blue\  = \ RGBColor[0, 0, 1]; \nGreen\  = \ RGBColor[0, 1, 0]; \n
    Red\  = \ RGBColor[1, 0, 0]; \nYellow\  = \ RGBColor[1, 1, 0]; \n
    Cyan\  = \ RGBColor[0, 1, 1]; \nMagenta\  = \ RGBColor[1, 0, 1]; \n
    White\  = \ RGBColor[1, 1, 1]; \nBlack\  = \ RGBColor[0, 0, 0]; \)], 
  "Input",
  CellLabel->"In[4]:="]
}, Closed]],

Cell[CellGroupData[{

Cell["Define function PathGraph2D", "Subsubsection"],

Cell[BoxData[
    \(\(PathGraph2D\  = \ 
      Function[\n\t\t{l, pathnumber, \ varx, \ vary, \ color}, \n\t\t
        If[\((pathnumber\  <= \ 0)\)\  || \ \((pathnumber\  > \ Length[l])\), 
          \n\t\t\tPrint["\<Error: There are \>", \ Length[l], "\< paths.\>"], 
          \ \n\t\t\t
          If[\((Min[varx, vary]\  <= \ 0)\)\  || 
              \((Max[varx, \ vary]\  > \ 
                  Length[\(l[\([pathnumber]\)]\)[\([1]\)]])\)\ , \n
            \t\t\t\t\t\t
            Print["\<Error: There are \>", \ 
              Length[\(l[\([pathnumber]\)]\)[\([1]\)]], \ "\< variables.\>"], 
            \n\t\t\t\t\t
            ListPlot[
              Map[Function[q, \ q[\([{varx, \ vary}]\)]], \n\t\t\t\t\t\t\t\ 
                l[\([pathnumber]\)]], \n\t\t\t\t\t\tPlotJoined -> True, \ \n
              \t\t\t\t\t\tPlotStyle\  -> \ color, \n\t\t\t\t\t\t
              DisplayFunction -> Identity]]]]; \)\)], "Input",
  CellLabel->"In[5]:="]
}, Closed]],

Cell[CellGroupData[{

Cell["Define function PathGraph3D", "Subsubsection"],

Cell[BoxData[
    \(\(PathGraph3D\  = \ 
      Function[{l, \ pathnumber, \ varx, \ vary, \ varz, \ color}, \ \t\t
        If[\((pathnumber\  <= \ 0)\)\  || \ \((pathnumber\  > \ Length[l])\), 
          \n\t\t\tPrint["\<Error: There are \>", \ Length[l], "\< paths.\>"], 
          \ \n\t\t\t
          If[\((Min[varx, vary, \ varz]\  <= \ 0)\)\  || 
              \((Max[varx, \ vary, \ varz]\  > \ 
                  Length[\(l[\([pathnumber]\)]\)[\([1]\)]])\)\ , \n
            \t\t\t\t\t\t
            Print["\<Error: There are \>", \ 
              Length[\(l[\([pathnumber]\)]\)[\([1]\)]], \ "\< variables.\>"], 
            \n\t\t\t\t
            ScatterPlot3D[
              Map[Function[q, \ q[\([{varx, \ vary, \ varz}]\)]], \ 
                l[\([pathnumber]\)]], \ \n\t\t\t\t\tPlotJoined -> True, \ \n
              \t\t\t\t\tPlotStyle -> color, \ \n\t\t\t\t\t
              DisplayFunction -> Identity]]]]; \)\)], "Input",
  CellLabel->"In[6]:="]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Open parameters file", "Section"],

Cell["\<\
We first need to open the file which contains the parameters which \
determine the system to solve and how we go about solving it.  The file is \
called \"stoparams.c\" and is a C file.  It must be in your current \
directory.  The command I included here opens the file using emacs, but it is \
not important which text editor is used.  The next step is to edit this file \
appropriately and MAKE SURE TO SAVE  the file before continuing.\
\>", "Text"],

Cell[BoxData[
    \(\(! "\<emacs stoparams2.c&\>"\)\)], "Input",
  CellLabel->"In[6]:="]
}, Closed]],

Cell[CellGroupData[{

Cell["Generate new noise", "Section"],

Cell["\<\
In order to generate new noise files, you must delete the old ones. \
 New noise files (called bpaths1, bpaths2, ...) will be generated when you \
run stosolve.  If you change the seed for the random number generator, you \
need to run this line or the change will not take effect.\
\>", "Text"],

Cell[BoxData[
    \(\(! "\</bin/rm bpaths*\>"\)\)], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Compile and run stosolve", "Section"],

Cell[CellGroupData[{

Cell["\<\
This command compiles the C code with the parameters you have just \
entered.  Any time you change the parameters be sure to save the file and \
then run this command.  It is important that \"makefile\" be in the current \
directory.\
\>", "Text"],

Cell[BoxData[
    \(\(! make\)\)], "Input",
  CellLabel->"In[7]:="]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
This command runs the C program.  It reads noise from the (binary) \
file \"bpaths\" and stores the solutions in a (text) file called \
\"soln.\"\
\>", "Text"],

Cell[BoxData[
    \(\(! stosolve2\)\)], "Input",
  CellLabel->"In[8]:="]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Load solutions", "Section"],

Cell[CellGroupData[{

Cell["\<\
We need to create a list of the points in the solution paths.  This \
produces list (called p) of each path, reading from the file \"soln.\"  Each \
path consists of  a list of points.  Each point has first coordinate time and \
then the coordinates of the path in order.\
\>", "Text"],

Cell[BoxData[
    \(\(\nsolnfile\  = \ OpenRead["\<soln\>"]; \nml\  = \ {}; \n
    r\  = \ empty; \n
    While[r\  =!= \ EndOfFile, \n\tr\  = \ empty; \n\tl\  = \ {}; \n\t
      While[\((r\  =!= \ EndOfFile)\)\  && \ \((r\  =!= \ "\<:\>")\), \n\t\t
        r\  = \ Read[solnfile, \ Word]; \n\t\t
        If[\((r\  === \ "\<:\>")\)\  || \ \((r\  === \ EndOfFile)\), \ 
          ml\  = \ Append[ml, l], l\  = \ Append[l, ToExpression[r]]]\n\t]]; 
    \nClose[solnfile]; \np\  = \ {}; \n
    For[j\  = \ 1, \ j\  <= \ Length[ml], \ \(j++\), \n\tq\  = \ {}; \n\t
      For[i\  = \ 1, \ i < \ Length[ml[\([j]\)]], \ i\  += 3, \ 
        q\  = \ Append[q, \ \(ml[\([j]\)]\)[\([{i, \ i + 1, \ i + 2}]\)]\ ]]; 
      \n\t\tp\  = \ Append[p, q]]; \)\)], "Input",
  CellLabel->"In[7]:="]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Graph paths in 2D and 3D", "Section"],

Cell[CellGroupData[{

Cell["Introduction to graphing commands", "Subsection"],

Cell[CellGroupData[{

Cell["PathGraph2D, PathGraph3D", "Subsubsection"],

Cell["\<\
PathGraph2D and PathGraph3D graph the paths in 2D or 3D and have \
the following options:
PathGraph2D[l, pathnumber, varx, vary, color], 
PathGraph3D[l, pathnumber, varx, vary, varz, color]
\t1. The parameter l is the list of paths which are loaded above in the \
\"Load solutions\" section.
\t2. The parameter pathnumber is the number of the path which you want to \
graph (e.g. 1 is the first path).
\t3. The parameters varx, vary, varz are the numbers of the coordinates you \
want to graph, with 1 being time, 2 being the first coordinate of the \
solution, 3 being the second coordinate in the solution, etc. (e.g. for the \
phase space of the first two solution coordinates, we use varx = 2 and vary = \
3).
\t4. The parameter color is the color you want the path.  The options are \
Red, Blue, Green, Yellow, Cyan, Magenta, White, and Black.  You can also \
specify RGB values by setting the color to be RGBColor[a,b,c] where a,b,c are \
the RGB values.\
\>", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Show", "Subsubsection"],

Cell["\<\
The Show command displays the graphs.  The first parameters are the \
graphs (generated with PathGraph2D and PathGraph3D) which you want to show on \
the same axes.  The next parameters are options.  Among the options you must \
specify DisplayFunction->$DisplayFunction or the graphs will not be \
displayed.  
When viewing 3D graphs, it is often advisable to include the option \
ViewPoint->[a,b,c] where [a,b,c] is the point from which you are viewing the \
graph.  As this point is not at all obvious, the best way to use this option \
is to go to the Input Menu and select  \"3D ViewPoint selector\" (or \
<Shift>-<Mod1>-V).  This window allows you to select your viewpoint by the \
mouse.  Then press the Paste button to copy the viewpoint to the notebook and \
use it in the Show command.\
\>", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Some Tips and Shortcuts", "Subsubsection"],

Cell[TextData[{
  "Note that graphics objects can be assigned to variables, so the command \n\
\tgraph1 = PathGraph2D[p, 1, 1, 2, Blue];\nmakes perfect sense to ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " and allows you now to run commands like Show[graph1, \
DisplayFunction->$DisplayFunction].\nIf you want to show the same thing as \
you displayed in the previous command (perhaps with a change in ViewPoint or \
some other option), you can use the command Show[%, \
DisplayFunction->$DisplayFunction] (and you can change or add options) and \
the % simply takes same graphs that were shown last time."
}], "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Graph", "Subsection"],

Cell[CellGroupData[{

Cell["An example of graphing 2 paths in 2D.", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[PathGraph2D[p, 2, 2, 3, Blue], \ PathGraph2D[p, 1, 2, 3, Red], \n
      \tPathGraph2D[p, 3, 2, 3, Green], DisplayFunction -> $DisplayFunction]; 
    \)\)], "Input",
  CellLabel->"In[10]:="],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.467717 0.245179 0.145783 0.316429 [
[.22254 .13328 -6 -9 ]
[.22254 .13328 6 0 ]
[.7129 .13328 -3 -9 ]
[.7129 .13328 3 0 ]
[.95807 .13328 -3 -9 ]
[.95807 .13328 3 0 ]
[.45522 .06668 -30 -4.5 ]
[.45522 .06668 0 4.5 ]
[.45522 .22489 -24 -4.5 ]
[.45522 .22489 0 4.5 ]
[.45522 .304 -18 -4.5 ]
[.45522 .304 0 4.5 ]
[.45522 .38311 -24 -4.5 ]
[.45522 .38311 0 4.5 ]
[.45522 .46221 -6 -4.5 ]
[.45522 .46221 0 4.5 ]
[.45522 .54132 -24 -4.5 ]
[.45522 .54132 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.22254 .14578 m
.22254 .15203 L
s
[(-1)] .22254 .13328 0 1 Mshowa
.7129 .14578 m
.7129 .15203 L
s
[(1)] .7129 .13328 0 1 Mshowa
.95807 .14578 m
.95807 .15203 L
s
[(2)] .95807 .13328 0 1 Mshowa
.125 Mabswid
.27157 .14578 m
.27157 .14953 L
s
.32061 .14578 m
.32061 .14953 L
s
.36965 .14578 m
.36965 .14953 L
s
.41868 .14578 m
.41868 .14953 L
s
.51675 .14578 m
.51675 .14953 L
s
.56579 .14578 m
.56579 .14953 L
s
.61482 .14578 m
.61482 .14953 L
s
.66386 .14578 m
.66386 .14953 L
s
.76193 .14578 m
.76193 .14953 L
s
.81097 .14578 m
.81097 .14953 L
s
.86 .14578 m
.86 .14953 L
s
.90904 .14578 m
.90904 .14953 L
s
.1735 .14578 m
.1735 .14953 L
s
.12447 .14578 m
.12447 .14953 L
s
.07543 .14578 m
.07543 .14953 L
s
.0264 .14578 m
.0264 .14953 L
s
.25 Mabswid
0 .14578 m
1 .14578 L
s
.46772 .06668 m
.47397 .06668 L
s
[(-0.25)] .45522 .06668 1 0 Mshowa
.46772 .22489 m
.47397 .22489 L
s
[(0.25)] .45522 .22489 1 0 Mshowa
.46772 .304 m
.47397 .304 L
s
[(0.5)] .45522 .304 1 0 Mshowa
.46772 .38311 m
.47397 .38311 L
s
[(0.75)] .45522 .38311 1 0 Mshowa
.46772 .46221 m
.47397 .46221 L
s
[(1)] .45522 .46221 1 0 Mshowa
.46772 .54132 m
.47397 .54132 L
s
[(1.25)] .45522 .54132 1 0 Mshowa
.125 Mabswid
.46772 .0825 m
.47147 .0825 L
s
.46772 .09832 m
.47147 .09832 L
s
.46772 .11414 m
.47147 .11414 L
s
.46772 .12996 m
.47147 .12996 L
s
.46772 .1616 m
.47147 .1616 L
s
.46772 .17743 m
.47147 .17743 L
s
.46772 .19325 m
.47147 .19325 L
s
.46772 .20907 m
.47147 .20907 L
s
.46772 .24071 m
.47147 .24071 L
s
.46772 .25653 m
.47147 .25653 L
s
.46772 .27236 m
.47147 .27236 L
s
.46772 .28818 m
.47147 .28818 L
s
.46772 .31982 m
.47147 .31982 L
s
.46772 .33564 m
.47147 .33564 L
s
.46772 .35146 m
.47147 .35146 L
s
.46772 .36728 m
.47147 .36728 L
s
.46772 .39893 m
.47147 .39893 L
s
.46772 .41475 m
.47147 .41475 L
s
.46772 .43057 m
.47147 .43057 L
s
.46772 .44639 m
.47147 .44639 L
s
.46772 .47803 m
.47147 .47803 L
s
.46772 .49386 m
.47147 .49386 L
s
.46772 .50968 m
.47147 .50968 L
s
.46772 .5255 m
.47147 .5255 L
s
.46772 .55714 m
.47147 .55714 L
s
.46772 .57296 m
.47147 .57296 L
s
.46772 .58878 m
.47147 .58878 L
s
.46772 .60461 m
.47147 .60461 L
s
.46772 .05085 m
.47147 .05085 L
s
.46772 .03503 m
.47147 .03503 L
s
.46772 .01921 m
.47147 .01921 L
s
.46772 .00339 m
.47147 .00339 L
s
.25 Mabswid
.46772 0 m
.46772 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
0 0 1 r
.5 Mabswid
.06682 .53631 m
.10059 .54519 L
.13359 .55302 L
.14819 .55983 L
.1571 .56614 L
.1754 .5721 L
.18426 .57747 L
.20582 .58251 L
.21627 .58689 L
.23763 .59091 L
.25989 .59429 L
.27174 .59702 L
.28518 .59939 L
.30506 .60135 L
.33466 .60274 L
.36452 .60332 L
.39633 .60308 L
.43629 .60199 L
.48576 .59985 L
.53516 .59641 L
.59979 .5917 L
.66714 .58533 L
.74114 .57726 L
.82137 .56733 L
.87944 .55541 L
.92406 .54213 L
.94998 .52787 L
.9629 .51315 L
.96486 .49833 L
.96467 .4837 L
.9652 .46932 L
.95622 .45517 L
.95387 .44149 L
.95888 .4281 L
.94871 .41481 L
.94692 .402 L
.934 .38946 L
.92729 .37748 L
.92887 .36588 L
.92527 .35442 L
.92885 .34326 L
.92759 .33218 L
.92609 .32133 L
.92272 .31069 L
.91723 .30032 L
.9173 .29028 L
.91208 .28039 L
.90364 .27082 L
.89968 .26163 L
.89287 .2527 L
Mistroke
.89865 .2441 L
.88606 .23549 L
.89275 .22736 L
.88928 .21918 L
.88494 .21124 L
.88327 .20355 L
.87849 .19602 L
.88262 .18876 L
.88416 .1815 L
.87779 .17432 L
.86622 .16744 L
.86896 .16097 L
.85182 .15455 L
.85185 .14869 L
.84718 .14293 L
.84155 .13739 L
.84144 .13209 L
.84239 .12688 L
.84304 .12174 L
.8347 .11666 L
.82858 .11189 L
.82212 .10737 L
.8257 .1031 L
.83174 .0988 L
.83397 .09441 L
.82537 .09003 L
.81116 .08596 L
.8084 .08234 L
.80916 .07885 L
.81375 .0754 L
.81594 .07188 L
.82065 .06836 L
.81691 .06478 L
.79936 .06136 L
.79778 .05846 L
.79839 .05566 L
.79683 .05288 L
.79901 .05019 L
.7877 .04749 L
.77643 .04514 L
.77385 .04313 L
.76805 .04123 L
.77359 .03951 L
.77278 .03767 L
.76358 .03589 L
.74949 .03438 L
.75252 .03327 L
.74949 .03211 L
.75057 .03104 L
.74378 .02996 L
Mistroke
.74293 .02908 L
.74842 .02824 L
.74385 .02727 L
.7456 .02643 L
.74705 .02557 L
.74473 .02467 L
.73921 .02386 L
.73356 .0232 L
.71946 .02271 L
.71318 .02261 L
.7047 .02268 L
.70818 .02297 L
.70412 .02317 L
.7028 .02347 L
.70352 .0238 L
.6964 .02411 L
.69568 .0246 L
.68944 .02511 L
.68782 .02577 L
.68562 .02647 L
.67448 .02721 L
.67085 .02825 L
.67316 .02936 L
.66647 .03039 L
.65307 .03158 L
.64255 .03312 L
.63872 .03491 L
.63109 .03677 L
.62324 .03881 L
.61859 .04103 L
.61215 .04333 L
.60057 .04577 L
.59852 .04848 L
.58324 .0512 L
.57411 .05428 L
.56637 .05756 L
.56476 .06099 L
.55224 .06441 L
.54 .06811 L
.51562 .07208 L
.49233 .07663 L
.47821 .08174 L
.45087 .08714 L
.41164 .09318 L
.38524 .10018 L
.34511 .10777 L
.31956 .11632 L
.26797 .12541 L
.22762 .13573 L
.17978 .14697 L
Mistroke
.13348 .15931 L
.1004 .17268 L
.06393 .18673 L
.03974 .20151 L
.02873 .21671 L
.02543 .23194 L
.03474 .24701 L
.02381 .26158 L
.04419 .2762 L
.04555 .29003 L
.05045 .30359 L
.05053 .3168 L
.05847 .32978 L
.06658 .34233 L
.05658 .35445 L
.06079 .36665 L
.07467 .37852 L
.09118 .38982 L
.09908 .4005 L
.09664 .41078 L
.10578 .42095 L
.10628 .43071 L
.10888 .4403 L
.12229 .44965 L
.1256 .45849 L
.13336 .46709 L
.14442 .47534 L
.14596 .48315 L
.15968 .4908 L
.16311 .49794 L
.16924 .50488 L
.17819 .51153 L
.1846 .51784 L
.19157 .52386 L
.21264 .5296 L
.23519 .53468 L
.24966 .53906 L
.26782 .54299 L
.28118 .54636 L
.27844 .54931 L
.29561 .55229 L
.30922 .55476 L
.33238 .55682 L
.35987 .55822 L
.38739 .55886 L
.41423 .55875 L
.44318 .55792 L
.4813 .55632 L
.52904 .55373 L
.594 .5499 L
Mistroke
.65632 .54438 L
.72399 .53728 L
.79374 .52848 L
.84368 .51795 L
.89957 .50625 L
.9391 .49325 L
.95814 .4794 L
.96314 .46527 L
.94762 .45124 L
.94098 .43786 L
.92635 .42488 L
.93773 .41251 L
.94278 .40004 L
.94719 .38765 L
.94109 .37534 L
.93637 .3634 L
.93732 .35178 L
.921 .34034 L
.93132 .32952 L
.92749 .31861 L
.91967 .30798 L
.91126 .29774 L
.91158 .2879 L
.90718 .27821 L
.89505 .2688 L
.90116 .25988 L
.90264 .25093 L
.88655 .2421 L
.87782 .23385 L
.88476 .22597 L
.88559 .21803 L
.88602 .21021 L
.88222 .2025 L
.87765 .19503 L
.87371 .1878 L
.86465 .1808 L
.86129 .17416 L
.85694 .16772 L
.86708 .1615 L
.85944 .15511 L
.85874 .14904 L
.86315 .14309 L
.84986 .13711 L
.85444 .1316 L
.85011 .12605 L
.84239 .12071 L
.83202 .11567 L
.83293 .11099 L
.81792 .10637 L
.81672 .10222 L
Mistroke
.82173 .09818 L
.81739 .09407 L
.81643 .09014 L
.8023 .08631 L
.79563 .08292 L
.79652 .07976 L
.78994 .07664 L
.7818 .07374 L
.77298 .07111 L
.78094 .06876 L
.77399 .06624 L
.76956 .06395 L
.75908 .06181 L
.76214 .05999 L
.76154 .05812 L
.7595 .05629 L
.76698 .05455 L
.75996 .05264 L
.76304 .05095 L
.7626 .0492 L
.7741 .0475 L
.7737 .04552 L
.77432 .04357 L
.78184 .04165 L
.78244 .03955 L
.77584 .03747 L
.76372 .03561 L
.76419 .0341 L
.77126 .03261 L
.76577 .03095 L
.76964 .02946 L
.77202 .0279 L
.7706 .02629 L
.76652 .02476 L
.76447 .02335 L
.75888 .02203 L
.75768 .02088 L
.76637 .01978 L
.76786 .01846 L
.75696 .01713 L
.74751 .01611 L
.74125 .01536 L
.72357 .0148 L
.71767 .01472 L
.71718 .01479 L
.71313 .01488 L
.71102 .01508 L
.70333 .01533 L
.68666 .01579 L
.68904 .01668 L
Mistroke
.67905 .0175 L
.66931 .01857 L
.66615 .01988 L
.66099 .02126 L
.64797 .02275 L
.64223 .02457 L
.63933 .02651 L
.63176 .0285 L
.63154 .03066 L
.62587 .03279 L
.61335 .03503 L
.60462 .03758 L
.59018 .04032 L
.58484 .0434 L
.58174 .04657 L
.56099 .04977 L
.5501 .05348 L
.53337 .05741 L
.52201 .06174 L
.49893 .06629 L
.48259 .07139 L
.46258 .07685 L
.43195 .08275 L
.40195 .08938 L
.35588 .0967 L
.31312 .10514 L
.25432 .11459 L
.19155 .12546 L
.13933 .13784 L
.10435 .15142 L
.07198 .16571 L
.04461 .18063 L
.04255 .19604 L
.03546 .21125 L
.03266 .22639 L
.04405 .24136 L
.04724 .25578 L
.05056 .26986 L
.05087 .28363 L
.05578 .29715 L
.05846 .31032 L
.06148 .3232 L
.07405 .33578 L
.0725 .34781 L
.07589 .35968 L
.08261 .37127 L
.083 .38248 L
.0915 .39349 L
.09805 .40409 L
.09541 .41434 L
Mistroke
.10672 .42449 L
.11463 .43417 L
.12166 .44347 L
.12788 .45243 L
.13598 .46107 L
.15827 .46935 L
.15445 .47689 L
.15454 .48441 L
.16069 .4918 L
.16883 .4989 L
.1717 .50567 L
.1726 .51225 L
.17628 .51869 L
.18011 .52492 L
.18882 .53095 L
.18583 .53665 L
.1965 .54232 L
.20722 .54762 L
.22411 .55254 L
.242 .55693 L
.24788 .56076 L
.27622 .56437 L
.29022 .56716 L
.30197 .56952 L
.31042 .57153 L
.32872 .57328 L
.35124 .5745 L
.38002 .57511 L
.40685 .57492 L
.44276 .57402 L
.49269 .57217 L
.54022 .56901 L
.59492 .56462 L
.65824 .55883 L
.72845 .55144 L
.79784 .54228 L
.86645 .53141 L
.91439 .51888 L
.94266 .50527 L
.95661 .49112 L
.95729 .47683 L
.96903 .46277 L
.95418 .44862 L
.94729 .43511 L
.95005 .422 L
.95258 .40905 L
.9411 .39624 L
.94438 .38395 L
.95321 .37178 L
.94481 .35957 L
Mistroke
Mfstroke
1 0 0 r
.06034 .53631 m
.10508 .54537 L
.11517 .55307 L
.13289 .56038 L
.15739 .56709 L
.17189 .57302 L
.17976 .57847 L
.19965 .58362 L
.22778 .58814 L
.24475 .59184 L
.26272 .59501 L
.27371 .59765 L
.30438 .59995 L
.3201 .60139 L
.3382 .60238 L
.37292 .60287 L
.40838 .60241 L
.44552 .60101 L
.48234 .59863 L
.5376 .59531 L
.6036 .59055 L
.67312 .5841 L
.74969 .57589 L
.83332 .56575 L
.91118 .55354 L
.94704 .53943 L
.97619 .5246 L
.97434 .50923 L
.96508 .49416 L
.96229 .4796 L
.95403 .46535 L
.95703 .45156 L
.95464 .43793 L
.94823 .42458 L
.95134 .41163 L
.95239 .39881 L
.94907 .38617 L
.92343 .37384 L
.93454 .3624 L
.93425 .35085 L
.93464 .3395 L
.92497 .32834 L
.93175 .31761 L
.91655 .30689 L
.91443 .29675 L
.9145 .28684 L
.90926 .27709 L
.90625 .26764 L
.90764 .25843 L
.90428 .24934 L
Mistroke
.89496 .24049 L
.88958 .23204 L
.87867 .22388 L
.88282 .21614 L
.87578 .20842 L
.87597 .20102 L
.87352 .19374 L
.87149 .18664 L
.86526 .17972 L
.86359 .17308 L
.85577 .16659 L
.85405 .16043 L
.85355 .15441 L
.8519 .14851 L
.84915 .14275 L
.85287 .13715 L
.84657 .13156 L
.834 .12622 L
.82741 .12132 L
.82586 .11667 L
.82504 .11214 L
.82967 .1077 L
.83628 .10322 L
.83916 .09864 L
.83084 .09405 L
.82706 .08976 L
.80554 .08565 L
.79098 .08218 L
.79839 .07916 L
.80562 .076 L
.79773 .07269 L
.80716 .06965 L
.80676 .06641 L
.80259 .06323 L
.79975 .06022 L
.79367 .05733 L
.79219 .05465 L
.78942 .05206 L
.78421 .04959 L
.78213 .0473 L
.77794 .0451 L
.77949 .04305 L
.77689 .04099 L
.7727 .03904 L
.77112 .03724 L
.76758 .0355 L
.75679 .03389 L
.76716 .0326 L
.75348 .03105 L
.7617 .02989 L
Mistroke
.75702 .02853 L
.75964 .02732 L
.7641 .02606 L
.76522 .0247 L
.76648 .02334 L
.7566 .02196 L
.75191 .02087 L
.7533 .01993 L
.7429 .01896 L
.73004 .01829 L
.72404 .01798 L
.71725 .01783 L
.7005 .01787 L
.7074 .01835 L
.69913 .01865 L
.69106 .01916 L
.67882 .01988 L
.66812 .02092 L
.66285 .02222 L
.65449 .02365 L
.66105 .02528 L
.65546 .0267 L
.65046 .02825 L
.64744 .02991 L
.64562 .03163 L
.64128 .03336 L
.63223 .03518 L
.63287 .03722 L
.63248 .0392 L
.63314 .04116 L
.63173 .04307 L
.6276 .04499 L
.62539 .04698 L
.62207 .04901 L
.62015 .05108 L
.61965 .05318 L
.62724 .05525 L
.61467 .05709 L
.61872 .05923 L
.6203 .06122 L
.61047 .06315 L
.60642 .0653 L
.59591 .06753 L
.59682 .07 L
.5912 .0724 L
.58487 .07492 L
.57525 .07757 L
.57747 .08043 L
.56899 .08318 L
.56163 .08612 L
Mistroke
.55285 .0892 L
.53941 .09247 L
.53381 .09604 L
.52144 .09971 L
.51421 .10365 L
.49979 .10771 L
.4874 .1121 L
.47061 .11675 L
.45366 .12177 L
.43334 .12716 L
.40793 .13301 L
.37711 .13945 L
.34353 .1466 L
.30528 .15454 L
.2723 .16338 L
.23526 .17295 L
.19834 .18337 L
.15692 .1946 L
.12375 .20675 L
.09659 .2196 L
.06194 .23296 L
.05058 .24704 L
.03941 .26118 L
.05212 .27539 L
.05507 .28902 L
.05736 .30234 L
.06576 .31538 L
.06403 .32797 L
.05392 .3404 L
.06797 .3529 L
.07731 .36481 L
.08047 .37627 L
.08024 .38746 L
.08814 .39846 L
.09108 .40907 L
.10098 .41943 L
.12483 .42934 L
.12963 .43845 L
.12148 .44728 L
.12813 .45618 L
.14193 .46475 L
.15283 .47281 L
.16321 .48044 L
.17855 .48766 L
.19416 .49435 L
.20287 .50051 L
.21706 .50634 L
.22598 .51168 L
.23606 .5167 L
.24859 .52136 L
Mistroke
.26727 .52561 L
.27837 .52928 L
.28877 .53259 L
.29862 .53557 L
.31072 .53824 L
.33368 .54053 L
.35862 .54217 L
.39182 .54311 L
.41159 .54315 L
.44058 .54265 L
.48639 .54138 L
.53234 .5389 L
.57754 .53522 L
.628 .53039 L
.6918 .52429 L
.76361 .51657 L
.84989 .50705 L
.90192 .49536 L
.93418 .48248 L
.9581 .46894 L
.96043 .45498 L
.95475 .4412 L
.95397 .42779 L
.95345 .41463 L
.95055 .40171 L
.92988 .38907 L
.9277 .37721 L
.92746 .3656 L
.93167 .35419 L
.93965 .34285 L
.92427 .33149 L
.93642 .32074 L
.91287 .30983 L
.91075 .29974 L
.90688 .28988 L
.90316 .28029 L
.89454 .27095 L
.90139 .262 L
.89432 .25302 L
.89467 .24438 L
.88406 .23587 L
.88442 .22779 L
.88778 .21983 L
.88645 .21192 L
.88428 .20417 L
.87938 .19661 L
.87501 .18931 L
.87588 .18225 L
.86069 .17528 L
.85831 .16884 L
Mistroke
.86828 .16257 L
.86211 .15613 L
.85575 .14997 L
.854 .14408 L
.84791 .13834 L
.84694 .13285 L
.84494 .12749 L
.83918 .12226 L
.8428 .11728 L
.83741 .11229 L
.83727 .10752 L
.8332 .10284 L
.83562 .09834 L
.83029 .09385 L
.83151 .08958 L
.82328 .08535 L
.81522 .08142 L
.81404 .07776 L
.81172 .0742 L
.81438 .07075 L
.81013 .0673 L
.80113 .06401 L
.78543 .06103 L
.78593 .05851 L
.78191 .05602 L
.7754 .05369 L
.77718 .05156 L
.77695 .04943 L
.76552 .04733 L
.76911 .04558 L
.78158 .04376 L
.77759 .04164 L
.77976 .03966 L
.77489 .03765 L
.77068 .03581 L
.76788 .03411 L
.76769 .03252 L
.76356 .03095 L
.75564 .02953 L
.75152 .02834 L
.75176 .02728 L
.7512 .02623 L
.74286 .02522 L
.7389 .02444 L
.73618 .02379 L
.73844 .02322 L
.72277 .02259 L
.71568 .0224 L
.72075 .02241 L
.71381 .02228 L
Mistroke
.70782 .02233 L
.7098 .02255 L
.70284 .02271 L
.69466 .02305 L
.68886 .02361 L
.6846 .02431 L
.68269 .02512 L
.67085 .02597 L
.66472 .02712 L
.67371 .02841 L
.68395 .02945 L
.6815 .03019 L
.6748 .03098 L
.67009 .03194 L
.66914 .03301 L
.6704 .03409 L
.67598 .03511 L
.67443 .03597 L
.68286 .03686 L
.67425 .0375 L
.66845 .03837 L
.66938 .03938 L
.67106 .04034 L
.67112 .04125 L
.66898 .04214 L
.66469 .04307 L
.66651 .0441 L
.66306 .04506 L
.66204 .0461 L
.65288 .04715 L
.63286 .04843 L
.61879 .05023 L
.61299 .05238 L
.61161 .05464 L
.59461 .05691 L
.58786 .05959 L
.5806 .06241 L
.57837 .06538 L
.55801 .06836 L
.54962 .07184 L
.53406 .07548 L
.52537 .07948 L
.50631 .08365 L
.48855 .08826 L
.47205 .09328 L
.45756 .09865 L
.43467 .10432 L
.40711 .11052 L
.37093 .11735 L
.33445 .12504 L
Mistroke
.29676 .13358 L
.24236 .143 L
.18774 .15372 L
.14196 .16573 L
.10854 .17877 L
.07711 .19249 L
.0579 .20683 L
.0525 .22145 L
.04223 .23597 L
.03615 .25052 L
.03739 .26499 L
.03174 .27919 L
.04276 .2933 L
.05553 .30688 L
.06936 .3199 L
.08168 .33232 L
.0814 .34421 L
.10059 .3559 L
.10106 .36688 L
.09853 .37767 L
.10689 .38835 L
.10184 .39862 L
.102 .40886 L
.10555 .41892 L
.11878 .42872 L
.13577 .438 L
.13057 .44667 L
.13743 .45534 L
.13962 .46368 L
.15463 .47181 L
.14811 .47941 L
.16313 .48706 L
.17403 .49418 L
.17189 .50088 L
.18465 .50753 L
.19476 .51373 L
.21562 .51955 L
.22924 .52471 L
.2426 .52942 L
.24479 .5337 L
.26095 .53784 L
.26972 .54148 L
.26587 .54482 L
.27263 .54821 L
.28472 .55137 L
.29006 .55414 L
.30608 .55673 L
.32806 .55884 L
.34693 .56033 L
.37295 .56128 L
Mistroke
Mfstroke
0 1 0 r
.05251 .53631 m
.09356 .54558 L
.11258 .55359 L
.1299 .56096 L
.15481 .56774 L
.17339 .57373 L
.18274 .57913 L
.20541 .58418 L
.2237 .58854 L
.24605 .59234 L
.27217 .59547 L
.29436 .59785 L
.32159 .59959 L
.34952 .60057 L
.39 .60078 L
.42631 .5999 L
.47227 .59806 L
.5251 .59502 L
.58641 .5906 L
.66613 .58461 L
.74481 .57658 L
.81884 .56656 L
.88907 .55472 L
.93157 .54119 L
.95819 .52674 L
.96988 .51182 L
.96256 .49683 L
.97148 .48229 L
.95443 .46775 L
.95453 .45392 L
.96899 .4403 L
.95501 .42653 L
.94415 .41337 L
.94343 .40071 L
.94887 .38829 L
.93425 .37592 L
.93843 .36416 L
.93782 .35248 L
.92291 .341 L
.92568 .33013 L
.91723 .31936 L
.91439 .30899 L
.91121 .29888 L
.9168 .28901 L
.91383 .27917 L
.8988 .26956 L
.90417 .26052 L
.90048 .25149 L
.89088 .24271 L
.88147 .23433 L
Mistroke
.8818 .22634 L
.87446 .21848 L
.88574 .21095 L
.88432 .20324 L
.87648 .19569 L
.87724 .18848 L
.86906 .18138 L
.85819 .17461 L
.86747 .16824 L
.86835 .16173 L
.86674 .15531 L
.87162 .14904 L
.86258 .14274 L
.85264 .13679 L
.8546 .1312 L
.84993 .12566 L
.8452 .12033 L
.8401 .11522 L
.83823 .11033 L
.84838 .10557 L
.84116 .10062 L
.82455 .09595 L
.80908 .0918 L
.80954 .08814 L
.8199 .08452 L
.81923 .08069 L
.82084 .07694 L
.81713 .0732 L
.81118 .06963 L
.80503 .06628 L
.80406 .06315 L
.79443 .0601 L
.7914 .05736 L
.77844 .05474 L
.7873 .05252 L
.78244 .0501 L
.77261 .04784 L
.77312 .04589 L
.76832 .04396 L
.77012 .04219 L
.78006 .0404 L
.77401 .03837 L
.77355 .03654 L
.76325 .03475 L
.76096 .03327 L
.75261 .03188 L
.74509 .03073 L
.74515 .02981 L
.73356 .02889 L
.72548 .02831 L
Mistroke
.72419 .02795 L
.72312 .02763 L
.72209 .02735 L
.71724 .0271 L
.71825 .02698 L
.72338 .02684 L
.7216 .02656 L
.7089 .02634 L
.70589 .02646 L
.69188 .02666 L
.68906 .02723 L
.6919 .02787 L
.68895 .02842 L
.68327 .02904 L
.6706 .02981 L
.66498 .0309 L
.65022 .03213 L
.65056 .03373 L
.65033 .03529 L
.65001 .03684 L
.64926 .03837 L
.64825 .03989 L
.64106 .04142 L
.62229 .04311 L
.62196 .04528 L
.62375 .04742 L
.62223 .04948 L
.62286 .05155 L
.61172 .05356 L
.60733 .05584 L
.59962 .0582 L
.59664 .06073 L
.58953 .06329 L
.58017 .06601 L
.5796 .06893 L
.57554 .07182 L
.57449 .07476 L
.57843 .07769 L
.56706 .08046 L
.56862 .0835 L
.56391 .08644 L
.55209 .08945 L
.54109 .09274 L
.53384 .09626 L
.5152 .09993 L
.51278 .10403 L
.50194 .10813 L
.4921 .11245 L
.4774 .11696 L
.45841 .1218 L
Mistroke
.44501 .12706 L
.42272 .1326 L
.39684 .13864 L
.36703 .14528 L
.33471 .15261 L
.29706 .16069 L
.25482 .16965 L
.21354 .17959 L
.17124 .19047 L
.13633 .20231 L
.10725 .2149 L
.07779 .22805 L
.05775 .24178 L
.06195 .25582 L
.06091 .26951 L
.04931 .283 L
.0493 .29658 L
.05446 .30993 L
.05831 .32292 L
.06271 .33559 L
.06722 .34793 L
.07887 .35994 L
.08962 .37144 L
.0903 .38246 L
.10175 .39328 L
.09827 .40361 L
.09674 .41386 L
.10977 .42398 L
.11665 .43358 L
.1236 .44284 L
.12861 .45176 L
.1365 .46039 L
.14786 .46867 L
.15928 .4765 L
.17067 .4839 L
.18617 .49086 L
.18895 .4973 L
.19234 .50355 L
.21165 .5096 L
.21582 .51504 L
.22497 .52027 L
.2349 .52517 L
.24493 .52972 L
.24479 .53393 L
.25131 .53807 L
.25845 .54196 L
.26989 .5456 L
.27997 .54887 L
.30642 .55182 L
.31775 .554 L
Mistroke
.33968 .55585 L
.35544 .55707 L
.38646 .55785 L
.40829 .55778 L
.45785 .55712 L
.49953 .55515 L
.55119 .55208 L
.60369 .54768 L
.6773 .54194 L
.74906 .53432 L
.82115 .52489 L
.88613 .51369 L
.92182 .50092 L
.95176 .48741 L
.95509 .47331 L
.95554 .45936 L
.96271 .44564 L
.96419 .43194 L
.95765 .41844 L
.94668 .40534 L
.9378 .39275 L
.93082 .38061 L
.93187 .36886 L
.9253 .35727 L
.92304 .34606 L
.92396 .33509 L
.93264 .32429 L
.92942 .31342 L
.90742 .30283 L
.90635 .293 L
.91022 .28337 L
.90128 .2738 L
.8951 .26462 L
.88406 .25576 L
.88322 .24735 L
.90074 .2391 L
.89031 .23052 L
.88907 .22236 L
.88437 .21436 L
.89395 .20663 L
.88743 .19877 L
.87224 .19122 L
.868 .1842 L
.86031 .17741 L
.86454 .17094 L
.87084 .16447 L
.86494 .15793 L
.86631 .15166 L
.86598 .14546 L
.86462 .13938 L
Mistroke
.86125 .13343 L
.8501 .12766 L
.83746 .1223 L
.83315 .11736 L
.83748 .11263 L
.82378 .10785 L
.83281 .10353 L
.81835 .09903 L
.80718 .09499 L
.80243 .09133 L
.80561 .08785 L
.81618 .08435 L
.81385 .08061 L
.81393 .07701 L
.81528 .07346 L
.80102 .06994 L
.79757 .06685 L
.7958 .06391 L
.78248 .06107 L
.78234 .05863 L
.78652 .05624 L
.78233 .05378 L
.77775 .05147 L
.77931 .04932 L
.77451 .04716 L
.77244 .04517 L
.7653 .04327 L
.75765 .04159 L
.7472 .04015 L
.75146 .03901 L
.74336 .03777 L
.75008 .03678 L
.74843 .03561 L
.7464 .03452 L
.74689 .03349 L
.73607 .03247 L
.73023 .03176 L
.73098 .03122 L
.73196 .03066 L
.7377 .03009 L
.74379 .02937 L
.73736 .02851 L
.73704 .02783 L
.73654 .02716 L
.74202 .02653 L
.73931 .02576 L
.73219 .02507 L
.73642 .02458 L
.73258 .02399 L
.72672 .02352 L
Mistroke
.71974 .0232 L
.716 .02309 L
.71623 .02307 L
.70532 .02305 L
.70723 .02332 L
.70754 .02353 L
.70249 .02374 L
.6984 .02407 L
.70681 .02451 L
.7052 .02472 L
.69765 .02497 L
.69339 .02541 L
.6912 .02597 L
.68111 .02657 L
.67042 .02743 L
.6728 .02857 L
.67689 .02963 L
.6811 .03055 L
.67312 .03135 L
.67246 .03235 L
.67593 .03335 L
.6827 .03424 L
.67311 .03493 L
.67242 .03587 L
.65898 .03681 L
.65144 .0381 L
.64026 .03957 L
.64831 .04132 L
.64385 .04282 L
.63593 .04441 L
.63831 .04619 L
.63037 .04788 L
.62576 .04975 L
.61401 .05172 L
.61523 .05397 L
.60963 .05615 L
.59282 .05844 L
.58491 .06115 L
.56048 .06402 L
.54687 .0675 L
.53989 .07129 L
.53063 .07521 L
.50744 .07931 L
.48751 .08396 L
.47806 .08908 L
.44815 .09436 L
.41492 .10036 L
.37787 .10715 L
.33621 .11482 L
.29558 .12349 L
Mistroke
.25501 .1331 L
.21073 .14365 L
.16813 .15521 L
.12303 .16772 L
.08268 .18124 L
.06061 .19561 L
.04454 .21034 L
.03835 .22526 L
.02959 .2401 L
.02792 .25492 L
.03649 .26954 L
.04233 .28369 L
.04166 .29744 L
.04463 .31098 L
.04495 .32422 L
.04762 .33723 L
.04865 .34995 L
.06708 .36243 L
.07075 .3742 L
.07449 .38568 L
.08727 .39687 L
.09167 .40753 L
.09057 .41789 L
.09868 .42811 L
.11132 .43795 L
.12297 .44727 L
.10909 .45613 L
.12347 .46522 L
.13534 .47376 L
.16196 .48185 L
.17055 .48908 L
.18847 .49597 L
.19116 .50225 L
.19261 .50836 L
.21732 .51433 L
.23061 .51954 L
.24259 .5243 L
.24903 .52866 L
.25623 .53277 L
.26614 .53662 L
.27784 .54014 L
.28438 .54329 L
.28357 .54621 L
.30291 .5491 L
.32156 .55142 L
.35152 .55321 L
.37225 .55416 L
.38887 .55453 L
.42463 .55445 L
.46371 .55342 L
Mistroke
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[10]:=",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/@P10`40000100000?ooog1`L<c<c;34
gZo^k[Ng]fIVI_Cdm42P`0000000E@00ZP00o`1E001EE@1EZP1Eo`2Z002ZE@2ZZP2Zo`3o003oE@3o
ZP3ooeD005D0EED0ZUD0oeEE05EEEEEEZUEEoeFZ05FZEEFZZUFZoeGo05GoEEGoZUGoojX00:X0EJX0
ZZX0ojYE0:YEEJYEZZYEojZZ0:ZZEJZZZZZZoj[o0:[oEJ[oZZ[oool00?l0EOl0Z_l0oomE0?mEEOmE
Z_mEoonZ0?nZEOnZZ_nZoooo0?ooEOooZ_ooo`0001lO7clo?emOGgmoOinOWkno_moOgoooon?U]knH
][OUiBgUiL30`9T0I_ooc>OWifm_Kh0003<cIVIVc820P9VIo`00P28R8X200020P4:JY`0g?0200?mV
<omVc1YOI`00kUDJRj>SXiVIVIT00000<`00IP00V@00c00c000c<`0cIP0cV@0cc00co`1V001V<`1V
IP1VV@1Vc01Vo`2I002I<`2IIP2IV@2Ic02Io`3<003<<`3<IP3<V@3<c03<o`3o<`3oIP3oV@3oc3<0
03<0<c<0IS<0VC<0c3<0oc<c03<c<c<cVC<cc3<coc=V03=V<c=VIS=VVC=Vc3=Voc>I03>I<c>IIS>I
VC>Ic3>Ioc?<03?<<c?<IS?<VC?<c3?<oc?o03?o<c?oIS?oVC?oc3?oofH006H0<fH0IVH0VFH0c6H0
ofHc06Hc<fHcIVHcVFHcc6HcofIV06IV<fIVVFIVofJI06JI<fJIIVJIVFJIc6JIofK<06K<<fK<IVK<
VFK<c6K<ofKo06Ko<fKoIVKoVFKoc6KooiT0<iT0VIT0c9T0oiTc09Tc<iTcIYTcVITcc9TcoiUV09UV
<iUVIYUVVIUVc9UVoiVI09VI<iVIIYVIc9W<09W<<iW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wo
ol`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc8I900<:BDT0UdT008I90PZHB@00QTT00`Y9
B@2GB@00QTT00`Y9B@2GB@00QTT00`Y9B@0kBAL=ADT008I90PXfB@<=63X014T=3@e3B@00QTT00`Y9
B@0`B@8=1SX4B@<j2aH013Y9BDT5>TA90026B@032TU902e90`d2B@Dj41H00dTF5P025P<j00E9>SXj
3@13B@00QTT00`Y9B@0[B@8=0dT2>P03BAHF00TF24T75P033CXj008j0`e2B@00QTT00`Y9B@0XB@<=
0TT4>P8F00@=BDTj11H=B@XF0SX00`e9B@10B@00QTT22R=91Pd3B@053CX=3DT02AH2>Q590aH23@@F
00<jBDT0?dT008I900<:BDT07dT33@E910d00dTF5P045PA90SXEB@@F0TT01@dF5QH=03m90026B@03
2TU901i900<=BDT014T23@990aH013XF5QH5B@8j6TT45P<j?dT008I900<:BDT06TT43@991@d4B@03
5SXF00<F1SXMB@8j0QH014Tj3@dmB@00QTT22QM910d3B@<=1dT45RI90Pd00dTj>P025P8j0`djB@00
QTT00`Y9B@0BB@@=14T33@M90SX25P<j9DT33@9900Lj5QHFBCX=03]90026B@032TU900i910d7B@03
3DU900A91QH00dTj>P0WB@8=14T01SXF5QHj3C]9001/B@8:14T00`Y9B@02B@@:0TT32PI900<:BDT0
2dT33@U90`d00dTF5P035P03BAHF00A90SXWB@8=1DT25P04>QHF>P8=>4T006]900@:BDT:24T00`Y9
B@06B@032TU900=900<:BDT024T33@Q90Pd65P8j11H2B@@j:dT23@E91AH014T=3@deB@00JdT010Y9
B@X9B@032TU900E900<:BDT00dT32PA910d6B@@=0aH00dTj>P02>P=900@FBDU90SXaB@<=14T00aI9
5P025P03B@d=03A9001UB@@:0TT010Y9B@X:B@042TU9B@<:1TT00`Y9B@043@=91`d2B@8F13X6B@03
5TU900@j<DT23@E900Hj5QHjBAHfB@00JdT010Y9B@X;B@042TU92PQ90`d3B@@=1DT45P8j2DT011I9
>SXgB@8=14T2>P@F=DT006]900@:BDT:24T010Y9B@X2B@032TU900H=00@:BDU90Pd4B@DF0cX<B@8F
>dT23@8j0dT010dFBAHdB@00K4T22PY90PX63@032TU900=900@:3@d=0dT35P=90SX=B@8F0SXlB@8j
00D=BDT=B@025SE9001fB@@=2DT43@=90aH2B@@j3DT25T9913X45P03>TU90359001`B@H=1TT73@=9
00@:5QHF1CX=B@04>QHF5TI911H5>S59001ZB@H=14T83@M911H3>Q190aI9B@@F00D=BDT=>P0aB@00
I4T63@I910d8B@LF00A9>SX:3dT45Te90QH00dTF5P035P03>TU902i9001PB@@=24T43@E91aH4B@@j
0TT22P]90aH00cY9B@1@B@<F00=93@d00aH^B@00G4T43@I91Pd3B@HF1DT6>PI900<:BDT024T2>P03
5TU905A900<=BAH00QH013Y9BAH]B@00ETT63@I910d3B@HF1dT4>Pa900<:BDT01DT3>P8FEdT23@03
5SY9008j00=95QH0;4T004i920d9B@<=00=95QH011H9B@@j44T00`Y9B@03B@8j00=95QH0FdT010dF
5TT25Ri90011B@@:00=93@d01Pd?B@033AHF00<F2TT5>QA90PX00dTj>P045S]910XNB@03>Pd=008F
7dT42Pa90010B@H=4TT23@@F2TT5>QQ900<j5QH00aI1B@032TU901e900HjBAH=5SXNB@032TU900e9
000jB@H=0dT00`Y9B@0;B@D=11H9B@Dj6DT00cXF5P025P032TU904=900<:BDT07TT013Y95Pd25Qi9
00<:BDT034T003I910d00dT:2P022PA900<:BDT01DT63@990aH9B@@j6dT2>P8F14T00`Y9B@13B@03
2TU901m900HjBAHF3AHNB@032TU900]9000bB@@=3DT010Y9BDT43@A911H8B@@j74T3>P8F1TT00`Y9
B@13B@032TU902190cX25Qm900<:BDT02TT002i910d?B@8:10d4B@@F24T4>Qa913X35PQ90PY2B@<:
94T2>P03BAHF01Y900@:BDT:34T002Y910dBB@<=00@:BDU911H8B@@j6TT4>PHF2dT00`Y9B@13B@03
2TU9029900@=>SX=0QHJB@8:3DT002M90`dBB@@=0dT45PI91SXHB@@j1QHAB@032TU906Q900<=BCX0
0QH00dTFB@0WB@009DT23A990`d3B@@F1DT5>QY913X45QM900<:BDT0JDT010dj>Pd35RM9000R2P<=
40X43@032QHF00<F10X5>Q`:0cX45XP:00<=>PX00Pd00aH:2P0T2P590007B@032TU900]900<:BDT0
24T23@03B@Y900a910d00dTF5P025PA91CX;B@032TU900]900@:BDU90cX45P=900<:BDT02dT00`Y9
B@0;B@032TU900]900<:BDT02dT00`Y9B@0;B@032TU900]900<:BDT02dT00`Y9B@0;B@032TU900]9
00<:BDT02DT2>P043DTF5PY900<:BDT02dT00`Y9B@0;B@007DT33@e910d00dTF5P025P=91CX2B@03
2TU901M91CX45R=900<:BDT0@dT00`Y9B@0UB@04>TTjB@8F64T00`Y9B@0;B@006dT23@a910d2B@<F
0TT5>Qa91CX00dTF5P025RM900<:BDT0JdT01CX=3CXF02M9000HB@<=2TT43@=90aH5>Qe913X2B@@F
:dT00`Y9B@1[B@03>TT=008F9dT001I90Pd;B@8=0dT45P<j7TT4>P03BAHF00<F;dT22Ve900DjBDTF
3@0VB@0054T23@]90Pd2B@<F0cXNB@<j1QHdB@032TU906a900<j3CX00QHVB@004TT23@]90Pd014TF
5QH2>Qi90SX55SY900<:BDT0K4T010dj5QH23BE9000@B@8=2TT23@@F00=9>SX07DT00cXF5P025Sm9
00<:BDT0KDT00cXFB@02B@033DU90299000>B@8=2DT00`dF5P025P=90SXKB@03>QHF008F@dT22Vi9
0QH01CY9BDT=02=9000=B@033DU900M900<=5QH014T3>QY911I7B@032TU906i90aH2B@033DU90259
000=B@033DU900A90Pd25PA90SXIB@@FBdT00`Y9B@1_B@065SXF>TT=8TT000e900<=BDT00TT23@03
BAI900=90SXGB@@FCdT00`Y9B@1`B@8F00<jB@d08TT000a900@=BDU90Pd2B@045TU9B@8j5DT013XF
5QICB@032TU907590QH00dT=B@0QB@0034T010e93@d2B@8F0TT2>QA911IFB@8:LTT00cXF5P0SB@00
34T23@=900<FBDT00SXCB@<FFTT00`Y9B@1bB@045SXF5R59000;B@8=0dT011I9BCXBB@<F00<jBDT0
FTT00`Y9B@1bB@05>QHjBAH084T000Y90Pd3B@045TTj>Q590QH00cY9B@1MB@032TU9079900Dj3AI9
5P0PB@002TT23@9900<FBCX04DT25P03>TU905m90PYcB@06>TTF>TTF7dT000U90Pd2B@045TU9>Pm9
0QITB@032TU907=900@j5QHF84T000U90Pd2B@045TU9>Pe90QIVB@032TU907990aH00cY9B@0OB@00
24T01@e93DTF009900<jBDT02DT00cXF5P1XB@032TU9075900DFBCX=5P0QB@0024T23@9900@FBDTj
2DT00cXF5P1@B@8:14T00`Y9B@02B@@:0TT32PI900<:BDT0LTT011Hj3AHQB@0024T23@05BAI9BCX0
1dT2>P8FDDT010Y9B@X8B@032TU900I900<:BDT00dT32W9900Dj5SXF3@0PB@001dT010e9BAH3B@03
>TU900=90SX25U=900@:BDT:2DT00`Y9B@05B@032TU900=900<:BDT0LdT011HjBAHPB@001dT010e9
BAH2B@03>TU900990SX25UE900@:BDT:2TT010Y9BDT32PI900<:BDT0LdT011Hj>QHPB@0024T01@dF
B@dj00A900<jBAH0EdT010Y9B@X;B@042TU92PQ900<:BDT0LTT01@dF>Pe9008F7TT000Q900DF3DT=
>P03B@03>QHF05Q900@:BDT:24T010Y9B@X2B@032TU900I90PYdB@053AHF3CX00QHMB@0024T01AH=
BCX=009900<j5TT0FTT22PY90PX3B@@:1DT00`Y9B@1eB@043AHF>Qi90008B@075TT=>Pe9>P02B@03
5TU9079900<:BDT0M4T25P043AI9>Qe90008B@035Pdj009900DjBDU95P1dB@032TU907A900LFBDTj
5Pdj01a90008B@035Pdj009900<jBDT00TT00aI9B@1aB@032TU907A900LFB@djBAHj01a90008B@03
5TU9008j00<=BDT00TT00aI9B@1aB@8:MDT021I9B@dj3AHj6dT000M900L=BAI9>TT=00A900<FBDT0
LDT00`Y9B@1eB@8F00E9>PdF>P0KB@0024T00`dFB@02>P03B@e9009900<FBDT0LDT00`Y9B@1gB@05
5PdFBCX06dT000U900H=5SY9>Pd2B@035TU9079900<:BDT0N4T25P03>TU901Y9000:B@055Pe9BCX0
0TT00aI9B@1bB@032TU907M900@j5TTj74T000Y900Lj5Pe93CXF07E90PYiB@04>QHF>Q]90009B@04
>TTFB@8=00<j5TT0M4T00`Y9B@1hB@03>TTF008F6TT000Y900<jBAH00Pd00aHjB@1dB@032TU907U9
00@j5PdF6TT000Y900LjBAH=5TTj07E900<:BDT0NDT01CY95SXF01U9000;B@06>QH=5PdjMDT22W]9
00@j5PdF6DT000]900Hj5PdF3CYeB@032TU907Y900@FB@dF6DT000a900DF3AI9>P1eB@032TU907Y9
00@F>QHj6DT000a900DF>QI9>P1QB@8:14T00`Y9B@02B@<:1TT00`Y9B@1kB@8F00<jBDT05dT000a9
00DFBCXF>P1PB@042TU92P]900<:BDT00dT00`Y9B@1kB@8F00<jBDT05dT000e900<F3AH00SYOB@04
2TU92P]900<:BDT00dT32Wa90QH00`e9B@0FB@003DT011H=5TT2>Ui900@:BDT:24T32PI900<:BDT0
O4T013XF5QHGB@003DT01QH=BAH=>Ui900@:BDT:24T00`Y9B@06B@032TU907e900<F>SX00QHEB@00
3DT01aH=BAH=BCX0GDT010Y9B@X8B@032TU900I900<:BDT0ODT01QI93CXF>QA9000=B@065TT=BAI9
0SYMB@8:2DT42PE90PYnB@075TU93DTF>P0CB@003DT01QI9B@dF>P9900<jBDT0KTT00`Y9B@1nB@05
5TT=>QH054T000e900HFBDT=5SX3B@03>TU906e900<:BDT0OdT00aHj5P0EB@003DT01aI9BDTj5Pd0
0TT00cY9B@1]B@032TU907m900<j5Pd05DT000e900PFBDU9>QI93@9900<jBDT0K4T00`Y9B@1oB@8F
00<jBDT04dT000e900HFBDTjBAH2B@033DTj06i90PZ0B@045Pdj>QA9000=B@035TTj009900HF3DT=
BCY^B@032TU907m900<FB@d00SXCB@003TT00aI9>P02B@055TT=BCX0KTT00`Y9B@1oB@8F00@=>TTj
4TT000i900PFBDTj3AI93@990SY/B@032TU907m900DFBAHF>P0CB@003TT021I9B@djBAH=14T2>VY9
0PZ0B@055Pe9>QH04dT000m90QH2B@04>TTF3@E90SYXB@032TU9081900@FBCXF4dT0011900L=5QI9
>QH=00I900<jBDT0ITT00`Y9B@20B@055SXj3AH04TT0015900H=BAI9>QH6B@03>TU906I900<:BDT0
PDT01AHj3AH=0159000AB@033DTF009900<j5TT014T00cY9B@1VB@032TU9081900Lj5TT=5TT=0119
000AB@073DU95TTj3@025P=900<jBDT0ITT22X590SX25P=900<=BDT03DT0019900<=BAH00TT013X=
BAH2B@03>TU906I900<:BDT0P4T01CX=5QH=00=900<=BDT034T001=900X=5TU9>Pe95TTjJDT00`Y9
B@1oB@8j00A95TT=0TT00`e9B@0=B@0054T02AH=BCX=BAI9>P1?B@8:14T00`Y9B@03B@032TU90099
0`X6B@032TU9081900Dj5SXFB@023A19000EB@065Pdj3DTF0TT00cY9B@1;B@042TU92PU900<:BDT0
1DT00`Y9B@03B@<:P4T013XFBCX25Q59000EB@035TTj008=00A95TTjCDT010Y9B@X:B@032TU900A9
00<:BDT00dT00`Y9B@21B@07>QI9B@dF>P0?B@005TT00aI9>P023@04BAI9>Ta900@:BDT:2TT010Y9
BDT32PI900<:BDT0PDT013XFBDT23@035TU900e9000GB@8F00I93@e95SY<B@042TU92P]900@:BDT:
24T00`Y9B@22B@07>QH=BAI9>P0>B@006DT01AI93DTF04e900@:BDT:24T010Y9B@X2B@032TU900I9
00<:BDT0Q4T01AH=5TTj00i9000IB@055TT=5SX0CTT22PU910X2B@@:1DT22XI900@F>TTj3TT001U9
00Dj5TTF>P1XB@032TU908E900@F3CXj3TT001Y900@FBAHjJ4T00`Y9B@24B@043AHF>Pm9000JB@04
5PdF>VQ900<:BDT0Q4T01@dF3AHj00i9000JB@055SXF3CX0IdT22XA900H=BAH=>QH>B@006TT01AI9
>QHj06M900<:BDT0PTT01@e9BDTF008j00<FBDT02dT001]900DFBCXF>P1VB@032TU9085900<=BDT0
0TT01CXF>TTF00a9000KB@055TU93AH00cYSB@032TU9081900<=BDT00dT01CXF>TTF00a9000LB@06
5TT=3AI90cYQB@032TU9085900<=BDT00dT01CXF3DTF00]9000MB@045Pe95P990SYQB@8:PdT01@e9
BDT=008j00<FBAH02dT001i90QH00dTFB@02B@8jGdT00`Y9B@23B@073DU93DTj3@025P]9000OB@04
3AI95P9900<jBCX0GTT00`Y9B@24B@053DU93CX00TT00aI9B@09B@007dT020e95PdF>TTjGdT00`Y9
B@25B@083DT=>TTFBAH:B@0084T23@045PdF>V190PZ6B@053DU93CX00QH;B@008TT35P03BCY905i9
00<:BDT0QTT01@e93CXF00a9000PB@8F00E93CXF>P1OB@032TU908I900D=BDTF>P0<B@007dT00aI9
B@02B@043CXF>Ui900<:BDT0QdT010dF5SX<B@0084T02AI9BDT=BCXF>P1DB@@:1DT00`Y9B@28B@04
5SXj3@]9000QB@8F0TT010e9>QIFB@032TU900A90`Z8B@055TTjB@d02TT002=900LFBDT=BCXF008=
DdT00`Y9B@04B@032TU908Q900@F>TT=2dT002A90QH01TT=BCXF>P8=DDT00`Y9B@04B@032TU908Q9
0QH00dT=B@0:B@009TT35P990QH00`e9B@1?B@032TU900A900<:BDT0QdT01CXF>QH=00]9000YB@8F
00@j3AHjCTT32PI90PZ6B@8j00E95Pdj5P0;B@00:TT013XF5TT25Tm900<:BDT014T00`Y9B@23B@8j
0dT01AH=>PdF00Y9000ZB@033DTj008F00=95QH0E4T00`Y9B@21B@8j14T01aH=3CX=BAH02DT002]9
0`d013XF5TT25U9900<:BDT0OTT3>PE90QH01DT=BCXF00Y9000]B@8=00E9>QHFB@025U1900<:BDT0
NdT3>PI90QH3B@043DTj5PY9000_B@043DU9>P8F00=95SX0CTT22WU90cX7B@8F00D=BDU93@02B@03
5TU900U9000_B@8=00A9>TU90QH00dTj>P1<B@032TU907A913X6B@8=0QH5B@063DU95Pdj2TT00319
00<=>TT014T011I9BCY;B@032TU9071913X6B@@=0QH6B@033DU9009900<FBCX02DT003190Pd2>P=9
0aH00dTj>P19B@032TU906a913X6B@@=0TT25PI90Pd5B@033AHj00U9000aB@8=00A9>SY90QH014TF
5QH3>TI90PYUB@Pj1TT43@=90aH6B@8=1dT010dFBCX8B@00<DT00`e93@02B@8j00=95QH00aH2B@8j
A4T00`Y9B@1HB@`j2TT43@E90QH7B@8=2DT010e95SX8B@00<TT00`e93@03B@<j0TT35P990SY2B@03
2TU904i92SXBB@@=14T55PM90Pd:B@053DTFBCX024T0039900@=BDT=1DT4>P@F00=9>SX0@4T00`Y9
B@17B@Lj64T43@991QH:B@8=2dT00`e95P02B@03>TU900I9000cB@043DU93@Q900@j5QI90QH00dTj
>P02>Sa900<:BDT0@DT6>QY91@d65Pa910d=B@033AI9009900<jBDT01TT003=900@=BDU90`d7B@03
>QHF008F14T4>SQ90PXeB@dj74T23@LF3dT33A1900<=5TT00TT2>PQ9000dB@8=14T53@A90SX35PI9
13XdB@032TU902Q933XQB@<=1aHBB@@=4DT00`dF5P02B@8j2TT003I900<=BDT01TT33@A900<F>QH0
0QH7B@8j<TT00`Y9B@0PB@Pj8dT63@LF54T53A=900<=5QH00TT2>Pa9000?B@8F0cX00`e9B@0NB@03
3DU900Y90`d2B@045SXjB@<F1TT3>QA910X3B@032TU9009910X2B@<:1TT00`Y9B@0FB@Xj8TT63@TF
54T73AI900<=5QH00TT2>Pi9000AB@@F0cXNB@8=3DT33@045Pdj>P990QH7B@`j2TT00`Y9B@07B@03
2TU900I900<:BDT00dT00`Y9B@09B@dj8TT63@XF5dT63A]900<=5QH00TT2>Q19000EB@@F0cXLB@8=
3dT35P8j00@=BAHF4DT_>RI93aHHB@T=7dT35P990SXBB@006DT45P8j00<=BDT064T33@m90aH23@<F
6DT00`Y9B@09B@042TU9B@<:1TT00`Y9B@0QB@hF7DT:3BA911H2B@<j54T001e90QH2>P<=6DT23@i9
0SXA5Pe900<:BDT02TT010Y9B@X8B@032TU901Q900<=5QH01QH53Ai920d[B@<F00@=BDU90SXGB@00
7dT35P03>TT=008=64T43@m90SX:5P999QH00`Y9B@0I5RA91`d_B@@F00<=BDT013XIB@008TT25P03
>TU9008=6TT33@m91SX00dT=3@0[5PP=;dT63C591AH00`dj>P04>Qe9000TB@<F00@jB@d=6dT53A19
4CX?3@]900<:BDT0:TT=3C991AH6>R=9000WB@@F00<j3DT07DT63CE900<:BDT06TT@3CY91AH5>RU9
000[B@<F0PdQB@@=<DT00`Y9B@0;B@l=A4T65P@j;dT002i90QH33B991`dZB@h=B4T4>PLF00<jBDT0
=TT003190aH23BM9:Pd22T]91SX75T19000bB@04>QHF5P8=CTT00`Y9B@13B@@j2AI7B@00=DT00cXF
5P025P8=BTT00`Y9B@0lB@/FD4T003Q90SX45TQ900<:BDT0<DT;5U]9000mB@03>QHF00<F00<jBDT0
@4T22RI931H2>VA90011B@8=1aHlB@032TU901I93aH4>Vi90017B@8=00<j5QH01AHeB@032TU900I9
41H5>Wa9001=B@@j6AH;BAXF0cZ>B@00EDT43A4j2aH@>P03B@Y909Q90026B@8:V4T008I900<:BDT0
UdT008I900<:BDT0UdT008I900<:BDT0UdT00001\
\>"],
  ImageRangeCache->{{{0, 287}, {176.938, 0}} -> {-1.91271, -0.460723, 
  0.0142466, 0.0110387}}]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["An example of a graph in 3D", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[PathGraph3D[p, 1, 1, 2, 3, Magenta], 
      ViewPoint -> {2.199, \ \(-1.020\), \ 2.361}, \n\t
      DisplayFunction -> $DisplayFunction]; \)\)], "Input",
  CellLabel->"In[11]:="],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1.29571 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics3D
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.254495 1.41914 0 1.41914 [
[.26204 1.27883 0 -3.20089 ]
[.26204 1.27883 6 5.79911 ]
[.52346 .90367 0 -3.08706 ]
[.52346 .90367 12 5.91294 ]
[.82536 .47039 0 -2.95137 ]
[.82536 .47039 12 6.04863 ]
[.79497 .00634 -2.12919 -9 ]
[.79497 .00634 9.87081 0 ]
[.86048 .03054 -1.02696 -9 ]
[.86048 .03054 4.97304 0 ]
[.9255 .05457 -0.98933 -9 ]
[.9255 .05457 5.01067 0 ]
[.99004 .07842 -0.95169 -9 ]
[.99004 .07842 5.04831 0 ]
[1.00967 .12445 0 -2.84506 ]
[1.00967 .12445 6 6.15494 ]
[1.012 .14818 0 -2.851 ]
[1.012 .14818 18 6.149 ]
[1.01435 .17212 0 -2.85699 ]
[1.01435 .17212 6 6.14301 ]
[ 0 0 0 0 ]
[ 1 1.29571 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.22981 1.29571 m
1 .18976 L
s
.245 1.27391 m
.23647 1.27145 L
s
[(0)] .26204 1.27883 -1 -0.28869 Mshowa
.50653 .89835 m
.49807 .8957 L
s
[(10)] .52346 .90367 -1 -0.31399 Mshowa
.80859 .46461 m
.8002 .46173 L
s
[(20)] .82536 .47039 -1 -0.34414 Mshowa
.125 Mabswid
.29446 1.20288 m
.28935 1.20138 L
s
.34528 1.1299 m
.34018 1.12838 L
s
.39753 1.05487 m
.39244 1.05333 L
s
.45126 .97772 m
.44617 .97615 L
s
.56342 .81667 m
.55835 .81505 L
s
.62199 .73256 m
.61693 .73092 L
s
.68232 .64593 m
.67727 .64426 L
s
.74449 .55666 m
.73945 .55495 L
s
.8747 .36967 m
.86968 .36791 L
s
.94294 .2717 m
.93792 .26991 L
s
.25 Mabswid
.72778 0 m
.99095 .0971 L
s
.78536 .02124 m
.78055 .0287 L
s
[(-1)] .79497 .00634 -0.64513 1 Mshowa
.85073 .04536 m
.84586 .05277 L
s
[(0)] .86048 .03054 -0.65768 1 Mshowa
.91563 .0693 m
.91069 .07667 L
s
[(1)] .9255 .05457 -0.67022 1 Mshowa
.98004 .09307 m
.97504 .10039 L
s
[(2)] .99004 .07842 -0.68277 1 Mshowa
.125 Mabswid
.79847 .02608 m
.79558 .03055 L
s
.81157 .03091 m
.80867 .03537 L
s
.82464 .03574 m
.82173 .04019 L
s
.8377 .04055 m
.83478 .045 L
s
.86375 .05017 m
.86082 .05461 L
s
.87675 .05496 m
.87381 .0594 L
s
.88973 .05975 m
.88678 .06418 L
s
.90269 .06453 m
.89973 .06896 L
s
.92855 .07407 m
.92558 .07849 L
s
.94145 .07883 m
.93847 .08324 L
s
.95433 .08358 m
.95134 .08799 L
s
.96719 .08833 m
.9642 .09273 L
s
.77222 .0164 m
.76935 .02087 L
s
.75907 .01154 m
.7562 .01603 L
s
.7459 .00668 m
.74303 .01117 L
s
.7327 .00182 m
.72985 .00631 L
s
.25 Mabswid
.99095 .0971 m
1 .18976 L
s
.99303 .11833 m
.9847 .11527 L
s
[(0)] 1.00967 .12445 -1 -0.36777 Mshowa
.99535 .14208 m
.98702 .13903 L
s
[(0.5)] 1.012 .14818 -1 -0.36644 Mshowa
.99768 .16604 m
.98935 .163 L
s
[(1)] 1.01435 .17212 -1 -0.36511 Mshowa
.125 Mabswid
.99349 .12306 m
.98849 .12123 L
s
.99395 .1278 m
.98896 .12597 L
s
.99441 .13255 m
.98942 .13072 L
s
.99488 .13731 m
.98988 .13548 L
s
.99581 .14686 m
.99081 .14503 L
s
.99628 .15164 m
.99128 .14981 L
s
.99675 .15643 m
.99175 .1546 L
s
.99721 .16123 m
.99222 .1594 L
s
.99815 .17086 m
.99316 .16903 L
s
.99863 .17568 m
.99363 .17386 L
s
.9991 .18051 m
.9941 .17869 L
s
.99957 .18536 m
.99457 .18354 L
s
.99256 .1136 m
.98757 .11176 L
s
.9921 .10888 m
.98711 .10704 L
s
.99164 .10417 m
.98665 .10233 L
s
.99118 .09947 m
.98619 .09763 L
s
.25 Mabswid
.00633 1.13765 m
0 1.2297 L
s
0 1.2297 m
.22981 1.29571 L
s
.22981 1.29571 m
.2329 1.20389 L
s
.2329 1.20389 m
.00633 1.13765 L
s
.72778 0 m
.99095 .0971 L
s
.99095 .0971 m
1 .18976 L
s
1 .18976 m
.73246 .09243 L
s
.73246 .09243 m
.72778 0 L
s
.00633 1.13765 m
0 1.2297 L
s
0 1.2297 m
.73246 .09243 L
s
.73246 .09243 m
.72778 0 L
s
.72778 0 m
.00633 1.13765 L
s
.2329 1.20389 m
.99095 .0971 L
s
.99095 .0971 m
1 .18976 L
s
1 .18976 m
.22981 1.29571 L
s
.22981 1.29571 m
.2329 1.20389 L
s
0 0 m
1 0 L
1 1.29571 L
0 1.29571 L
closepath
clip
newpath
1 0 1 r
.5 Mabswid
.28154 1.19841 m
.28273 1.19371 L
s
.27318 1.20092 m
.28154 1.19841 L
s
.28273 1.19371 m
.2822 1.18856 L
s
.2822 1.18856 m
.28317 1.18392 L
s
.26326 1.20287 m
.27318 1.20092 L
s
.28317 1.18392 m
.28287 1.17895 L
s
.28287 1.17895 m
.28518 1.17482 L
s
.2436 1.20167 m
.26326 1.20287 L
s
.28518 1.17482 m
.28623 1.17034 L
s
.28623 1.17034 m
.28636 1.16563 L
s
.02894 1.19632 m
.04081 1.19848 L
s
.28636 1.16563 m
.28869 1.16163 L
s
.04081 1.19848 m
.04456 1.19809 L
s
.28869 1.16163 m
.29054 1.15751 L
s
.29054 1.15751 m
.29138 1.15312 L
s
.29138 1.15312 m
.28705 1.14723 L
s
.22258 1.19977 m
.2436 1.20167 L
s
.04456 1.19809 m
.05011 1.19816 L
s
.28705 1.14723 m
.29123 1.14401 L
s
.29123 1.14401 m
.29277 1.13998 L
s
.05011 1.19816 m
.05726 1.1986 L
s
.29277 1.13998 m
.29447 1.13603 L
s
.05726 1.1986 m
.06207 1.19824 L
s
.29447 1.13603 m
.29383 1.13141 L
s
.20317 1.19806 m
.22258 1.19977 L
s
.29383 1.13141 m
.29701 1.12798 L
s
.06207 1.19824 m
.06534 1.19735 L
s
.29701 1.12798 m
.29509 1.12305 L
s
.29509 1.12305 m
.2962 1.11909 L
s
.06534 1.19735 m
.07143 1.19724 L
s
.18539 1.19655 m
.20317 1.19806 L
s
.2962 1.11909 m
.29782 1.11533 L
s
.07143 1.19724 m
.07947 1.19759 L
s
.29782 1.11533 m
.29821 1.11122 L
s
.07947 1.19759 m
.08489 1.19706 L
s
.29821 1.11122 m
.29911 1.1073 L
s
.08489 1.19706 m
.09056 1.19651 L
s
.16843 1.19503 m
.18539 1.19655 L
s
.29911 1.1073 m
.30103 1.10373 L
s
.09056 1.19651 m
.0946 1.19541 L
s
.30103 1.10373 m
.30186 1.09984 L
s
.15397 1.19401 m
.16843 1.19503 L
s
.0946 1.19541 m
.10327 1.1956 L
s
.30186 1.09984 m
.30129 1.09557 L
s
.10327 1.1956 m
.10843 1.19463 L
s
.14382 1.19412 m
.15397 1.19401 L
s
.10843 1.19463 m
.11416 1.19376 L
s
.30129 1.09557 m
.30165 1.09163 L
s
.13361 1.19405 m
.14382 1.19412 L
s
.1238 1.19395 m
.13361 1.19405 L
s
.11416 1.19376 m
.1238 1.19395 L
s
.30165 1.09163 m
.30071 1.08735 L
s
.30071 1.08735 m
.30328 1.08418 L
s
.30328 1.08418 m
.30324 1.08023 L
s
.30324 1.08023 m
.30489 1.07683 L
s
.30489 1.07683 m
.30592 1.07326 L
s
.30592 1.07326 m
.30704 1.06974 L
s
.30704 1.06974 m
.3072 1.06596 L
s
.3072 1.06596 m
.30841 1.06253 L
s
.30841 1.06253 m
.30819 1.0587 L
s
.30819 1.0587 m
.30939 1.05533 L
s
.30939 1.05533 m
.31087 1.05208 L
s
.31087 1.05208 m
.31209 1.04875 L
s
.31209 1.04875 m
.31305 1.04537 L
s
.31305 1.04537 m
.31551 1.04247 L
s
.31551 1.04247 m
.31565 1.03886 L
s
.31565 1.03886 m
.31432 1.03484 L
s
.31432 1.03484 m
.31438 1.03131 L
s
.31438 1.03131 m
.31562 1.02817 L
s
.31562 1.02817 m
.31703 1.02509 L
s
.31703 1.02509 m
.31972 1.02242 L
s
.31972 1.02242 m
.32286 1.01987 L
s
.32286 1.01987 m
.32514 1.01705 L
s
.32514 1.01705 m
.32481 1.01342 L
s
.32481 1.01342 m
.32553 1.01017 L
s
.32553 1.01017 m
.3221 1.00567 L
s
.3221 1.00567 m
.3203 1.00175 L
s
.3203 1.00175 m
.32364 .99947 L
s
.32364 .99947 m
.32694 .99716 L
s
.32694 .99716 m
.3267 .99374 L
s
.3267 .99374 m
.33051 .9916 L
s
.33051 .9916 m
.33203 .98872 L
s
.33203 .98872 m
.33266 .98558 L
s
.33266 .98558 m
.33361 .98255 L
s
.33361 .98255 m
.33379 .97931 L
s
.33379 .97931 m
.33506 .97643 L
s
.33506 .97643 m
.33602 .97347 L
s
.33602 .97347 m
.33641 .97035 L
s
.33641 .97035 m
.33754 .96747 L
s
.33754 .96747 m
.33817 .96446 L
s
.33817 .96446 m
.34015 .96188 L
s
.34015 .96188 m
.34115 .95899 L
s
.34115 .95899 m
.34178 .95601 L
s
.34178 .95601 m
.34303 .95323 L
s
.34303 .95323 m
.34381 .95032 L
s
.34381 .95032 m
.34289 .9469 L
s
.34289 .9469 m
.34696 .94506 L
s
.34696 .94506 m
.34535 .94143 L
s
.34535 .94143 m
.34891 .93945 L
s
.34891 .93945 m
.34943 .9365 L
s
.34943 .9365 m
.35168 .9341 L
s
.35168 .9341 m
.35435 .93183 L
s
.35435 .93183 m
.35625 .9293 L
s
.35625 .9293 m
.35817 .92677 L
s
.35817 .92677 m
.35747 .92343 L
s
.35747 .92343 m
.35799 .9205 L
s
.35799 .9205 m
.35995 .91804 L
s
.35995 .91804 m
.35913 .91471 L
s
.35913 .91471 m
.35772 .91123 L
s
.35772 .91123 m
.35793 .90832 L
s
.35793 .90832 m
.35795 .90536 L
s
.35795 .90536 m
.35561 .9017 L
s
.35561 .9017 m
.35888 .89984 L
s
.35888 .89984 m
.35856 .89683 L
s
.35856 .89683 m
.35827 .89387 L
s
.35827 .89387 m
.357 .89062 L
s
.357 .89062 m
.35609 .88754 L
s
.35609 .88754 m
.35647 .88489 L
s
.35647 .88489 m
.35612 .88203 L
s
.35612 .88203 m
.35932 .8803 L
s
.35932 .8803 m
.35963 .87764 L
s
.35963 .87764 m
.36007 .87504 L
s
.36007 .87504 m
.36099 .8726 L
s
.36099 .8726 m
.3622 .87025 L
s
.3622 .87025 m
.36281 .86772 L
s
.36281 .86772 m
.36229 .86484 L
s
.36229 .86484 m
.36409 .86272 L
s
.36409 .86272 m
.36564 .86051 L
s
.36564 .86051 m
.36745 .85838 L
s
.36745 .85838 m
.36876 .85608 L
s
.36876 .85608 m
.36942 .85357 L
s
.36942 .85357 m
.37054 .85122 L
s
.37054 .85122 m
.3714 .84878 L
s
.3714 .84878 m
.37259 .84646 L
s
.37259 .84646 m
.37413 .84424 L
s
.37413 .84424 m
.37761 .84263 L
s
.37761 .84263 m
.37625 .83946 L
s
.37625 .83946 m
.37889 .83759 L
s
.37889 .83759 m
.38093 .83551 L
s
.38093 .83551 m
.38023 .83255 L
s
.38023 .83255 m
.38092 .83006 L
s
.38092 .83006 m
.38006 .82709 L
s
.38006 .82709 m
.38194 .82501 L
s
.38194 .82501 m
.38226 .82243 L
s
.38226 .82243 m
.3824 .81981 L
s
.3824 .81981 m
.38175 .81696 L
s
.38175 .81696 m
.38396 .81504 L
s
.38396 .81504 m
.38358 .81228 L
s
.38358 .81228 m
.38348 .80963 L
s
.38348 .80963 m
.38303 .80689 L
s
.38303 .80689 m
.38146 .80382 L
s
.38146 .80382 m
.38177 .80139 L
s
.38177 .80139 m
.38045 .79845 L
s
.38045 .79845 m
.38038 .79594 L
s
.38038 .79594 m
.37856 .7929 L
s
.37856 .7929 m
.37722 .79006 L
s
.37722 .79006 m
.37482 .78691 L
s
.37482 .78691 m
.37238 .7838 L
s
.37238 .7838 m
.3691 .78048 L
s
.3691 .78048 m
.36459 .77683 L
s
.36459 .77683 m
.35874 .77283 L
s
.35874 .77283 m
.3522 .76873 L
s
.3522 .76873 m
.34451 .76437 L
s
.34451 .76437 m
.33809 .76054 L
s
.33809 .76054 m
.33065 .7565 L
s
.33065 .7565 m
.32322 .75259 L
s
.32322 .75259 m
.31467 .74843 L
s
.31467 .74843 m
.30812 .74506 L
s
.30812 .74506 m
.30304 .74225 L
s
.30304 .74225 m
.29609 .73892 L
s
.29609 .73892 m
.29489 .73754 L
s
.29489 .73754 m
.29374 .73618 L
s
.29374 .73618 m
.29852 .73671 L
s
.29852 .73671 m
.30087 .73639 L
s
.30087 .73639 m
.30307 .73596 L
s
.30307 .73596 m
.30679 .73598 L
s
.30679 .73598 m
.308 .73513 L
s
.308 .73513 m
.30713 .73358 L
s
.30713 .73358 m
.31227 .73397 L
s
.31227 .73397 m
.31624 .73389 L
s
.31624 .73389 m
.31869 .73325 L
s
.31869 .73325 m
.32029 .7323 L
s
.32029 .7323 m
.32392 .73196 L
s
.32392 .73196 m
.32633 .73117 L
s
.32633 .73117 m
.33047 .7309 L
s
.33047 .7309 m
.3381 .73168 L
s
.3381 .73168 m
.34098 .7308 L
s
.34098 .7308 m
.34063 .72885 L
s
.34063 .72885 m
.34399 .72809 L
s
.34399 .72809 m
.34913 .72786 L
s
.34913 .72786 m
.35356 .72731 L
s
.35356 .72731 m
.35786 .72666 L
s
.35786 .72666 m
.36341 .72634 L
s
.36341 .72634 m
.36904 .72596 L
s
.36904 .72596 m
.37293 .72495 L
s
.60368 .71709 m
.60604 .71249 L
s
.59598 .7199 m
.60368 .71709 L
s
.60604 .71249 m
.60643 .70726 L
s
.58621 .72193 m
.59598 .7199 L
s
.37293 .72495 m
.37821 .72432 L
s
.60643 .70726 m
.60803 .70249 L
s
.60803 .70249 m
.60969 .69777 L
s
.57152 .72218 m
.58621 .72193 L
s
.60969 .69777 m
.61077 .6929 L
s
.37821 .72432 m
.38218 .7232 L
s
.61077 .6929 m
.60745 .68662 L
s
.54832 .71935 m
.57152 .72218 L
s
.38218 .7232 m
.38643 .72211 L
s
.60745 .68662 m
.60871 .68196 L
s
.60871 .68196 m
.61045 .6775 L
s
.61045 .6775 m
.6133 .67342 L
s
.38643 .72211 m
.39131 .72117 L
s
.6133 .67342 m
.61707 .66967 L
s
.61707 .66967 m
.61507 .664 L
s
.39131 .72117 m
.39773 .72067 L
s
.52866 .7174 m
.54832 .71935 L
s
.61507 .664 m
.61988 .66067 L
s
.61988 .66067 m
.61586 .6544 L
s
.39773 .72067 m
.40226 .71946 L
s
.61586 .6544 m
.61714 .65 L
s
.40226 .71946 m
.40661 .71814 L
s
.51098 .71585 m
.52866 .7174 L
s
.61714 .65 m
.618 .64549 L
s
.40661 .71814 m
.41084 .71672 L
s
.618 .64549 m
.61889 .64103 L
s
.4966 .71518 m
.51098 .71585 L
s
.61889 .64103 m
.61857 .6362 L
s
.41084 .71672 m
.41562 .71544 L
s
.48352 .71476 m
.4966 .71518 L
s
.61857 .6362 m
.62208 .6327 L
s
.41562 .71544 m
.42314 .71497 L
s
.62208 .6327 m
.62215 .62805 L
s
.47025 .71409 m
.48352 .71476 L
s
.42314 .71497 m
.43116 .71457 L
s
.43116 .71457 m
.44124 .71473 L
s
.44124 .71473 m
.44797 .71366 L
s
.62215 .62805 m
.62406 .62406 L
s
.457 .71325 m
.47025 .71409 L
s
.44797 .71366 m
.457 .71325 L
s
.62406 .62406 m
.62325 .61919 L
s
.62325 .61919 m
.62516 .61528 L
s
.62516 .61528 m
.62782 .61163 L
s
.62782 .61163 m
.62931 .6076 L
s
.62931 .6076 m
.63061 .60353 L
s
.63061 .60353 m
.63122 .59925 L
s
.63122 .59925 m
.63197 .59506 L
s
.63197 .59506 m
.63402 .59133 L
s
.63402 .59133 m
.63208 .58628 L
s
.63208 .58628 m
.63332 .58237 L
s
.63332 .58237 m
.63764 .57951 L
s
.63764 .57951 m
.63795 .57528 L
s
.63795 .57528 m
.63821 .57107 L
s
.63821 .57107 m
.63962 .56729 L
s
.63962 .56729 m
.63995 .56316 L
s
.63995 .56316 m
.64156 .5595 L
s
.64156 .5595 m
.64291 .55576 L
s
.64291 .55576 m
.64333 .55173 L
s
.64333 .55173 m
.64609 .54852 L
s
.64609 .54852 m
.6466 .54455 L
s
.6466 .54455 m
.64842 .54105 L
s
.64842 .54105 m
.64927 .53723 L
s
.64927 .53723 m
.65173 .53398 L
s
.65173 .53398 m
.65226 .53008 L
s
.65226 .53008 m
.65443 .52676 L
s
.65443 .52676 m
.65425 .52264 L
s
.65425 .52264 m
.6541 .51858 L
s
.6541 .51858 m
.65568 .51514 L
s
.65568 .51514 m
.65697 .51162 L
s
.65697 .51162 m
.65951 .50853 L
s
.65951 .50853 m
.66032 .50485 L
s
.66032 .50485 m
.65995 .50079 L
s
.65995 .50079 m
.6579 .4962 L
s
.6579 .4962 m
.65991 .49306 L
s
.65991 .49306 m
.66079 .48953 L
s
.66079 .48953 m
.66104 .48581 L
s
.66104 .48581 m
.66338 .48283 L
s
.66338 .48283 m
.66521 .47966 L
s
.66521 .47966 m
.66423 .47555 L
s
.66423 .47555 m
.66703 .47277 L
s
.66703 .47277 m
.67206 .47073 L
s
.67206 .47073 m
.67296 .46724 L
s
.67296 .46724 m
.6754 .4643 L
s
.6754 .4643 m
.67608 .46074 L
s
.67608 .46074 m
.67693 .45726 L
s
.67693 .45726 m
.67813 .45392 L
s
.67813 .45392 m
.67999 .45082 L
s
.67999 .45082 m
.68086 .44738 L
s
.68086 .44738 m
.68078 .44363 L
s
.68078 .44363 m
.68165 .44024 L
s
.68165 .44024 m
.68363 .43725 L
s
.68363 .43725 m
.68541 .43418 L
s
.68541 .43418 m
.68522 .43044 L
s
.68522 .43044 m
.68614 .42712 L
s
.68614 .42712 m
.68738 .42391 L
s
.68738 .42391 m
.68988 .42115 L
s
.68988 .42115 m
.68784 .41681 L
s
.68784 .41681 m
.68798 .41329 L
s
.68798 .41329 m
.6912 .41085 L
s
.6912 .41085 m
.69138 .40734 L
s
.69138 .40734 m
.6918 .40393 L
s
.6918 .40393 m
.69424 .40124 L
s
.69424 .40124 m
.69441 .39776 L
s
.69441 .39776 m
.69428 .3942 L
s
.69428 .3942 m
.69476 .39087 L
s
.69476 .39087 m
.69562 .3877 L
s
.69562 .3877 m
.69709 .38474 L
s
.69709 .38474 m
.69603 .38091 L
s
.69603 .38091 m
.69643 .37763 L
s
.69643 .37763 m
.70067 .3757 L
s
.70067 .3757 m
.70524 .37384 L
s
.70524 .37384 m
.70657 .37081 L
s
.70657 .37081 m
.70683 .36741 L
s
.70683 .36741 m
.7076 .36421 L
s
.7076 .36421 m
.70932 .36135 L
s
.70932 .36135 m
.71161 .35869 L
s
.71161 .35869 m
.715 .3564 L
s
.715 .3564 m
.71658 .35345 L
s
.71658 .35345 m
.72071 .35139 L
s
.72071 .35139 m
.72048 .34778 L
s
.72048 .34778 m
.72098 .34444 L
s
.72098 .34444 m
.7232 .34172 L
s
.7232 .34172 m
.72561 .33906 L
s
.72561 .33906 m
.72761 .33625 L
s
.72761 .33625 m
.72904 .33323 L
s
.72904 .33323 m
.72993 .33002 L
s
.72993 .33002 m
.73239 .32737 L
s
.73239 .32737 m
.7335 .32424 L
s
.7335 .32424 m
.73523 .32133 L
s
.73523 .32133 m
.73487 .31768 L
s
.73487 .31768 m
.73173 .31309 L
s
.73173 .31309 m
.73011 .3091 L
s
.73011 .3091 m
.73063 .30591 L
s
.73063 .30591 m
.73228 .30314 L
s
.73228 .30314 m
.72991 .29894 L
s
.72991 .29894 m
.73019 .29573 L
s
.73019 .29573 m
.73033 .29249 L
s
.73033 .29249 m
.73177 .28972 L
s
.73177 .28972 m
.72853 .28531 L
s
.72853 .28531 m
.72839 .28205 L
s
.72839 .28205 m
.72639 .27815 L
s
.72639 .27815 m
.72617 .27494 L
s
.72617 .27494 m
.72326 .27079 L
s
.72326 .27079 m
.72068 .26682 L
s
.72068 .26682 m
.71844 .26302 L
s
.71844 .26302 m
.71671 .25945 L
s
.71671 .25945 m
.7128 .25514 L
s
.7128 .25514 m
.70766 .25048 L
s
.70766 .25048 m
.70027 .2451 L
s
.70027 .2451 m
.69278 .23981 L
s
.69278 .23981 m
.68497 .23452 L
s
.68497 .23452 m
.67275 .22779 L
s
.67275 .22779 m
.66045 .22122 L
s
.66045 .22122 m
.65044 .21565 L
s
.65044 .21565 m
.64368 .21137 L
s
.64368 .21137 m
.63743 .20737 L
s
.63743 .20737 m
.6344 .20461 L
s
.6344 .20461 m
.63503 .20318 L
s
.63503 .20318 m
.63437 .20128 L
s
.63437 .20128 m
.63482 .19977 L
s
.63482 .19977 m
.63722 .19895 L
s
.63722 .19895 m
.63779 .19742 L
s
.63779 .19742 m
.6428 .19746 L
s
.6428 .19746 m
.64828 .19759 L
s
.64828 .19759 m
.65405 .19773 L
s
.65405 .19773 m
.65942 .19764 L
s
.65942 .19764 m
.66143 .19627 L
s
.66143 .19627 m
.66864 .19673 L
s
.66864 .19673 m
.67085 .19529 L
s
.67085 .19529 m
.67227 .19353 L
s
.67227 .19353 m
.67661 .1928 L
s
.67661 .1928 m
.67735 .19072 L
s
.67735 .19072 m
.6795 .18913 L
s
.6795 .18913 m
.68255 .18783 L
s
.68255 .18783 m
.6882 .18743 L
s
.6882 .18743 m
.69486 .1873 L
s
.69486 .1873 m
.69557 .18495 L
s
.69557 .18495 m
.69953 .18375 L
s
.69953 .18375 m
.70223 .18206 L
s
.70223 .18206 m
.70838 .18157 L
s
.70838 .18157 m
.70874 .17891 L
s
.70874 .17891 m
.7149 .17834 L
s
.7149 .17834 m
.71995 .17729 L
s
.71995 .17729 m
.7215 .17492 L
s
.7215 .17492 m
.72705 .17398 L
s
.72705 .17398 m
.7319 .17271 L
s
.7319 .17271 m
.73964 .17242 L
s
.73964 .17242 m
.74544 .17133 L
s
.74544 .17133 m
.75116 .17014 L
s
.75116 .17014 m
.75388 .16781 L
s
.75388 .16781 m
.76037 .1668 L
s
.76037 .1668 m
.76486 .16501 L
s
.76486 .16501 m
.76596 .16194 L
s
.76596 .16194 m
.76992 .1599 L
s
.76992 .1599 m
.77532 .15834 L
s
.77532 .15834 m
.7789 .15607 L
s
.7789 .15607 m
.78536 .15481 L
s
.78536 .15481 m
.79343 .15406 L
s
.79343 .15406 m
.80065 .1529 L
s
.80065 .1529 m
.8098 .15236 L
s
0 g
.25 Mabswid
.72778 0 m
.99095 .0971 L
s
.99095 .0971 m
1 .18976 L
s
1 .18976 m
.73246 .09243 L
s
.73246 .09243 m
.72778 0 L
s
.00633 1.13765 m
0 1.2297 L
s
0 1.2297 m
.73246 .09243 L
s
.73246 .09243 m
.72778 0 L
s
.72778 0 m
.00633 1.13765 L
s
.99095 .0971 m
1 .18976 L
s
.99303 .11833 m
.9847 .11527 L
s
[(0)] 1.00967 .12445 -1 -0.36777 Mshowa
.99535 .14208 m
.98702 .13903 L
s
[(0.5)] 1.012 .14818 -1 -0.36644 Mshowa
.99768 .16604 m
.98935 .163 L
s
[(1)] 1.01435 .17212 -1 -0.36511 Mshowa
.125 Mabswid
.99349 .12306 m
.98849 .12123 L
s
.99395 .1278 m
.98896 .12597 L
s
.99441 .13255 m
.98942 .13072 L
s
.99488 .13731 m
.98988 .13548 L
s
.99581 .14686 m
.99081 .14503 L
s
.99628 .15164 m
.99128 .14981 L
s
.99675 .15643 m
.99175 .1546 L
s
.99721 .16123 m
.99222 .1594 L
s
.99815 .17086 m
.99316 .16903 L
s
.99863 .17568 m
.99363 .17386 L
s
.9991 .18051 m
.9941 .17869 L
s
.99957 .18536 m
.99457 .18354 L
s
.99256 .1136 m
.98757 .11176 L
s
.9921 .10888 m
.98711 .10704 L
s
.99164 .10417 m
.98665 .10233 L
s
.99118 .09947 m
.98619 .09763 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[11]:=",
  ImageSize->{167.875, 217.5},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40002W0000f@P10`4000010?ooo`000<c<c?l00;Ng
]fIVI_Cdm42P`0000000E@00ZP00o`1E001EE@1EZP1Eo`2Z002ZE@2ZZP2Zo`3o003oE@3oZP3ooeD0
05D0EED0ZUD0oeEE05EEEEEEZUEEoeFZ05FZEEFZZUFZoeGo05GoEEGoZUGoojX00:X0EJX0ZZX0ojYE
0:YEEJYEZZYEojZZ0:ZZEJZZZZZZoj[o0:[oEJ[oZZ[oool00?l0EOl0Z_l0oomE0?mEEOmEZ_mEoonZ
0?nZEOnZZ_nZoooo0?ooEOooZ_ooo`0001lO7clo?emOGgmoOinOWkno_moOgoooon?U]knH][OUiBgU
iLC4a>[ZjWAdM:NWYj0Pl=WIfJj^[WMgMiVIVJo^k[jn_[C^]<30`9T0I_ooc>OWifm_Kh0003<cIVIV
c820P9VIo`00P28R8X200020P4:JY`0g?0200?mV<omVc1YOI`00kUDJRj>SXiT00000<`00IP00V@00
c00c000c<`0cIP0cV@0cc00co`1V001V<`1VIP1VV@1Vc01Vo`2I002I<`2IIP2IV@2Ic02Io`3<003<
<`3<IP3<V@3<c03<o`3o<`3oIP3oV@3oc3<003<0<c<0IS<0VC<0c3<0oc<c03<c<c<cVC<cc3<coc=V
03=V<c=VIS=VVC=Vc3=Voc>I03>I<c>IIS>IVC>Ic3>Ioc?<03?<<c?<IS?<VC?<c3?<oc?o03?o<c?o
IS?oVC?oc3?oofH006H0<fH0IVH0VFH0c6H0ofHc06Hc<fHcIVHcVFHcc6HcofIV06IV<fIVVFIVofJI
06JI<fJIIVJIVFJIc6JIofK<06K<<fK<IVK<VFK<c6K<ofKo06Ko<fKoIVKoVFKoc6KooiT0<iT0VIT0
c9T0oiTc09Tc<iTcIYTcVITcc9TcoiUV09UV<iUVIYUVVIUVc9UVoiVI09VI<iVIIYVIc9W<09W<<iW<
IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`00<`0<jM7002WA`00YdL00:M7002WA`00YdL00:M7
002WA`00YdL00:M7002WA`00NDL422Y7001kA`0324M702U7001kA`0324M702U7001cA`@814L00`Q7
A`0YA`00JdL220i700@8AdM70PPVA`00JTL22003A`P800Y70`P2A`0424M722E7001ZA`8800=724L0
0`P9A`0424M7209700<8AdL08dL006U700<8A`P00dL00`Q7200220Y700@8AdL81TL421]7001XA`04
24M720M710P7A`0424M720Q700<8AdL06TL006Q700@8AdL81TL01@Q7A`Q700881DL010Q7A`P8A`03
24M701Y7001WA`0524M7A`P02TL00`Q7200320=70PP9A`0324M701Y7001WA`0524M7A`P03TL010Q7
20P<A`0324M700A710PBA`00ITL00`Q7A`02A`0324M700i700<8A`P00PP7A`<81TL00`Q7A`0CA`00
IDL00`Q7A`03A`0324M7019710P6A`0324M700E700<8AdL04TL006E700<8AdL00dL00`Q7A`0DA`04
24L820e700<8AdL04DL006A700<8AdL014L00`Q7A`0HA`<82dL00`Q7A`0@A`00HdL00`Q7A`05A`03
24M701M700<8AdL00`P6A`0424M72197001SA`0324M700E700<8AdL074L520=70PPCA`00HTL00`Q7
A`06A`8884L421I7001QA`0324M700I700<8AdL00`POA`<85DL0065700<8AdL01TL00`Q7A`03A`<8
7DL221E7001PA`0324M700I700<8AdL01dL221]70`PDA`00H4L00`Q7A`05A`0324M700Y70`PGA`04
24L821A7001OA`0324M700I700<8AdL03DL321=700<8AdL00PP3A`883dL005i700<8AdL01TL00`Q7
A`0AA`<844L00`Q7A`02209700@8AdL83TL005i700<8AdL01TL00`Q7A`0DA`<834L010Q7AdL22097
00@8AdL83TL005e700<8AdL01TL00`Q7A`0HA`<824L00`Q7A`02A`880TL420=700<8AdL00TL320=7
001LA`0324M700I700<8AdL02TL6>`a70`P5A`0324M700970PP2A`0424M720]70@P1A`57001LA`03
24M700I700<8AdL024L3>aA70PP2A`0324M700=70PP2A`0424M720]70@P1A`57001KA`0324M700I7
00<8AdL01dL3>aQ70`P6A`8800=720P00PP8A`<80dL005Y700<8AdL01TL00`Q7A`03A`Dk6dL00`Q7
A`0320=70PP00dL8200320M700<8AdL00dL005Y700<8AdL01TL00`Q7A`04>am700<8AdL00dL52005
A`Q7A`P024L00`Q7A`03A`00FDL00`Q7A`06A`0323/k008k8DL00`Q7A`07A`880TL320Q710P2A`00
FDL010Q7AdL4>`0323/k02=700<8AdL024L00`Q7A`03A`0324M700a7001HA`0324M7009700DkAdM7
200UA`0324M700Q700<8AdL00dL00`Q7A`0<A`00EdL00`Q7A`03A`8k00=724L094L00`Q7A`08A`03
24M700970`P>A`00EdL00`Q7A`05A`0324M702=700<8AdL024L00`Q7A`05A`0324M700a7001FA`03
24M700I700<8>c/08dL00`Q7A`08A`0324M701A7001EA`0324M700I700D8AdM7>`0QA`0324M700Q7
00<8AdL05DL005E700<8AdL01TL00`Q7A`02A`8k7TL00`Q7A`08A`0324M701I7001DA`0324M700I7
00<8AdL01DL3>a]700<8AdL01dL00`Q7A`0GA`00DdL00`Q7A`06A`0324M700U700<kAdL05dL00`Q7
A`08A`0324M701M7001CA`0324M700I700<8AdL02DL2>aM700<8AdL024L00`Q7A`0HA`00DTL00`Q7
A`06A`0324M700]700<kAdL05DL00`Q7A`07A`0324M701U7001BA`0324M700E700<8AdL03DL00c]7
A`0CA`0324M700I70`PKA`00DDL00`Q7A`06A`0324M700e700<kAdL04TL00`Q7A`08A`0324M701Y7
001@A`0324M700I700<8AdL03dL00c]7A`0AA`0324M700M700<8AdL06dL0051700<8AdL01DL00`Q7
A`0@A`03>dM7011700<8AdL024L00`Q7A`0KA`00CdL00`Q7A`06A`0324M7011700<kAdL03dL00`Q7
A`08A`0324M701a7001>A`0324M700I700<8AdL04DL00c]7A`0>A`0324M700Q700<8AdL07DL004i7
00<8AdL01DL00`Q7A`0BA`03>dM700i700<8AdL01dL00`Q7A`0NA`00CDL00`Q7A`06A`0324M701=7
00<kAdL034L00`Q7A`08A`0324M701i7001<A`0324M700I700<8AdL04dL2>`e700<8AdL024L00`Q7
A`0OA`00C4L00`Q7A`06A`0324M701=700<kAdL034L00`Q7A`07A`0324M70217001;A`0324M700I7
00<8AdL04dL2>`e700<8AdL024L00`Q7A`0PA`00BdL00`Q7A`05A`0324M701=70S/=A`0324M700Q7
00<8AdL08DL004Y700<8AdL01TL00`Q7A`0CA`8k3DL00`Q7A`07A`0324M702970019A`0324M700I7
00<8AdL04dL2>`e700<8AdL024L00`Q7A`0RA`00BDL00`Q7A`05A`0324M701=70S/=A`0324M700I7
00<8A`P09DL004Q700<8AdL01TL00`Q7A`0CA`03>dM700a700<8AdL01dL00`Q7A`0TA`00AdL00`Q7
A`06A`0324M701=70S/=A`0324M700M700<8AdL09DL004M700<8AdL01DL00`Q7A`0DA`03>dM700]7
00<8AdL024L00`Q7A`0UA`00ATL00`Q7A`06A`0324M701A700<kAdL02dL00`Q7A`07A`0324M702I7
0015A`0324M700I700<8AdL05DL00c]7A`0:A`0324M700M700<8AdL09dL004E700<8AdL01TL00`Q7
A`0DA`03>dM700Y700<8AdL024L00`Q7A`0WA`00A4L00`Q7A`06A`0324M701E700<kAdL02DL00`Q7
A`08A`0324M702Q70014A`0324M700E700<8AdL05TL00c]7A`09A`0324M700M700<8AdL0:DL004=7
00<8AdL01TL00`Q7A`0EA`8k2TL00`Q7A`08A`0324M702U70012A`0324M700I700<8AdL05TL00c]7
A`08A`0324M700Q700<8AdL0:TL0049700<8AdL01DL00`Q7A`0GA`03>dM700Q700<8AdL01dL00`Q7
A`0[A`00@DL00`Q7A`06A`0324M701M700<kAdL01dL00`Q7A`07A`0324M702a70010A`0324M700I7
00<8AdL05dL2>`Q700<8AdL024L00`Q7A`0/A`00@4L00`Q7A`05A`0324M701Q700<kAdL01dL00`Q7
A`05A`0324L802m7000oA`0324M700I700<8AdL064L00c]7A`06A`0324M700M700<8AdL0;TL003i7
00<8AdL01TL00`Q7A`0GA`8k24L00`Q7A`08A`0424M7A`@80dL222A7000nA`0324M700E700<8AdL0
64L00c]7A`07A`0324M700M700<8AdL00TL00`Q7A`03A`0424M722=7000mA`0324M700I700<8AdL0
64L00c]7A`06A`0324M700M700<8AdL014L00`Q7A`02A`0424M722=7000mA`0324M700E700<8AdL0
64L2>`M700<8AdL024L00`Q7A`05A`0524M7A`P00TL00`Q7A`0QA`00?4L00`Q7A`06A`0324M701U7
00<kAdL01DL00`Q7A`07A`0324M700M700@8AdL80TL00`Q7A`0QA`00>dL00`Q7A`06A`0324M701Y7
00<kAdL014L00`Q7A`07A`0324M700E700@8AdL80TL010Q7A`PSA`00>dL00`Q7A`05A`0324M701Y7
00<kAdL014L00`Q7A`07A`0324M700M70PP4A`8894L003Y700<8AdL01TL00`Q7A`0JA`03>dM700A7
00<8AdL01dL00`Q7A`0cA`00>DL00`Q7A`06A`0324M701Y70S/5A`0324M700M700<8AdL0=4L003U7
00<8AdL01DL00`Q7A`0KA`03>dM700=700<8AdL01dL00`Q7A`0eA`00>4L00`Q7A`06A`0324M701]7
00<kAdL00TL00`Q7A`08A`0324M703E7000gA`0324M700I700<8AdL06dL2>`A700<8AdL01dL00`Q7
A`0fA`00=dL00`Q7A`05A`0324M701a700<kAdL00TL00`Q7A`06A`88>DL003I700<8AdL01TL00`Q7
A`0LA`05>dM7A`P02TL00`Q7A`0gA`00=TL00`Q7A`05A`0324M701a70S/3A`0324M700M700<8AdL0
>4L003E700<8AdL01TL00`Q7A`0LA`05>dM7A`P02DL00`Q7A`0iA`00=4L00`Q7A`06A`0324M701e7
00@kAdL82DL00`Q7A`0jA`00=4L00`Q7A`05A`0324M701e70S/2A`0324M700M700<8AdL0>TL003=7
00<8AdL01TL00`Q7A`0MA`04>dM720U700<8AdL0>dL0039700<8AdL01TL00`Q7A`0NA`03>dL800U7
00<8AdL0?4L0039700<8AdL01DL00`Q7A`0OA`03>dL800U700<8AdL0?4L0035700<8AdL01TL00`Q7
A`0NA`03>dL800U700<8AdL0?DL0031700<8AdL01TL00`Q7A`0OA`03>`Q700Q700<8AdL0?TL00317
00<8AdL01DL00`Q7A`0PA`03>`Q700Q700<8AdL0?TL002m700<8AdL01TL00`Q7A`0OA`03>`Q700Q7
00<8AdL0?dL002i700<8AdL01TL00`Q7A`0PA`8k1dL22497000^A`0324M700E700<8AdL08DL00c]7
A`06A`0324M70457000]A`0324M700I700<8AdL084L00`PkA`07A`0324M70457000]A`0324M700E7
00<8AdL084L010Q7>c/6A`0324M70497000/A`0324M700I700<8AdL07dL010Q7Ac/6A`0324M704=7
000[A`0324M700I700<8AdL084L00`Q7>`07A`0324M704=7000[A`0324M700E700<8AdL084L010Q7
Ac/6A`0324M704A7000ZA`0324M700I700<8AdL07dL01@Q7AdLk00E700<8AdL0ADL002U700<8AdL0
1TL00`Q7A`0PA`0524M7Ac/01DL00`Q7A`15A`00:DL00`Q7A`05A`0324M7021700<8AdL00TL00c]7
A`02A`0324M704I7000XA`0324M700I700<8AdL02TL:>`]700<8AdL00dL01C]7AdL804U7000WA`03
24M700I700<8AdL01TL5>`Y72S/00dL8A`02A`8k0dL00`Q7A`18A`009dL00`Q7A`05A`0324M700Lk
6DL5>`E700<8AdL0B4L002I700<8AdL014L2>`0423/k>ai700<8AdL01TL224]7000VA`0324M70097
0c/00`Q7A`0PA`0324M700I700<8AdL0BTL002E700<8AdL014L00c/8>`0QA`0324M700M700<8AdL0
BTL002A700<8AdL01TL00`Q7>`0OA`0324M700M700<8AdL0BdL002A700<8AdL01DL010Q7AdL2>ae7
00<8AdL01TL00`Q7A`1<A`008dL00`Q7A`06A`0324M700=70S/JA`0324M700M700<8AdL0C4L00297
00<8AdL01TL00`Q7A`06A`03>dM701I700<8AdL01dL00`Q7A`1=A`008TL00`Q7A`05A`0324M700Q7
0S/EA`0324M700M700<8AdL0CTL0025700<8AdL01TL00`Q7A`09A`8k54L00`Q7A`06A`0324M704m7
000PA`0324M700I700<8AdL02dL00c]7A`0AA`0324M700M700<8AdL0CdL0021700<8AdL01DL00`Q7
A`0<A`8k4DL00`Q7A`07A`0324M70517000OA`0324M700I700<8AdL03DL00c]7A`0?A`0324M700I7
00<8AdL0DDL001m700<8AdL01DL00`Q7A`0>A`03>dM700i700<8AdL01TL225=7000NA`0324M700E7
00<8AdL03dL00c]7A`0=A`0324M700M700<8AdL0DTL001e700<8AdL01TL00`Q7A`0>A`8k3dL00`Q7
A`06A`0324M705=7000MA`0324M700E700<8AdL03dL2>`i700<8AdL01dL00`Q7A`1CA`0074L00`Q7
A`06A`0324M700m700<kAdL034L00`Q7A`07A`0324M705A7000KA`0324M700I700<8AdL03dL2>`i7
00<8AdL01TL00`Q7A`1EA`006dL00`Q7A`05A`0324M7011700<kAdL034L00`Q7A`06A`0324M705I7
000JA`0324M700I700<8AdL03dL2>`e700<8AdL01dL00`Q7A`1FA`006DL00`Q7A`06A`0324M70117
00<kAdL034L00`Q7A`06A`0324M705M7000IA`0324M700E700<8AdL044L2>`e700<8AdL01TL00`Q7
A`1HA`0064L00`Q7A`06A`0324M7011700<kAdL02dL00`Q7A`07A`0324M705Q7000HA`0324M700E7
00<8AdL04DL00c]7A`0;A`0324M700I700<8AdL0FDL001M700<8AdL01DL00`Q7A`0BA`03>dM700Y7
00<8AdL01DL225a7000FA`0324M700I700<8AdL04TL00c]7A`09A`0324M700M700@8AdM710P3A`88
D4L001I700<8AdL01DL00`Q7A`0CA`03>dM700Q700<8AdL01dL00`Q7A`04A`0524M7A`P00TL00`Q7
A`1=A`005DL00`Q7A`05A`0324M701A700<kAdL024L00`Q7A`06A`0324M700E700D8AdM72002A`03
24M704e7000DA`0324M700I700<8AdL054L00c]7A`07A`0324M700I700<8AdL01TL01@Q7AdL80097
00<8AdL0CDL001A700<8AdL01DL00`Q7A`0EA`03>dM700I700<8AdL01dL00`Q7A`06A`0524M7A`P0
0TL00`Q7A`1=A`004dL00`Q7A`06A`0324M701A700<kAdL01dL00`Q7A`06A`0324M700E70`P3A`04
24M724m7000BA`0324M700I700<8AdL054L2>`Q700<8AdL01TL00`Q7A`08A`0324M700970PQ@A`00
4TL00`Q7A`05A`0324M701E700<kAdL01TL00`Q7A`07A`0324M705m7000AA`0324M700I700<8AdL0
5DL00c]7A`06A`0324M700I700<8AdL0H4L0015700<8AdL01DL00`Q7A`0EA`8k1dL00`Q7A`06A`03
24M70657000@A`0324M700E700<8AdL05TL00c]7A`05A`0324M700M700<8AdL0HDL000m700<8AdL0
1TL00`Q7A`0FA`03>dM700E700<8AdL01DL226A7000?A`0324M700E700<8AdL05TL2>`I700<8AdL0
1TL00`Q7A`1SA`003TL00`Q7A`05A`0324M701M700<kAdL014L00`Q7A`06A`0324M706A7000=A`03
24M700I700<8AdL05TL2>`I700<8AdL01TL00`Q7A`1TA`003DL00`Q7A`05A`0324M701M700<kAdL0
14L00`Q7A`06A`0324M706E7000<A`0324M700I700<8AdL05dL00c]7A`03A`0324M700I700<8AdL0
ITL000]700<8AdL01TL00`Q7A`0HA`03>dM700=700<8AdL01TL00`Q7A`1VA`002dL00`Q7A`05A`03
24M701Q70S/4A`0324M700I700<8AdL0IdL000Y700<8AdL01TL00`Q7A`0HA`03>dM7009700<8AdL0
1TL00`Q7A`1XA`002TL00`Q7A`05A`0324M701U700DkAdM72009A`0324M706Q70009A`0324M700E7
00<8AdL06TL01C]7AdL800Q700<8AdL0JDL000Q700<8AdL01TL00`Q7A`0IA`05>dM7A`P01dL226a7
0008A`0324M700E700<8AdL06TL013]7A`P8A`0324M706]70007A`0324M700E700<8AdL06dL013]7
A`P8A`0324M706]70006A`0324M700I700<8AdL06dL00c]72008A`0324M706a70006A`0324M700E7
00<8AdL06dL00c]72008A`0324M706e70005A`0324M700E700<8AdL074L00c]72008A`0324M706e7
0004A`0324M700I700<8AdL074L00c/8A`07A`0324M706i70004A`0324M700E700<8AdL07DL00c]7
A`06A`0324M706m70003A`0324M700I700<8AdL07DL00c]7A`06A`0324M706m70003A`0324M700E7
00<8AdL07DL00c]7A`06A`0324M707170002A`0324M700E700<8AdL07DL00`PkA`06A`0324M70757
00000dL8A`07A`0324M701e700<8>dL01DL00`Q7A`1bA`0000=724L01TL00`Q7A`0MA`0324Lk00I7
00<8AdL0LTL000881TL00`Q7A`0MA`0424M7>`E700<8AdL0LdL0000324L800<80TL00`Q7A`0LA`04
24M7>`E700<8AdL0M4L0000324M700=710PMA`0424M7>`E700<8AdL0M4L0000324M700=700@8AdM7
10PHA`0524M7Ac/014L00`Q7A`1eA`0000<8AdL00dL00`Q7A`05A`<854L00`Q7A`02A`05>dM7A`P0
N4L0000324M7009700<8AdL02DL4211700D8AdM7>`04A`0324M707I700001@Q7AdL8011710P;A`03
24M7009700DkAdM7201iA`0000D8AdM7200DA`@81TL00`Q7A`03A`04>dM727Y700001@Q7A`Pk00A7
4c/00dL82002209700<8AdL00dL00c]7201kA`0000D8AdL8A`05>a973S/227]700000`Q7200OA`03
24M700E700<8AdL0NTL0008884L00`Q7A`04A`0324M707]70002221700<8AdL014L00`Q7A`1kA`00
0PPPA`0324M700=700<8AdL0O4L000970`PMA`0324M7009700<8AdL0ODL000E710PIA`0324M70097
00<8AdL0ODL000U70`PFA`0524M7A`P0P4L000a70`PBA`0524M7A`P0PDL000m710P>A`0424M72897
000CA`<82dL22003A`Q70857000FA`<824L010Q720P3A`88ODL001U710P4A`8814L010Q7A`QlA`00
7DL32003A`P800A700@8AdL8O4L002170PP5A`0424M727a7000WA`0424M727a7000WA`0424M727a7
000XA`88ODL00:M7002WA`00YdL00:M7002WA`00YdL00:M70000\
\>"],
  ImageRangeCache->{{{0, 166.875}, {216.5, 0}} -> {0.179327, -0.0733876, 
  0.00479003, 0.00479003}}]
}, Open  ]]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Statistics", "Section"],

Cell[CellGroupData[{

Cell["Get final values (lastpcoords)", "Subsection"],

Cell["\<\
We make a list called lastpcoords such that listpcoords[[i]] is a \
list of the final values of the ith coordinate (as described above, i.e. 1 \
means time, 2 means first coordinate, etc.) of the solutions (one entry for \
each path). \
\>", "Text"],

Cell[BoxData[
    \(lastp\  = \ Map[Last, \ p]; \nlastpcoords\  = \ {}; \n
    For[i = 0, \ i < \ Length[lastp[\([1]\)]], \ \(i++\), \n\t
      lastpcoords\  = \ 
        Append[lastpcoords, \ Map[Function[q, q[\([i + 1]\)]], \ lastp]]]; 
    \)], "Input",
  CellLabel->"In[8]:="]
}, Closed]],

Cell[CellGroupData[{

Cell["Get statistics", "Subsection"],

Cell[TextData[{
  "We can have functions Mean, Median, Mode, GeometricMean, HarmonicMean, \
RootMeanSquare, SampleRange, Variance, StandardDeviation, MeanDeviation, \
MedianDeviation, InterquartileRange, and others (see ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " help, under Add-ons)."
}], "Text"],

Cell[CellGroupData[{

Cell["Example: find mean of first coordinate of solutions", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Mean[lastpcoords[\([2]\)]]\)], "Input",
  CellLabel->"In[9]:="],

Cell[BoxData[
    \(0.10082000000000002`\)], "Output",
  CellLabel->"Out[9]="]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Example: find median of second coordinate of solutions", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Median[lastpcoords[\([3]\)]]\)], "Input",
  CellLabel->"In[58]:="],

Cell[BoxData[
    \(1.07147799999999993`\)], "Output",
  CellLabel->"Out[58]="]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Example: find mean solution vector", "Subsubsection"],

Cell["\<\
The command Map takes a function and a list and applies the \
function to each element of the list, so Map[f, {a,b,c}] is {f(a), f(b), \
f(c)}.  This allows us to apply statistics to whole vectors.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Map[Mean, lastpcoords]\)], "Input",
  CellLabel->"In[12]:="],

Cell[BoxData[
    \({25.`, 0.10082000000000002`, 0.693147800000000025`}\)], "Output",
  CellLabel->"Out[12]="]
}, Open  ]]
}, Closed]]
}, Closed]]
}, Closed]]
},
FrontEndVersion->"X 3.0",
ScreenRectangle->{{0, 1280}, {0, 1024}},
CellGrouping->Manual,
WindowSize->{520, 600},
WindowMargins->{{Automatic, 16}, {Automatic, 139}},
StyleDefinitions -> Notebook[{
  
  Cell[CellGroupData[{
  Cell["Style Definitions", "Subtitle"],
  
  Cell["\<\
These styles are carefully tuned for all aspects of the online \
help. It's recommended that they remain unchanged.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell["Style Environment Names", "Subsection"],
  
  Cell[StyleData[All, "Working"],
    CellBracketOptions->{"Color"->RGBColor[0, 0.2, 1]},
    ScriptMinSize->9],
  
  Cell[StyleData[All, "Printout"],
    PrivateFontOptions->{"FontType"->"Outline"}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Notebook Options", "Subsection"],
  
  Cell["\<\
  The options defined for the style below will be used at the \
Notebook level.  \
\>", "Text"],
  
  Cell[StyleData["Notebook"],
    PageHeaders->{{Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"], None, Cell[ 
            TextData[ {
              OptionValueBox[ "FileName"]}], "Header"]}, {Cell[ 
            TextData[ {
              OptionValueBox[ "FileName"]}], "Header"], None, Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"]}},
    PrintingOptions->{"FirstPageHeader"->False,
    "FacingPages"->True},
    CellLabelAutoDelete->False,
    CellFrameLabelMargins->6,
    StyleMenuListing->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Headings", "Subsection"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Title"],
    CellMargins->{{5, Inherited}, {2, 0}},
    CellGroupingRules->{"TitleGrouping", 0},
    PageBreakBelow->False,
    CounterIncrements->"Title",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subtitle", 0}, {"Subsubtitle", 0}},
    FontFamily->"Times",
    FontSize->36,
    FontWeight->"Bold",
    FontColor->GrayLevel[1],
    Background->RGBColor[0, 0, 1]],
  
  Cell[StyleData["Title", "Printout"],
    CellMargins->{{0, 10}, {2, 0}},
    FontSize->28,
    Background->GrayLevel[0.6]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subtitle"],
    CellMargins->{{12, Inherited}, {2, 2}},
    CellGroupingRules->{"TitleGrouping", 10},
    PageBreakBelow->False,
    CounterIncrements->"Subtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subsubtitle", 0}},
    FontFamily->"Helvetica",
    FontSize->24],
  
  Cell[StyleData["Subtitle", "Printout"],
    CellMargins->{{7, 10}, {2, 4}},
    FontSize->18]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubtitle"],
    CellMargins->{{12, Inherited}, {2, 12}},
    CellGroupingRules->{"TitleGrouping", 20},
    PageBreakBelow->False,
    CounterIncrements->"Subsubtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}},
    FontFamily->"Helvetica",
    FontSize->14,
    FontSlant->"Italic"],
  
  Cell[StyleData["Subsubtitle", "Printout"],
    CellMargins->{{7, 10}, {2, 8}},
    FontSize->12]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Section"],
    CellFrame->{{6, 0.5}, {0.5, 0.5}},
    CellMargins->{{8, 4}, {2, 50}},
    CellElementSpacings->{"ClosedGroupTopMargin"->18},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CellFrameColor->RGBColor[0, 0, 1],
    CounterIncrements->"Section",
    CounterAssignments->{{"Subsection", 0}, {"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->18,
    FontWeight->"Bold",
    FontColor->RGBColor[0, 0, 1]],
  
  Cell[StyleData["Section", "Printout"],
    CellFrame->{{6, 0.5}, {0.5, 0.5}},
    CellMargins->{{6, 4}, {2, 70}},
    CellFrameColor->GrayLevel[0],
    FontSize->14,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsection"],
    CellDingbat->"\[FilledSquare]",
    CellMargins->{{24, 4}, {2, 18}},
    CellElementSpacings->{"ClosedGroupTopMargin"->12},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CellFrameLabelMargins->6,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->15,
    FontWeight->"Bold",
    FontColor->RGBColor[0, 0, 1]],
  
  Cell[StyleData["Subsection", "Printout"],
    CellMargins->{{19, 4}, {2, 18}},
    FontSize->12,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubsection"],
    CellMargins->{{10, 4}, {2, 18}},
    CellElementSpacings->{"ClosedGroupTopMargin"->12},
    CellGroupingRules->{"SectionGrouping", 50},
    PageBreakBelow->False,
    CounterIncrements->"Subsubsection",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontColor->RGBColor[0, 0, 1]],
  
  Cell[StyleData["Subsubsection", "Printout"],
    CellMargins->{{7, 4}, {2, 18}},
    FontSize->10,
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Text", "Subsection"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Text"],
    CellMargins->{{10, 4}, {0, 8}},
    CellHorizontalScrolling->True,
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 12},
    CounterIncrements->"Text",
    FontSize->14],
  
  Cell[StyleData["Text", "Printout"],
    CellMargins->{{7, 0}, {0, 8}},
    FontSize->11]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MathCaption"],
    CellFrame->{{3, 0}, {0, 0.5}},
    CellMargins->{{47, 14}, {0, 14}},
    CellFrameColor->RGBColor[0, 0, 1],
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 8},
    FontFamily->"Helvetica",
    FontSize->12],
  
  Cell[StyleData["MathCaption", "Printout"],
    CellMargins->{{34, 54}, {0, 14}},
    CellFrameColor->GrayLevel[0],
    FontSize->10]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Input/Output", "Subsection"],
  
  Cell["\<\
The cells in this section define styles used for input and output \
to the kernel.  Be careful when modifying, renaming, or removing these \
styles, because the front end associates special meanings with these style \
names.    \
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Input"],
    CellMargins->{{54, 14}, {3, 8}},
    Evaluatable->True,
    CellGroupingRules->"InputGrouping",
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelAutoDelete->False,
    CellLabelMargins->{{21, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultInputFormatType,
    AutoItalicWords->{},
    FormatType->StandardForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    CounterIncrements->"Input",
    FontSize->12,
    FontWeight->"Bold"],
  
  Cell[StyleData["Input", "Printout"],
    ShowCellBracket->False,
    CellMargins->{{43, 4}, {3, 8}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Output"],
    CellMargins->{{55, 4}, {8, 3}},
    CellEditDuplicate->True,
    CellGroupingRules->"OutputGrouping",
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    CellLabelAutoDelete->False,
    CellLabelMargins->{{21, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultOutputFormatType,
    AutoItalicWords->{},
    FormatType->StandardForm,
    CounterIncrements->"Output",
    FontSize->12],
  
  Cell[StyleData["Output", "Printout"],
    ShowCellBracket->False,
    CellMargins->{{43, 4}, {4, 2}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Message"],
    CellMargins->{{56, 4}, {3, 8}},
    CellGroupingRules->"OutputGrouping",
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    CellLabelMargins->{{21, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultOutputFormatType,
    AutoItalicWords->{},
    FormatType->InputForm,
    CounterIncrements->"Message",
    StyleMenuListing->None,
    FontColor->RGBColor[0, 0.2, 1]],
  
  Cell[StyleData["Message", "Printout"],
    ShowCellBracket->False,
    CellMargins->{{44, 4}, {3, 8}},
    FontSize->10,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Print"],
    CellMargins->{{56, 4}, {3, 8}},
    CellGroupingRules->"OutputGrouping",
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    CellLabelMargins->{{21, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultOutputFormatType,
    AutoItalicWords->{},
    FormatType->InputForm,
    CounterIncrements->"Print",
    StyleMenuListing->None],
  
  Cell[StyleData["Print", "Printout"],
    ShowCellBracket->False,
    CellMargins->{{44, 4}, {3, 8}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Graphics"],
    CellMargins->{{55, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Graphics",
    StyleMenuListing->None],
  
  Cell[StyleData["Graphics", "Printout"],
    CellMargins->{{42, 4}, {3, 8}},
    ImageSize->{250, 250},
    FontSize->9]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["CellLabel"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontColor->RGBColor[0, 0.2, 1]],
  
  Cell[StyleData["CellLabel", "Printout"],
    FontSize->7,
    FontSlant->"Oblique",
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Open  ]],
  
  Cell[CellGroupData[{
  
  Cell["Unique Styles", "Subsection"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemizedText"],
    CellMargins->{{26, 4}, {0, 8}},
    ShowSpecialCharacters->Automatic,
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 8},
    ParagraphIndent->-19,
    CounterIncrements->"Text",
    FontSize->14],
  
  Cell[StyleData["ItemizedText", "Printout"],
    CellMargins->{{20, 4}, {0, 8}},
    ParagraphIndent->-14,
    FontSize->11]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemizedTextNote"],
    CellMargins->{{45, 4}, {0, 5}},
    ShowSpecialCharacters->Automatic,
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 5},
    CounterIncrements->"Text",
    FontSize->14],
  
  Cell[StyleData["ItemizedTextNote", "Printout"],
    CellMargins->{{34, 4}, {0, 4}},
    ParagraphSpacing->{0, 4},
    FontSize->11]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["IndentedText"],
    CellMargins->{{26, 4}, {0, 8}},
    ShowSpecialCharacters->Automatic,
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 8},
    CounterIncrements->"Text",
    FontSize->14],
  
  Cell[StyleData["IndentedText", "Printout"],
    CellMargins->{{20, 4}, {0, 8}},
    FontSize->11]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Note"],
    CellFrame->True,
    CellMargins->{{10, 4}, {0, 8}},
    CellHorizontalScrolling->True,
    CellFrameColor->RGBColor[0, 0, 1],
    ShowSpecialCharacters->Automatic,
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 8},
    CounterIncrements->"Text",
    FontFamily->"Helvetica",
    FontSize->10],
  
  Cell[StyleData["Note", "Printout"],
    CellMargins->{{7, 4}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->8]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Picture"],
    CellMargins->{{10, Inherited}, {0, 8}},
    CellHorizontalScrolling->True],
  
  Cell[StyleData["Picture", "Printout"],
    CellMargins->{{7, Inherited}, {0, 8}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Caption"],
    CellMargins->{{10, 50}, {0, 3}},
    FontFamily->"Helvetica",
    FontSize->9,
    FontColor->RGBColor[0, 0, 1]],
  
  Cell[StyleData["Caption", "Printout"],
    CellMargins->{{7, 50}, {2, 4}},
    FontSize->7,
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Tables", "Subsection"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["2ColumnTable"],
    CellMargins->{{10, 4}, {0, 8}},
    CellHorizontalScrolling->True,
    GridBoxOptions->{ColumnWidths->{0.39, 0.59},
    ColumnAlignments->{Left}}],
  
  Cell[StyleData["2ColumnTable", "Printout"],
    CellMargins->{{7, 4}, {0, 8}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["3ColumnTable"],
    CellMargins->{{10, 4}, {0, 8}},
    CellHorizontalScrolling->True,
    StyleMenuListing->None,
    GridBoxOptions->{ColumnWidths->0.325,
    ColumnAlignments->{Left}}],
  
  Cell[StyleData["3ColumnTable", "Printout"],
    CellMargins->{{7, 4}, {0, 8}},
    FontSize->10]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Formulas and Programming", "Subsection"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ChemicalFormula"],
    CellMargins->{{10, 4}, {0, 8}},
    CellHorizontalScrolling->True,
    AutoSpacing->False,
    ScriptLevel->1,
    ScriptBaselineShifts->{0.6, Automatic},
    SingleLetterItalics->False,
    ZeroWidthTimes->True],
  
  Cell[StyleData["ChemicalFormula", "Printout"],
    CellMargins->{{7, 4}, {0, 8}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["DisplayFormula"],
    CellMargins->{{55, Inherited}, {Inherited, Inherited}},
    CellHorizontalScrolling->True,
    ScriptLevel->0,
    SingleLetterItalics->True,
    UnderoverscriptBoxOptions->{LimitsPositioning->True},
    GridBoxOptions->{ColumnWidths->Automatic}],
  
  Cell[StyleData["DisplayFormula", "Printout"],
    CellMargins->{{43, Inherited}, {Inherited, Inherited}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Program"],
    CellMargins->{{10, Inherited}, {Inherited, Inherited}},
    CellHorizontalScrolling->True,
    FontFamily->"Courier"],
  
  Cell[StyleData["Program", "Printout"],
    CellMargins->{{7, Inherited}, {Inherited, Inherited}},
    FontSize->9.5]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Hyperlink Styles", "Subsection"],
  
  Cell["\<\
The cells below define styles useful for making hypertext \
ButtonBoxes.  The \"Hyperlink\" style is for links within the same Notebook, \
or between Notebooks.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Hyperlink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontColor->GrayLevel[1],
    Background->RGBColor[0, 0.6, 1],
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`NotebookLocate[ #2]}]&),
    Active->True,
    ButtonFrame->"None",
    ButtonNote->ButtonData}],
  
  Cell[StyleData["Hyperlink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell["\<\
The following styles are for linking automatically to the on-line \
help system.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MainBookLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontColor->GrayLevel[1],
    Background->RGBColor[0, 0.6, 1],
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "MainBook", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["MainBookLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["AddOnsLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontColor->GrayLevel[1],
    Background->RGBColor[0, 0.6, 1],
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "AddOns", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["AddOnLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["RefGuideLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontColor->GrayLevel[1],
    Background->RGBColor[0, 0.6, 1],
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "RefGuideLink", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["RefGuideLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["GettingStartedLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontColor->GrayLevel[1],
    Background->RGBColor[0, 0.6, 1],
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "GettingStarted", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["GettingStartedLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["OtherInformationLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontColor->GrayLevel[1],
    Background->RGBColor[0, 0.6, 1],
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "OtherInformation", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["OtherInformationLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1]]
  }, Closed]]
  }, Closed]]
  }, Open  ]]
  }]
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1709, 49, 82, 1, 116, "Title"],
Cell[1794, 52, 79, 1, 33, "Subtitle"],
Cell[1876, 55, 652, 11, 141, "Text"],

Cell[CellGroupData[{
Cell[2553, 70, 78, 1, 92, "Section"],
Cell[2634, 73, 338, 6, 84, "Text"],
Cell[2975, 81, 671, 10, 160, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[3683, 96, 72, 1, 60, "Section"],
Cell[3758, 99, 430, 12, 282, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[4225, 116, 31, 0, 60, "Section"],
Cell[4259, 118, 970, 34, 84, "Text"],
Cell[5232, 154, 447, 11, 84, "Text"],
Cell[5682, 167, 142, 3, 46, "Text"],
Cell[5827, 172, 46, 0, 27, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[5910, 177, 33, 0, 60, "Section"],
Cell[5946, 179, 206, 5, 46, "Text"],

Cell[CellGroupData[{
Cell[6177, 188, 38, 0, 36, "Subsubsection"],
Cell[6218, 190, 80, 2, 26, "Input"],
Cell[6301, 194, 93, 2, 26, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[6431, 201, 54, 0, 30, "Subsubsection"],
Cell[6488, 203, 83, 2, 26, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[6608, 210, 35, 0, 30, "Subsubsection"],
Cell[6646, 212, 336, 6, 138, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[7019, 223, 52, 0, 30, "Subsubsection"],
Cell[7074, 225, 952, 18, 218, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[8063, 248, 52, 0, 30, "Subsubsection"],
Cell[8118, 250, 961, 18, 218, "Input"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[9128, 274, 39, 0, 60, "Section"],
Cell[9170, 276, 463, 7, 103, "Text"],
Cell[9636, 285, 88, 2, 26, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[9761, 292, 37, 0, 60, "Section"],
Cell[9801, 294, 305, 5, 84, "Text"],
Cell[10109, 301, 60, 1, 26, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[10206, 307, 43, 0, 60, "Section"],

Cell[CellGroupData[{
Cell[10274, 311, 257, 5, 65, "Text"],
Cell[10534, 318, 67, 2, 26, "Input"]
}, Open  ]],

Cell[CellGroupData[{
Cell[10638, 325, 169, 4, 46, "Text"],
Cell[10810, 331, 72, 2, 26, "Input"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[10931, 339, 33, 0, 60, "Section"],

Cell[CellGroupData[{
Cell[10989, 343, 295, 5, 84, "Text"],
Cell[11287, 350, 778, 13, 314, "Input"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[12114, 369, 43, 0, 60, "Section"],

Cell[CellGroupData[{
Cell[12182, 373, 55, 0, 40, "Subsection"],

Cell[CellGroupData[{
Cell[12262, 377, 49, 0, 36, "Subsubsection"],
Cell[12314, 379, 984, 18, 365, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[13335, 402, 29, 0, 30, "Subsubsection"],
Cell[13367, 404, 819, 13, 210, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[14223, 422, 48, 0, 30, "Subsubsection"],
Cell[14274, 424, 632, 11, 196, "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[14955, 441, 27, 0, 34, "Subsection"],

Cell[CellGroupData[{
Cell[15007, 445, 62, 0, 36, "Subsubsection"],

Cell[CellGroupData[{
Cell[15094, 449, 216, 4, 74, "Input"],
Cell[15313, 455, 34563, 1630, 186, 22505, 1476, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[49925, 2091, 52, 0, 30, "Subsubsection"],

Cell[CellGroupData[{
Cell[50002, 2095, 205, 4, 58, "Input"],
Cell[50210, 2101, 30266, 1725, 226, 18513, 1575, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[80549, 3834, 29, 0, 60, "Section"],

Cell[CellGroupData[{
Cell[80603, 3838, 52, 0, 40, "Subsection"],
Cell[80658, 3840, 259, 5, 65, "Text"],
Cell[80920, 3847, 280, 6, 90, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[81237, 3858, 36, 0, 34, "Subsection"],
Cell[81276, 3860, 313, 7, 84, "Text"],

Cell[CellGroupData[{
Cell[81614, 3871, 76, 0, 36, "Subsubsection"],

Cell[CellGroupData[{
Cell[81715, 3875, 83, 2, 26, "Input"],
Cell[81801, 3879, 78, 2, 26, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[81928, 3887, 79, 0, 30, "Subsubsection"],

Cell[CellGroupData[{
Cell[82032, 3891, 86, 2, 26, "Input"],
Cell[82121, 3895, 79, 2, 26, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[82249, 3903, 59, 0, 30, "Subsubsection"],
Cell[82311, 3905, 221, 4, 65, "Text"],

Cell[CellGroupData[{
Cell[82557, 3913, 80, 2, 26, "Input"],
Cell[82640, 3917, 110, 2, 26, "Output"]
}, Open  ]]
}, Closed]]
}, Closed]]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

